/***************************************************************************
 *   Copyright (C) 2005-2006 Nicolas Hadacek <hadacek@kde.org>             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "cmdline.h"

#include <qfile.h>
#include <kapplication.h>
#include <klocale.h>

#include "devices/base/device_list.h"
#include "devices/base/device_group.h"
#include "common/log.h"
#include "common/about.h"
#include "progs/base/prog_config.h"
#include "devices/pic/pic_memory.h"

//-----------------------------------------------------------------------------
Log::ConsoleView *CommandLine::_log = 0;
KCmdLineArgs *CommandLine::_args = 0;
const Programmer::Group *CommandLine::_group = 0;
CommandLine::Command CommandLine::_command = CommandLine::Nb_Commands;
QString CommandLine::_device;
HexBuffer::Format CommandLine::_format = HexBuffer::Nb_Formats;

const CommandLine::CommandData CommandLine::COMMAND_DATA[Nb_Commands] = {
  { "connect",     NeedProgrammer | NeedDevice },
  { "run",         NeedProgrammer | NeedDevice },
  { "program",     NeedProgrammer | InputHex | NeedDevice },
  { "verify",      NeedProgrammer | InputHex | NeedDevice },
  { "read",        NeedProgrammer | OutputHex | NeedDevice },
  { "erase",       NeedProgrammer | NeedDevice },
  { "blank_check", NeedProgrammer | NeedDevice },
  { "checksum",    InputHex | NeedDevice },
  { "upload_firmware", NeedProgrammer | InputHex }
};

const KCmdLineOptions CommandLine::OPTIONS[] = {
  { "c", 0, 0 },
  { "command <name>", I18N_NOOP("Perform the requested command."), 0 },
  { "command-list", I18N_NOOP("Return the list of recognized commands."), 0 },
  { "p", 0, 0 },
  { "programmer <name>", I18N_NOOP("Programmer to use."), 0 },
  { "programmer-list", I18N_NOOP("Return the list of supported programmers."), 0 },
  { "d", 0, 0 },
  { "device <name>", I18N_NOOP("Target device."), 0 },
  { "device-list", I18N_NOOP("Return the  list of supported devices."), 0 },
  { "f", 0, 0 },
  { "format <name>", I18N_NOOP("Hex output file format."), "inhx32" },
  { "format-list", I18N_NOOP("Return the list of supported hex file formats."), 0 },
  { "debug", I18N_NOOP("Display debug messages."), 0 },
  { "extra-debug", I18N_NOOP("Display extra debug messages."), 0 },
  { "max-debug", I18N_NOOP("Display all debug messages."), 0 },
  { "quiet", I18N_NOOP("Do not display messages."), 0 },
  { "+[file]", I18N_NOOP("Hex filename for reading, programming or verifying."), 0 },
  { 0, 0, 0}
};

int CommandLine::commandList()
{
  _log->log(Log::Data, i18n("Supported commands:"));
  for (uint i=0; i<Nb_Commands; i++)
    _log->log(Log::Data, QString(" ") + COMMAND_DATA[i].name);
  return OK;
}

int CommandLine::formatList()
{
  _log->log(Log::Data, i18n("Supported hex file formats:"));
  for (uint i=0; i<HexBuffer::Nb_Formats; i++)
    _log->log(Log::Data, QString(" ") + HexBuffer::FORMATS[i]);
  return OK;
}

int CommandLine::programmerList()
{
  _log->log(Log::Data, i18n("Supported programmers:"));
  for (uint i=0; i<Programmer::lister().nbGroups(); i++)
    _log->log(Log::Data, " " + QString(Programmer::lister().group(i).name()));
  return OK;
}

int CommandLine::deviceList()
{
  QStringList names;
  if ( _group==0 ) {
    _log->log(Log::Data, i18n("Supported devices:"));
    names = Programmer::lister().supportedDevices();
  } else {
    _log->log(Log::Data, i18n("Supported devices for \"%1\":")
      .arg(_group->label()));
    names = _group->supportedDevices();
  }
  QString s;
  for (uint i=0; i<names.count(); i++) s += " " + names[i];
  _log->log(Log::Data, s + "\n");
  return OK;
}

int CommandLine::process()
{
  _args = KCmdLineArgs::parsedArgs();

  _log = new Log::ConsoleView;
  if ( _args->isSet("debug") ) _log->setDebugLevel(Log::NormalDebug);
  if ( _args->isSet("extra-debug") ) _log->setDebugLevel(Log::ExtraDebug);
  if ( _args->isSet("max-debug") ) _log->setDebugLevel(Log::MaxDebug);
  if (_args->isSet("quiet")) _log->setQuiet();

  if ( _args->isSet("command-list") ) return commandList();
  if ( _args->isSet("format-list") ) return formatList();
  if ( _args->isSet("programmer-list") ) return programmerList();

  // find programmer type
  QString option = _args->getOption("programmer");
  if ( !option.isEmpty() ) {
    for (uint i=0; i<Programmer::lister().nbGroups(); i++) {
      const Programmer::Group &group = Programmer::lister().group(i);
      if ( option==group.name() ) {
        _group = &group;
        break;
      }
    }
    if ( _group==0 )
      return errorExit(i18n("Unknown programmer \"%1\".").arg(option), ARG_ERROR);
  }

  // device list
  if ( _args->isSet("device-list") ) return deviceList();

  // find command
  option = _args->getOption("command");
  if ( option.isEmpty() )
    return errorExit(i18n("No command specified."), ARG_ERROR);
  for (uint k=0; k<Nb_Commands; k++) {
    if ( option==COMMAND_DATA[k].name ) {
      _command = Command(k);
      break;
    }
  }
  if ( _command==Nb_Commands )
    return errorExit(i18n("Unknown command \"%1\".").arg(option), ARG_ERROR);
  CommandProperties properties = COMMAND_DATA[_command].properties;

  // check programmer
  if (_group) {
    if ( _command==UploadFirmware && !(_group->properties() & ::Programmer::CanUploadFirmware) )
      return errorExit(i18n("Uploading firmware is not supported for specified programmer."), NOT_SUPPORTED_ERROR);
  } else if ( properties & NeedProgrammer ) return errorExit(i18n("Programmer not specified."), ARG_ERROR);

  // find device name
  const Device::Data *data = 0;
  _device = _args->getOption("device");
  if ( !_device.isEmpty() ) {
    data = Device::lister().data(_device);
    if ( data==0 ) return errorExit(i18n("Unknown device \"%1\".").arg(_device), ARG_ERROR);
    if ( _group && !_group->isSupported(_device) )
      return errorExit(i18n("Device not supported by the specified programmer."), NOT_SUPPORTED_ERROR);
  } else if ( properties & NeedDevice ) return errorExit(i18n("Device not specified."), ARG_ERROR);

  // check hex file
  if ( _args->count()>1 ) return errorExit(i18n("Too many arguments."), ARG_ERROR);
  KURL url;
  if ( _args->count()==1 ) {
    url = _args->url(0);
    if ( !url.isLocalFile() ) return errorExit(i18n("Only local files are supported."), ARG_ERROR);
  }
  QFile file(url.path());
  Device::Memory *memory = 0;
  if ( properties & NeedDevice && (properties & InputHex || properties & OutputHex) )
    memory = data->group().createMemory(*data);
  if ( properties & InputHex ) {
    if ( _args->count()==0 ) return errorExit(i18n("Input hex filename not specified."), ARG_ERROR);
    if (memory) {
      if ( !file.open(IO_ReadOnly) ) return errorExit(i18n("Could not open hex file."), FILE_ERROR);
      QString error;
      Device::Memory::WarningTypes warningTypes;
      QStringList warnings;
      if ( !memory->load(file, error, warningTypes, warnings) )
        return errorExit(i18n("Could not load hex file: %1.").arg(error), FILE_ERROR);
      if ( warningTypes!=Device::Memory::NoWarning )
        _log->log(Log::Warning, i18n("Hex file seems incompatible with device: %1").arg(warnings.join(" ")));
      file.close();
    }
  } else if ( properties & OutputHex ) {
    if ( _args->count()==0 ) return errorExit(i18n("Output hex filename not specified."), ARG_ERROR);
    if ( file.exists() ) return errorExit(i18n("Output hex filename already exists."), FILE_ERROR);
  } else if ( _args->count()==1 ) return errorExit(i18n("Too many arguments."), ARG_ERROR);

  // check format
  option = _args->getOption("format");
  for (uint i=0; i<HexBuffer::Nb_Formats; i++)
    if ( option==HexBuffer::FORMATS[i] ) {
      _format = HexBuffer::Format(i);
      break;
    }
  if ( _format==HexBuffer::Nb_Formats )
    return errorExit(i18n("Unknown hexfile format."), ARG_ERROR);

  // execute action
  Programmer::Base *programmer = 0;
  if ( properties & NeedProgrammer ) {
    Port::Type portType = Programmer::Config::readPortType(*_group);
    QString portDevice = Programmer::Config::readPortDevice(*_group, portType);
    programmer = _group->createProgrammer(data, portType, portDevice);
    programmer->setView(_log);
  }
  switch (_command) {
    case Connect: return (programmer->connectDevice() ? OK : PROG_ERROR);
    case Run: return (programmer->run() ? OK : PROG_ERROR);
    case Program: return (programmer->program(*memory) ? OK : PROG_ERROR);
    case Verify: return (programmer->verify(*memory) ? OK : PROG_ERROR);
    case Read: {
      if ( !programmer->read(*memory) ) return PROG_ERROR;
      if ( !file.open(IO_WriteOnly) )
        return errorExit(i18n("Could not open file \"%1\" for writing.").arg(file.name()), FILE_ERROR);
      if ( !memory->save(file, _format) )
        return errorExit(i18n("Error while writing file \"%1\" (%2).")
                         .arg(file.name()).arg(file.errorString()), FILE_ERROR);
      return OK;
    }
    case Erase: return (programmer->erase() ? OK : PROG_ERROR);
    case BlankCheck: return (programmer->blankCheck() ? OK : PROG_ERROR);
    case Checksum: {
      uint cs = memory->checksum();
      _log->log(Log::Information, i18n("Cheksum: %1").arg(toHexLabel(cs, 4)));
      if ( data->group().name()=="pic" ) {
        uint ucs = static_cast<Pic::Memory *>(memory)->unprotectedChecksum();
        if ( ucs!=cs ) _log->log(Log::Information, i18n("Unprotected checksum: %1").arg(toHexLabel(ucs, 4)));
      } }
      return OK;
    case UploadFirmware: return (programmer->uploadFirmware(file.name()) ? OK : PROG_ERROR);
    case Nb_Commands: break;
  }
  Q_ASSERT(false);
  return PROG_ERROR;
}

int CommandLine::errorExit(const QString &message, int code)
{
  _log->log(Log::Error, i18n("Error: ") + message.latin1() + "\n");
  return code;
}

//-----------------------------------------------------------------------------
int main(int argc, char **argv)
{
  KApplication::disableAutoDcopRegistration();
  AboutData ad;
  KCmdLineArgs::init(argc, argv, &ad);
  KCmdLineArgs::addCmdLineOptions(CommandLine::OPTIONS);
  KApplication::addCmdLineOptions();
  KApplication a(false, false);
  return CommandLine::process();
}
