/***************************************************************************
 *   Copyright (C) 2006 Nicolas Hadacek <hadacek@kde.org>                  *
 *   Copyright (C) 1992-2003 Trolltech AS.                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "misc_gui.h"

#include <qpushbutton.h>
#include <qapplication.h>
#include <qtimer.h>
#include <qwidgetstack.h>
#include <qtabwidget.h>
#include <qobjectlist.h>
#include <qpainter.h>
#include <qheader.h>
#include <qmetaobject.h>
#include <qvariant.h>
#include <qpopupmenu.h>

#include <kcursor.h>
#include <kiconloader.h>
#include <kmessagebox.h>
#include <kaction.h>

//-----------------------------------------------------------------------------
bool BusyCursor::_overridePaused = false;

void BusyCursor::start()
{
  QApplication::setOverrideCursor(KCursor::waitCursor(), true);
}

void BusyCursor::stop()
{
  QApplication::restoreOverrideCursor();
}

void BusyCursor::pause()
{
  _overridePaused = QApplication::overrideCursor();
  stop();
}

void BusyCursor::restore()
{
  if (_overridePaused) start();
}

//-----------------------------------------------------------------------------
Container::Container(QWidget *parent, Type type, const char *name)
  : QFrame(parent, name), _type(type), _topLayout(0), _gridLayout(0)
{
  initLayout();
}

Container::Container(QWidgetStack *stack, uint index, Type type,
                     const char *name)
  : QFrame(stack, name), _type(type), _topLayout(0), _gridLayout(0)
{
  initLayout();
  stack->addWidget(this, index);
}

Container::Container(QTabWidget *tabw, const QString &title, Type type,
                     const char *name)
  : QFrame(tabw, name), _type(type), _topLayout(0), _gridLayout(0)
{
  initLayout();
  tabw->addTab(this, title);
}

void Container::setFrame(Type type)
{
  _type = type;
  switch (type) {
  case Flat:
    _topLayout->setMargin(parent()->inherits("QTabWidget") ? 10 : 0);
    setFrameStyle(QFrame::NoFrame);
    break;
  case Sunken:
    _topLayout->setMargin(10);
    setFrameStyle(QFrame::StyledPanel | QFrame::Sunken);
    break;
  }
}

void Container::initLayout()
{
  delete _topLayout;
  _topLayout = new QGridLayout(this, 1, 1, 0, 10);
  _gridLayout = new QGridLayout(1, 1, 10);
  _topLayout->addLayout(_gridLayout, 0, 0);
  _topLayout->setRowStretch(1, 1);
  setFrame(_type);
  QObjectList *list = queryList("QWidget", 0, true, false);
  QObjectListIterator it(*list);
  QObject *o;
  while ( (o = it.current()) ) {
    ((QWidget *)o)->hide();
    ++it;
  }
}

void Container::addWidget(QWidget *w, int x, int y, int alignment)
{
  w->show();
  _gridLayout->addWidget(w, x, y, alignment);
}

void Container::addMultiCellWidget(QWidget *w, int xmin, int xmax,
                                   int ymin, int ymax, int alignment)
{
  w->show();
  _gridLayout->addMultiCellWidget(w, xmin, xmax, ymin, ymax, alignment);
}

//-----------------------------------------------------------------------------
void MessageBox::information(const QString &text, Log::ShowMode show)
{
  if ( show==Log::DontShow ) return;
  BusyCursor::pause();
  KMessageBox::information(qApp->mainWidget(), text, QString::null, QString::null, KMessageBox::Notify | KMessageBox::AllowLink);
  BusyCursor::restore();
}

void MessageBox::sorry(const QString &text, Log::ShowMode show)
{
  if ( show==Log::DontShow ) return;
  BusyCursor::pause();
  KMessageBox::sorry(qApp->mainWidget(), text, QString::null, KMessageBox::Notify | KMessageBox::AllowLink);
  BusyCursor::restore();
}

void MessageBox::detailedSorry(const QString &text, const QString &details, Log::ShowMode show)
{
  if ( show==Log::DontShow ) return;
  BusyCursor::pause();
  KMessageBox::detailedSorry(qApp->mainWidget(), text, details, QString::null, KMessageBox::Notify | KMessageBox::AllowLink);
  BusyCursor::restore();
}

bool MessageBox::askContinue(const QString &text, const QString &title, const KGuiItem &buttonContinue)
{
  ::BusyCursor::pause();
  int res = KMessageBox::warningContinueCancel(qApp->mainWidget(), text, title, buttonContinue);
  ::BusyCursor::restore();
  return ( res==KMessageBox::Continue );
}

//----------------------------------------------------------------------------
ListView::ListView(QWidget *parent, const char *name)
  : KListView(parent, name)
{}

void ListView::clear()
{
  _editItems.clear();
  KListView::clear();
}

bool ListView::eventFilter(QObject *o, QEvent *e)
{
  QValueList<EditListViewItem *>::const_iterator it;
  for (it=_editItems.begin(); it!=_editItems.end(); ++it) {
    for (uint i=0; i<(*it)->_editWidgets.count(); i++) {
      if ( (*it)->_editWidgets[i]==o ) {
        switch (e->type()) {
          case QEvent::KeyPress: {
            QKeyEvent *ke = static_cast<QKeyEvent *>(e);
            switch (ke->key()) {
              case Key_Enter:
              case Key_Return:
                (*it)->renameDone(true);
                return true;
              case Key_Escape:
                (*it)->removeEditBox();
                return true;
            }
            break;
          }
          case QEvent::FocusOut: {
            //qDebug("focus out %i %i=%i", qApp->focusWidget(), focusWidget(), (*it)->_editWidgets[i]);
            if ( qApp->focusWidget() && focusWidget()==(*it)->_editWidgets[i] ) break;
            //qDebug("ext");
            QCustomEvent *e = new QCustomEvent(9999);
            QApplication::postEvent(o, e);
            return true;
          }
          case 9999:
            //qDebug("9999");
            (*it)->renameDone(false);
            return true;
          default: break;
        }
      }
    }
  }
  return KListView::eventFilter(o, e);
}

void ListView::stopRenaming()
{
  QValueList<EditListViewItem *>::const_iterator it;
  for (it=_editItems.begin(); it!=_editItems.end(); ++it)
    if ( (*it)->isRenaming() ) (*it)->renameDone(false);
}

//----------------------------------------------------------------------------
EditListViewItem::EditListViewItem(ListView *list)
  : KListViewItem(list), _renaming(false)
{
  setRenameEnabled(0, true);
  list->_editItems.append(this);
}

EditListViewItem::EditListViewItem(KListViewItem *item)
  : KListViewItem(item), _renaming(false)
{
  setRenameEnabled(0, true);
  static_cast<ListView *>(listView())->_editItems.append(this);
}

EditListViewItem::~EditListViewItem()
{
  if ( listView() ) static_cast<ListView *>(listView())->_editItems.remove(this);
  for (uint i=0; i<_editWidgets.count(); i++)
    delete _editWidgets[i];
}

void EditListViewItem::paintCell(QPainter *p, const QColorGroup &cg, int column, int width, int align)
{
  if ( column<int(_editWidgets.count()) && _editWidgets[column] )
    p->fillRect(0, 0, width, height(), cg.color(QColorGroup::Background));
  else KListViewItem::paintCell(p, cg, column, width, align);
}

void EditListViewItem::startRename()
{
  if ( !renameEnabled(0) ) return;
  QListView *lv = listView();
  if ( !lv ) return;
  KListViewItem::startRename(0);
  if (renameBox) {
    renameBox->removeEventFilter(lv);
    renameBox->hide();
    lv->removeChild(renameBox);
  }
  _renaming = true;
  _editWidgets.resize(lv->columns());
  for (uint i=0; i<_editWidgets.count(); i++) {
    QRect r = lv->itemRect(this);
    r = QRect(lv->viewportToContents(r.topLeft()), r.size());
    r.setLeft(lv->header()->sectionPos(i));
    r.setWidth(lv->header()->sectionSize(i) - 1);
    if ( i==0 ) r.setLeft(r.left() + lv->itemMargin() + (depth() + (lv->rootIsDecorated() ? 1 : 0)) * lv->treeStepSize() - 1);
    if ( (lv->contentsX() + lv->visibleWidth())<(r.x() + r.width()) )
      lv->scrollBy((r.x() + r.width() ) - ( lv->contentsX() + lv->visibleWidth() ), 0);
    if ( r.width()>lv->visibleWidth() ) r.setWidth(lv->visibleWidth());

    _editWidgets[i] = editWidgetFactory(i);
    if ( _editWidgets[i]==0 ) continue;
    _editWidgets[i]->installEventFilter(lv);
    lv->addChild(_editWidgets[i], r.x(), r.y());
    uint w = QMIN(r.width(), _editWidgets[i]->sizeHint().width());
    _editWidgets[i]->resize(w, r.height());
    lv->viewport()->setFocusProxy(_editWidgets[i]);
    _editWidgets[i]->setFocus();
    _editWidgets[i]->show();
  }
}

void EditListViewItem::removeEditBox()
{
  QListView *lv = listView();
  if ( !lv ) return;
  _renaming = false;
  bool resetFocus = false;
  for (uint i=0; i<_editWidgets.count(); i++) {
    if ( lv->viewport()->focusProxy()==_editWidgets[i] ) resetFocus = true;
    delete _editWidgets[i];
  }
  _editWidgets.clear();
  delete renameBox;
  renameBox = 0;
  if (resetFocus) {
    lv->viewport()->setFocusProxy(lv);
    lv->setFocus();
  }
}

void EditListViewItem::editDone(int col, const QWidget *edit)
{
  if ( edit->metaObject()->findProperty("text", true)!=-1 )
    emit listView()->itemRenamed(this, col, edit->property("text").toString());
  else if ( edit->metaObject()->findProperty("currentText", true)!=-1 )
    emit listView()->itemRenamed(this, col, edit->property("currentText").toString());
}

void EditListViewItem::renameDone(bool force)
{
  QListView *lv = listView();
  if ( !lv || !_renaming ) return;
  _renaming = false;
  for (uint i=0; i<_editWidgets.count(); i++) {
    if ( !force && !alwaysAcceptEdit(i) ) continue;
    emit lv->itemRenamed(this, i);
    if ( _editWidgets[i] ) editDone(i, _editWidgets[i]);
  }
  removeEditBox();
}

int EditListViewItem::width(const QFontMetrics &fm, const QListView *lv, int col) const
{
  int w = KListViewItem::width(fm, lv, col);
  QWidget *edit = editWidgetFactory(col);
  if ( edit==0 ) return w;
  w = QMAX(w, edit->sizeHint().width());
  delete edit;
  return w;
}

//----------------------------------------------------------------------------
PopupButton::PopupButton(QWidget *parent, const char *name)
  : KPushButton(parent, name)
{
  init();
}

PopupButton::PopupButton(const QString &text, QWidget *parent, const char *name)
  : KPushButton(text, parent, name)
{
  init();
}

void PopupButton::init()
{
  _separator = false;
  setFlat(true);
  QPopupMenu *popup = new QPopupMenu(this);
  setPopup(popup);
}

void PopupButton::appendAction(KAction *action)
{
  if ( _separator && popup()->count()!=0 ) popup()->insertSeparator();
  action->plug(popup());
  _separator = false;
}
