/***************************************************************************
 *   Copyright (C) 2005-2006 Nicolas Hadacek <hadacek@kde.org>             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "purl.h"

#include <qapplication.h>
#include <qfileinfo.h>
#include <qdatetime.h>
#include <qdir.h>
#include <qregexp.h>

#include <klocale.h>
#include <kio/netaccess.h>
#include <kfileitem.h>
#include <kfiledialog.h>
#include <kdirselectdialog.h>
#include <ktempfile.h>
#include <kicontheme.h>

#include "misc_gui.h"

#include "data/xpms/project.xpm"
#include "data/xpms/sourcefile.xpm"
#include "data/xpms/includefile.xpm"
#include "data/xpms/objectfile.xpm"

//-----------------------------------------------------------------------------
PURLBase::PURLBase(const QString &filepath)
{
  _empty = filepath.isEmpty();
  if ( !_empty ) {
    _relative = QDir::isRelativePath(filepath);
    _url = KURL::fromPathOrURL('/' + filepath);
    _url.cleanPath();
  } else _relative = true;
}

PURLBase::PURLBase(const KURL &url)
  : _relative(false), _empty(url.isEmpty()), _url(url)
{
  _url.cleanPath();
}

QString PURLBase::path() const
{
  return _url.directory(false, false);
}

PDirectory PURLBase::directory() const
{
  return PDirectory(path());
}

bool PURLBase::isInto(const PDirectory &dir) const
{
  return path().startsWith(dir.path());
};

//-----------------------------------------------------------------------------
PDirectory::PDirectory(const QString &path)
  : PURLBase(path.isEmpty() ? QString::null : path + '/')
{}

bool PDirectory::create(Log::ShowMode show) const
{
  // assume dir do no exist if ioslave cannot tell...
  if ( KIO::NetAccess::exists(_url, false, qApp->mainWidget()) ) return true;
  if ( !KIO::NetAccess::mkdir(_url, qApp->mainWidget()) ) {
    _error = KIO::NetAccess::lastErrorString();
    MessageBox::detailedSorry(i18n("Could not create directory"), _error, show);
    return false;
  }
  return true;
}

PDirectory PDirectory::getExisting(const QString &startDir, QWidget *widget,
                                   const QString &caption)
{
  KURL kurl = KDirSelectDialog::selectDirectory(startDir, false, widget, caption);
  if ( kurl.isEmpty() ) return PDirectory();
  return PDirectory(kurl.path(1));
}

QString PDirectory::expandMacros(const QString &s) const
{
  QString r = s;
  r.replace(QRegExp("\\$\\(SRCPATH\\)"), path());
  r.replace(QRegExp("//"), "/");
  return r;
}

//----------------------------------------------------------------------------
const PURL::Data PURL::DATA[Nb_FileTypes] = {
  { Source | Editable, Asm, I18N_NOOP("ASM File"),            { "asm", "src", "pic", 0 }, sourcefile_xpm,  0,               "XPicAsm" },
  { Source | Editable, Asm, I18N_NOOP("ASM File for PIC30"),  { "s", 0 },                 sourcefile_xpm,  0,               "XPicAsm" },
  { Source | Editable, Asm, I18N_NOOP("ASM File for PICC"),   { "as", 0 },                sourcefile_xpm,  0,               "XPicAsm" },
  { Header | Editable, Asm, I18N_NOOP("Include File"),        { "inc", 0 },               includefile_xpm, 0,               "XPicAsm" },
  { Source | Editable, C,   I18N_NOOP("C Source File"),       { "c", 0 },                 0,               "text/x-csrc",   "C"       },
  { Header | Editable, C,   I18N_NOOP("C Header File"),       { "h", 0 },                 0,               "text/x-chdr",   "C"       },

  { ReadOnly,            Nb_SourceTypes, I18N_NOOP("Object File"),         { "o", "obj", 0 },          objectfile_xpm,  0,               0         },
  { ReadOnly,            Nb_SourceTypes, I18N_NOOP("Library File"),        { "a", "lib", 0 },          objectfile_xpm,  0,               0         },
  { Editable,            Nb_SourceTypes, I18N_NOOP("Hex File"),            { "hex", 0 },               0,               "text/x-hex",    0         },
  { ReadOnly,            Nb_SourceTypes, I18N_NOOP("Elf File"),            { "elf", 0 },               0,               0,               0         },
  { NoProperty,          Nb_SourceTypes, I18N_NOOP("Project File"),        { "piklab", 0 },            project_xpm,     0,               0         },
  { Editable | ReadOnly, Nb_SourceTypes, I18N_NOOP("List File"),           { "lst", 0 },               0,               "text/plain",    0         },
  { Editable | ReadOnly, Nb_SourceTypes, I18N_NOOP("Map File"),            { "map", 0 },               0,               "text/plain",    0         },
  { ReadOnly,            Nb_SourceTypes, I18N_NOOP("COD File"),            { "cod", 0 },               0,               0,               0         },
  { Editable | ReadOnly, Nb_SourceTypes, I18N_NOOP("COFF File"),           { "cof", 0 },               0,               "application/x-object", "XPicAsm" },
  { Editable,            Nb_SourceTypes, I18N_NOOP("Linker Script"),       { "lkr", 0 },               includefile_xpm, 0,               0         },
  { Editable,            Nb_SourceTypes, I18N_NOOP("PIC30 Linker Script"), { "gld", 0 },               includefile_xpm, 0,               0         },
  { Editable | ReadOnly, Nb_SourceTypes, I18N_NOOP("SDCC List File"),      { "rst", 0 },               0,               "text/plain",    0         },
  { Editable | ReadOnly, Nb_SourceTypes, I18N_NOOP("Symbol File"),         { "sym", 0 },               0,               "text/plain",    0         },
  { Editable | ReadOnly, Nb_SourceTypes, I18N_NOOP("SDCC Memory File"),    { "mem", 0 },               0,               "text/plain",    0         },
  { Editable | ReadOnly, Nb_SourceTypes, I18N_NOOP("SDCC Debug File"),     { "adb", 0 },               0,               "text/plain",    0         },
  { Editable | ReadOnly, Nb_SourceTypes, I18N_NOOP("SDCC Linker File"),    { "lnk", 0 },               0,               "text/plain",    0         },
  { Editable | ReadOnly, Nb_SourceTypes, I18N_NOOP("SDCC P File"),         { "p", 0 },                 0,               "text/plain",    0         },
  { Editable | ReadOnly, Nb_SourceTypes, I18N_NOOP("SDCC D File"),         { "d", 0 },                 0,               "text/plain",    0         },
  { Editable | ReadOnly, Nb_SourceTypes, I18N_NOOP("PICC Symbol File"),    { "sdb", 0 },               0,               "text/plain",    0         },
  { Editable | ReadOnly, Nb_SourceTypes, I18N_NOOP("Hexmate Log File"),    { "hxl", 0 },               0,               "text/plain",    0         },
  { Editable | ReadOnly, Nb_SourceTypes, I18N_NOOP("Rlf File"),            { "rlf", 0 },               0,               "text/plain",    0         },

  { NoProperty,          Nb_SourceTypes, I18N_NOOP("Unknown File"),        { 0 },                      0,               0,               0         },
  { NoProperty,          Nb_SourceTypes, I18N_NOOP("Pikdev Project File"), { "pikprj", 0 },            0,               0,               0         }
};

const char * const PURL::SOURCE_TYPE_LABELS[Nb_SourceTypes] = {
  I18N_NOOP("Asm"), I18N_NOOP("C")
};

QPixmap PURL::icon(FileType type)
{
  if (DATA[type].xpm_icon) return QPixmap(PURL::DATA[type].xpm_icon);
  if ( hasMimetype(type) ) return KMimeType::mimeType(DATA[type].mimetype)->pixmap(KIcon::Small);
  return QPixmap();
}

PURL::PURL(const QString &filepath)
  : PURLBase(filepath)
{}

PURL::PURL(const PDirectory &dir, const QString &filename, FileType type)
  : PURLBase(dir.path() + '/' + addExtension(filename, type))
{}

PURL::PURL(const PDirectory &dir, const QString &filepath)
  : PURLBase(dir.path() + '/' + filepath)
{}

PURL::FileType PURL::fileType() const
{
  QFileInfo info(filename());
  for (uint k=0; k<Nb_FileTypes; k++)
    for (uint i=0; DATA[k].extensions[i]; i++)
      if ( info.extension(false).lower()==DATA[k].extensions[i] ) return FileType(k);
  return Unknown;
}

PURL PURL::toFileType(FileType type, ExtensionOptions options) const
{
  return toExtension(DATA[type].extensions[0], options);
}

PURL PURL::toExtension(const QString &extension, ExtensionOptions options) const
{
  QFileInfo info(filename());
  QString fname = (options & AddExtension ? info.fileName() : info.baseName());
  QString ext = extension;
  if ( options & UpperCaseExtension ) ext = ext.upper();
  return PURL(directory().path() + fname + '.' + ext);
}

QString PURL::basename() const
{
  QFileInfo info(_url.fileName());
  return info.baseName();
}

QString PURL::filename() const
{
  QFileInfo info(_url.fileName());
  return info.fileName();
}

QString PURL::filepath() const
{
  //qDebug("%s: %s + %s", _url.prettyURL().latin1(), path().latin1(), filename().latin1());
  return path() + filename();
}

QString PURL::relativeTo(const PDirectory &dir) const
{
  Q_ASSERT ( isInto(dir) );
  return filepath().right(filepath().length() - dir.path().length());
}

bool PURL::probablyExists() const
{
  // assume file exists if ioslave cannot tell...
  return KIO::NetAccess::exists(_url, true, qApp->mainWidget());
}

bool PURL::copyTo(const PURL &destination, Log::ShowMode show) const
{
  // do not overwrite
  bool ok = KIO::NetAccess::file_copy(_url, destination._url, -1, false, false, qApp->mainWidget());
  if ( !ok ) {
    _error = KIO::NetAccess::lastErrorString();
    MessageBox::detailedSorry(i18n("Could not copy file"), _error, show);
  }
  return ok;
}

bool PURL::create(Log::ShowMode show) const
{
  // assume file do no exist if ioslave cannot tell...
  if ( KIO::NetAccess::exists(_url, false, qApp->mainWidget()) ) return true;
  KTempFile tmp;
  tmp.setAutoDelete(true);
  // do not overwrite
  bool ok = KIO::NetAccess::file_copy(tmp.name(), _url, -1, false, false, qApp->mainWidget());
  if ( !ok ) {
    _error = KIO::NetAccess::lastErrorString();
    MessageBox::detailedSorry(i18n("Could not create file"), _error, show);
  }
  return ok;
}

bool PURL::write(const QString &text, Log::ShowMode show) const
{
  PFile file(*this, show);
  if ( file.openForWrite() ) file.stream() << text;
  bool ok = file.close();
  _error = file.errorString();
  return ok;
}

bool PURL::del(Log::ShowMode show) const
{
  bool ok = KIO::NetAccess::del(_url, qApp->mainWidget());
  if ( !ok ) {
    _error = KIO::NetAccess::lastErrorString();
    MessageBox::detailedSorry(i18n("Could not delete file."), _error, show);
  }
  return ok;
}

bool PURL::isDestinationUpToDate(FileType type) const
{
  if ( !data().properties & Source ) return true;
  KIO::UDSEntry suds, duds;
  if ( !KIO::NetAccess::stat(_url, suds, qApp->mainWidget()) ) return true; // source does not exists
  PURL url = toFileType(type);
  if ( !KIO::NetAccess::stat(url._url, duds, qApp->mainWidget()) ) return false; // destination does not exists
  KFileItem sitem(suds, _url), ditem(duds, url._url);
  return ( sitem.time(KIO::UDS_MODIFICATION_TIME)<ditem.time(KIO::UDS_MODIFICATION_TIME) );
}

QString PURL::addExtension(const QString &filename, FileType type)
{
  QFileInfo info(filename);
  if ( !info.extension().isEmpty() ) return filename;
  return filename + '.' + extension(type);
}

QString PURL::extensions(FileType type)
{
  QString s;
  for (uint i=0; DATA[type].extensions[i]; i++) {
    if ( i!=0 ) s += ' ';
    s += QString("*.") + DATA[type].extensions[i];
  }
  return s;
}

bool PURL::hasMimetype(FileType type)
{
  if ( DATA[type].mimetype==0 ) return false;
  KMimeType::Ptr ptr = KMimeType::mimeType(DATA[type].mimetype);
  return ( ptr!=KMimeType::defaultMimeTypePtr() );
}

QString PURL::filter(FileType type)
{
  //if ( hasMimetype(type) ) return DATA[type].mimetype; // #### we cannot mix mimetype and regular filters in KFileDialog...
  QString s = extensions(type);
  return s + ' ' + s.upper() + '|' + i18n(DATA[type].label) + " (" + s + ")";
}

bool PURL::isDosFile() const
{
  PFile file(*this, Log::DontShow);
  if( !file.openForRead() ) return false;
  int oldc = 0;
  for (;;) {
    int c = file.qfile()->getch();
    if ( c==-1 )  break;
    if( c=='\n' && oldc=='\r' ) return true;
    oldc = c;
  }
  return false;
}

QString PURL::extensions(FileProperty property)
{
  QString e;
  for (uint i=0; i<Nb_FileTypes; i++) {
    if ( !(DATA[i].properties & property) ) continue;
    if ( i!=0 ) e += ' ';
    QString s = extensions(FileType(i));
    e += s + ' ' + s.upper();
  }
  return e;
}

QString PURL::sourceFilter()
{
  QString f = extensions(Source) + ' ' + extensions(Header) + '|' + i18n("All Source Files");
  for (uint i=0; i<Nb_FileTypes; i++) {
    if ( !(DATA[i].properties & (Source | Header)) ) continue;
    f += '\n' + filter(FileType(i));
  }
  return f;
}

QString PURL::objectFilter()
{
  QString f = extensions(Object) + ' ' + extensions(Library) + '|' + i18n("All Object Files");
  f += '\n' + filter(Object);
  f += '\n' + filter(Library);
  return f;
}

QString PURL::projectFilter()
{
  QString f = extensions(Project) + ' ' + extensions(PikdevProject) + '|' + i18n("Project Files");
  f += '\n' + filter(Project);
  f += '\n' + filter(PikdevProject);
  return f;
}

PURL PURL::getOpen(const QString &startDir, const QString &filter,
                   QWidget *widget, const QString &caption)
{
  return KFileDialog::getOpenURL(startDir, filter, widget, caption);
}

PURL PURL::getSave(const QString &startDir, const QString &filter,
                   QWidget *widget, const QString &caption, SaveAction action)
{
  PURL url = KFileDialog::getSaveURL(startDir, filter, widget, caption);
  if ( url.isEmpty() ) return PURL();
  switch (action) {
    case NoSaveAction: break;
    case AskOverwrite:
      if ( url.probablyExists() ) {
        if ( !MessageBox::askContinue(i18n("File \"%1\" already exists. Overwrite ?").arg(url.pretty()),
                                      i18n("Warning"), i18n("Overwrite")) ) return PURL();
      }
      break;
    case CancelIfExists:
      if ( url.probablyExists() ) return PURL();
      break;
  }
  return url;
}

//-----------------------------------------------------------------------------
PFileBase::~PFileBase()
{
  delete _stream;
  delete _tmp;
}

void PFileBase::flush()
{
  if ( qfile() ) qfile()->flush();
}

QTextStream &PFileBase::stream()
{
  if ( _stream==0 ) _stream = new QTextStream(qfile());
  return *_stream;
}

bool PFileBase::hasError() const
{
  if ( qfile()==0 || !_error.isEmpty() ) return true;
  return ( qfile()->status()!=IO_Ok );
}

QString PFileBase::errorString() const
{
  if ( _error.isEmpty() ) {
    if ( qfile()==0 ) return i18n("File not open.");
    else return qfile()->errorString();
  }
  return _error;
}

QStringList PFileBase::readLines()
{
  QStringList lines;
  for (;;) {
    QString s = stream().readLine();
    if ( s.isNull() ) break;
    lines.append(s);
  }
  return lines;
}

bool PFileBase::openForWrite()
{
  close();
  if (_tmp) delete _tmp;
  _tmp = new KTempFile;
  _tmp->setAutoDelete(true);
  if ( _tmp->status()!=0 ) {
    _error = i18n("Could not create temporary file.");
    MessageBox::detailedSorry(_error, i18n("File: %1").arg(_tmp->name()), _show);
    return false;
  }
  return true;
}

//-----------------------------------------------------------------------------
PFile::PFile(const PURL &url, Log::ShowMode show)
  : PFileBase(show), _url(url)
{}

void PFile::setUrl(const PURL &url)
{
  close();
  _url = url;
}

bool PFile::close()
{
  if (_tmp) _tmp->close();
  else _file.close();
  bool ok = (_tmp ? _tmp->status() : _file.status())==IO_Ok;
  if ( !_file.name().isEmpty() ) {
    KIO::NetAccess::removeTempFile(_file.name());
    _file.setName(QString::null);
  }
  delete _stream;
  _stream = 0;
  if ( ok && _tmp && !_url.isEmpty() && !KIO::NetAccess::upload(_tmp->name(), _url.kurl(), qApp->mainWidget()) ) {
    _error = KIO::NetAccess::lastErrorString();
    ok = false;
    MessageBox::detailedSorry(i18n("Could not save file."), errorString(), _show);
  }
  delete _tmp;
  _tmp = 0;
  return ok;
}

bool PFile::openForRead()
{
  close();
  QString tmp;
  if ( !KIO::NetAccess::download(_url.kurl(), tmp, qApp->mainWidget()) ) {
    _error = KIO::NetAccess::lastErrorString();
    MessageBox::detailedSorry(i18n("Could not open file for reading."), errorString(), _show);
    return false;
  }
  _file.setName(tmp);
  if ( !_file.open(IO_ReadOnly) ) {
    _error = i18n("Could not open temporary file.");
    MessageBox::detailedSorry(_error, i18n("File: %1").arg(tmp), _show);
    return false;
  }
  return true;
}

const QFile *PFile::qfile() const
{
  return (_tmp ? _tmp->file() : &_file);
}

QFile *PFile::qfile()
{
  return (_tmp ? _tmp->file() : &_file);
}

//-----------------------------------------------------------------------------
QFile *PTempFile::qfile()
{
  return (_tmp ? _tmp->file() : 0);
}

const QFile *PTempFile::qfile() const
{
  return (_tmp ? _tmp->file() : 0);
}

PURL PTempFile::url() const
{
  return (_tmp ? PURL(_tmp->name()) : PURL());
}

bool PTempFile::close()
{
  delete _stream;
  _stream = 0;
  if (_tmp) {
    _tmp->close();
    if ( _tmp->status()!=IO_Ok ) {
      _error = i18n("Could not write to temporary file.");
      MessageBox::detailedSorry(_error, i18n("File: %1").arg(_tmp->name()), _show);
      return false;
    }
  }
  return true;
}
