/***************************************************************************
 *   Copyright (C) 2005-2006 Nicolas Hadacek <hadacek@kde.org>             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#ifndef PURL_H
#define PURL_H

#include <qfile.h>
#include "common/qflags.h"
#include <qpixmap.h>

#include <kurl.h>
class KTempFile;
class PDirectory;

#include "common/log.h"

//----------------------------------------------------------------------------
class PURLBase
{
public:
  PURLBase(const QString &filepath = QString::null);
  PURLBase(const KURL &url);
  bool operator <(const PURLBase &url) const { return _url<url._url; }
  bool operator ==(const PURLBase &url) const { return _url==url._url; }
  bool operator !=(const PURLBase &url) const { return _url!=url._url; }
  const KURL &kurl() const { return _url; }
  QString pretty() const { return _url.prettyURL(0, KURL::StripFileProtocol); }
  bool isEmpty() const { return _empty; }
  bool isLocal() const { return _url.isLocalFile(); }
  QString path() const; // with ending '/' unless empty path
  PDirectory directory() const;
  bool isInto(const PDirectory &dir) const;
  bool isRelative() const { return _relative; }

protected:
  bool    _relative, _empty;
  KURL    _url;
  mutable QString _error;
};

//----------------------------------------------------------------------------
class PURL : public PURLBase
{
public:
  enum FileType { AsmGPAsm = 0, AsmPIC30, AsmPICC, Inc, CSource, CHeader,
                  Object, Library, Hex, Elf, Project,
                  Lst, Map, Cod, Coff, Lkr, Gld, Rst, Sym, Mem, Adb, Lnk, SDCC_P, SDCC_D,
                  Sdb, Hxl, Rlf,
                  Unknown, PikdevProject, Nb_FileTypes };
  enum SourceType { Asm = 0, C, Nb_SourceTypes };
  enum FileProperty { NoProperty = 0, Source = 1, Header = 2, Editable = 4, ReadOnly = 8 };
  Q_DECLARE_FLAGS(FileProperties, FileProperty)
  struct Data {
    FileProperties properties;
    SourceType sourceType;
    const char *label;
    const char *extensions[10];
    const char **xpm_icon;
    const char *mimetype;
    const char *highlightModeName;
  };
  static const Data DATA[Nb_FileTypes];
  static const char * const SOURCE_TYPE_LABELS[Nb_SourceTypes];
  static bool hasMimetype(FileType type);
  static QPixmap icon(FileType type);

public:
  PURL(const QString &filePath = QString::null);
  PURL(const KURL &url) : PURLBase(url) {}
  // add correct extension if filename has no extension
  PURL(const PDirectory &path, const QString &filename, FileType type);
  PURL(const PDirectory &path, const QString &filepath);

  enum ExtensionOption { NoExtensionOption = 0, AddExtension = 1,
                         UpperCaseExtension = 2 };
  Q_DECLARE_FLAGS(ExtensionOptions, ExtensionOption)
  PURL toFileType(FileType type, ExtensionOptions options = NoExtensionOption) const;
  PURL toExtension(const QString &extension, ExtensionOptions options = NoExtensionOption) const;
  const Data &data() const { return DATA[fileType()]; }
  FileType fileType() const;
  QString basename() const; // filename without extension
  QString filename() const; // filename without path
  QString filepath() const; // filename with path
  bool isDestinationUpToDate(FileType type) const;
  QString relativeTo(const PDirectory &dir) const;
  bool probablyExists() const;
  bool create(Log::ShowMode show) const; // do not overwrite
  bool isDosFile() const;
  bool copyTo(const PURL &destination, Log::ShowMode show) const; // do not overwrite
  bool write(const QString &text, Log::ShowMode show) const;
  bool del(Log::ShowMode show) const;

  static QString extension(FileType type) { return DATA[type].extensions[0]; }
  static QString extensions(FileType type);
  static QString filter(FileType type);
  static QString sourceFilter();
  static QString objectFilter();
  static QString projectFilter();
  static QString extensions(FileProperty property);

  static PURL getOpen(const QString &startDir, const QString &filter, QWidget *widget,
                      const QString &caption);
  enum SaveAction { NoSaveAction, AskOverwrite, CancelIfExists };
  static PURL getSave(const QString &startDir, const QString &filter, QWidget *widget,
                      const QString &caption, SaveAction action);
  // add correct extension if filename has no extension
  static QString addExtension(const QString &filename, FileType type);

  class List : public QValueList<PURL> {
  public:
    List() {}
    List(const PURL &url) { append(url); }
    List(const QValueList<PURL> &list) : QValueList<PURL>(list) {}
  };
};
Q_DECLARE_OPERATORS_FOR_FLAGS(PURL::FileProperties)
Q_DECLARE_OPERATORS_FOR_FLAGS(PURL::ExtensionOptions)

//----------------------------------------------------------------------------
class PDirectory : public PURLBase
{
public:
  PDirectory(const QString &path = QString::null);
  bool create(Log::ShowMode show) const;
  QString expandMacros(const QString &s) const;

  static PDirectory getExisting(const QString &startDir, QWidget *widget,
                                const QString &caption);
};

//-----------------------------------------------------------------------------
class PFileBase
{
public:
  PFileBase(Log::ShowMode show) : _tmp(0), _stream(0), _show(show) {}
  virtual ~PFileBase();
  virtual bool close() = 0;
  virtual QFile *qfile() = 0;
  virtual const QFile *qfile() const = 0;
  virtual PURL url() const = 0;
  QTextStream &stream();
  QString readText() { return stream().read(); }
  QString readLine() { return stream().readLine(); }
  QStringList readLines();
  void appendText(const QString &text) { stream() << text; }
  void flush();
  bool hasError() const;
  QString errorString() const;
  bool openForWrite();

protected:
  KTempFile    *_tmp;
  QTextStream  *_stream;
  QString       _error;
  Log::ShowMode _show;
};

class PFile : public PFileBase
{
public:
  PFile(const PURL &url, Log::ShowMode show);
  virtual ~PFile() { close(); }
  void setUrl(const PURL &url); // close file too
  virtual PURL url() const { return _url; }
  bool openForRead();
  virtual bool close();
  virtual QFile *qfile();
  virtual const QFile *qfile() const;

private:
  PURL  _url;
  QFile _file;
};

//-----------------------------------------------------------------------------
class PTempFile : public PFileBase
{
public:
  PTempFile(Log::ShowMode show) : PFileBase(show) {}
  virtual QFile *qfile();
  virtual const QFile *qfile() const;
  virtual PURL url() const;
  virtual bool close();
};

#endif
