/***************************************************************************
 *   Copyright (C) 2005 Nicolas Hadacek <hadacek@kde.org>                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "log.h"

#include <qpopupmenu.h>
#include <qeventloop.h>
#include <qapplication.h>

#include <klocale.h>

//-----------------------------------------------------------------------------
const Log::LogData Log::TYPE_DATA[Nb_Types] = {
  { 0, "red"    }, // error
  { 0, "red"    }, // soft error
  { 0, "orange" }, // warning
  { 0, "black"  }, // data
  { 0, "blue"   } // info
};

const Log::LogData Log::DEBUG_DATA[Nb_DebugLevels+1] = {
  { I18N_NOOP("Normal debug messages"), "darkGreen" },
  { I18N_NOOP("Extra debug messages"),  "darkGreen" },
  { I18N_NOOP("All debug messages"),    "darkGreen" },
  { I18N_NOOP("No debug message"),      0           }
};

//-----------------------------------------------------------------------------
Log::View::View()
  : _debugLevel(Nb_DebugLevels)
{
 for (uint i=0; i<Nb_Types; i++) _modes[i] = Show;
}

void Log::View::setQuiet()
{
  _debugLevel = Nb_DebugLevels;
  for (uint i=0; i<Nb_Types; i++) _modes[i] = DontShow;
}

void Log::View::log(Type type, const QString &text, Action action)
{
  if ( _modes[type]==DontShow || text.isEmpty() ) return;
  doLog(type, text, action);
}

void Log::View::log(DebugLevel level, const QString &text, Action action)
{
  Q_ASSERT( level!=Nb_DebugLevels );
  updateDebugLevel();
  if ( _debugLevel==Nb_DebugLevels || level>_debugLevel || text.isEmpty() ) return;
  doLog(level, text, action);
}

//-----------------------------------------------------------------------------
void Log::ConsoleView::doLog(Type type, const QString &text, Action)
{
  QString s = text + "\n";
  if ( type==Error || type==SoftError ) fprintf(stderr, "%s", s.latin1());
  else fprintf(stdout, "%s", s.latin1());
}

void Log::ConsoleView::doLog(DebugLevel, const QString &text, Action)
{
  QString s = text + "\n";
  fprintf(stdout, "%s", s.latin1());
}

void Log::ConsoleView::appendToLastLine(const QString &text)
{
  fprintf(stdout, "%s", text.latin1());
}

void Log::ConsoleView::detailedSorry(const QString &message, const QString &details)
{
  log(Error, message + ": " + details, Immediate);
}

bool Log::ConsoleView::askContinue(const QString &message)
{
  log(Warning, message, Immediate);
  return false; // always fail
}

//-----------------------------------------------------------------------------
void Log::Base::logUserAbort()
{
  log(Log::SoftError, i18n("Operation aborted by user."));
}
void Log::Base::log(Type type, const QString &message, Action action) { manager().log(type, message, action); }
void Log::Base::log(DebugLevel level, const QString &message, Action action) { manager().log(level, message, action); }
void Log::Base::appendToLastLine(const QString &text) { manager().appendToLastLine(text); }
bool Log::Base::hasError() const { return _parent->manager().hasError(); }
QString Log::Base::error() const { return _parent->manager().error(); }
void Log::Base::resetError() { manager().resetError(); }
void Log::Base::addProgress(uint steps) { manager().addProgress(steps); }
void Log::Base::detailedSorry(const QString &message, const QString &details) { manager().detailedSorry(message, details); }
bool Log::Base::askContinue(const QString &message) { return manager().askContinue(message); }

//-----------------------------------------------------------------------------
Log::Manager::Manager(const char *name)
  : Base(0), _view(0)
{
  _proxy = new ManagerProxy(name);
}

Log::Manager::~Manager()
{
  delete _proxy;
}

void Log::Manager::log(Type type, const QString &message, Action action)
{
  if ( type==Error ) _error = message;
  if (_view) _view->log(type, message, action);
}

void Log::Manager::log(DebugLevel level, const QString &message, Action action)
{
  if (_view) _view->log(level, message, action);
}

void Log::Manager::detailedSorry(const QString &message, const QString &details)
{
  if (_view) _view->detailedSorry(message, details);
}

bool Log::Manager::askContinue(const QString &message)
{
  if ( _view==0 ) return false;
  return _view->askContinue(message);
}
