/***************************************************************************
 *   Copyright (C) 2005 Nicolas Hadacek <hadacek@kde.org>                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#ifndef LOG_H
#define LOG_H

#include <qobject.h>

namespace Log
{
  // immediat visibily by calling processEvent... BEWARE of side effects
  enum Action { Immediate, Delayed };
  enum ShowMode { DontShow, Show };
  enum DebugLevel { NormalDebug = 0, ExtraDebug, MaxDebug, Nb_DebugLevels };
  enum Type { Error = 0, SoftError, Warning, Data, Information, Nb_Types };
  struct LogData {
    const char *label, *color;
  };
  extern const LogData TYPE_DATA[Nb_Types];
  extern const LogData DEBUG_DATA[Nb_DebugLevels+1];

//-----------------------------------------------------------------------------
class View;
class Manager;

class Generic
{
public:
  Generic() {}
  virtual ~Generic() {}
  virtual void log(Type type, const QString &text, Action action = Immediate) = 0;
  virtual void log(DebugLevel level, const QString &text, Action action = Immediate) = 0;
  virtual void appendToLastLine(const QString &text) = 0;
  virtual void detailedSorry(const QString &message, const QString &details) = 0;
  virtual bool askContinue(const QString &message) = 0;
};

class View : public Generic
{
public:
  View();
  ShowMode showMode(Type type) const { return _modes[type]; }
  void setShowMode(Type type, ShowMode mode) { _modes[type] = mode; }
  void setDebugLevel(DebugLevel level) { _debugLevel = level; }
  void setQuiet();

  virtual void log(Type type, const QString &text, Action action = Immediate);
  virtual void log(DebugLevel level, const QString &text, Action action = Immediate);
  virtual void appendToLastLine(const QString &text) = 0;
  virtual void clear() = 0;
  virtual void detailedSorry(const QString &message, const QString &details) = 0;
  virtual bool askContinue(const QString &message) = 0;

protected:
  ShowMode _modes[Nb_Types];
  DebugLevel _debugLevel;

  virtual void updateDebugLevel() {}
  virtual void doLog(Type type, const QString &text, Action action) = 0;
  virtual void doLog(DebugLevel level, const QString &text, Action action) = 0;
};

//-----------------------------------------------------------------------------
class ConsoleView : public View
{
public:
  ConsoleView() {}
  virtual void appendToLastLine(const QString &text);
  virtual void clear() {}
  virtual void detailedSorry(const QString &message, const QString &details);
  virtual bool askContinue(const QString &message);

private:
  virtual void doLog(Type type, const QString &text, Action action);
  virtual void doLog(DebugLevel level, const QString &text, Action action);
};

//-----------------------------------------------------------------------------
class Base : public Generic
{
public:
  Base(Base *parent) : _parent(parent) {}
  virtual void log(Type type, const QString &message, Action action = Immediate);
  virtual void log(DebugLevel level, const QString &message, Action action = Immediate);
  virtual void appendToLastLine(const QString &text);
  virtual bool hasError() const;
  virtual QString error() const;
  virtual void resetError();
  virtual void addProgress(uint steps);
  virtual void detailedSorry(const QString &message, const QString &details);
  virtual bool askContinue(const QString &message);
  void logUserAbort();

protected:
  Base *_parent;
  virtual Manager &manager() { return _parent->manager(); }
};

class ManagerProxy : public QObject
{
Q_OBJECT
public:
  ManagerProxy(const char *name) : QObject(0, name) {}

signals:
  void addProgress(uint step);
  void setTotalSteps(uint total);
  void addToTotalSteps(uint add);

private:
  friend class Manager;
};

class Manager : public Base
{
public:
  Manager(const char *name = 0);
  virtual Manager &manager() { return *this; }
  virtual ~Manager();
  void setView(View *view) { _view = view; }
  View *view() { return _view; }
  ManagerProxy *proxy() { return _proxy; }

  virtual void log(Type type, const QString &message, Action action = Immediate);
  virtual void log(DebugLevel level, const QString &message, Action action = Immediate);
  virtual void appendToLastLine(const QString &text) { if (_view) _view->appendToLastLine(text); }
  virtual bool hasError() const { return !_error.isNull(); }
  virtual QString error() const { return _error; }
  virtual void resetError() { _error = QString::null; }
  virtual void addProgress(uint steps) { emit _proxy->addProgress(steps); }
  void setTotalSteps(uint total) { emit _proxy->setTotalSteps(total); }
  virtual void detailedSorry(const QString &message, const QString &details);
  virtual bool askContinue(const QString &message);

protected:
  QString _error;

private:
  ManagerProxy *_proxy;
  View         *_view;
};

class Client : public Base
{
public:
  Client(Base &base) : Base(&base) {}
};

} // namespace

#endif
