/***************************************************************************
 *   Copyright (C) 2005-2006 Nicolas Hadacek <hadacek@kde.org>             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "misc.h"

#include <ctype.h>
#include <unistd.h>
#include <qfontmetrics.h>
#include <qregexp.h>

//-----------------------------------------------------------------------------
QCString repeat(const char *c, uint nb)
{
  QCString s;
  for (uint i=0; i<nb; i++) s += c;
  return s;
}

uint nbBits(uint value)
{
  uint nb = 1;
  for (;;) {
    value /= 2;
    if ( value==0 ) break;
    nb++;
  }
  return nb;
}

uint nbChars(uint value) { return nbBits(value)/4 + (nbBits(value)%4 ? 1 : 0); }
uint nbBytes(uint value) { return nbBits(value)/8 + (nbBits(value)%8 ? 1 : 0); }

QCString toHex(uint value, uint nbChars)
{
  Q_ASSERT( nbChars>=1 && nbChars<=9 );
  // uint is 32 bits and apparently the compiler refuses to shift if 4*nbChars>=32...
  Q_ASSERT( nbChars>=8 || (value >> (4*nbChars))==0 );
  if ( nbChars<8 && (value >> (4*nbChars))!=0 ) qDebug("toHex %s %i", toHexLabel(value, 8).latin1(), nbChars);
  QCString format;
  format.setNum(nbChars);
  QCString s(nbChars);
  s.sprintf("%0" + format + "X", value);
  return s;
}

QString toHexLabel(uint value, uint nbChars)
{
  return "0x" + toHex(value, nbChars);
}

QString toHexLabels(const QMemArray<uint> &values, uint nbChars)
{
  QString s = "]";
  for (uint i=0; i<values.count(); i++) {
    if ( i!=0 ) s += ' ';
    s += toHexLabel(values[i], nbChars);
  }
  s += "]";
  return s;
}

uint fromHex(char c, bool *ok)
{
  c = toupper(c);
  if (ok) *ok = isxdigit(c);
  return (c>='A' ? 10+c-'A' : c-'0');
}

uint fromHex(const QString &s, bool *ok)
{
  if ( s.length()==0 ) {
    if (ok) *ok = false;
    return 0;
  }
  ulong v = 0;
  for (uint i=0; i<s.length(); i++) {
    v *= 16;
    v += fromHex(char(s[i]), ok);
    if ( ok && !*ok ) break;
  }
  return v;
}

uint fromHexLabel(const QString &s, bool *ok)
{
  if ( s.length()<3 || s[0]!='0' || s[1]!='x' ) {
    if (ok) *ok = false;
    return 0;
  }
  return fromHex(s.mid(2), ok);
}

uint fromHexLabel(const QString &s, uint nbChars, bool *ok)
{
  if ( s.length()!=nbChars+2 ) {
    if (ok) *ok = false;
    return 0;
  }
  return fromHexLabel(s, ok);
}

uchar bin2bcd(uchar bin)
{
  char h = bin / 10;
  char l = bin % 10;
  return (h*16) + l;
}

uchar bcd2bin(uchar bcd)
{
  char h = bcd / 16;
  char l = bcd % 16;
  return (h*10) + l;
}

int maxCharWidth(const QFont &font, const QString &s)
{
  QFontMetrics fm(font);
  int w = 0;
  for (uint i=0; i<s.length(); i++)
    w = QMAX(w, fm.width(s[i]));
  return w;
}

int maxHexCharWidth(const QFont &font)
{
  return maxCharWidth(font, "0123456789ABCDEF");
}

bool XOR(bool b1, bool b2)
{
  return ( (!b1 && b2) || (b1 && !b2) );
}

QString prettyCString(const QCString &cs)
{
  QString s;
  for (uint i=0; i<cs.length(); i++) {
    if ( isprint(cs[i]) ) s += cs[i];
    else s += "\\" + toHex(uchar(cs[i]), 2);
  }
  return s;
}

QString escapeXml(const QString &cs)
{
  QString s;
  for (uint i=0; i<cs.length(); i++) {
    if ( cs[i]=='<' ) s += "&lt;";
    else if ( cs[i]=='>' ) s += "&gt;";
    else s += cs[i];
  }
  return s;
}

QString htmlTableRow(const QString &title, const QString &value)
{
  return "<tr><td>" + title + "</td><td>" + value + "</td></tr>";
}

//-----------------------------------------------------------------------------
VersionData VersionData::fromString(const QString &s)
{
  VersionData vd;
  QRegExp re("([0-9]+)\\.([0-9]+)\\.([0-9]+)(.*)");
  if ( !re.exactMatch(s) ) return vd;
  vd._major = re.cap(1).toUInt();
  vd._mid = re.cap(2).toUInt();
  vd._minor = re.cap(3).toUInt();
  vd._sub = re.cap(4);
  return vd;
}

VersionData VersionData::fromHexString(const QString &s)
{
  VersionData vd;
  Q_ASSERT(s.length()==6);
  vd._major = fromHex(s.mid(0, 2));
  vd._mid =   fromHex(s.mid(2, 2));
  vd._minor = fromHex(s.mid(4, 2));
  return vd;
}

QString VersionData::pretty() const
{
  if ( !isValid() ) return "---";
  return QString::number(_major) + '.' + QString::number(_mid) + '.'
         + QString::number(_minor) + _sub;
}
