/***************************************************************************
 *   Copyright (C) 2005-2006 Nicolas Hadacek <hadacek@kde.org>             *
 *   Copyright (C) 2004 Alain Gibaud <alain.gibaud@free.fr>                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "config_center.h"

#include <qlabel.h>
#include <qlayout.h>
#include <qtooltip.h>
#include <qgroupbox.h>
#include <qtabwidget.h>
#include <kiconloader.h>
#include <klocale.h>

#include "global_config.h"
#include "device_gui.h"
#include "tools/tool.h"
#include "tools/tool_config.h"
#include "prog_config_widget.h"
#include "tools/misc_gui.h"

//----------------------------------------------------------------------------
GeneralConfigWidget::GeneralConfigWidget()
  : GridConfigWidget(0, "general_config_widget")
{
  _programAfterBuild = new QCheckBox(i18n("Program device after successful build."), this);
  _topLayout->addMultiCellWidget(_programAfterBuild, 0,0, 0,1);

  _showDebug = new QComboBox(this);
  _showDebug->insertItem(i18n(Log::DEBUG_DATA[Log::Nb_DebugLevels].label));
  for (uint i=0; i<Log::Nb_DebugLevels; i++)
    _showDebug->insertItem(i18n(Log::DEBUG_DATA[i].label));
  _topLayout->addWidget(_showDebug, 1, 0);

  _userIdSetToChecksum = new QCheckBox(i18n("Set User Ids to unprotected checksum (if User Ids are empty)."), this);
  _topLayout->addMultiCellWidget(_userIdSetToChecksum, 2,2, 0,1);
}

void GeneralConfigWidget::loadConfig()
{
  _programAfterBuild->setChecked(GlobalConfig::programAfterBuild());
  Log::DebugLevel level = GlobalConfig::debugLevel();
  if ( level!=Log::Nb_DebugLevels ) _showDebug->setCurrentItem(level+1);
  _userIdSetToChecksum->setChecked(GlobalConfig::isUserIdSetToChecksum());
}

void GeneralConfigWidget::saveConfig()
{
  GlobalConfig::writeProgramAfterBuild(_programAfterBuild->isChecked());
  uint i = _showDebug->currentItem();
  GlobalConfig::writeDebugLevel(i==0 ? Log::Nb_DebugLevels : Log::DebugLevel(i-1));
  GlobalConfig::writeUserIdSetToChecksum(_userIdSetToChecksum->isChecked());
}

QPixmap GeneralConfigWidget::pixmap() const
{
  KIconLoader loader;
  return loader.loadIcon("configure", KIcon::Toolbar, KIcon::SizeMedium);
}

//----------------------------------------------------------------------------
StandaloneConfigWidget::StandaloneConfigWidget()
  : ConfigWidget(0, "standalone_config_widget")
{
  QGridLayout *top = new QGridLayout(this, 1, 1, 10, 10);

  QLabel *label = new QLabel(i18n("Device:"), this);
  top->addWidget(label, 0, 0);
  _device = new DeviceChooser::Button(true, this);
  top->addWidget(_device, 0, 1);

  _tools = new ToolsConfigWidget(0, this);
  top->addMultiCellWidget(_tools, 1,1, 0,2);

  top->setRowStretch(1, 1);
  top->setColStretch(2, 1);
}

void StandaloneConfigWidget::loadConfig()
{
  _device->setDevice(Compile::Config::device(0));
  _tools->loadConfig();
}

void StandaloneConfigWidget::saveConfig()
{
  Compile::Config::setDevice(0, _device->device());
  _tools->saveConfig();
}

QPixmap StandaloneConfigWidget::pixmap() const
{
  KIconLoader loader;
  return loader.loadIcon("configure", KIcon::Toolbar, KIcon::SizeMedium);
}

//----------------------------------------------------------------------------
ToolchainConfigWidget::ToolchainConfigWidget(const Tool::Group &group, QWidget *parent)
  : GridConfigWidget(parent, "install_tool_config_widget"), _group(group)
{
  QLabel *label = new QLabel(i18n("Path:"), this);
  _topLayout->addWidget(label, 0, 0);
  _dir = new SelectDirectoryWidget(PDirectory(), this);
  connect(_dir, SIGNAL(changed()), SLOT(changed()));
  _topLayout->addMultiCellWidget(_dir, 0,0, 1,2);

  _withWine = new QCheckBox(i18n("Use with \"wine\""), this);
  connect(_withWine, SIGNAL(clicked()), SLOT(changed()));
  _topLayout->addMultiCellWidget(_withWine, 1,1, 0,2);

  uint row = 2;
  for (uint k=0; k<Tool::Nb_Categories; k++) {
    const Tool::Base *base = group.base(Tool::Category(k));
    if ( base==0 ) _founds[k] = 0;
    else {
      label = new QLabel(QString(Tool::CATEGORY_DATA[k].label) + ":", this);
      _topLayout->addWidget(label, row, 0);
      _founds[k] = new QLabel(this);
      _topLayout->addMultiCellWidget(_founds[k], row,row, 1,2);
      row++;
    }
  }
  _topLayout->setColStretch(2, 1);
  _topLayout->setRowStretch(row, 1);
}

void ToolchainConfigWidget::loadConfig()
{
  Compile::Config *config = _group.createConfig(0);
  _withWine->setChecked(config->withWine());
  _dir->setDirectories(config->executableDirectory().path());
  delete config;
  changed();
}

void ToolchainConfigWidget::saveConfig()
{
  Compile::Config *config = _group.createConfig(0);
  config->setWithWine(_withWine->isChecked());
  config->setExecutableDirectory(_dir->directories());
  delete config;
}

void ToolchainConfigWidget::changed()
{
  for (uint k=0; k<Tool::Nb_Categories; k++) {
    if ( _founds[k]==0 ) continue;
    const Tool::Base *base = _group.base(Tool::Category(k));
    QString exec = PDirectory(_dir->directories()).path() + base->baseExecutable();
    bool ok = base->checkExecutable(exec, _withWine->isChecked());
    QString s = (ok ? i18n("\"%2\" found") : i18n("\"%2\" not found")).arg(base->baseExecutable());
    _founds[k]->setText(s);
  }
}

//----------------------------------------------------------------------------
ToolchainsConfigWidget::ToolchainsConfigWidget()
  : GridConfigWidget(0, "toolchains_config_widget")
{
  QLabel *label = new QLabel(i18n("Toolchain:"), this);
  _topLayout->addWidget(label, 0, 0);
  _combo = new QComboBox(this);
  connect(_combo, SIGNAL(activated(int)), SLOT(changed(int)));
  _topLayout->addWidget(_combo, 0, 1);

  Container *container = new Container(this, Container::Sunken);
  _topLayout->addMultiCellWidget(container, 1,1, 0,2);
  _stack = new QWidgetStack(container);
  container->addWidget(_stack, 0, 0);

  _topLayout->setRowStretch(2, 1);
  _topLayout->setColStretch(2, 1);

  for (uint i=0; i<Tool::lister().nbGroups(); i++) {
    const Tool::Group &group = Tool::lister().group(i);
    _combo->insertItem(group.label());
    _stack->addWidget(new ToolchainConfigWidget(group, _stack), i);
  }
}

void ToolchainsConfigWidget::loadConfig()
{
  for (uint i=0; i<Tool::lister().nbGroups(); i++)
    static_cast<ToolchainConfigWidget *>(_stack->widget(i))->loadConfig();
}

void ToolchainsConfigWidget::saveConfig()
{
  for (uint i=0; i<Tool::lister().nbGroups(); i++)
    static_cast<ToolchainConfigWidget *>(_stack->widget(i))->saveConfig();
}

QPixmap ToolchainsConfigWidget::pixmap() const
{
  KIconLoader loader;
  return loader.loadIcon("configure", KIcon::Toolbar, KIcon::SizeMedium);
}

void ToolchainsConfigWidget::changed(int i)
{
  _stack->raiseWidget(i);
}

//----------------------------------------------------------------------------
ConfigWidget *ConfigCenter::factory(Type type)
{
  switch (type) {
    case General:     return new GeneralConfigWidget;
    case ProgSelect:  return new Programmer::SelectConfigWidget;
    case ProgOptions: return new Programmer::OptionsConfigWidget;
    case Standalone:  return new StandaloneConfigWidget;
    case Tools:       return new ToolchainsConfigWidget;
    case Nb_Types:    break;
  }
  Q_ASSERT(false);
  return 0;
}

ConfigCenter::ConfigCenter(Type showType, QWidget *parent)
  : Dialog(QSize(), IconList, i18n("Configure Piklab"), Ok|Cancel, Cancel,
           parent, "configure_dialog", true)
{
  for (uint i=0; i<Nb_Types; i++) {
    _configWidgets[i] = factory(Type(i));
    _configWidgets[i]->loadConfig();
    QFrame *page = addPage(_configWidgets[i]->title(), _configWidgets[i]->header(), _configWidgets[i]->pixmap());
    QVBoxLayout *vbox = new QVBoxLayout(page);
    _configWidgets[i]->reparent(page, QPoint(0,0));
    vbox->addWidget(_configWidgets[i]);
  }

  showPage(showType);
}

void ConfigCenter::slotApply()
{
  for (uint i=0; i<Nb_Types; i++)
    _configWidgets[i]->saveConfig();
}

void ConfigCenter::slotOk()
{
  slotApply();
  slotClose();
}
