/***************************************************************************
 *   Copyright (C) 2006 Nicolas Hadacek <hadacek@kde.org>                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "config_gen.h"

#include <qlayout.h>
#include <qlabel.h>
#include <qcombobox.h>
#include <klocale.h>

#include "device_gui.h"
#include "devices/base/generic_memory.h"
#include "devices/base/gui/hex_view.h"
#include "devices/base/gui/memory_editor.h"
#include "devices/base/device_group.h"
#include "devices/base/gui/device_group_ui.h"
#include "devices/base/device_list.h"
#include "text_editor.h"

ConfigGenerator::ConfigGenerator(const QString &device, uint group, PURL::SourceType stype, QWidget *parent)
  : Dialog(QSize(400, 300), parent, "config_generator", true, i18n("Config Generator"), Close, Close),
    _memory(0), _configEditor(0)
{
  QVBoxLayout *top = new QVBoxLayout(mainWidget(), 10, 10);

  QHBoxLayout *hbox = new QHBoxLayout(top);
  QLabel *label = new QLabel(i18n("Device:"), mainWidget());
  hbox->addWidget(label);
  _deviceChooser = new DeviceChooser::Button(false, mainWidget());
  _deviceChooser->setDevice(device);
  connect(_deviceChooser, SIGNAL(changed()), SLOT(set()));
  hbox->addWidget(_deviceChooser);
  hbox->addSpacing(20);

  label = new QLabel(i18n("Config Type:"), mainWidget());
  hbox->addWidget(label);
  _configType = new QComboBox(mainWidget());
  connect(_configType, SIGNAL(activated(int)), SLOT(typeChanged()));
  for (uint i=0; i<Tool::lister().nbGroups(); i++)
    _configType->insertItem(Tool::lister().group(i).label());
  hbox->addWidget(_configType);

  label = new QLabel(i18n("Source Type:"), mainWidget());
  hbox->addWidget(label);
  _sourceType = new QComboBox(mainWidget());
  for (uint i=0; i<PURL::Nb_SourceTypes; i++)
    _sourceType->insertItem(i18n(PURL::SOURCE_TYPE_LABELS[i]));
  _sourceType->setMinimumWidth(_sourceType->sizeHint().width());
  connect(_sourceType, SIGNAL(activated(int)), SLOT(computeConfig()));
  hbox->addWidget(_sourceType);
  hbox->addStretch(1);

  _hbox = new QHBoxLayout(top);

  _text = new SimpleTextEditor(Editor::NoCapability, PURL::Nb_FileTypes, mainWidget());
  _text->setReadOnly(true);
  top->addWidget(_text);

  hbox = new QHBoxLayout(top);
  QPushButton *button = new QPushButton(i18n("Copy to clipboard"), mainWidget());
  connect(button, SIGNAL(clicked()), SLOT(copyToClipboard()));
  hbox->addWidget(button);
  hbox->addSpacing(20);
  _warning = new QLabel(mainWidget());
  hbox->addWidget(_warning);
  hbox->addStretch(1);

  set(device, group, stype);
}

ConfigGenerator::~ConfigGenerator()
{
  delete _memory;
}

void ConfigGenerator::set(const QString &device, uint group, PURL::SourceType stype)
{
  _deviceChooser->setDevice(device);
  _configType->setCurrentItem(group);
  setSourceType(stype);
  set();
}

void ConfigGenerator::typeChanged()
{
  setSourceType(PURL::Nb_SourceTypes);
  computeConfig();
}

void ConfigGenerator::setSourceType(PURL::SourceType stype)
{
  uint group = _configType->currentItem();
  _sourceType->clear();
  uint k = 0;
  for (uint i=0; i<PURL::Nb_SourceTypes; i++) {
    if ( !Tool::lister().group(group).supports(PURL::SourceType(i)) ) continue;
    _sourceType->insertItem(i18n(PURL::SOURCE_TYPE_LABELS[i]));
    if ( i==uint(stype) ) _sourceType->setCurrentItem(k);
    k++;
  }
}

void ConfigGenerator::set()
{
  delete _memory;
  const Device::Data *data = Device::lister().data(_deviceChooser->device());
  _memory = data->group().createMemory(*data);
  delete _configEditor;
  _configEditor = data->groupui().createConfigEditor(*_memory, mainWidget());
  if (_configEditor) {
    _configEditor->show();
    connect(_configEditor, SIGNAL(modified()), SLOT(computeConfig()));
    _configEditor->updateDisplay();
    _hbox->addWidget(_configEditor);
  }
  computeConfig();
}

void ConfigGenerator::computeConfig()
{
  const Tool::Group &group = Tool::lister().group(_configType->currentItem());
  PURL::SourceType stype = PURL::Nb_SourceTypes;
  uint k = 0;
  for (uint i=0; i<PURL::Nb_SourceTypes; i++) {
    if ( !group.supports(PURL::SourceType(i)) ) continue;
    if ( k==uint(_sourceType->currentItem()) ) stype = PURL::SourceType(i);
    k++;
  }
  _warning->setText(QString::null);
  if ( group.isSupported(_memory->device().name()) ) {
    _text->setFileType(group.sourceFileType(stype));
    bool ok;
    _text->setText(group.generateDeviceConfig(stype, *_memory, ok));
    if ( !ok ) _warning->setText(i18n("Config generation is only partially supported for this device."));
  } else {
   _text->setFileType(PURL::Nb_FileTypes);
   _text->setText(i18n("Device not supported by the selected toolchain."));
  }
}

void ConfigGenerator::copyToClipboard()
{
  _text->selectAll();
  _text->copy();
  _text->deselect();
}
