/***************************************************************************
 *   Copyright (C) 2006 Nicolas Hadacek <hadacek@kde.org>                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "debug_manager.h"

#include <qtimer.h>
#include <klocale.h>

#include "object_view.h"
#include "toplevel.h"
#include "devices/base/device_list.h"
#include "text_editor.h"
#include "project.h"
#include "tools/tool_config.h"
#include "register_view.h"
#include "watch_view.h"

DebugManager::DebugManager(QObject *parent)
  :  QObject(parent, "debug_manager"), Log::Manager("debug_manager"),
     _data(0), _coffParser(0)
{
  connect(&_timer, SIGNAL(timeout()), SLOT(slotTimeout()));
  connect(&MainWindow::self(), SIGNAL(stateChanged()), SLOT(updateDevice()));
}

DebugManager::~DebugManager()
{
  delete _coffParser;
}

void DebugManager::addEditor(Editor *editor)
{
  if ( _editors.find(editor)!=_editors.end() ) return;
  _editors.append(editor);
  connect(editor, SIGNAL(destroyed()), SLOT(editorDestroyed()));
  if ( ::qt_cast<Register::View *>(editor) ) {
    connect(editor, SIGNAL(readSignal(uint)), SLOT(readRegister(uint)));
    connect(editor, SIGNAL(writeSignal(uint, uint)), SLOT(writeRegister(uint, uint)));
    connect(editor, SIGNAL(setWatchedSignal(uint, bool)), SLOT(setRegisterWatched(uint, bool)));
    QTimer::singleShot(0, this, SLOT(updateRegisterViews()));
  } else QTimer::singleShot(0, this, SLOT(updateAsmEditors()));
}

void DebugManager::clearEditors()
{
  QValueList<Editor *>::iterator it = _editors.begin();
  for (; it!=_editors.end(); ++it) (*it)->disconnect(this);
  _editors.clear();
}

void DebugManager::editorDestroyed()
{
  QValueList<Editor *>::iterator it = _editors.begin();
  for (; it!=_editors.end(); ++it) {
    if ( (*it)!=sender() ) continue;
    _editors.remove(it);
    break;
  }
}

void DebugManager::updateDevice()
{
  const Device::Data *data = MainWindow::deviceData();
  if ( data==_data ) return;
  _data = data;
  _watchedRegisters.clear();
  Register::View *rv = registerView();
  if (rv) rv->stopWatchAll();
  Watch::View *wv = MainWindow::watchView();
  if (wv) wv->init();
  clear();
}

void DebugManager::init()
{
  const Project *project = MainWindow::projectManager().project();
  delete _coffParser;
  const Tool::Group &group = Tool::lister().group(Compile::Config::toolType(project));
  if ( project==0 || (group.name()!="gputils" && group.name()!="sdcc") ) { // #### FIXME
    _coffParser = 0;
    return;
  }
  _coffParser = new Coff::SectionParser(*_data);
  _coffParser->parse(project->url().toFileType(PURL::Coff));
  computeBreakpoints();
  update();
}

QValueVector<uint> DebugManager::computeBreakpoints()
{
  _breakpoints.clear();
  if ( _coffParser==0 ) return _breakpoints;

  QValueList<Editor *>::const_iterator it = _editors.begin();
  for (; it!=_editors.end(); ++it) {
    KBEditor *e = ::qt_cast<KBEditor *>(*it);
    if ( e==0 ) continue;
    QValueVector<uint> lines = e->breakpoints();
    for (uint i=0; i<lines.count(); i++) {
      QValueVector<uint> addresses = _coffParser->addresses(e->url(), lines[i]);
      if ( addresses.isEmpty() )
        log(Log::Warning, i18n("Invalid breakpoint at line %1 in %2.").arg(lines[i]).arg(e->url().filename()));
      else {
        if ( addresses.count()>1 )
          log(Log::Warning, i18n("Breakpoint corresponds to several addresses. Using first found."));
        _breakpoints.append(addresses[0]);
      }
    }
  }
  return _breakpoints;
}

void DebugManager::clear()
{
  delete _coffParser;
  _coffParser = 0;
  update();
}

bool DebugManager::updateEditorPC(KBEditor *editor) const
{
  if ( PURL::DATA[editor->fileType()].sourceType!=PURL::Asm && editor->fileType()!=PURL::Coff ) return false;
  int line = (_coffParser && _pc!=Device::INVALID ? _coffParser->lineForAddress(editor->url(), _pc) : -1);
  //qDebug("update PC %s: %i", editor->url().pretty().latin1(), line);
  editor->clearMarks(KTextEditor::MarkInterface::Execution);
  if ( line==-1 ) return false;
  editor->setState(Editor::IconBorderOn, true);
  editor->setMark(line, KTextEditor::MarkInterface::Execution);
  editor->setCursor(line, 0);
  return true;
}

bool DebugManager::update()
{
  _timer.stop();
  bool on = false;
  if ( MainWindow::programmer() ) {
    _pc = MainWindow::programmer()->pc();
    on = MainWindow::programmer()->isTargetVddOn();
  } else _pc = Device::INVALID;
  if ( hasError() ) return false;
  static_cast<KToggleAction &>(MainWindow::action("prog_power")).setChecked(on);
  updateAsmEditors();
  if ( !updateRegisterViews() ) return false;
  QString s;
  if ( isActive() ) s = MainWindow::programmer()->statusString();
  emit statusChanged(s);
  return true;
}

void DebugManager::updateAsmEditors()
{
  // update PC display
  bool currentHasPC = false;
  KBEditor *shown = 0;
  QValueList<Editor *>::iterator it = _editors.begin();
  for (; it!=_editors.end(); ++it) {
    KBEditor *e = ::qt_cast<KBEditor *>(*it);
    if ( e==0 ) continue;
    if ( !updateEditorPC(e) ) continue;
    shown = e;
    if ( shown==MainWindow::currentEditor() ) currentHasPC = true;
  }
  if ( shown==0 && _coffParser && _pc!=Device::INVALID ) {
    PURL url = _coffParser->urlForAddress(_pc);
    KBEditor *e = ::qt_cast<KBEditor *>(MainWindow::editorManager().openEditor(url));
    if ( e && updateEditorPC(e) ) {
      shown = e;
      if ( shown==MainWindow::currentEditor() ) currentHasPC = true;
    }
  }
  if ( shown && !currentHasPC && _pc!=Device::INVALID ) MainWindow::editorManager().showEditor(shown);

  // update disassembly breakpoints
  const Project *project = MainWindow::projectManager().project();
  if (project) {
    PURL url = project->url().toFileType(PURL::Coff);
    KBEditor *e = ::qt_cast<KBEditor *>(MainWindow::editorManager().findEditor(url));
    if (e) {
      e->clearBreakpoints();
      if ( _coffParser && !_breakpoints.isEmpty() ) {
        int line = _coffParser->lineForAddress(url, _breakpoints[0]);
        e->setMark(line, KTextEditor::MarkInterface::BreakpointActive);
      }
    }
  }
}

bool DebugManager::updateRegisterViews()
{
  Register::View *rv = registerView();
  if (rv) rv->setActive(isActive());
  Watch::View *wv = MainWindow::watchView();
  for (uint i=0; i<_watchedRegisters.count(); i++) {
    if (rv) rv->setWatched(_watchedRegisters[i], true);
    if (wv) wv->setWatched(_watchedRegisters[i], true);
    if ( isActive() && !readRegister(_watchedRegisters[i]) ) return false;
  }
  return true;
}

void DebugManager::run()
{
  if ( !MainWindow::programmer()->setBreakpoints(computeBreakpoints()) ) return;
  if ( !MainWindow::programmer()->run() ) return;
  if ( !update() ) return;
  _timer.start(300);
}

void DebugManager::halt()
{
  _timer.stop();
  if ( !MainWindow::programmer()->stop() ) return;
  update();
}

void DebugManager::reset()
{
  _timer.stop();
  if ( !MainWindow::programmer()->reset() ) return;
  update();
}

void DebugManager::step()
{
  if ( !MainWindow::programmer()->setBreakpoints(computeBreakpoints()) ) return;
  if ( !MainWindow::programmer()->step() ) return;
  update();
}

void DebugManager::slotTimeout()
{
  _timer.stop();
  if ( !MainWindow::programmer()->updateState() ) return;
  if ( MainWindow::programmer()->state()==Programmer::Running ) {
    _timer.start(300);
    return;
  }
  log(Log::Information, i18n("Reached breakpoint."));
  update();
  emit targetStateChanged();
}

void DebugManager::setRegisterWatched(uint i, bool watched)
{
  if (watched) {
    Q_ASSERT( _watchedRegisters.find(i)==_watchedRegisters.end() );
    _watchedRegisters.append(i);
  } else _watchedRegisters.remove(i);
  Register::View *rv = registerView();
  if (rv) rv->setWatched(i, watched);
  Watch::View *wv = MainWindow::watchView();
  if (wv) wv->setWatched(i, watched);
  if ( watched && isActive() ) readRegister(i);
}

void DebugManager::readAll()
{
  for (uint i=0; i<_data->nbRegisters(); i++)
    readRegister(i);
}

void DebugManager::stopWatchAll()
{
  Register::View *rv = registerView();
  Watch::View *wv = MainWindow::watchView();
  for (uint i=0; i<_watchedRegisters.count(); i++) {
    if (rv) rv->setWatched(i, false);
    if (wv) wv->setWatched(i, false);
  }
  _watchedRegisters.clear();
}

bool DebugManager::readRegister(uint i)
{
  Q_ASSERT( isActive() );
  uint value = MainWindow::programmer()->readRegister(i);
  if ( MainWindow::programmer()->hasError() ) return false;
  Register::View *rv = registerView();
  if (rv) rv->setRegisterValue(i, value);
  Watch::View *wv = MainWindow::watchView();
  if (wv) wv->setRegisterValue(i, value);
  return true;
}

bool DebugManager::writeRegister(uint i, uint value)
{
  Q_ASSERT( isActive() );
  if ( !MainWindow::programmer()->writeRegister(i, value) ) return false;
  return readRegister(i);
}

Register::View *DebugManager::registerView() const
{
  QValueList<Editor *>::const_iterator it = _editors.begin();
  for (; it!=_editors.end(); ++it) {
    Register::View *rv = ::qt_cast<Register::View *>(*it);
    if (rv) return rv;
  }
  return 0;
}

bool DebugManager::isActive() const
{
  return (MainWindow::programmer() ? MainWindow::programmer()->state()==Programmer::Halted : false);
}
