/***************************************************************************
 *   Copyright (C) 2005 Nicolas Hadacek <hadacek@kde.org>                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "pic_group_ui.h"

#include <kglobal.h>
#include <kpopupmenu.h>
#include <kaction.h>

#include "devices/pic/device_data.h"
#include "pic_hex_view.h"
#include "pic_register_view.h"
#include "pic_config_editor.h"

QString Pic::GroupUI::informationUrl(const Device::Data &data) const
{
  QString url = "http://microchip.com/stellent/idcplg?IdcService=SS_GET_PAGE&nodeId=";
  if ( data.document().isEmpty() ) {
    url += "2044&AllWords=";
    if ( data.name().startsWith("PS")) url += data.name();
    if ( data.name().startsWith("30") ) url += "dsPIC" + data.name();
    else url += "PIC" + data.name();
  } else url += "1335&dDocName=en" + data.document();
  return url;
}

QString Pic::GroupUI::informationHtml(const Device::Data &data) const
{
  const Pic::Data &pdata = static_cast<const Pic::Data &>(data);
  QString s = "<table>";
  s += htmlTableRow(i18n("Memory Type"), i18n(Pic::MEMORY_TECHNOLOGY_DATA[pdata.memoryTechnology()].label));
  if ( pdata.isPresent(Pic::MemoryCode) ) {
    uint nbWords = pdata.nbWords(Pic::MemoryCode);
    QString tmp = i18n("%1 words").arg(KGlobal::locale()->formatNumber(nbWords, 0));
    tmp += i18n(" (%2 bits)").arg(pdata.nbBitsWord(Pic::MemoryCode));
    s += htmlTableRow(i18n(Pic::MEMORY_RANGE_TYPE_DATA[Pic::MemoryCode].label), tmp);
  }
  if ( pdata.isPresent(Pic::MemoryEeprom) ) {
    uint nbWords = pdata.nbWords(Pic::MemoryEeprom);
    QString tmp = i18n("%1 bytes").arg(KGlobal::locale()->formatNumber(nbWords, 0));
    tmp += i18n(" (%2 bits)").arg(pdata.nbBitsWord(Pic::MemoryEeprom));
    if ( !pdata.isWritable(Pic::MemoryEeprom) ) tmp += i18n(" (not programmable)");
    s += htmlTableRow(i18n(Pic::MEMORY_RANGE_TYPE_DATA[Pic::MemoryEeprom].label), tmp);
  }
  s += "</table>";
  return s;
}

Device::HexView *Pic::GroupUI::createHexView(const HexEditor &editor, QWidget *parent) const
{
  return new HexView(editor, parent);
}

Device::RegisterView *Pic::GroupUI::createRegisterView(const Device::Data &data, bool debugging, QWidget *parent) const
{
  return new RegisterView(static_cast<const Data &>(data), debugging, parent);
}

Device::MemoryEditor *Pic::GroupUI::createConfigEditor(Device::Memory &memory, QWidget *parent) const
{
  return new MemoryConfigEditorWidget(static_cast<Memory &>(memory), parent);
}

class SData
{
public:
  SData() {}
  SData(const QString &n, uint i) : name(n), index(i) {}
  QString name;
  uint index;
  bool operator <(const SData &sd) const { return name<sd.name; };
};

KPopupMenu *Pic::GroupUI::createWatchPopup(const Device::Data &data, QWidget *parent) const
{
  const Pic::Data &pdata = static_cast<const Pic::Data &>(data);
  KPopupMenu *popup = new KPopupMenu(parent);
  popup->insertTitle(i18n("Watch Register"));
  uint nbRegisters = pdata.nbRegisters();
  uint nbChars = pdata.nbCharsRegisterAddress();
  QPopupMenu *spop = new QPopupMenu(popup);
  QPopupMenu *gpop = new QPopupMenu(popup);
  QValueList<SData> sfrs;
  for (uint i=0; i<nbRegisters; i++) {
    uint address = pdata.nbBytesRegister() * i;
    Pic::RegisterType type = pdata.registerType(address);
    RegisterProperties rp = pdata.registerProperties(address);
    if ( type==UnusedRegister || (rp & NotAccessible) || !(rp & Read) ) continue;
    if ( type==Pic::Sfr ) sfrs.append(SData(pdata.registerName(address), i));
    else if (type==Pic::Gpr ) gpop->insertItem(toHexLabel(address, nbChars), i);
  }
  qHeapSort(sfrs);
  for (uint i=0; i<sfrs.count(); i++) spop->insertItem(sfrs[i].name, sfrs[i].index);
  popup->insertItem(i18n("SFRs"), spop);
  popup->insertItem(i18n("GPRs"), gpop);
  popup->insertItem(i18n("Clear"), parent, SLOT(stopWatchAll()));
  return popup;
}

Device::RegisterItem *Pic::GroupUI::createRegisterItem(const Device::Data &data, uint index, KListViewItem *parent) const
{
  return new Pic::RegisterItem(data, index, parent);
}
