/***************************************************************************
 *   Copyright (C) 2006 Nicolas Hadacek <hadacek@kde.org>                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "pic_register_view.h"

#include <qlayout.h>
#include <qlabel.h>
#include <qpushbutton.h>
#include <qcheckbox.h>
#include <klocale.h>

#include "devices/base/gui/hex_word_editor.h"
#include "common/gui/misc_gui.h"
#include "devices/pic/pic.h"

//-----------------------------------------------------------------------------
Pic::BankWidget::BankWidget(uint i, const Data &data, bool debugging, QWidget *parent)
  : QFrame(parent, "bank_widget"), _data(data), _index(i)
{
  setFrameStyle(WinPanel | Sunken);
  QGridLayout *top = new QGridLayout(this, 1, 1, 5, 0);
  top->setColSpacing(1, 4);
  QFont f("courier", font().pointSize());

  uint rowOffset = 0;
  if ( data.nbBanks()!=0 ) {
    QLabel *label = new QLabel(i18n("Bank %1").arg(i), this);
    label->setAlignment(AlignCenter);
    top->addMultiCellWidget(label, 0,0, 0,2);
    top->setRowSpacing(1, 5);
    rowOffset += 2;
  }

  uint nbRegisters = data.nbRegistersPerBank();
  uint mask = (1 << data.nbBitsRegister()) - 1;
  uint nbChars = data.nbCharsRegisterAddress();
  _buttons.resize(nbRegisters);
  _watch.resize(nbRegisters);
  _words.resize(nbRegisters);
  for (uint k=0; k<nbRegisters; k++) {
    uint address = registerAddress(k);
    QLabel *label = new QLabel(toHexLabel(address, nbChars) + ":", this);
    label->setFont(f);
    top->addWidget(label, rowOffset+k, 0);
    QString name = data.registerName(address);
    if (debugging) {
      _buttons[k] = new QPushButton(name, this);
      connect(_buttons[k], SIGNAL(clicked()), SLOT(read()));
      top->addWidget(_buttons[k], rowOffset+k, 2);
      RegisterProperties rp = data.registerProperties(address);
      bool unused = ( data.registerType(address)==UnusedRegister );
      bool na = ( unused || rp & NotAccessible );
      _buttons[k]->setEnabled(!na && (rp & Read));
      if ( !na && (rp & Read) ) {
        _watch[k] = new QCheckBox(this);
        connect(_watch[k], SIGNAL(clicked()), SLOT(toggleWatch()));
        top->addWidget(_watch[k], rowOffset+k, 4);
      } else _watch[k] = 0;
      if ( !na && (rp & Write) ) {
        _words[k] = new Device::RegisterHexWordEditor(this, data.nbCharsRegister(), mask);
        _words[k]->setEnabled(rp & Read);
        connect(_words[k], SIGNAL(modified()), SLOT(write()));
        _words[k]->setFont(f);
        top->addWidget(_words[k], rowOffset+k, 6);
      } else _words[k] = 0;
    } else {
      QLabel *label = new QLabel(name, this);
      top->addWidget(label, rowOffset+k, 2);
      _buttons[k] = 0;
      _watch[k] = 0;
      _words[k] = 0;
    }
  }

  if (debugging) {
    top->setColSpacing(3, 5);
    top->setColSpacing(5, 5);
  }
}

void Pic::BankWidget::setActive(bool active)
{
  for (uint i=0; i<_words.count(); i++) {
    uint address = registerAddress(i);
    RegisterProperties rp = _data.registerProperties(address);
    bool unused = ( _data.registerType(address)==UnusedRegister );
    bool na = ( unused || rp & NotAccessible );
    if (_buttons[i]) _buttons[i]->setEnabled(!na && active && (rp & Read));
    if (_words[i]) _words[i]->setEnabled(active);
  }
}

void Pic::BankWidget::readAll()
{
  for (uint i=0; i<_words.count(); i++)
    if (_words[i]) read(i);
}

uint Pic::BankWidget::registerIndex(uint i) const
{
  return _index * _data.nbRegistersPerBank() + i;
}

uint Pic::BankWidget::registerAddress(uint i) const
{
  return registerIndex(i) * _data.nbBytesRegister();
}

void Pic::BankWidget::read()
{
  for (uint i=0; i<_buttons.count(); i++) {
    if ( sender()!=_buttons[i] ) continue;
    read(i);
    break;
  }
}

void Pic::BankWidget::write()
{
  for (uint i=0; i<_buttons.count(); i++) {
    if ( sender()!=_words[i] ) continue;
    emit writeSignal(registerIndex(i), _words[i]->value());
    break;
  }
}

void Pic::BankWidget::setRegisterValue(uint i, uint value)
{
  if ( _words[i]->value()!=value ) _words[i]->setColor(red);
  else _words[i]->unsetColor();
  _words[i]->setValue(value);
}

void Pic::BankWidget::stopWatchAll()
{
  for (uint i=0; i<_watch.count(); i++) {
    if ( _watch[i]==0 || !_watch[i]->isChecked() ) continue;
    emit setWatchedSignal(registerIndex(i), false);
  }
}

void Pic::BankWidget::setWatched(uint i, bool watched)
{
  _watch[i]->setChecked(watched);
  _words[i]->unsetColor();
}

void Pic::BankWidget::toggleWatch()
{
  for (uint i=0; i<_watch.count(); i++) {
    if ( sender()!=_watch[i] ) continue;
    emit setWatchedSignal(registerIndex(i), _watch[i]->isChecked());
    break;
  }
}

//-----------------------------------------------------------------------------
Pic::RegisterView::RegisterView(const Pic::Data &data, bool debugging, QWidget *parent)
  : Device::RegisterView(data, debugging, parent, "pic_register_view"), _readAllButton(0), _clearAllButton(0)
{
  QVBoxLayout *vbox = new QVBoxLayout(this, 10, 10);
  QHBoxLayout *hbox = new QHBoxLayout(vbox);

  if (debugging) {
    QWidget *w = new QWidget(this);
    hbox->addWidget(w);
    QGridLayout *grid = new QGridLayout(w, 1, 1, 0, 10);
    _readAllButton = new QPushButton(i18n("Read All"), w);
    connect(_readAllButton, SIGNAL(clicked()), SLOT(readAll()));
    grid->addWidget(_readAllButton, 0, 0);
    QLabel *label = new QLabel(i18n("To read a single register, click on the respective button."), w);
    grid->addWidget(label, 0, 1);
    _clearAllButton = new QPushButton(i18n("Clear all watching"), w);
    connect(_clearAllButton, SIGNAL(clicked()), SLOT(stopWatchAll()));
    grid->addWidget(_clearAllButton, 1, 0);
    label = new QLabel(i18n("To watch a single register, click on the respective checkbox."), w);
    grid->addWidget(label, 1, 1);
    label = new QLabel(i18n("Note: Watching many registers will significantly slow down debugging."), w);
    grid->addMultiCellWidget(label, 2,2, 0,1);
    grid->setColStretch(2, 2);
  }

  hbox = new QHBoxLayout(vbox);
  QHBoxLayout *hbox2 = new QHBoxLayout(hbox);
  hbox->addStretch(1);

  bool is18 = data.architecture()==Pic::Arch_18X;
  uint nb = (data.nbBanks()==0 ? 1 : data.nbBanks());
  _banks.resize(nb);
  for (uint i=0; i<nb; i++) {
    if ( is18 && i!=0 && i!=(nb-1) ) _banks[i] = 0;
    else {
      _banks[i] = new BankWidget(i, data, debugging, this);
      connect(_banks[i], SIGNAL(readSignal(uint)), SIGNAL(readSignal(uint)));
      connect(_banks[i], SIGNAL(writeSignal(uint, uint)), SIGNAL(writeSignal(uint, uint)));
      connect(_banks[i], SIGNAL(setWatchedSignal(uint, bool)), SIGNAL(setWatchedSignal(uint, bool)));
      _banks[i]->show();
      hbox2->addWidget(_banks[i]);
    }
  }
}

void Pic::RegisterView::setActive(bool active)
{
  if (_readAllButton) _readAllButton->setEnabled(active);
  for (uint i=0; i<_banks.count(); i++)
    if (_banks[i]) _banks[i]->setActive(active);
}

void Pic::RegisterView::readAll()
{
  ::BusyCursor bc;
  for (uint i=0; i<_banks.count(); i++)
    if (_banks[i]) _banks[i]->readAll();
}

void Pic::RegisterView::stopWatchAll()
{
  for (uint i=0; i<_banks.count(); i++)
    if (_banks[i]) _banks[i]->stopWatchAll();
}

void Pic::RegisterView::setWatched(uint index, bool watched)
{
  uint nb = static_cast<const Pic::Data &>(_data).nbRegistersPerBank();
  _banks[index/nb]->setWatched(index%nb, watched);
}

void Pic::RegisterView::setRegisterValue(uint index, uint value)
{
  uint nb = static_cast<const Pic::Data &>(_data).nbRegistersPerBank();
  _banks[index/nb]->setRegisterValue(index%nb, value);
}
