/***************************************************************************
 *   Copyright (C) 2005 Nicolas Hadacek <hadacek@kde.org>                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "pic.h"

#include <klocale.h>

#include "common/misc.h"

//------------------------------------------------------------------------------
const Pic::ArchitectureData Pic::ARCHITECTURE_DATA[Nb_Architectures] = {
//  name PC nbBytes packed codeAlignment nbBytesRegister registerBankLength
//  {Code, Cal, UserID, DevId, Conf, EEPROM, DebugVec, HardStack, CalBackup, Program Executive}
  { "10F2XX", 9, 2, false, 1, 8, 0x020,
    { 12, 12, 12, 12, 12,  8, 12,  0, 12,  0 } },
  // 16C5X: 9, 10, 11 or 12-bit program counter
  { "16C5X", 12, 2, false, 1, 8, 0x020,
    { 12, 12, 12, 12, 12,  8, 12,  0, 12,  0 } },
  { "16X",   13, 2, false, 1, 8, 0x080,
    { 14, 14, 14, 14, 14,  8, 14,  0, 14,  0 } }, // max eeprom: 256 words
  { "17X",   21, 2, false, 32, 8, 0x100,
    { 16,  0,  0,  0, 16,  8,  0,  0,  0,  0 } },
  { "18X",   21, 2, true,  32, 8, 0x100,
    { 16,  8,  8,  8,  8,  8, 16,  0,  8,  0 } },
  // dsPIC: eeprom max = 2 kwords = 4 kbytes
  { "30X",   23, 4, false, 4, 16, 0xA00,
    { 24,  0,  0, 16, 16, 16, 24,  0,  0, 24 } } // user id == unitid 24bits?
};

const Pic::ProgVoltageTypeData Pic::PROG_VOLTAGE_TYPE_DATA[Nb_ProgVoltageTypes] = {
  { "vpp"            },
  { "vdd_prog"       },
  { "vdd_prog_write" },
  { "vdd_prog_read"  }
};

const Pic::MemoryTechnologyData Pic::MEMORY_TECHNOLOGY_DATA[Nb_MemoryTechnologies] = {
  { "EEE",     I18N_NOOP("Enhanced Flash") },
  { "EE",      I18N_NOOP("Standard Flash") },
  { "EPROM",   I18N_NOOP("EPROM (OTP)")    },
  { "ROM",     I18N_NOOP("ROM")            },
  { "ROMLESS", I18N_NOOP("ROM-less")       }
};

//-----------------------------------------------------------------------------
Pic::Data::Data()
  : _config(*this)
{
  for (uint i=0; i<Nb_ProgVoltageTypes; i++) {
    _voltages[i].min = 0.0;
    _voltages[i].max = 0.0;
    _voltages[i].nominal = 0.0;
  }
  for (uint i=0; i<Nb_MemoryRangeTypes; i++) {
    _ranges[i].properties = NotPresent;
    _ranges[i].start = 0;
    _ranges[i].end = 0;
    _ranges[i].mask = 0;
    _ranges[i].rmask = 0;
    _ranges[i].hexFileOffset = 0;
  }
  _calibration.opcode = 0;
  _calibration.opcodeMask = 0;
}

Pic::Data::Data(const Device::ConstData &cdata)
  : Device::Data(cdata), _config(*this)
{}

bool Pic::Data::isWritable(MemoryRangeType type) const
{
  return ( MEMORY_RANGE_TYPE_DATA[type].properties & Pic::Writable
           && (range(type).properties & Programmable) );
}

uint Pic::Data::addressIncrement(MemoryRangeType type) const
{
  uint inc = ARCHITECTURE_DATA[_architecture].nbBytesWord;
  if ( ARCHITECTURE_DATA[_architecture].packed && type==MemoryCode ) return inc;
  return inc / 2;
}

uint Pic::Data::nbWords(MemoryRangeType type) const
{
  if ( !isPresent(type) ) return 0;
  return nbAddresses(type) / addressIncrement(type);
}

uint Pic::Data::nbAddresses(MemoryRangeType type) const
{
  if ( !isPresent(type) ) return 0;
  return (range(type).end - range(type).start + 1);
}

QString Pic::Data::fname(Device::FrequencyRange::Special special) const
{
  QString s = name();
  switch (special) {
    case Device::FrequencyRange::Normal: break;
    case Device::FrequencyRange::LowPower:
      // assume name is of form "NNX..."
      s.insert(2, 'L');
      break;
    case Device::FrequencyRange::LowVoltage:
      // assume name is of form "NNXN..."
      s.replace(2, 1, "LV");
      break;
    case Device::FrequencyRange::Nb_Specials: Q_ASSERT(false); break;
  }
  return s;
}

bool Pic::Data::matchId(uint rawId, uint deviceIdAddress, Device::IdData &idata) const
{
  if ( !isPresent(MemoryDeviceId) || deviceIdAddress!=range(MemoryDeviceId).start )
    return false;
  uint nid = 0;
  switch ( architecture() ) {
    case Arch_10F2XX:
    case Arch_16C5X:
    case Arch_16X:
    case Arch_17X:
    case Arch_18X:
      nid = rawId & ~0x1F;
      idata.revision = rawId & 0x1F;
      break;
    case Arch_30X:
      nid = (rawId >> 16) & 0xFFC0;
      idata.variant = (rawId >> 16) & 0x003F;
      idata.revision = (rawId >> 6) & 0x003F;
      idata.minorRevision = rawId & 0x003F;
      idata.process = (rawId >> 12) & 0x000F;
      break;
    case Nb_Architectures: Q_ASSERT(false); break;
  }
  return ( nid==id() );
}

QStringList Pic::Data::idNames(const Device::IdList &idList) const
{
  QStringList list;
  QMap<QString, Device::IdData>::const_iterator it = idList.begin();
  for (; it!=idList.end(); ++it) {
    switch (_architecture) {
      case Arch_10F2XX:
      case Arch_16C5X:
      case Arch_16X:
      case Arch_17X:
      case Arch_18X:
        list += i18n("%1 (rev. %2)").arg(it.key()).arg(it.data().revision);
        break;
      case Arch_30X:
        list += i18n("%1 (var. %1; proc. %2; rev. %3.%4").arg(it.key()).arg(it.data().variant)
                .arg(it.data().process).arg(it.data().revision).arg(it.data().minorRevision);
        break;
      case Nb_Architectures: Q_ASSERT(false); break;
    }
  }
  return list;
}

bool Pic::Data::checkCalibration(const Device::Array &data, QString *message) const
{
  Q_ASSERT( nbWords(MemoryCal)==data.count() );
  for (uint i=0; i<data.count(); i++) {
    QString address = toHexLabel(range(MemoryCal).start + i*addressIncrement(MemoryCal), nbCharsAddress());
    if ( data[i]==range(MemoryCal).mask ) {
      if (message) *message = i18n("Calibration word at address %1 is blank.").arg(address);
      return false;
    }
  }
  if ( data.count()==1 ) {
    if ( (data[0] & _calibration.opcodeMask)!=_calibration.opcode ) {
      if (message) *message = i18n("Calibration word is not a compatible opcode (%2).")
                              .arg(toHexLabel(_calibration.opcode, nbCharsWord(MemoryCode)));
      return false;
    }
  }
  return true;
}

uint Pic::Data::nbWordsAlignment(MemoryRangeType type) const
{
  if ( type!=MemoryCode ) return 1;
  return ARCHITECTURE_DATA[architecture()].nbWordsCodeAlignment;
}

QString Pic::Data::sfrName(uint address) const
{
  QMap<QString, uint>::const_iterator it;
  for (it=_sfrs.begin(); it!=_sfrs.end(); ++it)
    if ( it.data()==address ) return it.key();
  return QString::null;
}

QValueVector<uint> Pic::Data::mirrored(uint address) const
{
  QValueVector<uint> mirrors;
  for (uint i=0; i<_mirrored.count(); i++) {
    int delta = -1;
    for (uint k=0; k<_mirrored[i].count(); k++) {
      if ( address<_mirrored[i][k].start
           || address>=_mirrored[i][k].start+_mirrored[i][k].length ) continue;
      delta = address - _mirrored[i][k].start;
      break;
    }
    if ( delta==-1 ) continue;
    for (uint k=0; k<_mirrored[i].count(); k++) {
      if ( address==_mirrored[i][k].start+delta ) continue;
      mirrors.append(_mirrored[i][k].start+delta);
    }
    break;
  }
  return mirrors;
}

Pic::RegisterProperties Pic::Data::registerProperties(uint address) const
{
  QString name = registerName(address);
  if ( name=="INDF" ) return NotAccessible;
  if ( name=="PCLATH" ) return Read;
  return Read | Write;
}

Pic::RegisterType Pic::Data::registerType(uint address) const
{
  for (uint i=0; i<_unused.count(); i++)
    if ( address>=_unused[i].start && address<_unused[i].start+_unused[i].length ) return UnusedRegister;
  if ( !sfrName(address).isEmpty() ) return Sfr;
  QValueVector<uint> m = mirrored(address);
  bool first = true;
  for (uint l=0; l<m.count(); l++) {
    if ( !sfrName(m[l]).isEmpty() ) return MirroredSfr;
    if ( m[l]<address ) first = false;
  }
  if (first) return Gpr;
  return MirroredGpr;
}

QString Pic::Data::registerName(uint address) const
{
  if ( registerType(address)==UnusedRegister ) return "---";
  QString name = sfrName(address);
  if ( !name.isEmpty() ) return name;
  QValueVector<uint> m = mirrored(address);
  for (uint l=0; l<m.count(); l++) {
    name = sfrName(m[l]);
    if ( !name.isEmpty() ) return name;
  }
  return "<GPR>";
}
