/***************************************************************************
 *   Copyright (C) 2005 Nicolas Hadacek <hadacek@kde.org>                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#ifndef PIC_H
#define PIC_H

#include <qstringlist.h>
#include <qmap.h>
#include <qvaluevector.h>

#include "devices/base/generic_device.h"
#include "pic_config.h"

namespace Pic
{
class XmlToData;
class Group;

//-----------------------------------------------------------------------------
enum ResetMode { ResetHeld = 0, ResetReleased, Nb_ResetModes};

struct VoltageData {
  double min, max, nominal;
};
enum ProgVoltageType { Vpp = 0, VddBulkErase, VddWrite, VddRead, Nb_ProgVoltageTypes };
struct ProgVoltageTypeData {
  const char *name;
};
extern const ProgVoltageTypeData PROG_VOLTAGE_TYPE_DATA[Nb_ProgVoltageTypes];

enum MemoryTechnology { EFlash = 0, Flash, Eprom, Rom, Romless, Nb_MemoryTechnologies };
struct MemoryTechnologyData {
  const char *name, *label;
};
extern const MemoryTechnologyData MEMORY_TECHNOLOGY_DATA[Nb_MemoryTechnologies];

struct CalibrationData {
  uint opcode, opcodeMask;
};

enum Architecture { Arch_10F2XX = 0, Arch_16C5X, Arch_16X, Arch_17X, Arch_18X,
                    Arch_30X, Nb_Architectures };
struct ArchitectureData {
  const char *name;
  uint nbBitsPC;                     // nb bits max for PC
  uint nbBytesWord;                  // nb bytes per word (hex file and icd2)
  bool packed;
  uint nbWordsCodeAlignment;         // alignment for programming code memory
  uint nbBitsRegister;
  uint registerBankLength;
  uint nbBits[Nb_MemoryRangeTypes];  // nb bits per word
};
extern const ArchitectureData ARCHITECTURE_DATA[Nb_Architectures];

enum MemoryRangeProperty { NotPresent = 0, Present = 1, Programmable = 2 };
Q_DECLARE_FLAGS(MemoryRangeProperties, MemoryRangeProperty)
Q_DECLARE_OPERATORS_FOR_FLAGS(MemoryRangeProperties)
struct MemoryRangeData {
  MemoryRangeProperties properties;
  uint start, end, mask, rmask, hexFileOffset;
};

struct RangeData {
  uint start, length;
};

enum RegisterType { UnusedRegister, Sfr, MirroredSfr, Gpr, MirroredGpr };
enum RegisterProperty { NoProperty = 0, NotAccessible = 1, Read = 2, Write = 4 };
Q_DECLARE_FLAGS(RegisterProperties, RegisterProperty)
Q_DECLARE_OPERATORS_FOR_FLAGS(RegisterProperties)

//-----------------------------------------------------------------------------
class Data : public Device::Data
{
public:
  Data();
  Data(const Device::ConstData &cdata);
  virtual QString fname(Device::FrequencyRange::Special special) const;
  uint id() const { return _id; }
  bool isPresent(MemoryRangeType type) const { return (range(type).properties & Present); }
  bool isWritable(MemoryRangeType type) const;
  uint nbAddresses(MemoryRangeType type) const;
  uint nbWords(MemoryRangeType type) const;
  uint addressIncrement(MemoryRangeType type) const;
  uint nbWordsAlignment(MemoryRangeType type) const;
  const MemoryRangeData &range(MemoryRangeType type) const { return _ranges[type]; }
  virtual uint nbBitsAddress() const { return ARCHITECTURE_DATA[_architecture].nbBitsPC; }
  uint nbBitsWord(MemoryRangeType type) const { return ARCHITECTURE_DATA[_architecture].nbBits[type]; }
  uint nbBytesWord(MemoryRangeType type) const { return nbBitsWord(type)/8 + (nbBitsWord(type)%8 ? 1 : 0); }
  uint nbCharsWord(MemoryRangeType type) const { return nbBitsWord(type)/4 + (nbBitsWord(type)%4 ? 1 : 0); }
  const Config &config() const { return _config; }
  Architecture architecture() const { return Architecture(_architecture); }
  const VoltageData &voltage(ProgVoltageType type) const { return _voltages[type]; }
  virtual bool canWriteCalibration() const { return isPresent(MemoryCal) && isWritable(MemoryCal); }
  bool checkCalibration(const Device::Array &data, QString *message = 0) const;
  MemoryTechnology memoryTechnology() const { return _memoryTechnology; }
  bool matchId(uint rawId, uint deviceIdAddress, Device::IdData &data) const;
  QStringList idNames(const Device::IdList &idList) const;

  uint nbBitsRegister() const { return ARCHITECTURE_DATA[_architecture].nbBitsRegister; }
  uint nbBytesRegister() const { return nbBitsRegister()/8 + (nbBitsRegister()%8 ? 1 : 0); }
  uint nbCharsRegister() const { return nbBitsRegister()/4 + (nbBitsRegister()%4 ? 1 : 0); }
  uint nbBytesPerBank() const { return ARCHITECTURE_DATA[_architecture].registerBankLength; }
  uint nbRegistersPerBank() const { return nbBytesPerBank() / nbBytesRegister(); }
  uint nbBanks() const { return _nbBanks; }
  virtual uint nbRegisters() const { return _nbBanks * nbRegistersPerBank(); }
  uint nbCharsRegisterAddress() const { return nbChars(nbRegisters() - 1); }
  const QMap<QString, uint> &sfrs() const { return _sfrs; }
  QString sfrName(uint address) const;
  QValueVector<uint> mirrored(uint address) const;
  RegisterProperties registerProperties(uint address) const;
  RegisterType registerType(uint address) const;
  QString registerName(uint address) const;

private:
  uint _id;
  MemoryTechnology _memoryTechnology;
  VoltageData _voltages[Nb_ProgVoltageTypes];
  MemoryRangeData _ranges[Nb_MemoryRangeTypes];
  Config _config;
  CalibrationData _calibration;

  uint _nbBanks;
  QMap<QString, uint> _sfrs;
  QValueVector<QValueVector<RangeData> > _mirrored;
  QValueVector<RangeData> _unused;

  friend class XmlToData;
  friend class Group;
};

}; // namespace

#endif
