/***************************************************************************
 *   Copyright (C) 2005 Nicolas Hadacek <hadacek@kde.org>                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#ifndef PIC_CONFIG_H
#define PIC_CONFIG_H

#include <qmap.h>
#include <qvaluevector.h>
#include <qstringlist.h>
#include "common/qflags.h"
#include "common/misc.h"

namespace Pic
{
//-----------------------------------------------------------------------------
enum MemoryRangeType { MemoryCode = 0, MemoryCal, MemoryUserId, MemoryDeviceId,
                       MemoryConfig, MemoryEeprom, MemoryDebugVector, MemoryHardwareStack,
                       MemoryCalBackup, MemoryProgramExecutive, Nb_MemoryRangeTypes };
enum MemoryRangeTypeProperty { ReadOnly = 0, Writable = 1 };
Q_DECLARE_FLAGS(MemoryRangeTypeProperties, MemoryRangeTypeProperty)
Q_DECLARE_OPERATORS_FOR_FLAGS(MemoryRangeTypeProperties)
struct MemoryRangeTypeData {
  const char *name, *label;
  MemoryRangeTypeProperties properties;
};
extern const MemoryRangeTypeData MEMORY_RANGE_TYPE_DATA[Nb_MemoryRangeTypes];

class Data;

//-----------------------------------------------------------------------------
class Config
{
public:
  class Value {
  public:
    QString name, cname;
    QStringList extraCNames;
    uint value;
    bool operator <(const Value &cv) const { return value<cv.value; }
  };
  class Mask {
  public:
    QString name;
    QStringList ignoredCNames;
    uint value;
    QValueVector<Value> values; // ordered from lower to higher
    bool operator <(const Mask &cm) const { return value<cm.value; }
  };
  class Word {
  public:
    Word() : rmask(0), wmask(0) {}
    QString name;
    uint rmask, wmask, pmask, cmask; // read, write, protected, and checksum bits masks
    uint bvalue;              // blank value
    QValueVector<Mask> masks; // ordered from lower to higher
    uint usedMask() const;
  };

  enum BlockType { CodeBlock = 0, DataBlock, CalibrationBlock, BootBlock, ConfigBlock, Nb_BlockTypes };
  enum ProtectionType { ProgramProtected = 0, WriteProtected, ReadProtected, Nb_ProtectionTypes };
  QString maskName(ProtectionType ptype, BlockType btype, int block = -1) const;

public:
  Config(const Pic::Data &data) : _data(data) {}

  enum { MAX_NB_BLOCKS = 8 };
  QValueVector<Word>  _words;

  const Value *findValue(const QString &mask, const QString &value) const;
  const Mask *findMask(const QString &mask, uint *wordIndex = 0) const;
  static bool hasMaskName(const QString &mask);
  static QString maskLabel(const QString &mask);
  bool checkValueName(const QString &mask, const QString &name) const;
  bool hasValueName(const QString &mask, const QString &name) const;
  static QString valueLabel(const QString &mask, const QString &name);

  enum ConfigNameType { ConfigNameAsmMacro, ConfigNameCMacro, ConfigNameCVariable };
  QString configName(uint wordIndex, ConfigNameType type) const;
  QStringList ignoredMaskNames(uint wordIndex, const Mask &mask) const;
  QStringList extraMaskNames(uint wordIndex, const Value &value) const;
  QString maskName(uint wordIndex, const Value &value) const;
  QString noneProtectedValueName() const;

  bool hasBootBlock() const;
  uint nbBlocks() const;
  Range extractRange(const QString &name, BlockType type) const;

private:
  class MapData {
  public:
    MapData() {}
    MapData(int i, int b) : index(i), block(b) {}
    int index, block;
  };
  static QMap<QString, MapData> &masks();
  static QMap<QString, MapData> *_masks; // mask name -> index in DATA

  struct MaskData {
    const char *name, *label;
  };
  struct NameData {
    const char *name, *label, *configNames[5];
  };
  enum Type { Fixed, ValueDouble, ValueUInt, Ratio, MemoryRange, Toggle, Pin, Pins };
  class Data {
  public:
    const MaskData mask;
    Type type;
    const NameData values[50];
  };
  static const Data DATA[];

private:
  const Pic::Data &_data;

  static bool extractRanges(const QString &name, QValueVector<uint> &starts, uint &end, bool &ok);
  struct ProtectionData {
    const char *name[Nb_ProtectionTypes];
    MemoryRangeType type;
  };
  static const ProtectionData PROTECTION_DATA[Nb_BlockTypes];
  class ProtectedRange {
  public:
    QValueVector<uint> starts, ends;
  };
  ProtectedRange extractRange(const QString &mask, const QString &name, bool &ok) const;
  QStringList transformMaskNames(uint wordIndex, const QStringList &list) const;
  bool isAllProtectedValueName(const QString &name) const;
  bool isNoneProtectedValueName(const QString &name) const;
};

}; //namespace

#endif
