/***************************************************************************
 *   Copyright (C) 2005-2006 Nicolas Hadacek <hadacek@kde.org>             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "pic_prog.h"

#include <klocale.h>

#include "common/gui/misc_gui.h"
#include "devices/base/device_list.h"
#include "progs/base/prog_config.h"
#include "progs/base/prog_group.h"

//-----------------------------------------------------------------------------
uint Programmer::PicDeviceSpecific::findNonMaskStart(Pic::MemoryRangeType type, const Device::Array &data) const
{
  uint mask = device().range(type).mask;
  uint start = 0;
  for (; start<data.count(); start++)
    if ( data[start]!=mask ) break;
  const_cast<PicDeviceSpecific *>(this)->log(Log::ExtraDebug, QString("start before align: %1").arg(start));
  uint align = device().nbWordsAlignment(type);
  start -= start % align;
  const_cast<PicDeviceSpecific *>(this)->log(Log::ExtraDebug, QString("start after align: %1 (align=%2)").arg(start).arg(align));
  return start;
}

uint Programmer::PicDeviceSpecific::findNonMaskEnd(Pic::MemoryRangeType type, const Device::Array &data) const
{
  uint mask = device().range(type).mask;
  uint end = data.count()-1;
  for (; end>0; end--)
    if ( data[end]!=mask ) break;
  const_cast<PicDeviceSpecific *>(this)->log(Log::ExtraDebug, QString("end before align: %1").arg(end));
  uint align = device().nbWordsAlignment(type);
  if ( (end+1) % align ) end += align - (end+1) % align;
  const_cast<PicDeviceSpecific *>(this)->log(Log::ExtraDebug, QString("end after align: %1 (align=%2)").arg(end).arg(align));
  Q_ASSERT(end<data.count());
  return end;
}

Device::Array Programmer::PicDeviceSpecific::read(Pic::MemoryRangeType type)
{
  setPowerOn();
  Device::Array array = doRead(type);
  setPowerOff();
  return array;
}

bool Programmer::PicDeviceSpecific::write(Pic::MemoryRangeType mtype, const Device::Array &data, bool force)
{
  setPowerOn();
  doWrite(mtype, data, force);
  setPowerOff();
  return !hasError();
}

bool Programmer::PicDeviceSpecific::erase(bool isProtected)
{
  setPowerOn();
  doErase(isProtected);
  setPowerOff();
  return !hasError();
}

Programmer::Result Programmer::PicDeviceSpecific::eraseRange(Pic::MemoryRangeType type)
{
  Result result = Error;
  setPowerOn();
  switch (type) {
    case Pic::MemoryCode: result = doEraseCode();   break;
    case Pic::MemoryEeprom: result = doEraseEeprom(); break;
    case Pic::MemoryUserId: result = doEraseUserId(); break;
    case Pic::MemoryConfig: result = doEraseConfig(); break;
    case Pic::MemoryDeviceId:
    case Pic::MemoryCal:
    case Pic::MemoryCalBackup:
    case Pic::MemoryHardwareStack:
    case Pic::MemoryDebugVector:
    case Pic::MemoryProgramExecutive:
    case Pic::Nb_MemoryRangeTypes: Q_ASSERT(false); break;
  }
  setPowerOff();
  return result;
}

Programmer::Result Programmer::PicDeviceSpecific::doEraseConfig()
{
  Pic::Memory memory(device());
  if ( !doWrite(Pic::MemoryConfig, memory.arrayForWriting(Pic::MemoryConfig), true) ) return Error;
  if ( !doVerify(Pic::MemoryConfig, memory, BlankCheckVerify, RangeVector()) ) return Error;
  return Ok;
}

bool Programmer::PicDeviceSpecific::compareWords(Pic::MemoryRangeType type, uint index, uint v, uint d, Programmer::VerifyActions actions)
{
  if ( v==d ) return true;
  uint inc = device().addressIncrement(type);
  uint address = device().range(type).start + inc * index;
  if ( actions & Programmer::BlankCheckVerify )
    log(Log::SoftError, i18n("Device memory is not blank (in %1 at address 0x%2: reading 0x%3 and expecting 0x%4).")
      .arg(i18n(Pic::MEMORY_RANGE_TYPE_DATA[type].label)).arg(toHex(address, device().nbCharsAddress()))
      .arg(toHex(d, device().nbCharsWord(type))).arg(toHex(v, device().nbCharsWord(type))));
  else log(Log::SoftError, i18n("Device memory does not match hex file (in %1 at address 0x%2: reading 0x%3 and expecting 0x%4).")
      .arg(i18n(Pic::MEMORY_RANGE_TYPE_DATA[type].label)).arg(toHex(address, device().nbCharsAddress()))
      .arg(toHex(d, device().nbCharsWord(type))).arg(toHex(v, device().nbCharsWord(type))));
  return false;
}

bool Programmer::PicDeviceSpecific::verifyWord(uint i, uint word,
                                              Pic::MemoryRangeType type, const Pic::Memory &memory,
                                              Programmer::VerifyActions actions, const RangeVector &rv)
{
  if ( !(actions & Programmer::IgnoreProtectedVerify) && rv.contains(i) ) return true; // protected
  uint v = memory.normalizedWord(type, i);
  uint d = memory.normalizeWord(type, i, word);
  return compareWords(type, i, v, d, actions);
}

bool Programmer::PicDeviceSpecific::verify(Pic::MemoryRangeType type, const Pic::Memory &memory,
                                           Programmer::VerifyActions actions, const RangeVector &rv)
{
  setPowerOn();
  bool ok = doVerify(type, memory, actions, rv);
  setPowerOff();
  return ok;
}

bool Programmer::PicDeviceSpecific::doVerify(Pic::MemoryRangeType type, const Pic::Memory &memory,
                                             Programmer::VerifyActions actions, const RangeVector &rv)
{
  Device::Array data = doRead(type);
  if ( hasError() ) return false;
  for (uint i=0; i<data.count(); i++)
    if ( !verifyWord(i, data[i], type, memory, actions, rv) ) return false;
  return true;
}

//----------------------------------------------------------------------------
Programmer::PicBase::PicBase(const Group &group, const Pic::Data *data, const char *name)
  : Base(group, data, name), _singleType(Pic::Nb_MemoryRangeTypes), _deviceMemory(0)
{
  if (data) _deviceMemory = new Pic::Memory(*data);
}

Programmer::PicBase::~PicBase()
{
  delete _deviceMemory;
}

uint Programmer::PicBase::nbSteps(Operation operation) const
{
  uint length = Base::nbSteps(operation);
  if ( operation==Connect || operation==Erase ) return length;
  for (uint i=0; i<Pic::Nb_MemoryRangeTypes; i++) {
    Pic::MemoryRangeType type = Pic::MemoryRangeType(i);
    if ( type!=Pic::MemoryCode && type!=Pic::MemoryEeprom ) continue;
    if ( operation==Write && !device()->isWritable(type) ) continue;
    if ( operation==Write ) {
      length += device()->nbWords(type);
      if ( device()->memoryTechnology()==Pic::Eprom ) length += nbSteps(Verify);
    }
    length += device()->nbWords(type);
  }
  return length;
}

bool Programmer::PicBase::initProgramming(Pic::ProgVoltageType type)
{
  if ( vpp()!=UNKNOWN_VOLTAGE ) {
    const Pic::VoltageData &tvpp = device()->voltage(Pic::Vpp);
    if ( vpp()<tvpp.min )
      log(Log::Warning, i18n("Vpp (%1 V) is lower than the minimum required voltage (%2 V).")
                        .arg(vpp()).arg(tvpp.min));
    if ( vpp()>tvpp.max ) {
      QString s = i18n("Vpp (%1 V) is higher than the maximum voltage (%2 V). You may damage the device.")
                  .arg(vpp()).arg(tvpp.max);
      log(Log::Warning, s);
      if ( !askContinue(s) ) return false;
    }
  }
  if ( vdd()!=UNKNOWN_VOLTAGE ) {
    Q_ASSERT( type!=Pic::Vpp );
    const Pic::VoltageData &tvdd = device()->voltage(type);
    if ( vdd()<tvdd.min ) {
      if ( type==Pic::VddBulkErase && device()->voltage(Pic::VddWrite).min!=tvdd.min )
        log(Log::Warning, i18n("Vdd (%1 V) is too low for high-voltage programming\n(piklab only supports high-voltage programming at the moment).\nMinimum required is %2 V.")
                          .arg(vdd()).arg(tvdd.min));
      else if ( type==Pic::VddRead && device()->voltage(Pic::VddWrite).min!=tvdd.min )
        log(Log::Warning, i18n("Vdd (%1 V) is too low for reading\nMinimum required is %2 V.")
                          .arg(vdd()).arg(tvdd.min));
      else log(Log::Warning, i18n("Vdd (%1 V) is too low for programming\nMinimum required is %2 V.")
                            .arg(vdd()).arg(tvdd.min));
    } else if ( vdd()>tvdd.max ) {
      QString s = i18n("Vdd (%1 V) is higher than the maximum voltage (%2 V). You may damage the device.")
                  .arg(vdd()).arg(tvdd.max);
      log(Log::Warning, s);
      if ( !askContinue(s) ) return false;
    }
  }

  Device::Array data = specific()->read(Pic::MemoryConfig);
  if ( hasError() ) return false;
  _deviceMemory->setArray(Pic::MemoryConfig, data);
  _hasProtectedCode = _deviceMemory->isProtected(Pic::Config::ProgramProtected, Pic::MemoryCode);
  _hasProtectedEeprom = _deviceMemory->isProtected(Pic::Config::ProgramProtected, Pic::MemoryEeprom);
  log(Log::NormalDebug, QString("  protected: code=%1 data=%2")
                        .arg(_hasProtectedCode ? "true" : "false").arg(_hasProtectedEeprom ? "true" : "false"));

  return initProgramming();
}

bool Programmer::PicBase::preserveCode()
{
  if ( _hasProtectedCode && !askContinue(i18n("All or part of code memory is protected so cannot be preserved. Continue anyway?")) )
    return false;
  return readRange(Pic::MemoryCode, *_deviceMemory);
}

bool Programmer::PicBase::preserveEeprom()
{
  if ( _hasProtectedEeprom && !askContinue(i18n("All or part of data eeprom is protected so cannot be preserved. Continue anyway?")) )
    return false;
  return readRange(Pic::MemoryEeprom, *_deviceMemory);
}

//-----------------------------------------------------------------------------
uint Programmer::PicBase::readDeviceId()
{
  uint id = 0;
  Device::Array data = specific()->read(Pic::MemoryDeviceId);
  if ( hasError() || data.size()==0 ) return false;
  switch ( device()->architecture()) {
    case Pic::Arch_10F2XX:
    case Pic::Arch_16C5X:
    case Pic::Arch_16X:
    case Pic::Arch_17X: id = data[0]; break;
    case Pic::Arch_18X: id = data[0] | (data[1] << 8); break;
    case Pic::Arch_30X: id = data[1] | (data[0] << 16); break;
    case Pic::Nb_Architectures: Q_ASSERT(false); break;
  }
  return id;
}

bool Programmer::PicBase::verifyDeviceId()
{
  if ( !device()->isPresent(Pic::MemoryDeviceId) ) {
    log(Log::Warning, i18n("Device not autodetectable: continuing with the specified device name \"%1\"...").arg(device()->name()));
    return true;
  }
  uint rawId = readDeviceId();
  if ( rawId==0 || rawId==device()->range(Pic::MemoryDeviceId).mask ) {
    log(Log::Error, i18n("Missing or incorrect device (Id read as zero)."));
    return false;
  }
  Device::IdList ids = Device::lister().findIds(device()->range(Pic::MemoryDeviceId).start, rawId);
  QString message;
  if ( ids.count()!=0 ) {
    log(Log::Information, i18n("Read id: %1").arg(device()->idNames(ids).join("; ")));
    if ( ids.contains(device()->name()) ) return true;
    message = i18n("Read id does not match the specified device name \"%1\".").arg(device()->name());
  } else {
    uint nb = device()->nbWords(Pic::MemoryDeviceId) * device()->nbCharsWord(Pic::MemoryDeviceId);
    log(Log::Warning, i18n("  Unknown or incorrect device (Read id: %1).").arg(toHexLabel(rawId, nb)));
    message = i18n("Unknown device.");
  }
  if ( !askContinue(message) ) {
    logUserAbort();
    return false;
  }
  log(Log::Information, i18n("Continue with the specified device name: \"%1\"...").arg(device()->name()));
  return true;
}

//-----------------------------------------------------------------------------
QString Programmer::PicBase::prettyCalibration(const Device::Array &data) const
{
  QString s;
  for (uint i=0; i<data.count(); i++) {
    if ( i!=0 ) s += ", ";
    s += toHexLabel(data[i], device()->nbCharsWord(Pic::MemoryCal));
  }
  return s;
}

bool Programmer::PicBase::saveCalibration()
{
  if ( device()->isPresent(Pic::MemoryCal) && device()->isWritable(Pic::MemoryCal) ) {
    Device::Array data = specific()->read(Pic::MemoryCal);
    if ( hasError() ) return false;
    _deviceMemory->setArray(Pic::MemoryCal, data);
    if ( device()->isPresent(Pic::MemoryCal) ) {
      log(Log::NormalDebug, QString("  Read calibration word: %1").arg(prettyCalibration(data)));
      QString message;
      bool calok = device()->checkCalibration(data, &message);
      if ( !calok ) log(Log::Warning, "  " + message);
      if ( device()->isPresent(Pic::MemoryCalBackup) ) {
        data = specific()->read(Pic::MemoryCalBackup);
        if ( hasError() ) return false;
        _deviceMemory->setArray(Pic::MemoryCalBackup, data);
        log(Log::NormalDebug, QString("  Read calibration backup word: %1").arg(prettyCalibration(data)));
        if ( device()->checkCalibration(data, &message) ) {
          if ( !calok ) _deviceMemory->setArray(Pic::MemoryCal, data); // only replace osccal with backup if backup correct and osccal not correct
        } else log(Log::Warning, "  " + message);
      }
    } else if ( !askContinue(i18n("Reading calibration word is not supported. It may become lost...")) ) {
      logUserAbort();
      return false;
    }
  }
  return true;
}

bool Programmer::PicBase::restoreCalibration()
{
  Device::Array data = _deviceMemory->arrayForWriting(Pic::MemoryCal);
  if ( !device()->isPresent(Pic::MemoryCal)
       || !device()->isWritable(Pic::MemoryCal)
       || !device()->checkCalibration(data) ) return true;
  Device::Array cdata = specific()->read(Pic::MemoryCal);
  if ( hasError() ) return false;
  if ( cdata==data ) {
    log(Log::Information, i18n("  Calibration word is unchanged."));
    return true;
  }
  if ( !programRange(Pic::MemoryCal, data) ) return false;
  cdata = specific()->read(Pic::MemoryCal);
  if ( hasError() ) return false;
  if ( cdata==data ) log(Log::Information, i18n("  Calibration word has been preserved."));

  data = _deviceMemory->arrayForWriting(Pic::MemoryCalBackup);
  if ( !device()->isPresent(Pic::MemoryCalBackup)
       || !device()->isWritable(Pic::MemoryCalBackup)
       || !device()->checkCalibration(data) ) return true;
  cdata = specific()->read(Pic::MemoryCalBackup);
  if ( hasError() ) return false;
  if ( cdata==data ) {
    log(Log::Information, i18n("  Calibration backup word is unchanged."));
    return true;
  }
  if ( !programRange(Pic::MemoryCalBackup, data) ) return false;
  cdata = specific()->read(Pic::MemoryCalBackup);
  if ( hasError() ) return false;
  if ( cdata==data ) log(Log::Information, i18n("  Calibration backup word has been preserved."));
  return true;
}

bool Programmer::PicBase::restoreBandGapBits()
{
  bool hasProtectedBits = false;
  for (uint i=0; i<device()->nbWords(Pic::MemoryConfig); i++)
    if ( device()->config()._words[i].pmask!=0 ) hasProtectedBits = true;
  if ( !hasProtectedBits ) return true;
  Device::Array cdata = specific()->read(Pic::MemoryConfig);
  Device::Array data = _deviceMemory->arrayForWriting(Pic::MemoryConfig);
  for (uint i=0; i<cdata.count(); i++) {
    uint pmask = device()->config()._words[i].pmask;
    if ( pmask==0 ) continue;
    cdata[i] &= ~pmask;
    cdata[i] |= (data[i] & pmask);
  }
  log(Log::NormalDebug, QString("  Write config with band gap bits: %2").arg(toHexLabel(cdata[0], device()->nbCharsWord(Pic::MemoryConfig))));
  if ( !programRange(Pic::MemoryConfig, cdata) ) return false;
  data = specific()->read(Pic::MemoryConfig);
  if ( data==cdata ) log(Log::Information, i18n("  Band gap bits have been preserved."));
  return true;
}

bool Programmer::PicBase::eraseAll()
{
  if ( !saveCalibration() ) return false;
  // read config for protection state
  if ( !specific()->erase(_hasProtectedCode || _hasProtectedEeprom) ) return false;
  addProgress(OPERATION_DATA[Erase].nbSteps);
  if ( !restoreCalibration() ) return false;
  return true;
}

bool Programmer::PicBase::checkErase()
{
  if ( device()->memoryTechnology()==Pic::Rom || device()->memoryTechnology()==Pic::Romless
       || device()->memoryTechnology()==Pic::Eprom ) {
    log(Log::SoftError, i18n("Cannot erase ROM or EPROM device."));
    return false;
  }
  return true;
}

bool Programmer::PicBase::internalErase()
{
  initProgramming(Pic::VddBulkErase);
  bool ok = true;
  if ( _singleType!=Pic::Nb_MemoryRangeTypes ) ok = eraseRange(_singleType);
  else ok = eraseAll();
  if ( !restoreBandGapBits() ) return false;
  return ok;
}

bool Programmer::PicBase::eraseSingle(Pic::MemoryRangeType type)
{
  _singleType = type;
  bool ok = erase();
  _singleType = Pic::Nb_MemoryRangeTypes;
  return ok;
}

bool Programmer::PicBase::eraseRange(Pic::MemoryRangeType type)
{
  if ( !saveCalibration() ) return false;
  bool ok = internalEraseRange(type);
  if ( !restoreCalibration() ) return false;
  if ( ok && Config::blankCheckAfterErase() ) {
    Pic::Memory memory(*device());
    return verifyRange(type, memory, BlankCheckVerify);
  }
  return ok;
}

bool Programmer::PicBase::internalEraseRange(Pic::MemoryRangeType type)
{
  if ( type==Pic::MemoryCode && _hasProtectedCode ) {
    log(Log::SoftError, i18n("Cannot erase protected code memory. Consider erasing the whole chip."));
    return false;
  }
  if ( type==Pic::MemoryEeprom && _hasProtectedEeprom ) {
    log(Log::SoftError, i18n("Cannot erase protected data eeprom. Consider erasing the whole chip."));
    return false;
  }
  Result res = specific()->eraseRange(type);
  if ( res==Ok ) return true;
  if ( res==Error ) return false;
  if ( !askContinue(i18n("%1: Erasing this range only is not supported with this programmer. This will erase the whole chip and restore the other memory ranges.").arg(i18n(Pic::MEMORY_RANGE_TYPE_DATA[type].label))) ) {
    logUserAbort();
    return false;
  }
  if ( type!=Pic::MemoryCode && !preserveCode() ) return false;
  if ( type!=Pic::MemoryEeprom && !preserveEeprom() ) return false;
  if ( type!=Pic::MemoryUserId && !readRange(Pic::MemoryUserId, *_deviceMemory) ) return false;
  specific()->erase(_hasProtectedCode || _hasProtectedEeprom);
  if ( type!=Pic::MemoryCode && !programAndVerifyRange(Pic::MemoryCode, *_deviceMemory) ) return false;
  if ( type!=Pic::MemoryEeprom && !programAndVerifyRange(Pic::MemoryEeprom, *_deviceMemory) ) return false;
  if ( type!=Pic::MemoryUserId && !programAndVerifyRange(Pic::MemoryUserId, *_deviceMemory) ) return false;
  if ( !programAndVerifyRange(Pic::MemoryConfig, *_deviceMemory) ) return false;
  return true;
}

//-----------------------------------------------------------------------------
bool Programmer::PicBase::readSingle(Pic::MemoryRangeType type, Pic::Memory &memory)
{
  _singleType = type;
  Pic::Memory tmp(*device());
  bool ok = read(tmp);
  _singleType = Pic::Nb_MemoryRangeTypes;
  if (ok) {
    memory.copyFrom(type, tmp);
    if ( type==Pic::MemoryCal ) memory.copyFrom(Pic::MemoryCalBackup, tmp);
  }
  return ok;
}

bool Programmer::PicBase::readRange(Pic::MemoryRangeType type, Pic::Memory &memory)
{
  if ( !device()->isPresent(type) ) return true;
  log(Log::Information, i18n("  Read memory: %1").arg(i18n(Pic::MEMORY_RANGE_TYPE_DATA[type].label)));
  Device::Array data = specific()->read(type);
  if ( hasError() ) return false;
  memory.setArray(type, data);
  return true;
}

bool Programmer::PicBase::checkRead()
{
  if ( device()->memoryTechnology()==Pic::Romless ) {
    log(Log::SoftError, i18n("Cannot read ROMless device."));
    return false;
  }
  return true;
}

bool Programmer::PicBase::internalRead(Device::Memory &memory)
{
  initProgramming(Pic::VddRead);
  Pic::Memory &pmemory = static_cast<Pic::Memory &>(memory);
  if ( _singleType!=Pic::Nb_MemoryRangeTypes ) {
    if ( _singleType==Pic::MemoryCal ) {
      if ( !readRange(Pic::MemoryCal, pmemory) ) return false;
      return readRange(Pic::MemoryCalBackup, pmemory);
    }
    return readRange(_singleType, pmemory);
  }
  if ( !readRange(Pic::MemoryConfig,    pmemory) ) return false;
  if ( !readRange(Pic::MemoryUserId,    pmemory) ) return false;
  if ( !readRange(Pic::MemoryCal,       pmemory) ) return false;
  if ( !readRange(Pic::MemoryCalBackup, pmemory) ) return false;
  if ( !readRange(Pic::MemoryCode,      pmemory) ) return false;
  if ( !readRange(Pic::MemoryEeprom,    pmemory) ) return false;
  return true;
}

//-----------------------------------------------------------------------------
bool Programmer::PicBase::programSingle(Pic::MemoryRangeType type, const Pic::Memory &memory)
{
  _singleType = type;
  if ( type==Pic::MemoryCode || type==Pic::MemoryEeprom )
    log(Log::Warning, i18n("The range is not erased first so programming may fail..."));
  bool ok = program(memory);
  _singleType = Pic::Nb_MemoryRangeTypes;
  return ok;
}

bool Programmer::PicBase::programRange(Pic::MemoryRangeType mtype, const Device::Array &data)
{
  log(Log::Information, i18n("  Write memory: %1").arg(i18n(Pic::MEMORY_RANGE_TYPE_DATA[mtype].label)));
  bool only = ( Programmer::Config::onlyProgramNonMask() && (mtype==Pic::MemoryCode || mtype==Pic::MemoryEeprom) );
  return specific()->write(mtype, data, !only);
}

bool Programmer::PicBase::programAndVerifyRange(Pic::MemoryRangeType mtype, const Pic::Memory &memory)
{
  if ( !device()->isPresent(mtype) ) return true;
  Device::Array data = memory.arrayForWriting(mtype);
  if ( !programRange(mtype, data) ) return false;
  VerifyActions actions = IgnoreProtectedVerify;
  if ( mtype==Pic::MemoryCode && Programmer::Config::onlyVerifyProgrammed() )
    actions |= OnlyProgrammedVerify;
  return verifyRange(mtype, memory, actions);
}

bool Programmer::PicBase::programAll(const Pic::Memory &memory)
{
  if ( !programAndVerifyRange(Pic::MemoryCode, memory) ) return false;
  const Pic::Memory &tmp = (Config::preserveEeprom() ? *_deviceMemory : memory);
  if ( !programAndVerifyRange(Pic::MemoryEeprom, tmp) ) return false;
  if ( !programAndVerifyRange(Pic::MemoryUserId, memory) ) return false;
  if ( device()->nbWords(Pic::MemoryConfig)>1 ) {
    RangeVector rv = memory.protectedRanges(Pic::Config::ProgramProtected, Pic::MemoryConfig);
    if ( !rv.isEmpty() ) log(Log::Warning, i18n("Configuration bits are set to be program protected. Due to limitations in piklab, it can happen that not all configuration bits will be correctly written."));
  }
  if ( !programAndVerifyRange(Pic::MemoryConfig, memory) ) return false;
  return true;
};

bool Programmer::PicBase::checkProgram(const Device::Memory &memory)
{
  if ( device()->memoryTechnology()==Pic::Rom || device()->memoryTechnology()==Pic::Romless ) {
    log(Log::SoftError, i18n("Cannot write ROM or ROMless device."));
    return false;
  }
  if ( !group().isDebugger() && static_cast<const Pic::Memory &>(memory).hasDebugOn() ) {
    if ( !askContinue(i18n("DEBUG configuration bit is on. Are you sure you want to continue programming the chip?")) ) {
      logUserAbort();
      return false;
    }
  }
  return true;
}

bool Programmer::PicBase::internalProgram(const Device::Memory &memory)
{
  initProgramming(Pic::VddBulkErase);
  const Pic::Memory &pmemory = static_cast<const Pic::Memory &>(memory);

  // blank check if OTP device
  bool eprom = ( device()->memoryTechnology()==Pic::Eprom );
  if (eprom) {
    log(Log::Information, i18n("  EPROM device: blanck checking first..."));
    Pic::Memory memory(*device());
    if ( !verifyAll(pmemory, BlankCheckVerify) ) return false;
    log(Log::Information, i18n("  Blanck check successful"));
    // check if protecting device
    bool protectedCode = pmemory.isProtected(Pic::Config::ProgramProtected, Pic::MemoryCode);
    bool protectedEeprom = pmemory.isProtected(Pic::Config::ProgramProtected, Pic::MemoryEeprom);
    if ( protectedCode || protectedEeprom ) {
      log(Log::SoftError, i18n("Protecting code memory or data eeprom on OTP devices is disabled as a security..."));
      return false;
    }
  }

  // programming
  bool ok = true;
  if ( _singleType!=Pic::Nb_MemoryRangeTypes ) {
    if ( (_singleType==Pic::MemoryCode && _hasProtectedCode)
         || (_singleType==Pic::MemoryEeprom && _hasProtectedEeprom) ) {
      log(Log::SoftError, i18n("This memory range is programming protected."));
      return false;
    }
    addProgress(OPERATION_DATA[Erase].nbSteps);
    ok = programRange(_singleType, pmemory.arrayForWriting(_singleType));
    if (ok) ok = verifyRange(_singleType, pmemory, Programmer::NormalVerify);
  } else {
    log(Log::Information, i18n("  Erasing device"));
    if (eprom) addProgress(OPERATION_DATA[Erase].nbSteps);
    else {
      ok = ( !Config::preserveEeprom() || preserveEeprom() );
      if (ok) ok = eraseAll();
    }
    if (ok) ok = programAll(pmemory);
  }
  if ( !restoreBandGapBits() ) return false;
  return ok;
}

//-----------------------------------------------------------------------------
bool Programmer::PicBase::checkProgramCalibration(const Device::Array &data)
{
  QString message, s = prettyCalibration(data);
  if ( !device()->checkCalibration(data, &message) ) {
    detailedSorry(i18n("The calibration word %1 is not valid.").arg(s), message);
    return false;
  }
  return askContinue(i18n("Do you want to overwrite the device calibration with %1?").arg(s));
}

bool Programmer::PicBase::tryProgramCalibration(const Device::Array &data, bool &success)
{
  log(Log::Information, i18n("  Write memory: %1").arg(i18n(Pic::MEMORY_RANGE_TYPE_DATA[Pic::MemoryCal].label)));
  success = true;
  if ( !specific()->write(Pic::MemoryCal, data, true) ) return false;
  Device::Array read = specific()->read(Pic::MemoryCal);
  for (uint i=0; i<data.count(); i++)
    if ( data[i]!=read[i] ) success = false;
  if ( !success ) return true;
  if ( device()->isPresent(Pic::MemoryCalBackup) ) {
    read = specific()->read(Pic::MemoryCalBackup);
    if ( device()->checkCalibration(read) ) return true; // do not overwrite correct backup value
    log(Log::Information, i18n("  Write memory: %1").arg(i18n(Pic::MEMORY_RANGE_TYPE_DATA[Pic::MemoryCalBackup].label)));
    if ( !specific()->write(Pic::MemoryCalBackup, data, true) ) return false;
    read = specific()->read(Pic::MemoryCalBackup);
    for (uint i=0; i<data.count(); i++)
      if ( data[i]!=read[i] ) success = false;
  }
  return true;
}

bool Programmer::PicBase::internalProgramCalibration(const Device::Array &data)
{
  initProgramming(Pic::VddWrite);

  // try without erase
  bool success;
  if ( !tryProgramCalibration(data, success) ) return false;
  if (success) return true;
  if ( !askContinue(i18n("Programming calibration data needs a chip erase. Other memory ranges will be restored.")) ) {
    logUserAbort();
    return false;
  }

  // read all, erase and program all
  if ( !readRange(Pic::MemoryUserId, *_deviceMemory) ) return false;
  if ( !preserveCode() ) return false;
  if ( !preserveEeprom() ) return false;
  if ( !readRange(Pic::MemoryCode,   *_deviceMemory) ) return false;
  if ( !readRange(Pic::MemoryEeprom, *_deviceMemory) ) return false;
  log(Log::Information, i18n("  Erasing device"));
  bool ok = specific()->erase(_hasProtectedCode || _hasProtectedEeprom);
  if (ok) ok = programAll(*_deviceMemory);
  if ( !restoreBandGapBits() ) return false;
  if ( !ok ) return false;

  // retry
  if ( !tryProgramCalibration(data, success) ) return false;
  return success;
}

bool Programmer::PicBase::programCalibration(const Pic::Memory &memory)
{
  prepare(Connect);
  emit actionMessage(i18n("Programming calibration..."));
  bool ok = doProgramCalibration(memory.arrayForWriting(Pic::MemoryCal));
  endProgramming();
  return ok;
}

bool Programmer::PicBase::doProgramCalibration(const Device::Array &data)
{
  if ( !checkProgramCalibration(data) ) return false;
  if ( !doConnectDevice() ) return false;
  log(Log::Information, i18n("Programming calibration..."));
  if ( !internalProgramCalibration(data) ) return false;
  log(Log::Information, i18n("Programming calibration successful"));
  return true;
}

//-----------------------------------------------------------------------------
bool Programmer::PicBase::verifySingle(Pic::MemoryRangeType type, const Pic::Memory &memory)
{
  _singleType = type;
  bool ok = verify(memory);
  _singleType = Pic::Nb_MemoryRangeTypes;
  return ok;
}

bool Programmer::PicBase::blankCheckSingle(Pic::MemoryRangeType type)
{
  _singleType = type;
  bool ok = blankCheck();
  _singleType = Pic::Nb_MemoryRangeTypes;
  return ok;
}

bool Programmer::PicBase::verifyRange(Pic::MemoryRangeType type, const Pic::Memory &memory,
                                      VerifyActions actions)
{
  if ( !device()->isPresent(type) || !device()->isWritable(type) ) return true;
  log(Log::Information, i18n("  Verify memory: %1").arg(i18n(Pic::MEMORY_RANGE_TYPE_DATA[type].label)));
  RangeVector rv;
  if ( !(actions & IgnoreProtectedVerify) ) {
    rv = memory.protectedRanges(Pic::Config::ProgramProtected, type);
    if ( !rv.isEmpty() ) log(Log::Warning, i18n("  Part of device memory is protected (in %1) and cannot be verified.")
                             .arg(i18n(Pic::MEMORY_RANGE_TYPE_DATA[type].label)));
  }
  return specific()->verify(type, memory, actions, rv);
}

bool Programmer::PicBase::verifyAll(const Pic::Memory &memory, VerifyActions actions)
{
  if ( !verifyRange(Pic::MemoryConfig, memory, actions) ) return false;
  if ( !verifyRange(Pic::MemoryUserId, memory, actions) ) return false;
  if ( !verifyRange(Pic::MemoryCode,   memory, actions) ) return false;
  if ( !verifyRange(Pic::MemoryEeprom, memory, actions) ) return false;
  return true;
}

bool Programmer::PicBase::internalVerify(const Device::Memory &memory, VerifyActions actions)
{
  initProgramming(Pic::VddRead);
  const Pic::Memory &pmemory = static_cast<const Pic::Memory &>(memory);
  if ( _singleType!=Pic::Nb_MemoryRangeTypes ) {
    if ( _singleType==Pic::MemoryCal ) {
      if ( !verifyRange(Pic::MemoryCal, pmemory, actions) ) return false;
      return verifyRange(Pic::MemoryCalBackup, pmemory, actions);
    }
    return verifyRange(_singleType, pmemory, actions);
  }
  return verifyAll(pmemory, actions);
}
