/***************************************************************************
 *   Copyright (C) 2005-2006 Nicolas Hadacek <hadacek@kde.org>             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#ifndef PIC_PROG_H
#define PIC_PROG_H

#include "common/log.h"
#include "port/port_base.h"
#include "progs/base/generic_prog.h"
#include "pic_memory.h"

namespace Programmer
{
enum Result { Error = 0, Ok, NotImplemented };

//-----------------------------------------------------------------------------
class PicDeviceSpecific : public DeviceSpecific
{
public:
  PicDeviceSpecific(Programmer::Base &base) : DeviceSpecific(base) {}
  const Pic::Data &device() const { return static_cast<const Pic::Data &>(*_base.device()); }
  Result eraseRange(Pic::MemoryRangeType type);
  bool erase(bool isProtected);
  Device::Array read(Pic::MemoryRangeType type);
  bool write(Pic::MemoryRangeType type, const Device::Array &data, bool force);
  bool verify(Pic::MemoryRangeType type, const Pic::Memory &memory,
              VerifyActions actions, const RangeVector &rv);
  bool compareWords(Pic::MemoryRangeType type, uint index, uint v, uint d, VerifyActions actions);
  bool verifyWord(uint index, uint word, Pic::MemoryRangeType type,
                  const Pic::Memory &memory, VerifyActions actions, const RangeVector &rv);
  uint findNonMaskStart(Pic::MemoryRangeType type, const Device::Array &data) const;
  uint findNonMaskEnd(Pic::MemoryRangeType type, const Device::Array &data) const;

protected:
  virtual bool doErase(bool isProtected) = 0;
  virtual Result doEraseCode() { return NotImplemented; }
  virtual Result doEraseEeprom() { return NotImplemented; }
  virtual Result doEraseUserId() { return NotImplemented; }
  virtual Result doEraseConfig();
  virtual Device::Array doRead(Pic::MemoryRangeType type) = 0;
  virtual bool doWrite(Pic::MemoryRangeType type, const Device::Array &data, bool force) = 0;
  virtual bool doVerify(Pic::MemoryRangeType type, const Pic::Memory &memory,
                        VerifyActions actions, const RangeVector &rv);
};

//-----------------------------------------------------------------------------
class PicHardware : public Hardware
{
public:
  PicHardware(Programmer::Base &base, Port::Base *port) : Hardware(base, port) {}
  const Pic::Data &device() const { return static_cast<const Pic::Data &>(*_base.device()); }
};

//-----------------------------------------------------------------------------
class PicBase : public Base
{
public:
  PicBase(const Group &group, const Pic::Data *data, const char *name);
  virtual ~PicBase();
  const Pic::Data *device() const { return static_cast<const Pic::Data *>(_device); }
  bool eraseSingle(Pic::MemoryRangeType type);
  bool readSingle(Pic::MemoryRangeType type, Pic::Memory &memory);
  bool programSingle(Pic::MemoryRangeType type, const Pic::Memory &memory);
  bool verifySingle(Pic::MemoryRangeType type, const Pic::Memory &memory);
  bool blankCheckSingle(Pic::MemoryRangeType type);
  bool programCalibration(const Pic::Memory &memory);

protected:
  PicDeviceSpecific *specific() const { return static_cast<PicDeviceSpecific *>(_specific); }
  virtual double vdd() const { return Programmer::UNKNOWN_VOLTAGE; }
  virtual double vpp() const { return Programmer::UNKNOWN_VOLTAGE; }
  virtual bool verifyDeviceId();
  virtual uint nbSteps(Operation operation) const;
  bool initProgramming(Pic::ProgVoltageType type);
  virtual bool initProgramming() { return true; }

  virtual bool checkErase();
  virtual bool internalErase();

  virtual bool checkRead();
  virtual bool internalRead(Device::Memory &memory);
  bool readRange(Pic::MemoryRangeType type, Pic::Memory &memory);

  virtual bool checkProgram(const Device::Memory &memory);
  virtual bool internalProgram(const Device::Memory &memory);
  virtual bool programAll(const Pic::Memory &memory);
  bool programAndVerifyRange(Pic::MemoryRangeType type, const Pic::Memory &memory);
  bool programRange(Pic::MemoryRangeType type, const Device::Array &array);

  virtual bool checkVerify() { return checkRead(); }
  virtual bool internalVerify(const Device::Memory &memory, VerifyActions actions);
  virtual bool verifyAll(const Pic::Memory &memory, VerifyActions actions);
  bool verifyRange(Pic::MemoryRangeType type, const Pic::Memory &memory,
                   VerifyActions actions);

private:
  Pic::MemoryRangeType _singleType;
  Pic::Memory         *_deviceMemory;
  bool _hasProtectedCode, _hasProtectedEeprom;

  uint readDeviceId();
  bool eraseAll();
  bool eraseRange(Pic::MemoryRangeType type);
  bool saveCalibration();
  bool restoreCalibration();
  bool restoreBandGapBits();
  bool doProgramCalibration(const Device::Array &data);
  bool checkProgramCalibration(const Device::Array &data);
  bool internalProgramCalibration(const Device::Array &data);
  QString prettyCalibration(const Device::Array &data) const;
  bool tryProgramCalibration(const Device::Array &data, bool &success);
  bool preserveCode();
  bool preserveEeprom();
  bool internalEraseRange(Pic::MemoryRangeType type);
};

}; // namespace

#endif
