/***************************************************************************
 *   Copyright (C) 2005-2006 Nicolas Hadacek <hadacek@kde.org>             *
 *   Copyright (C) 2003-2004 Alain Gibaud <alain.gibaud@free.fr>           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "editor_manager.h"

#include <qiconset.h>
#include <qdragobject.h>
#include <qpainter.h>

#include <klocale.h>
#include <kiconloader.h>
#include <kpopupmenu.h>
#include <kaction.h>

#include "text_editor.h"
#include "hex_editor.h"
#include "object_view.h"
#include "new_dialogs.h"
#include "toplevel.h"
#include "devices/base/device_list.h"

//-----------------------------------------------------------------------------
void EditorTabBar::paintLabel(QPainter *p, const QRect &br, QTab *t, bool has_focus) const
{
  QFont f = p->font();
  f.setItalic(_readOnly[t]);
  p->setFont(f);
  QTabBar::paintLabel(p, br, t, has_focus);
}

//-----------------------------------------------------------------------------
EditorManager::EditorManager(QWidget *parent)
  : QTabWidget(parent, "editor_manager"), _current(0)
{
  setTabBar(new EditorTabBar(this));
  tabBar()->installEventFilter(this);
  connect(this, SIGNAL(currentChanged(QWidget *)), SLOT(showEditor(QWidget *)));
}

bool EditorManager::openFile(const PURL &url)
{
  if ( url.isEmpty() ) return false;
  Editor *e = findEditor(url);
  if (e) { // document already loaded
    if ( !MessageBox::askContinue(i18n("File \"%1\" already loaded. Reload?").arg(url.kurl().prettyURL()),
                                 i18n("Warning"), i18n("Reload")) ) return true;
    if ( !e->slotLoad() ) {
      closeEditor(e, false);
      return false;
    }
    return true;
  }
  if ( !openEditor(url) ) {
    static_cast< KRecentFilesAction &>(MainWindow::action("file_open_recent")).removeURL(url.kurl());
    return false;
  }
  static_cast<KRecentFilesAction &>(MainWindow::action("file_open_recent")).addURL(url.kurl());
  return true;
}

void EditorManager::connectEditor(Editor *editor)
{
    disconnectEditor(currentEditor());
    if ( editor==0 ) return;
    for (uint i=0; Editor::actionName(i); i++) {
      const Editor::ActionData &d = Editor::ACTION_DATA[i];
      if ( d.capability!=Editor::NoCapability && !(d.capability & editor->capabilities()) ) continue;
      connect(&MainWindow::action(Editor::actionName(i)), SIGNAL(activated()), editor, d.slot);
    }
    connect(editor, SIGNAL(modified()), SLOT(modifiedSlot()));
    connect(editor, SIGNAL(guiChanged()), SIGNAL(guiChanged()));
    connect(editor, SIGNAL(dropEventPass(QDropEvent *)), SLOT(slotDropEvent(QDropEvent *)));
    connect(editor, SIGNAL(statusTextChanged(const QString &)), SIGNAL(statusChanged(const QString &)));
}

void EditorManager::modifiedSlot()
{
  emit modified(currentEditor()->url());
}

void EditorManager::disconnectEditor(Editor *editor)
{
  if ( editor==0 ) return;
  for (uint i=0; Editor::actionName(i); i++)
    MainWindow::action(Editor::actionName(i)).disconnect(editor);
  editor->disconnect(this);
}

QString EditorManager::title(const Editor *e) const
{
  return (e->url().isEmpty() ? "<" + e->name() + ">" : e->url().filename());
}

void EditorManager::updateDisplay()
{
  // update titles
  QValueList<Editor *>::iterator it = _editors.begin();
  for (; it!=_editors.end(); ++it) {
    static_cast<EditorTabBar *>(tabBar())->setReadOnly(indexOf(*it), (*it)->isReadOnly());
    KIconLoader loader;
    QIconSet iset = loader.loadIconSet("filesave", KIcon::Small, 0), nulliset;
    changeTab(*it, (*it)->isModified() ? iset : nulliset, title(*it));
  }

  // update actions
  for (uint i=0; Editor::actionName(i); i++) {
    const Editor::ActionData &d = Editor::ACTION_DATA[i];
    KAction &a = MainWindow::action(Editor::actionName(i));
    if ( _current==0 ) a.setEnabled(false);
    else if ( d.capability!=Editor::NoCapability
              && !(d.capability & currentEditor()->capabilities()) ) a.setEnabled(false);
    else {
      bool on = _current->state(d.types);
      if (d.toggle) {
        a.setEnabled(true);
        static_cast<KToggleAction &>(a).setChecked(on);
      } else a.setEnabled(on);
    }
  }
}

bool EditorManager::closeCurrentEditor()
{
  if ( !closeEditor(_current, true) ) return false;
  emit guiChanged();
  return true;
}

bool EditorManager::closeAllEditors()
{
  if ( currentEditor()==0 ) return true;
  while ( currentEditor() )
    if ( !closeEditor(currentEditor(), true) ) break;
  emit guiChanged();
  return ( currentEditor()==0 );
}

bool EditorManager::closeAllOtherEditors()
{
  if ( nbEditors()==1 ) return true;
  QValueList<Editor *> list = _editors;
  list.remove(currentEditor());
  QValueList<Editor *>::iterator it = list.begin();
  bool ok = true;
  for (; it!=list.end(); ++it) {
    if ( !closeEditor(*it, true) ) {
      ok = false;
      break;
    }
  }
  emit guiChanged();
  return ok;
}

bool EditorManager::closeEditor(const PURL &url)
{
  Editor *e = findEditor(url);
  if ( e==0 ) return true;
  if ( !closeEditor(e, true) ) return false;
  emit guiChanged();
  return true;
}

bool EditorManager::closeEditor(Editor *e, bool ask)
{
  if ( e==0 ) return true;
  if ( ask && !e->checkSaved() ) return false;
  removePage(e);
  _editors.remove(e);
  Editor *g = static_cast<Editor *>(currentPage());
  changeToEditor(g);
  delete e;
  return true;
}

void EditorManager::showEditor(Editor *e)
{
  changeToEditor(e);
  emit guiChanged();
}

void EditorManager::changeToEditor(Editor *e)
{
  if ( e==_current ) return;
  connectEditor(e);
  _current = e;
  if ( e==0 ) return;
  showPage(e);
  e->clearFocus();   // force a got focus signal
  e->setFocus();
  e->statusChanged();
}

Editor *EditorManager::findEditor(const QString &name)
{
  QValueList<Editor *>::iterator it = _editors.begin();
  for (; it!=_editors.end(); ++it) if ( (*it)->name()==name ) return *it;
  return 0;
}

Editor *EditorManager::findEditor(const PURL &url)
{
  QValueList<Editor *>::iterator it = _editors.begin();
  for (; it!=_editors.end(); ++it) if ( (*it)->url()==url ) return *it;
  return 0;
}

Editor *EditorManager::createEditor(PURL::FileType type, const PURL &url)
{
  Editor *e = findEditor(url);
  if (e) closeEditor(e, false);
  switch (type) {
    case PURL::Hex: {
      HexEditor *he = new HexEditor(url, this);
      he->setDevice(MainWindow::device());
      return he;
    }
    case PURL::CSource:
    case PURL::CHeader:
    case PURL::AsmGPAsm:
    case PURL::AsmPIC30:
    case PURL::AsmPICC:
    case PURL::Inc: return new KBEditor(Editor::Breakpoints, url, this);
    case PURL::Lkr:
    case PURL::Gld:
    case PURL::Map:
    case PURL::Lst:
    case PURL::Rst:
    case PURL::Sym:
    case PURL::Mem:
    case PURL::Adb:
    case PURL::Lnk:
    case PURL::SDCC_P:
    case PURL::SDCC_D:
    case PURL::Cod:
    case PURL::Hxl:
    case PURL::Sdb:
    case PURL::Rlf:
    case PURL::Unknown:
      return new KBEditor(Editor::LoadSave, url, this);
    case PURL::Coff:
      return new Coff::Editor(url, *MainWindow::deviceData(), this);
    case PURL::Object:
    case PURL::Elf:
    case PURL::Library:
    case PURL::Project:
    case PURL::PikdevProject:
    case PURL::Nb_FileTypes: break;
  }
  return 0;
}

void EditorManager::addEditor(Editor *e)
{
  QValueList<Editor *>::iterator it = _editors.begin();
  for (; it!=_editors.end(); ++it) if ( *it==e ) return;
  _editors.append(e);
  addTab(e, QString::null);
  setTabEnabled(e, true);
  showEditor(e);
}

void EditorManager::slotDropEvent(QDropEvent *event)
{
  QStringList urls;
  if ( !QUriDrag::decodeLocalFiles(event, urls)) return;
  QStringList::const_iterator it = urls.begin();
  for(; it!=urls.end(); ++it) openEditor(PURL(*it));
}

Editor *EditorManager::openEditor(const PURL &url)
{
  Editor *e = findEditor(url);
  if ( e==0 ) {
    e = createEditor(url.fileType(), url);
    if ( e==0 ) return 0;
    if ( !e->slotLoad() ) {
      closeEditor(e, false);
      return 0;
    }
    addEditor(e);
  } else showEditor(e);
  return e;
}

void EditorManager::saveAllFiles()
{
  QValueList<Editor *>::iterator it = _editors.begin();
  for (; it!=_editors.end(); ++it) {
    if ( !(*it)->isModified() ) continue;
    (*it)->save();
  }
  updateDisplay();
  emit guiChanged();
}

PURL::List EditorManager::files() const
{
  PURL::List names;
  QValueList<Editor *>::const_iterator it = _editors.begin();
  for(; it!=_editors.end(); ++it) {
    if ( (*it)->url().isEmpty() ) continue;
    names.push_back((*it)->url());
  }
  return names;
}

bool EditorManager::eventFilter(QObject *watched, QEvent *e)
{
  if ( watched!=tabBar() || e->type()!=QEvent::ContextMenu ) return false;
  QContextMenuEvent *ce = static_cast<QContextMenuEvent *>(e);
  QPoint p = tabBar()->mapFromGlobal(ce->globalPos());
  QTab *tab = tabBar()->selectTab(p);
  if ( tab==0 ) return false;

  KIconLoader loader;
  QPixmap closeIcon = loader.loadIcon("fileclose", KIcon::Small);
  QPixmap saveIcon = loader.loadIcon("filesave", KIcon::Small);
  QPixmap saveAsIcon = loader.loadIcon("filesaveas", KIcon::Small);
  QPixmap reloadIcon = loader.loadIcon("reload", KIcon::Small);
  KPopupMenu *popup = new KPopupMenu;

  for (int i=0; i<tabBar()->count(); i++)
    if ( tabBar()->tabAt(i)==tab ) {
      Editor *editor = static_cast<Editor *>(page(i));
      popup->insertTitle(title(editor));
      popup->insertItem(closeIcon, i18n("Close"), 1);
      if ( nbEditors()>1 ) popup->insertItem(i18n("Close All Others"), 2);
      if ( editor->capabilities() & Editor::LoadSave ) {
        if ( editor->isModified() ) popup->insertItem(saveIcon, i18n("Save"), 3);
        popup->insertItem(saveAsIcon, i18n("Save As..."), 4);
        if ( !editor->url().isEmpty() ) popup->insertItem(reloadIcon, i18n("Reload"), 5);
      }
      switch (popup->exec(ce->globalPos())) {
        case 1: closeEditor(editor, true); break;
        case 2: closeAllOtherEditors(); break;
        case 3: editor->save(); break;
        case 4: editor->saveAs(); break;
        case 5: editor->reload(); break;
      }
      emit guiChanged();
    }
  delete popup;
  return true;
}

void EditorManager::clearMarks(KBEditor::MarkTypes types)
{
  QValueList<Editor *>::iterator it = _editors.begin();
  for (; it!=_editors.end(); ++it) {
    KBEditor *e = ::qt_cast<KBEditor *>(*it);
    if (e) e->clearMarks(types);
  }
}
