/***************************************************************************
 *   Copyright (C) 2006 Nicolas Hadacek <hadacek@kde.org>                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "object_view.h"

#include <qregexp.h>
#include <ktempfile.h>
#include <kate/document.h>
#include <klocale.h>

#include "devices/base/device_list.h"
#include "tools/tool.h"
#include "tools/process.h"
#include "tools/gputils.h"
#include "tools/gputils_config.h"
#include "common/gui/misc_gui.h"
#include "hex_editor.h"

//-----------------------------------------------------------------------------
PURL::List Coff::includedFiles(const PURL &url)
{
  Log::ConsoleView cview;
  SymbolParser parser;
  parser.setView(&cview);
  if ( parser.parse(url) ) return parser.includedFiles();
  return PURL::List();
}

QString Coff::device(const PURL &url)
{
  Log::ConsoleView cview;
  FileHeaderParser parser;
  parser.setView(&cview);
  if ( !parser.parse(url) ) return parser.device();
  return QString::null;
}

//-----------------------------------------------------------------------------
bool Coff::ParserBase::parse(const PURL &url)
{
  _url = url;
  if ( !url.probablyExists() ) return false;
  internalParse();
  return true;
}

QString Coff::ParserBase::content(const PURL &url, ContentType type) const
{
  // ### FIXME
  const Tool::Group *group = Tool::lister().group("gputils");
  const Tool::Base *tool = group->base(Tool::ObjectViewer);
  if ( !tool->check() ) return QString::null;
  QStringList args;
  switch (type) {
    case Sections:   args += "-s"; break;
    case Symbols:    args += "-t"; break;
    case FileHeader: args += "-f"; break;
  }
  args += url.filepath();
  QString s = tool->run(tool->executable(), args, group->withWine());
  return s;
}

//-----------------------------------------------------------------------------
// from gputils (0.13)
/* Section contains executable code */
#define STYP_TEXT     0x0020
/* Section contains initialized data */
#define STYP_DATA     0x0040
/* Section contains uninitialized data */
#define STYP_BSS      0x0080
/* Section contains initialized data for ROM */
#define STYP_DATA_ROM 0x0100
/* Section is absolute */
#define STYP_ABS      0x1000
/* Section is shared across banks */
#define STYP_SHARED   0x2000
/* Section is overlaid with other sections of the same name from different
   objects modules */
#define STYP_OVERLAY  0x4000
/* Section is available using access bit */
#define STYP_ACCESS   0x8000
/* Section contains the activation record for a function */
#define STYP_ACTREC   0x10000
/* Section has been relocated.  This is a temporary flag used by the linker */
#define STYP_RELOC    0x20000

Coff::SectionParser::SectionParser(const Device::Data &device)
  : ParserBase("coff_section_parser"), _device(device)
{}

bool Coff::SectionParser::parseData(QStringList::const_iterator &it, Section &section)
{
  for (;;) {
    ++it;
    if ( (*it).isNull() ) break;
    //qDebug("parse data: %s", (*it).latin1());
    if ( (*it).stripWhiteSpace().isEmpty() ) break;
    QRegExp re("([a-f\\d]+):\\s+([a-f\\d]+)\\s+(.*)");
    if ( !re.exactMatch(*it) ) {
      log(Log::Error, i18n("Could not parse data line \"%1\".").arg(*it));
      return false;
    }
    CodeLine cl;
    cl.address = fromHex(re.cap(1));
    cl.opcode = re.cap(2).stripWhiteSpace();
    cl.disasm = re.cap(3).stripWhiteSpace();
    section.codeLines.append(cl);
  }
  return true;
}

bool Coff::SectionParser::parseLines(QStringList::const_iterator &it, Section &section)
{
  //qDebug("Parse line info lines (%i)", section.codeLines.count());
  ++it; // skip header line
  for (uint i=0; i<section.codeLines.count(); i++) {
    ++it;
    QRegExp re("(\\d+)\\s+((?:0|0x[a-f\\d]+))\\s+(.*)");
    if ( !re.exactMatch(*it) ) {
      log(Log::Error, i18n("Could not parse line info line \"%1\".").arg(*it));
      return false;
    }
    section.codeLines[i].line = re.cap(1).toUInt();
    uint address = 0;
    if ( re.cap(2)!="0" ) address = fromHexLabel(re.cap(2), 0);
    if ( section.codeLines[i].address!=address ) {
      log(Log::Error, i18n("Data and line info addresses do not match."));
      return false;
    }
    section.codeLines[i].filename = re.cap(3).stripWhiteSpace();
  }
  return true;
}

bool Coff::SectionParser::parseSection(const QStringList &lines, QStringList::const_iterator &it)
{
  if ( (*it)!="Section Header" ) {
    log(Log::Error, i18n("Could not find section start."));
    return false;
  }
  ++it; // skip section header
  //qDebug("-- section start ------------");
  Section section;
  uint nbDataLines = 0;
  for (; it!=lines.end(); ++it) {
    //qDebug("parse line: %s", (*it).latin1());
    if ( (*it).stripWhiteSpace().isEmpty() ) continue;
    if ( (*it)=="Section Header" ) break;
    if ( (*it).startsWith(" ") ) continue; // flag description
    // name
    QRegExp re("Name(\\s+)(.*)");
    if ( re.exactMatch(*it) ) {
      //qDebug("name");
      section.name = re.cap(2);
      if ( section.name==".config" ) section.type = Section::Config;
      else if ( section.name==".devid" ) section.type = Section::DevId;
      else if ( section.name==".idlocs" ) section.type = Section::IdLocs;
      continue;
    }
    // address
    re.setPattern("Address(\\s+)(.*)");
    if ( re.exactMatch(*it) ) {
      //qDebug("address");
      if ( re.cap(2)=="0" ) section.address = 0;
      else {
        bool ok;
        section.address = fromHexLabel(re.cap(2), &ok);
        if ( !ok ) {
          log(Log::Error, i18n("Unrecognized address field \"%1\".").arg(re.cap(2)));
          return false;
        }
      }
      continue;
    }
    // flags
    re.setPattern("Flags(\\s+)(.*)");
    if ( re.exactMatch(*it) ) {
      //qDebug("flags");
      bool ok;
      section.flags = fromHexLabel(re.cap(2), &ok);
      if ( !ok ) {
        log(Log::Error, i18n("Unrecognized flags field."));
        return false;
      }
      if ( section.type==Section::Unknown ) {
        if ( section.flags & STYP_TEXT ) section.type = Section::Code;
        else if ( section.flags & STYP_BSS ) section.type = Section::UData;
        else if ( section.flags & STYP_DATA ) section.type = Section::Data;
      }
      continue;
    }
    // nb lines
    re.setPattern("Number of Line Numbers(\\s+)(.*)");
    if ( re.exactMatch(*it) ) {
      //qDebug("nb lines");
      bool ok;
      nbDataLines = re.cap(2).toUInt(&ok);
      if ( !ok ) {
        log(Log::Error, i18n("Unrecognized nb lines field."));
        return false;
      }
      continue;
    }
    // data
    if ( (*it)=="Data" ) {
      //qDebug("data");
      if ( !parseData(it, section) ) return false;
      continue;
    }
    // lines
    if ( (*it)=="Line Number Table" ) {
      //qDebug("line number");
      if ( !parseLines(it, section) ) return false;
      continue;
    }
    //qDebug("Unknown line \"%s\"", (*it).latin1());
  }
  if ( section.type==Section::Unknown ) {
    log(Log::Error, i18n("Unknown section name=\"%1\".").arg(section.name));
    return false;
  }
  if ( section.name.isEmpty() ) {
    log(Log::Error, i18n("Section has no name."));
    return false;
  }
  if ( section.type==Section::Code && nbDataLines!=section.codeLines.count() ) {
    log(Log::Error, i18n("Nb lines different from parsed code lines."));
    return false;
  }
  _sections.append(section);
  return true;
}

void Coff::SectionParser::internalParse()
{
  _sections.clear();
  QString text = content(url(), Sections);

  // parse sections
  QStringList lines = QStringList::split('\n', text, true);
  QStringList::const_iterator it = lines.begin();
  for (; it!=lines.end();) {
    //qDebug("--> start %s", (*it).latin1());
    if ( !parseSection(lines, it) && it!=lines.end() ) ++it;
  }

  // compute urls
  for (uint i=0; i<_sections.count(); i++) {
    if ( _sections[i].type!=Section::Code ) continue;
    for (uint k=0; k<_sections[i].codeLines.count(); k++) {
      QString filename = _sections[i].codeLines[k].filename;
      PURL furl(filename);
      if ( furl.isRelative() ) furl = PURL(url().directory(), furl.filepath());
      _sections[i].codeLines[k].url = furl;
    }
  }

  // generate text
  generateText();
}

QValueVector<Coff::SectionParser::CodeLine> Coff::SectionParser::findCodeLines(const PURL &url, int line) const
{
  QValueVector<CodeLine> list;
  for (uint i=0; i<_sections.count(); i++) {
    if ( _sections[i].type!=Section::Code ) continue;
    for (uint k=0; k<_sections[i].codeLines.count(); k++) {
      const CodeLine &cl = _sections[i].codeLines[k];
      if ( cl.url==url && (line==-1 || int(cl.line)==line) ) list.append(cl);
    }
  }
  return list;
}

int Coff::SectionParser::lineForAddress(const PURL &purl, uint address) const
{
  if ( purl==url() && _lines.contains(address) ) return _lines[address]-1;
  QValueVector<CodeLine> lines = findCodeLines(purl, -1);
  for (uint i=0; i<lines.count(); i++)
    if ( lines[i].address==address ) return lines[i].line-1;
  return -1;
}

PURL Coff::SectionParser::urlForAddress(uint address) const
{
  for (uint i=0; i<_sections.count(); i++) {
    if ( _sections[i].type!=Section::Code ) continue;
    for (uint k=0; k<_sections[i].codeLines.count(); k++) {
      if ( _sections[i].codeLines[k].address!=address ) continue;
      return _sections[i].codeLines[k].url;
    }
  }
  return url();
}

QValueVector<uint> Coff::SectionParser::addresses(const PURL &purl, uint line) const
{
  QValueVector<uint> ad;
  if ( purl==url() ) {
    QMap<uint, uint>::const_iterator it = _lines.begin();
    for (; it!=_lines.end(); ++it)
      if ( line+1==it.data() ) ad.append(it.key());
    return ad;
  }
  for (uint i=0; i<_sections.count(); i++) {
    if ( _sections[i].type!=Section::Code ) continue;
    for (uint k=0; k<_sections[i].codeLines.count(); k++) {
      if ( purl!=_sections[i].codeLines[k].url ) continue;
      if ( line+1!=_sections[i].codeLines[k].line ) continue;
      ad.append(_sections[i].codeLines[k].address);
    }
  }
  return ad;
}

QString Coff::SectionParser::replaceDisasmTab(const QString &s)
{
  int i = s.find('\t');
  if ( i==-1 ) return s;
  return s.mid(0, i).leftJustify(8) + s.mid(i+1);
}

class CodeData {
public:
  QString address, opcode, disasm;
};
class LineData {
public:
  QValueVector<CodeData> codes;
  QString lineNumber, lineText;
};
class FileData {
public:
  QValueVector<LineData> lines;
};

void Coff::SectionParser::generateText()
{
  // open and read files
  QMap<PURL, FileData> fd;
  for (uint i=0; i<_sections.count(); i++) {
    if ( _sections[i].type!=Section::Code ) continue;
    for (uint k=0; k<_sections[i].codeLines.count(); k++) {
      PURL url = _sections[i].codeLines[k].url;
      if ( fd.contains(url) ) continue;
      FileData fdata;
      PFile file(url, Log::DontShow);
      if ( file.openForRead() ) {
        QStringList lines = file.readLines();
        fdata.lines.resize(lines.count());
        for (uint i=0; i<lines.count(); i++)
          fdata.lines[i].lineText = lines[i];
      }
      fd[url] = fdata;
    }
  }

  // create strings (for later justification)
  uint addressWidth = _device.nbCharsAddress();
  uint opcodeWidth = 0, disasmWidth = 0, lineNumberWidth = 0, lineTextWidth = 0;
  QMap<PURL, FileData>::iterator it = fd.begin();
  for (; it!=fd.end(); ++it) {
    for (uint i=0; i<it.data().lines.count(); i++) {
      LineData &ldata = it.data().lines[i];
      QValueVector<CodeLine> lines = findCodeLines(it.key(), i+1);
      ldata.codes.resize(lines.count());
      for (uint k=0; k<lines.count(); k++) {
        ldata.codes[k].address = toHex(lines[k].address, _device.nbCharsAddress()).upper();
        ldata.codes[k].opcode = lines[k].opcode.upper();
        opcodeWidth = QMAX(opcodeWidth, ldata.codes[k].opcode.length());
        ldata.codes[k].disasm = replaceDisasmTab(lines[k].disasm);
        disasmWidth = QMAX(disasmWidth, ldata.codes[k].disasm.length());
      }
      ldata.lineNumber = QString::number(i+1);
      lineNumberWidth = QMAX(lineNumberWidth, ldata.lineNumber.length());
      lineTextWidth = QMAX(lineTextWidth, ldata.lineText.length());
    }
  }
  uint spacingWidth = 4;
  uint asmWidth = addressWidth + spacingWidth + opcodeWidth
    + spacingWidth + disasmWidth;
  uint totalWidth = asmWidth + spacingWidth + lineNumberWidth + 2 + lineTextWidth;
  QString spacing = repeat(" ", spacingWidth);

  // create text
  _lines.clear();
  uint line = 1;
  _text = QString::null;
  for (it = fd.begin(); it!=fd.end(); ++it) {
    QString s = QString(";---  ") + it.key().pretty() + "  ";
    _text +=  s.leftJustify(totalWidth, '-') + '\n';
    line++;
    for (uint i=0; i<it.data().lines.count(); i++) {
      const LineData &ldata = it.data().lines[i];
      if ( ldata.codes.count()==0 ) _text += repeat(" ", asmWidth);
      else {
        for (uint k=0; k<ldata.codes.count(); k++) {
          if ( k!=0 ) {
            _text += '\n';
            line++;
          }
          _text += ldata.codes[k].address + spacing;
          _text += ldata.codes[k].opcode.leftJustify(opcodeWidth) + spacing;
          _text += ldata.codes[k].disasm.leftJustify(disasmWidth);
          _lines[fromHex(ldata.codes[k].address)] = line;
        }
      }
      _text += spacing + ldata.lineNumber.leftJustify(lineNumberWidth) + ":";
      if ( !ldata.lineText.stripWhiteSpace().isEmpty() )
        _text += " " + ldata.lineText;
      _text += '\n';
      line++;
    }
  }
}

//-----------------------------------------------------------------------------
Coff::SymbolParser::SymbolParser()
  : ParserBase("coff_symbol_parser")
{}

bool Coff::SymbolParser::parseFile(QStringList::const_iterator &it)
{
  ++it;
  PURL url;
  for (; ; ++it) {
    if ( !(*it).startsWith("  ") ) {
      --it;
      break;
    }
    QRegExp re("  file = (.*)");
    if ( re.exactMatch(*it) ) {
      url = PURL(re.cap(1));
      continue;
    }
    re.setPattern("  line included = (\\d+)");
    if ( re.exactMatch(*it) ) {
      if ( re.cap(1).toUInt()==0 ) url = PURL(); // not included
      continue;
    }
  }
  if ( !url.isEmpty() ) _includedFiles += url;
  return true;
}

void Coff::SymbolParser::internalParse()
{
  _includedFiles.clear();
  QString text = content(url(), Symbols);

  QStringList lines = QStringList::split('\n', text, true);
  QStringList::const_iterator it = lines.begin();
  if ( (*it)!="Symbol Table" ) {
    log(Log::Error, i18n("Could not find symbol table start."));
    return;
  }
  for ( ; it!=lines.end(); ++it)
    if ( (*it).startsWith(".file") ) parseFile(it);
}

//-----------------------------------------------------------------------------
Coff::FileHeaderParser::FileHeaderParser()
  : ParserBase("coff_fileheader_parser")
{}

void Coff::FileHeaderParser::internalParse()
{
  _device = QString::null;
  QString text = content(url(), FileHeader);
  QStringList lines = QStringList::split('\n', text, true);
  QStringList::const_iterator it = lines.begin();
  for ( ; it!=lines.end(); ++it) {
    QRegExp re("Processor Type\\s+(.*)");
    if ( re.exactMatch(*it) ) _device = re.cap(1).stripWhiteSpace();
  }
}

//-----------------------------------------------------------------------------
Coff::Editor::Editor(const PURL &url, const Device::Data &device, QWidget *parent)
  : SimpleTextEditor(Breakpoints, url, parent), _device(device)
{
  setReadOnly(true);
  setState(AutoWrapOn, false);
}

bool Coff::Editor::load()
{
  Log::ConsoleView cview;
  SectionParser parser(_device);
  parser.setView(&cview);
  if ( !parser.parse(url()) ) return false;
  return setText(parser.text());
}

//-----------------------------------------------------------------------------
DisassemblyEditor::DisassemblyEditor(const PURL &url, const PURL &hexUrl,
                                     const QString &device, QWidget *parent)
  : SimpleTextEditor(Breakpoints, url, parent),
    _tmp(Log::DontShow), _hexUrl(hexUrl), _device(device)
{
  setReadOnly(true);
}

DisassemblyEditor::DisassemblyEditor(const PURL &url, const HexEditor &e,
                                     const QString &device, QWidget *parent)
  : SimpleTextEditor(Breakpoints, url, parent),
    _tmp(Log::DontShow), _device(device)
{
  setReadOnly(true);
  _tmp.openForWrite();
  if ( !e.save(_tmp.url()) ) return;
  _hexUrl = _tmp.url();
}

bool DisassemblyEditor::load()
{
  ::BusyCursor bc;
  // ### FIXME
  const Tool::Group *group = Tool::lister().group("gputils");
  const Tool::Base *tool = group->base(Tool::Disassembler);
  if ( !tool->check(_device) ) return false;
  if ( _device==Device::AUTO_DATA.name ) {
    MessageBox::sorry(i18n("You need to specify the device for disassembling."), Log::Show);
    return false;
  }
  QStringList args;
  args += "-p";
  args += GPUtils::toDeviceName(_device);
  GPUtils::Config *config = static_cast<GPUtils::Config *>(group->createConfig(0));
  if ( config->outputShortFormat() ) args += "-s";
  args += config->customOptions(Tool::Disassembler, _hexUrl.directory());
  args += _hexUrl.filepath();
  delete config;
  QString s = tool->run(tool->executable(), args, group->withWine());
  if ( s.isNull() ) return false;
  return setText(s);
}
