/***************************************************************************
 *   Copyright (C) 2006 Nicolas Hadacek <hadacek@kde.org>                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#ifndef OBJECT_VIEW_H
#define OBJECT_VIEW_H

#include <qvaluevector.h>

#include "text_editor.h"
#include "common/log.h"
namespace Device { class Data; };
namespace Compile { class DisassembleFile; }
class HexEditor;

namespace Coff
{
  extern QString device(const PURL &url);
  extern PURL::List includedFiles(const PURL &url);

//-----------------------------------------------------------------------------
class ParserBase : public Log::Manager
{
public:
  ParserBase(const char *name) : Log::Manager(name) {}
  bool parse(const PURL &url);

protected:
  PURL url() const { return _url; }
  enum ContentType { Sections, Symbols, FileHeader };
  QString content(const PURL &url, ContentType type) const;
  virtual void internalParse() = 0;

private:
  PURL _url;
};

//-----------------------------------------------------------------------------
class SectionParser : public ParserBase
{
public:
  SectionParser(const Device::Data &device);
  QString text() const { return _text; }
  int lineForAddress(const PURL &url, uint address) const; // return -1 if not in file
  PURL urlForAddress(uint address) const;
  QValueVector<uint> addresses(const PURL &url, uint line) const;

private:
  class CodeLine {
  public:
    uint address, line;
    QString opcode, disasm, filename;
    PURL url;
  };

  class Section {
  public:
    enum Type { Config, DevId, IdLocs, UData, Data, Code, Unknown };
    Section() : type(Unknown) {}
    Type    type;
    QString name;
    uint    address;
    uint    flags;
    QValueVector<CodeLine> codeLines;
  };

private:
  const Device::Data &_device;
  QValueVector<Section>  _sections;
  QMap<uint, uint>       _lines; // address -> line
  QString                _text;

  bool parseSection(const QStringList &lines, QStringList::const_iterator &it);
  bool parseData(QStringList::const_iterator &it, Section &section);
  bool parseLines(QStringList::const_iterator &it, Section &section);
  QValueVector<CodeLine> findCodeLines(const PURL &url, int line) const; // line==-1: all lines
  static QString replaceDisasmTab(const QString &s);
  virtual void internalParse();
  void generateText();
};

//-----------------------------------------------------------------------------
class SymbolParser : public ParserBase
{
public:
  SymbolParser();
  const PURL::List &includedFiles() const { return _includedFiles; }

private:
  PURL::List _includedFiles;

  bool parseFile(QStringList::const_iterator &it);
  virtual void internalParse();
};

//-----------------------------------------------------------------------------
class FileHeaderParser : public ParserBase
{
public:
  FileHeaderParser();
  QString device() const { return _device; }

private:
  QString _device;

  virtual void internalParse();
};

//-----------------------------------------------------------------------------
class Editor : public SimpleTextEditor
{
Q_OBJECT
public:
  Editor(const PURL &url, const Device::Data &device, QWidget *parent);

private:
  const Device::Data &_device;

  virtual bool load();
};

}; // namespace

//-----------------------------------------------------------------------------
class DisassemblyEditor : public SimpleTextEditor
{
Q_OBJECT
public:
  DisassemblyEditor(const PURL &url, const PURL &hexUrl, const QString &device, QWidget *parent);
  DisassemblyEditor(const PURL &url, const HexEditor &e, const QString &device, QWidget *parent);
  virtual PURL::FileType fileType() const { return PURL::AsmGPAsm; }

private:
  PTempFile _tmp;
  PURL      _hexUrl;
  QString   _device;

  virtual bool load();
};


#endif
