/***************************************************************************
 *   Copyright (C) 2005 Nicolas Hadacek <hadacek@kde.org>                  *
 *   Copyright (C) 2003-2004 Alain Gibaud <alain.gibaud@free.fr>           *
 *   Copyright (C) 2002-2003 Stephen Landamore <stephen@landamore.com>     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "parallel.h"

#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h> // needed on some system

#include "config.h"
#ifdef HAVE_PPDEV
#include <linux/ppdev.h>
#include <linux/parport.h>
#else // dummy
  #define PPCLAIM 0
  #define PPRELEASE 0
  #define PPYIELD 0
  #define PPRCONTROL 0
  #define PPWCONTROL 0
  #define PPRDATA 0
  #define PPWDATA 0
  #define PPRSTATUS 0
#endif

#include <klocale.h>

#include "common/misc.h"

//-----------------------------------------------------------------------------
QStringList *Port::Parallel::_list = 0;

Port::IOTypes Port::Parallel::probe(const QString &device)
{
  int fd = ::open(device.latin1(), O_RDWR);
  if ( fd<0) return NoIO;
#ifdef HAVE_PPDEV
  if ( ioctl(fd, PPCLAIM)<0 ) {
    ::close(fd) ;
    return Input;
  }
  ioctl(fd, PPRELEASE);
  ::close(fd);
  return (Input | Output);
#else
  return NoIO;
#endif
}

QStringList Port::Parallel::deviceList()
{
  QStringList list;
  // standard parport in user space
  for(int i = 0; i<8; ++i) list.append(QString("/dev/parport%1").arg(i));
  // new devfs parport flavour
  for(int i = 0; i<8; ++i) list.append(QString("/dev/parports/%1").arg(i));
  return list;
}

const QStringList &Port::Parallel::probedDeviceList()
{
  if ( _list==0 ) {
    QStringList all = deviceList();
    _list = new QStringList;
    for (uint i=0; i<all.count(); i++)
      if( probe(all[i]) & (Input | Output) ) _list->append(all[i]);
  }
  return *_list;
}

bool Port::Parallel::isAvailable()
{
#ifdef HAVE_PPDEV
  return true;
#else
  return false;
#endif
}

//-----------------------------------------------------------------------------
const Port::Parallel::PPinData Port::Parallel::PIN_DATA[Nb_Pins] = {
  { Control, 0x01, Output, "/DS"    }, // !strobe
  { Data,    0x01, Output, "D0"     }, // data 0
  { Data,    0x02, Output, "D1"     }, // data 1
  { Data,    0x04, Output, "D2"     }, // data 2
  { Data,    0x08, Output, "D3"     }, // data 3
  { Data,    0x10, Output, "D4"     }, // data 4
  { Data,    0x20, Output, "D5"     }, // data 5
  { Data,    0x40, Output, "D6"     }, // data 6
  { Data,    0x80, Output, "D7"     }, // data 7
  { Status,  0x40, Input,  "/ACK"   }, // !ack
  { Status,  0x80, Input,  "BUSY"   }, // busy
  { Status,  0x20, Input,  "PAPER"  }, // pout
  { Status,  0x10, Input,  "SELin"  }, // select
  { Control, 0x02, Output, "LF"     }, // !feed
  { Status,  0x08, Input,  "/ERROR" }, // !error
  { Control, 0x04, Output, "PRIME"  }, // !init
  { Control, 0x08, Output, "SELout" }, // !si
  { Nb_RequestTypes, 0x00, NoIO, "GND" }, // GND
  { Nb_RequestTypes, 0x00, NoIO, "GND" }, // GND
  { Nb_RequestTypes, 0x00, NoIO, "GND" }, // GND
  { Nb_RequestTypes, 0x00, NoIO, "GND" }, // GND
  { Nb_RequestTypes, 0x00, NoIO, "GND" }, // GND
  { Nb_RequestTypes, 0x00, NoIO, "GND" }, // GND
  { Nb_RequestTypes, 0x00, NoIO, "GND" }, // GND
  { Nb_RequestTypes, 0x00, NoIO, "GND" }  // GND
};

QValueVector<Port::PinData> Port::Parallel::pinData(IOType type) const
{
  QValueVector<PinData> v;
  for (uint i=0; i<Nb_Pins; i++) {
    if ( PIN_DATA[i].ioType!=type ) continue;
    PinData pd = { i, PIN_DATA[i].label };
    v.append(pd);
  }
  return v;
}

bool Port::Parallel::isGroundPin(uint pin) const
{
  return ( PIN_DATA[pin].label=="GND" );
}

Port::IOType Port::Parallel::ioType(uint pin) const
{
  return PIN_DATA[pin].ioType;
}

const Port::Parallel::RequestData Port::Parallel::REQUEST_DATA[Nb_RequestTypes] = {
  { PPRCONTROL, PPWCONTROL },
  { PPRSTATUS,  0          },
  { PPRDATA,    PPWDATA    }
};

Port::Parallel::Parallel(const QString &device, Log::Base &base)
  : Base(base), _fd(-1), _device(device)
{
  for (uint i=0; i<Nb_RequestTypes; i++) _images[i] = 0;
}

bool Port::Parallel::internalOpen()
{
  _fd = ::open(_device.latin1(), O_RDWR);
  if ( _fd<0 ) {
    setLError(i18n("Could not open device \"%1\""));
    return false;
  }
  if ( ioctl(_fd, PPCLAIM)<0 ) {
    setLError(i18n("Could not claim device \"%1\": check it is read/write enabled"));
    return false;
  }
  return true;
}

void Port::Parallel::internalClose()
{
  if ( _fd<0 ) return;
  ioctl(_fd, PPRELEASE);
  ::close(_fd);
  _fd = -1;
}

//bool Port::Parallel::yield()
//{
//  if ( _fd<0 ) return false;
//  return ( ioctl(_fd, PPYIELD)>=0 );
//}

bool Port::Parallel::setPinOn(uint pin, bool on, LogicType type)
{
  if ( _fd<0 ) return false;
  Q_ASSERT( pin<Nb_Pins );
  Q_ASSERT( PIN_DATA[pin].ioType==Output );
  RequestType rtype = PIN_DATA[pin].rType;
  Q_ASSERT( rtype!=Nb_RequestTypes );
  uchar mask = PIN_DATA[pin].mask;
  uchar c = (XOR(type==NegativeLogic, on) ? _images[rtype] | mask : _images[rtype] &  ~mask);
  int request = REQUEST_DATA[rtype].write;
  Q_ASSERT( request!=0 );
  if ( ioctl(_fd, request, &c)<0 ) {
    setLError(i18n("Error setting pin %1 to %2").arg(PIN_DATA[pin].label).arg(on));
    return false;
  }
  _images[rtype] = c;
  return true;
}

bool Port::Parallel::readPin(uint pin, LogicType type)
{
  if ( _fd<0 ) return false;
  Q_ASSERT( pin<Nb_Pins );
  Q_ASSERT( PIN_DATA[pin].ioType==Input );
  RequestType rtype = PIN_DATA[pin].rType;
  Q_ASSERT( rtype!=Nb_RequestTypes );
  int request = REQUEST_DATA[rtype].read;
  Q_ASSERT( request!=0 );
  uchar c = 0;
  if ( ioctl(_fd, request, &c)<0 ) {
    setLError(i18n("Error reading bit on pin %1").arg(PIN_DATA[pin].label));
    return 0;
  }
  _images[rtype] = c;
  return (type==NegativeLogic ? c : ~c ) & PIN_DATA[pin].mask;
}
