/***************************************************************************
 *   Copyright (C) 2005 Nicolas Hadacek <hadacek@kde.org>                  *
 *   Copyright (C) 2003-2004 Alain Gibaud <alain.gibaud@free.fr>           *
 *   Copyright (C) 2002-2003 Stephen Landamore <stephen@landamore.com>     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "serial.h"

#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <unistd.h> // need on some system

#include <klocale.h>

//-----------------------------------------------------------------------------
QStringList *Port::Serial::_list = 0;

Port::IOTypes Port::Serial::probe(const QString &device)
{
  int id = ::open(device.latin1(), O_RDONLY|O_NOCTTY|O_NONBLOCK);
  if ( id==-1 ) return NoIO;
  ::close(id);
  id = ::open(device.latin1(), O_RDWR|O_NOCTTY|O_NONBLOCK);
  if ( id==-1 ) return Input;
  ::close(id);
  return (Input | Output);
}

QStringList Port::Serial::deviceList()
{
  QStringList list;
  // standard serport in user space
  for(int i = 0; i < 8; ++i) list.append(QString("/dev/ttyS%1").arg(i));
  // new devfs serport flavour
  for(int i = 0; i < 8; ++i) list.append(QString("/dev/tts/%1").arg(i));
  // standard USB serport in user space
  for(int i = 0; i < 8; ++i) list.append(QString("/dev/ttyUSB%1").arg(i));
  // new devfs USB serport flavour
  for(int i = 0; i < 8; ++i) list.append(QString("/dev/usb/tts/%1").arg(i));
  // FreeBSD
  for(int i = 0; i < 8; ++i) list.append(QString("/dev/ttyd%1").arg(i));
  return list;
}

const QStringList &Port::Serial::probedDeviceList()
{
  if ( _list==0 ) {
    QStringList all = deviceList();
    _list = new QStringList;
    for (uint i=0; i<all.count(); i++)
      if( probe(all[i]) & (Input | Output) ) _list->append(all[i]);
  }
  return *_list;
}

//-----------------------------------------------------------------------------
const Port::Serial::SpeedData Port::Serial::SPEED_DATA[Nb_Speeds] = {
  { B0,         0 }, { B50,         50 }, { B75,       75 }, { B110,     110 },
  { B134,     134 }, { B150,       150 }, { B200,     200 }, { B300,     300 },
  { B600,     600 }, { B1200,     1200 }, { B1800,   1800 }, { B2400,   2400 },
  { B4800,   4800 }, { B9600,     9600 }, { B19200, 19200 }, { B38400, 38400 },
  { B57600, 57600 }, { B115200, 115200 }
};

const Port::Serial::SPinData Port::Serial::PIN_DATA[Nb_Pins] = {
  { Input,  "CD"  }, { Input,  "RX"  }, { Output, "TX"  }, { Output, "DTR" },
  { NoIO,   "GND" }, { Output, "DSR" }, { Output, "RTS" }, { Input,  "CTS" },
  { Input,  "RI"  }
};

QValueVector<Port::PinData> Port::Serial::pinData(IOType type) const
{
  QValueVector<PinData> v;
  for (uint i=0; i<Nb_Pins; i++) {
    if ( PIN_DATA[i].ioType!=type ) continue;
    PinData pd = { i, PIN_DATA[i].label };
    v.append(pd);
  }
  return v;
}
bool Port::Serial::isGroundPin(uint pin) const
{
  return ( PIN_DATA[pin].label=="GND" );
}

Port::IOType Port::Serial::ioType(uint pin) const
{
  return PIN_DATA[pin].ioType;
}

Port::Serial::Serial(const QString &device, Properties properties, Log::Base &base)
  : Base(base), _fd(-1), _device(device), _properties(properties)
{}

bool Port::Serial::getParameters(termios &info)
{
  if ( _fd<0 ) return false;
  if ( tcgetattr(_fd, &info)<0 ) {
    setLError(i18n("Could not get file descriptor parameters"));
    return false;
  }
  return true;
}

bool Port::Serial::setParameters(termios newInfo)
{
  if ( _fd<0 ) return false;
  if ( tcsetattr(_fd, TCSANOW, &newInfo)<0 ) {
    setLError(i18n("Could not set file descriptor parameters"));
    return false;
  }
  return true;
}

bool Port::Serial::internalOpen()
{
  // open non blocking: avoid missing DCD (comment from xwisp2)
  _fd = ::open(_device.latin1(), O_NONBLOCK|O_RDWR|O_NOCTTY);
  if ( _fd<0 ) {
    setLError(i18n("Could not open device \"%1\" read-write"));
    return false;
  }

  if ( !getParameters(_oldInfo) ) return false; // save configuration

  if ( _properties & Blocking ) {
    int flags = fcntl(_fd, F_GETFL, 0);
    if ( fcntl(_fd, F_SETFL, flags & ~O_NONBLOCK)<0 ) {
      setLError(i18n("Could not modify file descriptor flags"));
      return false;
    }
  }

  if ( _properties & NeedFlush ) {
    if ( tcflush(_fd, TCIOFLUSH)<0 ) {
      setLError(i18n("Could not flush device"));
      return false;
    }
  }

  return true;
}

void Port::Serial::internalClose()
{
  if ( _fd<0 ) return;
  if ( _properties & NeedFlush ) flush();
  if ( _properties & NeedBreak ) doBreak(1);
  tcsetattr(_fd, TCSANOW, &_oldInfo);
  ::close(_fd);
  _fd = -1;
}

bool Port::Serial::send(const char *data, uint size)
{
  if ( _fd<0 ) return false;
  uint retries = 10;
  int todo = size;
  for (;;) {
    int res = write(_fd, data + size - todo, todo);
    if ( res==todo ) break;
    if ( res<0 ) {
      if ( errno!=EAGAIN ) {
        setLError(i18n("Error sending data"));
        return false;
      }// else log(Log::Debug, i18n("Port not ready for write: retrying..."));
    }
    if ( res==0 ) {
      retries--;
      if ( retries==0 ) {
        if ( todo==int(size) ) log(Log::Error, i18n("Timeout sending data."));
        else log(Log::Error, i18n("Timeout: only some data sent (%1/%2 bytes).").arg(size-todo).arg(size));
        return false;
      }
      //if ( res==0 ) log(Log::Debug, i18n("Nothing sent: retrying..."));
    }
    if ( res>0 ) todo -= res;
    msleep(100);
  }
  if ( _properties & NeedDrain && tcdrain(_fd)<0 ) {
    setLError(i18n("Error while draining after sending data"));
    return false;
  }
  return true;
}

bool Port::Serial::checkForInput()
{
  uint retries = 20;
  for (;;) {
    int nb;
    if ( ioctl(_fd, FIONREAD, &nb)<0 ) {
      setLError(i18n("Error checking input buffer"));
      return false;
    }
    if ( nb!=0 ) break;
    retries--;
    if ( retries==0 ) {
      log(Log::Error, i18n("Timeout waiting for data."));
      return false;
    }
    msleep(100);
  }
  return true;
}

QCString Port::Serial::receive(uint size)
{
  if ( _fd<0 ) return false;
  // if serial cable disconnect: this can block
  // so we wait to get something in incoming buffer or timeout
  if ( !checkForInput() ) return false;
  QCString cs(size+1);
  uint retries = 10;
  int todo = size;
  while ( todo!=0 ) {
    int res = read(_fd, cs.data() + size - todo, todo);
    if ( res==todo ) break;
    if ( res<0 ) {
      if ( errno!=EAGAIN ) {
        setLError(i18n("Error receiving data"));
        return false;
      }// else log(Log::Debug, i18n("Port not ready for read: retrying..."));
    }
    if ( res==0 ) {
      retries--;
      if ( retries==0 ) {
        if ( todo==int(size) ) log(Log::Error, i18n("Timeout receiving data."));
        else log(Log::Error, i18n("Timeout: only some data received (%1/%2 bytes).").arg(size-todo).arg(size));
        return false;
      }
      //if ( res==0 ) log(Log::Debug, i18n("Nothing received: retrying..."));
    }
    if ( res>0 ) todo -= res;
    msleep(100);
  }
  return cs;
}

bool Port::Serial::flush()
{
  if ( _fd<0 ) return false;
  if ( _properties & NeedDrain && tcdrain(_fd)<0 ) {
    setLError(i18n("Could not drain device"));
    return false;
  }
  if ( tcflush(_fd, TCIFLUSH)<0 ) {
    setLError(i18n("Could not flush device"));
    return false;
  }
  return true;
}

bool Port::Serial::setPinOn(uint pin, bool on, LogicType type)
{
  if ( _fd<0 ) return false;
  if ( type==NegativeLogic ) on = !on;
  Q_ASSERT( pin<Nb_Pins );
  Q_ASSERT( PIN_DATA[pin].ioType==Output );
  int bit = 0;
  switch (Pin(pin)) {
    case TX:
      // bit = TIOCM_ST;
      if ( ioctl(_fd, on ? TIOCSBRK : TIOCCBRK, 0)<0 ) {
        setLError(i18n("Error setting bit %1 of serial port to %2").arg(PIN_DATA[pin].label).arg(on));
        return false;
      }
      return true;
    case DTR: bit = TIOCM_DTR; break;
    case DSR: bit = TIOCM_DSR; break;
    case RTS: bit = TIOCM_RTS; break;
    default: Q_ASSERT(false); return false;
  }
  if ( ioctl(_fd, on ? TIOCMBIS : TIOCMBIC, &bit)<0 ) {
    setLError(i18n("Error setting pin %1 of serial port to %2").arg(PIN_DATA[pin].label).arg(on));
    return false;
  }
  return true;
}

bool Port::Serial::readPin(uint pin, LogicType type)
{
  if ( _fd<0 ) return false;
  Q_ASSERT( pin<Nb_Pins );
  Q_ASSERT( PIN_DATA[pin].ioType==Input );
  int mask = 0;
  switch (Pin(pin)) {
    case CD:  mask = TIOCM_CD;  break;
    case RX:  mask = TIOCM_SR;  break;
    case CTS: mask = TIOCM_CTS; break;
    case RI:  mask = TIOCM_RI;  break;
    default: Q_ASSERT(false); return false;
  }
  int bits;
  if ( ioctl(_fd, TIOCMGET, &bits)<0 ) {
    setLError(i18n("Error reading serial pin %1").arg(PIN_DATA[pin].label));
    return false;
  }
  return ((type==NegativeLogic ? ~bits : bits) & mask);
}

bool Port::Serial::setMode(int iflag, int cflag, Speed speed, uint readTimeout)
{
  if ( !flush() ) return false;
  struct termios info;
  memset((void *)&info, 0, sizeof(termios));      // clear all
  info.c_iflag = iflag;
  info.c_oflag = 0;                               // out flags off
  info.c_cflag = cflag;
  info.c_ispeed = SPEED_DATA[speed].value;
  info.c_ospeed = SPEED_DATA[speed].value;
  info.c_cflag |= SPEED_DATA[speed].cflag;        // not for MacOSX (from xwisp2)
  info.c_lflag = 0;                               // all flags off
  info.c_cc[VMIN]  = 0;                           // wait for 1 char or timeout
  info.c_cc[VTIME] = readTimeout/100;             // wait for in deciseconds
  return setParameters(info);
}

bool Port::Serial::doBreak(uint duration)
{
  if ( !flush() ) return false;
  if ( ioctl(_fd, TIOCSBRK)<0 ) {
    setLError(i18n("Error setting up break"));
    return false;
  }
  msleep(duration);
  if ( ioctl(_fd, TIOCCBRK) ) {
    setLError(i18n("Error clearing up break"));
    return false;
  }
  msleep(1);
  return flush();
}
