/***************************************************************************
 * Copyright (C) 2005 Lorenz Msenlechner & Matthias Kranz                 *
 *                    <icd2linux@hcilab.org>                               *
 * Copyright (C) 2003-2005 Orion Sky Lawlor <olawlor@acm.org>              *
 * Copyright (C) 2005-2006 Nicolas Hadacek <hadacek@kde.org>               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "usb_port.h"

#include <klocale.h>
#include "config.h"
#include "common/misc.h"
#ifdef HAVE_USB
#include <usb.h>
#endif

Port::USB::USB(uint vendorId, uint productId, Log::Base &base)
  : Base(base), _vendorId(vendorId), _productId(productId), _handle(0), _device(0)
{
#ifdef HAVE_USB
  VersionData vd = VersionData::fromString(LIBUSB_VERSION);
  QString s = QString("libusb %1").arg(vd.pretty());
  if ( vd.toUInt()<VersionData(0, 1, 8).toUInt() ) qWarning("%s: may be too old (you need at least version 0.1.8)", s.latin1());
  _needInterruptMode = ( vd.toUInt()>=VersionData(0, 1, 12).toUInt() ); // #### 0.1.8 is without and 0.1.12 needs it: not sure about intermediate versions
  if (_needInterruptMode) qDebug("%s: using interrupt mode", s.latin1());
  else qDebug("%s: using bulk mode", s.latin1());
#endif
}

void Port::USB::setUError(const QString &message)
{
#ifdef HAVE_USB
  log(Log::Error, message + QString(" (err=%1).").arg(usb_strerror()));
#else
  Q_UNUSED(message);
#endif
}

void Port::USB::tryToDetachDriver()
{
  // try to detach an already existing driver... only on linux
#if defined (LIBUSB_HAS_GET_DRIVER_NP) && LIBUSB_HAS_GET_DRIVER_NP
  char dname[32] = "";
  if ( usb_get_driver_np(_handle, interface(*_device), dname, 31)<0 ) return;
  log(Log::NormalDebug, "a driver is already installed...");
#  if defined (LIBUSB_HAS_DETACH_KERNEL_DRIVER_NP) && LIBUSB_HAS_DETACH_KERNEL_DRIVER_NP
  usb_detach_kernel_driver_np(_handle, interface(*_device));
  log(Log::NormalDebug, "...try to detach it.");
  if ( usb_get_driver_np(_handle, interface(*_device), dname, 31)<0 ) return;
  log(Log::NormalDebug, "failed to detach it ?");
#  endif
#endif
}

bool Port::USB::internalOpen()
{
  _device = findDevice(_vendorId, _productId);
  if ( _device==0 ) {
    log(Log::Error, i18n("Could not find USB device (vendor=%1 product=%2).")
             .arg(toHexLabel(_vendorId, 4)).arg(toHexLabel(_productId, 4)));
    return false;
  }
#ifdef HAVE_USB
  _handle = usb_open(_device);
  if ( _handle==0 ) {
    log(Log::Error, i18n("Error opening USB device."));
    return false;
  }
  if ( usb_reset(_handle)<0 ) {
    log(Log::Error, i18n("Error resetting USB device."));
    return false;
  }
  usb_close(_handle);
  _handle = usb_open(_device);
  if ( _handle==0 ) {
    log(Log::Error, i18n("Error opening USB device."));
    return false;
  }
  tryToDetachDriver();
  if ( configuration()>=0 && usb_set_configuration(_handle, configuration())<0 ) {
    log(Log::Error, i18n("Error setting USB configuration."));
    return false;
  }
  if ( usb_claim_interface(_handle, interface(*_device))<0 ) {
    log(Log::Error, i18n("Could not claim USB interface:\n"
                  "Do a `dmesg` to see which kernel driver has claimed it.\n"
                  "You may need to `rmmod hid` or patch your kernel's hid driver."));
    return false;
  }
#else
  return false;
#endif
  return true;
}

void Port::USB::internalClose()
{
  if ( _handle==0 ) return;
#ifdef HAVE_USB
  usb_release_interface(_handle, interface(*_device));
  usb_close(_handle);
#endif
  _device = 0;
  _handle = 0;
}

struct usb_device *Port::USB::findDevice(uint vendorId, uint productId)
{
#ifdef HAVE_USB
  usb_init();
  // refresh libusb structures
  usb_find_busses();
  usb_find_devices();
  usb_bus *busses = usb_get_busses();
  // find device
  for (usb_bus *bus=busses; bus; bus=bus->next) {
    for (struct usb_device *dev=bus->devices; dev; dev=dev->next) {
      if ( dev->descriptor.idVendor==vendorId && dev->descriptor.idProduct==productId )
        return dev;
    }
  }
#else
  Q_UNUSED(vendorId); Q_UNUSED(productId);
#endif
  return 0;
}

bool Port::USB::sendControlMessage(const ControlMessageData &data)
{
  if ( hasError() ) return false;
  QString s = data.bytes;
  uint length = strlen(data.bytes) / 2;
  QByteArray ba(length);
  for (uint i=0; i<length; i++)
    ba[i] = fromHex(s.mid(2*i, 2), 0);
#ifdef HAVE_USB
  int res = usb_control_msg(_handle, data.type, data.request, data.value, 0, ba.data(), length, 1000); // 1s
  if ( res<0 ) {
    log(Log::Error, i18n("Error sending control message to USB port."));
    return false;
  }
#endif
  return true;
}

uint Port::USB::timeout(uint size) const
{
  return QMAX(size*5, 1000); // 5ms per byte or 1s
}

bool Port::USB::write(uint ep, const char *data, uint size)
{
  if ( hasError() ) return false;
#ifdef HAVE_USB
  uint retries = 10;
  int todo = size;
  for (;;) {
    int res = 0;
    //qDebug("write ep=%i todo=%i/%i", ep, todo, size);
    if (_needInterruptMode) res = usb_interrupt_write(_handle, ep, (char *)data + size - todo, todo, timeout(todo));
    else res = usb_bulk_write(_handle, ep, (char *)data + size - todo, todo, timeout(todo));
    //qDebug("res: %i", res);
    if ( res==todo ) break;
    retries--;
    if ( retries==0 ) {
      if ( res<0 ) setUError(i18n("Error sending data (ep=%1 res=%2)").arg(ep).arg(res));
      else log(Log::Error, i18n("Timeout: only some data sent (%1/%2 bytes).").arg(size-todo).arg(size));
      return false;
    }
    if ( res==0 ) log(Log::NormalDebug, i18n("Nothing sent: retrying..."));
    if ( res>0 ) todo -= res;
    msleep(100);
  }
#else
  Q_UNUSED(ep); Q_UNUSED(data); Q_UNUSED(size);
#endif
  return true;
}

bool Port::USB::read(uint ep, char *data, uint size)
{
  if ( hasError() ) return false;
#ifdef HAVE_USB
  uint retries = 10;
  int todo = size;
  for (;;) {
    int res = 0;
    //qDebug("read ep=%i size=%i", ep, todo);
    if (_needInterruptMode) res = usb_interrupt_read(_handle, ep, data + size - todo, todo, timeout(todo));
    else res = usb_bulk_read(_handle, ep, data + size - todo, todo, timeout(todo));
    //qDebug("res: %i", res);
    if ( res==todo ) break;
    retries--;
    if ( retries==0 ) {
      if ( res<0 ) setUError(i18n("Error receiving data (ep=%1 res=%2)").arg(ep).arg(res));
      else log(Log::Error, i18n("Timeout: only some data received (%1/%2 bytes).").arg(size-todo).arg(size));
      return false;
    }
    if ( res==0 ) log(Log::NormalDebug, i18n("Nothing received: retrying..."));
    if ( res>0 ) todo -= res;
    msleep(100);
  }
#else
  Q_UNUSED(ep); Q_UNUSED(data); Q_UNUSED(size);
#endif
  return true;
}

bool Port::USB::isAvailable()
{
#ifdef HAVE_USB
  return true;
#else
  return false;
#endif
}
