/***************************************************************************
 *   Copyright (C) 2005-2006 Nicolas Hadacek <hadacek@kde.org>             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "port_selector.h"

#include <klocale.h>

#include "port/serial.h"
#include "port/parallel.h"
#include "common/about.h"
#include "tools/misc_gui.h"
#include "progs/base/prog_config.h"
#include "progs/base/prog_list.h"

const char * const PortSelector::LABELS[Port::Nb_Types] = {
  I18N_NOOP("Serial"), I18N_NOOP("Parallel"), I18N_NOOP("USB")
};

PortSelector::PortSelector(QWidget *parent)
  : QFrame(parent, "port_selector"), _group(0), _main(0)
{
  _top = new QGridLayout(this, 1, 1, 0, 10);
  _top->setRowStretch(1, 1);

  _bgroup = new QButtonGroup;
  connect(_bgroup, SIGNAL(clicked(int)), SIGNAL(changed()));
}

void PortSelector::setGroup(const Programmer::Group &group)
{
  _group = &group;
  delete _main;
  _main = new QWidget(this);
  _top->addWidget(_main, 0, 0);
  _grid = new QGridLayout(_main, 1, 1, 0, 10);
  _grid->setColStretch(3, 1);
  for (uint i=0; i<Port::Nb_Types; i++) {
    _combos[i] = 0;
    _status[i] = 0;
    if ( !group.isPortSupported(Port::Type(i)) ) continue;
    Port::Type ptype = Port::Type(i);
    addPortType(ptype, Programmer::Config::readPortDevice(group, ptype));
  }
  _bgroup->setButton(Programmer::Config::readPortType(group));
  _main->show();
}

void PortSelector::addPortType(Port::Type type, const QString &device)
{
  QString noDeviceMessage, notAvailableMessage;
  switch (type) {
  case Port::SerialType:
    noDeviceMessage = i18n("<qt>Your computer might not have any serial port.</qt>");
    break;
  case Port::ParallelType:
    noDeviceMessage = i18n("<qt>Your computer might not have any parallel port"
                           "or the /dev/parportX device has not been created.<br>"
                           "Use \"mknod /dev/parport0 c 99 0\" to create it<br>"
                           "and \"chmod a+rw /dev/parport0\" to make it RW enabled.</qt>");
    notAvailableMessage = i18n("<qt>Piklab has been compiled without support for parallel port.</qt");
    break;
  case Port::USBType:
    notAvailableMessage = i18n("<qt>Piklab has been compiled without support for USB port.</qt>");
    break;
  case Port::Nb_Types: Q_ASSERT(false); break;
  }

  QRadioButton *rb = new QRadioButton(i18n(LABELS[type]), _main);
  _bgroup->insert(rb, type);
  if ( _bgroup->count()==1 ) _bgroup->setButton(type);
  _grid->addWidget(rb, 3*(_bgroup->count()-1), 0);

  QStringList list = Port::probedDeviceList(type);
  bool noDevice = ( list.isEmpty() && Port::DATA[type].withDevice );
  bool notAvailable = !Port::isAvailable(type);
  if ( noDevice || notAvailable ) {
    rb->setEnabled(false);
    QLabel *label = new QLabel(notAvailable ? notAvailableMessage : noDeviceMessage, _main);
    _grid->addMultiCellWidget(label, 3*(_bgroup->count()-1)+1,3*(_bgroup->count()-1)+1, 0,3);
    UrlLabel *ul = new UrlLabel(AboutData::url(AboutData::Support).url(), i18n("See Piklab homepage for help."), _main);
    _grid->addMultiCellWidget(ul, 3*(_bgroup->count()-1)+2,3*(_bgroup->count()-1)+2, 0,3);
  } else {
    _status[type] = new QLabel(_main);
    _grid->addWidget(_status[type], 3*(_bgroup->count()-1), 2);
  }

  if ( !list.isEmpty() ) {
    _combos[type] = new QComboBox(_main);
    for (uint i=0; i<list.count(); i++) _combos[type]->insertItem(list[i]);
    if ( list.contains(device) ) _combos[type]->setCurrentText(device);
    connect(_combos[type], SIGNAL(activated(int)), SIGNAL(changed()));
    _grid->addWidget(_combos[type], 3*(_bgroup->count()-1), 1);
  }
}

void PortSelector::setStatus(Port::Type type, const QString &message)
{
  for (uint i=0; i<Port::Nb_Types; i++) {
    if ( _status[i]==0 ) continue;
    if ( Port::Type(i)!=type ) _status[i]->hide();
    else {
      _status[i]->show();
      _status[i]->setText(message);
    }
  }
}

QString PortSelector::device(Port::Type type) const
{
  if ( _combos[type]==0 ) return QString::null;
  return _combos[type]->currentText();
}

void PortSelector::saveConfig()
{
  Programmer::Config::writePortType(*_group, type());
  for (uint i=0; i<Port::Nb_Types; i++) {
    if ( !_group->isPortSupported(Port::Type(i)) ) continue;
    Port::Type ptype = Port::Type(i);
    Programmer::Config::writePortDevice(*_group, ptype, device(ptype));
  }
}
