/***************************************************************************
 *   Copyright (C) 2005 Nicolas Hadacek <hadacek@kde.org>                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "prog_config_widget.h"

#include <qlayout.h>
#include <qlabel.h>
#include <qwidgetstack.h>
#include <qcombobox.h>
#include <qcheckbox.h>
#include <qtimer.h>
#include <qtabwidget.h>
#include <kiconloader.h>

#include "global_config.h"
#include "port_selector.h"
#include "progs/base/prog_list.h"
#include "progs/base/prog_config.h"
#include "progs/base/gui/prog_config_widget.h"
#include "progs/base/gui/prog_group_ui.h"

//----------------------------------------------------------------------------
Programmer::OptionsConfigWidget::OptionsConfigWidget()
{
  QGridLayout *top = new QGridLayout(this, 1, 1, 10, 10);

  _powerDown = new QCheckBox(i18n("Power down target after programming."), this);
  top->addMultiCellWidget(_powerDown, 0,0, 0,1);

  _targetSelfPowered = new QCheckBox(i18n("Target is self-powered."), this);
  top->addMultiCellWidget(_targetSelfPowered, 1,1, 0,1);

  _programAll = new QCheckBox(i18n("Program all memory even if not needed (slower)."), this);
  top->addMultiCellWidget(_programAll, 2,2, 0,1);

  _verifyProgrammed = new QCheckBox(i18n("Only verify programmed words in code memory (faster)."), this);
  top->addMultiCellWidget(_verifyProgrammed, 3,3, 0,1);

  _blankCheckAfterErase = new QCheckBox(i18n("Blank check after erase."), this);
  top->addMultiCellWidget(_blankCheckAfterErase, 4,4, 0,1);

  _preserveEeprom = new QCheckBox(i18n("Preserve data eeprom when programming."), this);
  top->addMultiCellWidget(_preserveEeprom, 5,5, 0,1);

  top->setRowStretch(6, 1);
}

void Programmer::OptionsConfigWidget::loadConfig()
{
  _powerDown->setChecked(Config::powerDownAfterProgramming());
  _targetSelfPowered->setChecked(Config::isTargetSelfPowered());
  _programAll->setChecked(!Config::onlyProgramNonMask());
  _verifyProgrammed->setChecked(Config::onlyVerifyProgrammed());
  _blankCheckAfterErase->setChecked(Config::blankCheckAfterErase());
  _preserveEeprom->setChecked(Config::preserveEeprom());
}

QPixmap Programmer::OptionsConfigWidget::pixmap() const
{
  KIconLoader loader;
  return loader.loadIcon("piklab_config_programmer", KIcon::Toolbar, KIcon::SizeMedium);
}

void Programmer::OptionsConfigWidget::saveConfig()
{
  Config::writePowerDownAfterProgramming(_powerDown->isChecked());
  Config::writeTargetSelfPowered(_targetSelfPowered->isChecked());
  Config::writeOnlyProgramNonMask(!_programAll->isChecked());
  Config::writeOnlyVerifyProgrammed(_verifyProgrammed->isChecked());
  Config::writeBlankCheckAfterErase(_blankCheckAfterErase->isChecked());
  Config::writePreserveEeprom(_preserveEeprom->isChecked());
}

//----------------------------------------------------------------------------
Programmer::SelectConfigWidget::SelectConfigWidget()
{
  QGridLayout *top = new QGridLayout(this, 1, 1, 10, 10);

  // programmer type
  QLabel *label = new QLabel(i18n("Programmer in use:"), this);
  top->addWidget(label, 0, 0);
  _combo = new QComboBox(this);
  connect(_combo, SIGNAL(activated(int)), SLOT(programmerChanged(int)));
  top->addWidget(_combo, 0, 1);

  QTabWidget *tw = new QTabWidget(this);
  top->addMultiCellWidget(tw, 1,1, 0,2);

  // port selector
  Container *container = new Container(tw, i18n("Port Selection"));
  _portSelector = new PortSelector(container);
  connect(_portSelector, SIGNAL(changed()), SLOT(portChanged()));
  container->addWidget(_portSelector, 0, 0);

  // specific programmer config
  _stack = new QWidgetStack(this);
  _stack->setMargin(10);
  tw->addTab(_stack, i18n("Specific"));
  uint current = 0;
  for (uint i=0; i<Programmer::lister().nbGroups(); i++) {
    const Programmer::Group &group = Programmer::lister().group(i);
    _combo->insertItem(group.label());
    Programmer::ConfigWidget *pcw = static_cast<Programmer::GroupUI *>(group.gui())->createConfigWidget(_stack);
    pcw->loadConfig();
    _stack->addWidget(pcw, i);
    if ( group.name()==GlobalConfig::programmerGroup().name() ) current = i;
  }

  // init
  _combo->setCurrentItem(current);
  _stack->raiseWidget(current);
  programmerChanged(current);
}

void Programmer::SelectConfigWidget::portChanged()
{
  ::BusyCursor bc; // can take a few seconds to connect programmer...
  Port::Type ptype = _portSelector->type();
  QString device = _portSelector->device(ptype);
  uint i = _combo->currentItem();
  bool ok = static_cast<Programmer::ConfigWidget *>(_stack->widget(i))->setPort(ptype, device);
  _portSelector->setStatus(_portSelector->type(), ok ? i18n("Connection: Ok") : i18n("Connection: Error"));
}

QPixmap Programmer::SelectConfigWidget::pixmap() const
{
  KIconLoader loader;
  return loader.loadIcon("piklab_config_programmer", KIcon::Toolbar, KIcon::SizeMedium);
}

void Programmer::SelectConfigWidget::saveConfig()
{
  _portSelector->saveConfig();
  uint i = _combo->currentItem();
  static_cast<ConfigWidget *>(_stack->widget(i))->saveConfig();
  GlobalConfig::writeProgrammerGroup(Programmer::lister().group(i));
}

void Programmer::SelectConfigWidget::programmerChanged(int i)
{
  _portSelector->setGroup(Programmer::lister().group(i));
  _stack->raiseWidget(_stack->widget(i));
  QTimer::singleShot(0, this, SLOT(portChanged()));
}
