/***************************************************************************
 *   Copyright (C) 2005-2006 Nicolas Hadacek <hadacek@kde.org>             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "generic_prog.h"

#include <klocale.h>
#include <kaction.h>

#include "prog_group.h"
#include "prog_config.h"
#include "devices/base/device_group.h"
#include "port/port_base.h"

//-----------------------------------------------------------------------------
Programmer::DeviceSpecific::DeviceSpecific(Programmer::Base &base)
 : Log::Client(base), _base(base)
{}

//-----------------------------------------------------------------------------
Programmer::Hardware::Hardware(Programmer::Base &base, Port::Base *port)
  : Log::Client(base), _port(port), _base(base)
{}

Programmer::Hardware::~Hardware()
{
  delete _port;
}

bool Programmer::Hardware::connectHardware()
{
  _port->close();
  return internalConnectHardware();
}

void Programmer::Hardware::disconnectHardware()
{
  _port->close();
}

//-----------------------------------------------------------------------------
const double Programmer::UNKNOWN_VOLTAGE = -1.0;

//----------------------------------------------------------------------------
const Programmer::Base::OperationData Programmer::Base::OPERATION_DATA[Nb_Operations] = {
  {  0, I18N_NOOP("Reading...")        },
  { 10, I18N_NOOP("Programming...")    },
  {  0, I18N_NOOP("Reading...")        },
  { 10, I18N_NOOP("Erasing...")        },
  {  0, I18N_NOOP("Blank Checking...") },
  { 10, I18N_NOOP("Connecting...")     }
};

//-----------------------------------------------------------------------------
Programmer::Base::Base(const Group &group, const Device::Data *device, const char *name)
  : Log::Manager(name), _hardware(0), _specific(0), _device(device), _state(NotConnected),
    _targetVddOn(false), _group(group)
{}

void Programmer::Base::init(Hardware *hardware, DeviceSpecific *ds)
{
  _hardware = hardware;
  _specific = ds;
}

Programmer::Base::~Base()
{
  delete _specific;
  delete _hardware;
}

bool Programmer::Base::simpleConnectHardware()
{
  Q_ASSERT(_hardware);
  disconnectHardware();
  _firmwareVersion.clear();
  initConnect();
  if (_device) {
    log(Log::Information, i18n("Connecting %1 on %2 with device %3...")
      .arg(_group.label()).arg(Port::DATA[Programmer::Config::readPortType(group())].label).arg(_device->name()));
  }
  return _hardware->connectHardware();
}

bool Programmer::Base::connectHardware()
{
  if ( !simpleConnectHardware() ) return false;
  if ( !internalSetupHardware() ) return false;
  log(Log::Information, i18n("Connected"));
  _state = Stopped;
  addProgress(OPERATION_DATA[Connect].nbSteps);
  return true;
}

void Programmer::Base::disconnectHardware()
{
  _hardware->disconnectHardware();
  _state = NotConnected;
  resetError();
}

void Programmer::Base::log(Log::Type type, const QString &message)
{
  if ( type==Log::Error ) _state = NotConnected;
  Log::Manager::log(type, message);
}

void Programmer::Base::log(Log::DebugLevel level, const QString &message)
{
  Log::Manager::log(level, message);
}

uint Programmer::Base::nbSteps(Operation operation) const
{
  uint length = OPERATION_DATA[Connect].nbSteps;
  if ( operation==Connect ) return length;
  length += OPERATION_DATA[operation].nbSteps;
  return length;
}

void Programmer::Base::prepare(Operation operation)
{
  emit actionMessage(i18n(OPERATION_DATA[operation].message));
  uint nb = nbSteps(operation);
  setTotalSteps(nb);
}

bool Programmer::Base::connectDevice()
{
  prepare(Connect);
  bool ok = doConnectDevice();
  endProgramming();
  return ok;
}

bool Programmer::Base::doConnectDevice()
{
  if ( _state==NotConnected ) {
    if ( !connectHardware() ) return false;
    if ( !verifyDeviceId() ) return false;
  } else {
    addProgress(OPERATION_DATA[Connect].nbSteps);
    setTargetVddOn(!Config::isTargetSelfPowered());
  }
  _state = Stopped;
  return true;
}

bool Programmer::Base::run()
{
  emit actionMessage(i18n("Running..."));
  if ( !doConnectDevice() ) return false;
  log(Log::Information, i18n("Run..."));
  internalRun();
  return !hasError();
}

bool Programmer::Base::step()
{
  return internalStep();
}

bool Programmer::Base::stop()
{
  emit actionMessage(i18n("Halting..."));
  return internalStop();
}

void Programmer::Base::endProgramming()
{
  if ( _state==Stopped && Config::powerDownAfterProgramming() )
    setTargetVddOn(false);
}

void Programmer::Base::setTargetVddOn(bool on)
{
  _targetVddOn = on;
  _specific->setTargetVddOn(on);
}

bool Programmer::Base::uploadFirmware(const QString &filepath)
{
  log(Log::NormalDebug, QString("  Firmware file: %1").arg(filepath));
  QFile file(filepath);
  if ( !file.open(IO_ReadOnly) ) {
    log(Log::Error, i18n("Could not open firmware file \"%1\".").arg(file.name()));
    return false;
  }
  prepare(Connect);
  bool ok  = doUploadFirmware(file);
  _firmwareVersion.clear();
  if (ok) readFirmware();
  endProgramming();
  return ok;
}

//-----------------------------------------------------------------------------
bool Programmer::Base::erase()
{
  prepare(Erase);
  bool ok = doErase();
  endProgramming();
  return ok;
}

bool Programmer::Base::doErase()
{
  if ( !checkErase() ) return false;
  if ( !doConnectDevice() ) return false;
  log(Log::Information, i18n("Erasing..."));
  if ( !internalErase() ) return false;
  log(Log::Information, i18n("Erasing done"));
  if ( Config::blankCheckAfterErase() ) {
    log(Log::Information, i18n("Blank checking..."));
    Device::Memory *memory = _device->group().createMemory(*_device);
    bool ok = internalVerify(*memory, BlankCheckVerify);
    delete memory;
    if ( !ok ) return false;
    log(Log::Information, i18n("Blank checking done"));
  }
  return true;
}

//-----------------------------------------------------------------------------
bool Programmer::Base::read(Device::Memory &memory)
{
  prepare(Read);
  bool ok = doRead(memory);
  endProgramming();
  return ok;
}

bool Programmer::Base::doRead(Device::Memory &memory)
{
  if ( !checkErase() ) return false;
  if ( !doConnectDevice() ) return false;
  log(Log::Information, i18n("Reading device memory..."));
  memory.clear();
  if ( !internalRead(memory) ) return false;
  log(Log::Information, i18n("Reading done"));
  return true;
}

//-----------------------------------------------------------------------------
bool Programmer::Base::program(const Device::Memory &memory)
{
  prepare(Write);
  bool ok = doProgram(memory);
  endProgramming();
  return ok;
}

bool Programmer::Base::doProgram(const Device::Memory &memory)
{
  if ( !checkProgram(memory) ) return false;
  if ( !doConnectDevice() ) return false;
  log(Log::Information, i18n("Programming device memory..."));
  if ( !internalProgram(memory) ) return false;
  log(Log::Information, i18n("Programming successful"));
  if ( group().isDebugger() && !initDebugging() ) {
    log(Log::Error, i18n("Failed to initialize device for debugging."));
    return false;
  }
  return true;
}

//-----------------------------------------------------------------------------
bool Programmer::Base::verify(const Device::Memory &memory)
{
  prepare(Verify);
  bool ok = doVerify(memory);
  endProgramming();
  return ok;
}

bool Programmer::Base::doVerify(const Device::Memory &memory)
{
  if ( !checkVerify() ) return false;
  if ( !doConnectDevice() ) return false;
  log(Log::Information, i18n("Verifying..."));
  if ( !internalVerify(memory, NormalVerify) ) return false;
  log(Log::Information, i18n("Verifying successful"));
  return true;
}

bool Programmer::Base::blankCheck()
{
  prepare(BlankCheck);
  bool ok = doBlankCheck();
  endProgramming();
  return ok;
}

bool Programmer::Base::doBlankCheck()
{
  if ( !checkVerify() ) return false;
  if ( !doConnectDevice() ) return false;
  log(Log::Information, i18n("Blank checking..."));
  Device::Memory *memory = _device->group().createMemory(*_device);
  bool ok = internalVerify(*memory, BlankCheckVerify);
  delete memory;
  if ( !ok ) return false;
  log(Log::Information, i18n("Blank checking successful"));
  return true;
}
