/***************************************************************************
 *   Copyright (C) 2005-2006 Nicolas Hadacek <hadacek@kde.org>             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#ifndef GENERIC_PROGRAMMER_H
#define GENERIC_PROGRAMMER_H

#include <qfile.h>

#include "common/log.h"
#include "port/port_base.h"
#include "devices/base/generic_device.h"
#include "devices/base/generic_memory.h"

namespace Programmer
{
  enum State { NotConnected = 0, Stopped, Running, Halted };
  class Hardware;
  class Base;
  class Group;
  extern const double UNKNOWN_VOLTAGE;

  enum VerifyAction { NormalVerify = 0, BlankCheckVerify = 1, IgnoreProtectedVerify = 2,
                      OnlyProgrammedVerify = 4 };
  Q_DECLARE_FLAGS(VerifyActions, VerifyAction)
  Q_DECLARE_OPERATORS_FOR_FLAGS(VerifyActions)

//-----------------------------------------------------------------------------
class DeviceSpecific : public Log::Client
{
public:
  DeviceSpecific(Programmer::Base &base);
  virtual bool init() = 0;
  virtual void setTargetVddOn(bool on) = 0;

protected:
  Programmer::Base &_base;

  virtual void setPowerOff() = 0;
  virtual void setPowerOn() = 0;
};

//-----------------------------------------------------------------------------
class Hardware : public Log::Client
{
public:
  Hardware(Programmer::Base &base, Port::Base *port);
  virtual ~Hardware();

  Port::Type portType() const { return _port->type(); }
  QString portDevice() const { return _port->device(); }
  bool connectHardware();
  void disconnectHardware();

protected:
  Port::Base       *_port;
  Programmer::Base &_base;

  virtual bool internalConnectHardware() = 0;
};

//-----------------------------------------------------------------------------
class Base : public QObject, public Log::Manager
{
Q_OBJECT
public:
  Base(const Group &group, const Device::Data *data, const char *name);
  virtual ~Base();

  virtual void log(Log::Type type, const QString &message);
  virtual void log(Log::DebugLevel level, const QString &message);
  void init(Hardware *hardware, DeviceSpecific *deviceSpecific);
  const Device::Data *device() const { return _device; }
  const Group &group() const { return _group; }
  Hardware *hardware() { return _hardware; }
  bool simpleConnectHardware();
  bool connectHardware();
  void disconnectHardware();
  bool connectDevice();
  void setTargetVddOn(bool on);
  bool isTargetVddOn() const { return _targetVddOn; }

  const VersionData &firmwareVersion() const { return _firmwareVersion; }
  virtual bool readFirmware() { qFatal("readFirmware not implemented"); return false; }
  bool uploadFirmware(const QString &filepath);

  bool erase();
  bool read(Device::Memory &memory);
  bool program(const Device::Memory &memory);
  bool verify(const Device::Memory &memory);
  bool blankCheck();

  State state() const { return _state; }
  bool isConnected() const { return _state!=NotConnected; }
  bool doConnectDevice();
  virtual bool setBreakpoints(const QValueVector<uint> &) { qFatal("setBreakpoints not implemented"); return false; }
  bool run();
  bool stop();
  virtual bool reset() { qFatal("reset not implemented"); return false; }
  bool step();
  virtual uint pc() const { return Device::INVALID; }
  virtual bool updateState() { qFatal("updateState not implemented"); return false; }
  virtual uint readRegister(uint index) { Q_UNUSED(index); return 0; }
  virtual bool writeRegister(uint index, uint value) { Q_UNUSED(index); Q_UNUSED(value); return false; }
  virtual QString statusString() const { return QString::null; }

signals:
  void actionMessage(const QString &message);

protected:
  Hardware           *_hardware;
  DeviceSpecific     *_specific;
  const Device::Data *_device;
  State               _state;
  bool                _targetVddOn;
  const Group        &_group;
  VersionData         _firmwareVersion;

  virtual void initConnect() {}
  virtual bool internalSetupHardware() = 0;
  virtual bool verifyDeviceId() = 0;
  enum Operation { Read = 0, Write, Verify, Erase, BlankCheck, Connect, Nb_Operations };
  struct OperationData {
    uint nbSteps;
    const char *message;
  };
  static const OperationData OPERATION_DATA[Nb_Operations];
  virtual uint nbSteps(Operation operation) const;
  void prepare(Operation operation);
  virtual bool doUploadFirmware(QFile &) { qFatal("doUploadFirmware not implemented"); return false; }

  virtual bool checkErase() = 0;
  virtual bool internalErase() = 0;
  virtual bool checkRead() = 0;
  virtual bool internalRead(Device::Memory &memory) = 0;
  virtual bool checkProgram(const Device::Memory &memory) = 0;
  virtual bool internalProgram(const Device::Memory &memory) = 0;
  virtual bool checkVerify() = 0;
  virtual bool internalVerify(const Device::Memory &memory, VerifyActions actions) = 0;
  virtual bool doProgram(const Device::Memory &memory);

  virtual bool initDebugging() { return true; }
  virtual bool internalRun() { qFatal("run not implemented"); return false; }
  virtual bool internalStop() { qFatal("stop not implemented"); return false; }
  virtual bool internalStep() { qFatal("step not implemented"); return false; }

  void endProgramming();
  bool doErase();
  bool doRead(Device::Memory &memory);
  bool doVerify(const Device::Memory &memory);
  bool doBlankCheck();

  friend class DeviceSpecific;
};

}; // namespace

#endif
