/***************************************************************************
 *   Copyright (C) 2006 Nicolas Hadacek <hadacek@kde.org>                  *
 *   Copyright (C) 2003-2005 Alain Gibaud <alain.gibaud@free.fr>           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "direct_18F.h"

//-----------------------------------------------------------------------------
bool Direct::P18F::skipMaskWords(Pic::MemoryRangeType type, const Device::Array &data,
                                 uint &i, uint nb, bool forceProgram)
{
  if (forceProgram) return false;
  uint mask = device().range(type).mask;
  for (uint k=0; k<nb; k++)
    if ( data[i+k]!=mask ) return false;
  i += nb;
  return true;
}

void Direct::P18F::program(Type type)
{
  switch (type) {
    case Code: {
      QString cmd = QString("d,C,c,C,c,C,c,Cw%1cw%2X0000,").arg(programHighTime(Code)).arg(programLowTime());
      pulseEngine(cmd.latin1());
      break;
    }
    case Erase: {
      pulseEngine("k0,X0000,"); // NOP
      QString cmd = QString("k0;w%1;w%2;X0000").arg(programHighTime(type)).arg(programLowTime());
      pulseEngine(cmd.latin1());
      break;
    }
    case Eeprom: {
      for (;;) {
        pulseEngine("k0,X50A6,");
        pulseEngine("k0,X6EF5,");
        pulseEngine("k0,X0000,"); // NOP
        uint b = get_byte();
        if ( (b & 0x02)==0 ) break; // WR bit clear
      }
      QString cmd = QString("w%1").arg(programLowTime());
      pulseEngine(cmd.latin1());
      break;
    }
  }
}

void Direct::P18F::setPointer(Pic::MemoryRangeType type, uint offset)
{
  if ( type==Pic::MemoryEeprom ) {
    pulseEngine("k0,S,k0,X6EA9,", 0x0E00 | (offset & 0x00FF));
    pulseEngine("k0,S,k0,X6EAA,", 0x0E00 | ((offset & 0xFF00) >> 8));
  } else {
    uint addr = device().range(type).start + offset;
    pulseEngine("k0,S,k0,X6EF8,", 0x0E00 | ((addr & 0xFF0000) >> 16));
    pulseEngine("k0,S,k0,X6EF7,", 0x0E00 | ((addr & 0x00FF00) >> 8));
    pulseEngine("k0,S,k0,X6EF6,", 0x0E00 | (addr & 0x0000FF));
  }
}

Device::Array Direct::P18F::doRead(Pic::MemoryRangeType type)
{
  uint mask = device().range(type).mask;
  //qDebug("read %s %i", Pic::MEMORY_RANGE_TYPE_DATA[type].label, device().nbWords(type));
  Device::Array data(device().nbWords(type));
  //pulseEngine("w300000"); // what for ?
  switch (type) {
    case Pic::MemoryEeprom:
      pulseEngine("k0,X9EA6,k0,X9CA6");
      for (uint i = 0; i<data.count(); i++) {
        setPointer(type, i);
        pulseEngine("k0,X80A6,k0,X50A8,k0,X6EF5,k0,X0000");
        data[i] = pulseEngine("r") & mask;
      }
      addProgress(data.count());
      break;
    case Pic::MemoryCode:
      setPointer(type, 0);
      //pulseEngine("w1000"); ??
      for (uint i = 0; i<data.count(); i++)
        data[i] = pulseEngine("R") & mask;
      addProgress(data.count());
      break;
    case Pic::MemoryUserId:
    case Pic::MemoryConfig:
    case Pic::MemoryDeviceId:
      setPointer(type, 0);
      for (uint i = 0; i<data.count(); i+=2) {
        uint w = pulseEngine("R");
        data[i] = w & mask;
        data[i+1] = w >> 8;
      }
      break;
    default: Q_ASSERT(false); break;
  }
  return data;
}

bool Direct::P18F::doVerify(Pic::MemoryRangeType type, const Pic::Memory &memory,
                            ::Programmer::VerifyActions actions, const RangeVector &rv)
{
  if ( type!=Pic::MemoryCode ) return P18::doVerify(type, memory, actions, rv);
  Device::Array data = memory.arrayForWriting(type);
  bool only = ( actions & ::Programmer::OnlyProgrammedVerify );
  uint offset = (only ? findNonMaskStart(type, data) : 0);
  uint nbWords = (only ? findNonMaskEnd(type, data)+1 : data.count());
  uint mask = device().range(type).mask;
  // #### TODO: only read blocks...
  setPointer(type, offset);
  for (uint i=0; i<nbWords; i++) {
    uint w = pulseEngine("R");
    if ( !verifyWord(offset+i, w & mask, type, memory, actions, rv) ) return false;
  }
  addProgress(data.count());
  return true;
}

bool Direct::P18F::doWrite(Pic::MemoryRangeType type, const Device::Array &data, bool force)
{
  uint inc = device().addressIncrement(type);
  //qDebug("write %s %i/%i %i", Pic::MEMORY_RANGE_TYPE_DATA[type].label, nbWords, data.count(), inc);
  //pulseEngine("w300000"); ??
  switch (type) {
    case Pic::MemoryUserId: {
      configureSinglePanel();
      pulseEngine("k0,X8EA6,k0,X9CA6,"); // direct access to code memory
      setPointer(type, 0);
      Q_ASSERT( device().nbWords(Pic::MemoryUserId)==8 );
      uint i = 0;
      for (uint k=0; k<4; k++) {
        uint word = data[i] | data[i+1] << 8;
        if ( (k+1)==4 ) pulseEngine("k15,S,", word);
        else pulseEngine("k13,S,", word);
        i += 2;
      }
      program(Code);
      break;
    }
    case Pic::MemoryCode: {
      configureSinglePanel();
      pulseEngine("k0,X8EA6,k0,X9CA6,"); // direct access to code memory
      const Data &d = Direct::data(device().name());
      for (uint i = 0; i<data.count(); ) {
        if ( i!=0 && i%0x100==0 ) addProgress(0x100);
        if ( skipMaskWords(type, data, i, d.progWidth, force) ) continue;
        setPointer(type, inc * i);
        for (uint k=0; k<d.progWidth; k++) {
          if ( (k+1)==d.progWidth ) pulseEngine("k15,S,", data[i]);
          else pulseEngine("k13,S,", data[i]);
          i++;
        }
        program(Code);
      }
      addProgress(data.count()%0x100);
      break;
    }
    case Pic::MemoryConfig:
      pulseEngine("k0,X8EA6,k0,X8CA6,"); // direct access to config memory
      pulseEngine("k0,XEF00,k0,XF800,"); // move PC somewhere else (0x100000)
      for (uint i = 0; i<data.count(); i+=2) {
        setPointer(type, inc * i);
        uint w = data[i] | (data[i+1] << 8);
        pulseEngine("k15,S,", w);
        program(Code);
        pulseEngine("k0,X2AF6,"); // increment address
        pulseEngine("k15,S,", w);
        program(Code);
        pulseEngine("k0,X0000,"); // NOP: some devices need 4 NOPS here...
        pulseEngine("k0,X0000,"); // NOP
        pulseEngine("k0,X0000,"); // NOP
        pulseEngine("k0,X0000,"); // NOP
      }
      break;
    case Pic::MemoryEeprom:
      pulseEngine("k0,X9EA6,k0,X9CA6,"); // direct access to eeprom memory
      for(uint i = 0; i<data.count(); ) {
        if ( i!=0 && i%0x100==0 ) addProgress(0x100);
        if ( skipMaskWords(type, data, i, 1, force) ) continue;
        setPointer(type, inc * i);
        pulseEngine("k0,S,k0,X6EA8,", 0x0E00 | data[i]); // load data
        pulseEngine("k0,X84A6,"); // enable memory writes
        unlockEeprom();
        pulseEngine("k0,X82A6,"); // write
        program(Eeprom);
        pulseEngine("k0,X94A6,"); // disable writes
        i++;
      }
      addProgress(data.count()%0x100);
      break;
    default: Q_ASSERT(false); break;
  }
  return true;
}

void Direct::P18F::unlockEeprom()
{
  pulseEngine("k0,X0E55,k0,X6EA7,k0,X0EAA,k0,X6EA7,");
}

bool Direct::P18F::doEraseCommand(uint cmd1, uint cmd2)
{
  if ( cmd1!=0 ) {
    pulseEngine("k0;X0E3C;k0;X6EF8;k0;X0E00;k0;X6EF7;k0;X0E05;k0;X6EF6"); // load 0x3C0005 address
    pulseEngine("k12;X" + toHex(cmd1, 4));
  }
  pulseEngine("k0;X0E3C;k0;X6EF8;k0;X0E00;k0;X6EF7;k0;X0E04;k0;X6EF6"); // load 0x3C0004 address
  pulseEngine("k12;X" + toHex(cmd2, 4));
  program(Erase);
  return true;
}

::Programmer::Result Direct::P18F::doEraseCode()
{
  doEraseCommand(0, 0x0083); // boot
  for (uint i=0; i<device().config().nbBlocks(); i++)
    doEraseCommand(0, 0x0088 + i);
  return ::Programmer::Ok;
}

::Programmer::Result Direct::P18F::doEraseEeprom()
{
  doEraseCommand(0, 0x0081);
  return ::Programmer::Ok;
}

bool Direct::P18F::doErase(bool)
{
  return doEraseCommand(0, 0x0080);
}

//-----------------------------------------------------------------------------
void Direct::P18F1220::program(Type type)
{
  if ( type==Eeprom ) {
    pulseEngine("k0,X0000,"); // NOP
    QString cmd = QString("k0;w%1;w%2;X0000").arg(programHighTime(type)).arg(programLowTime());
    pulseEngine(cmd.latin1());
  } else P18F::program(type);
}

//-----------------------------------------------------------------------------
void Direct::P18F242::configureSinglePanel()
{
  pulseEngine("k0,X8EA6,k0,X8CA6,"); // direct access to config memory
  pulseEngine("k0,X0E3C,k0,X6EF8,k0,X0E00,k0,X6EF7,k0,X0E06,k0,X6EF6,k12,X0000,"); // configure for single panel writes
}

//-----------------------------------------------------------------------------
::Programmer::Result Direct::P18F2221::doEraseCode()
{
  doEraseCommand(0, 0x8181); // boot
  for (uint i=0; i<device().config().nbBlocks(); i++) {
    uint v = (1 << i);
    doEraseCommand(v + v<<8, 0x8080);
  }
  return ::Programmer::Ok;
}

::Programmer::Result Direct::P18F2221::doEraseEeprom()
{
  doEraseCommand(0, 0x8484);
  return ::Programmer::Ok;
}

bool Direct::P18F2221::doErase(bool)
{
  return doEraseCommand(0x0F0F, 0x8787); // chip erase
}

//-----------------------------------------------------------------------------
bool Direct::P18F6527::doErase(bool)
{
  return doEraseCommand(0xFFFF, 0x8787); // chip erase
}
