/***************************************************************************
 *   Copyright (C) 2005-2006 Nicolas Hadacek <hadacek@kde.org>             *
 *   Copyright (C) 2002-2003 Stephen Landamore <stephen@landamore.com>     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "icd2.h"

#include <klocale.h>

#include "common/misc.h"
#include "port/port_base.h"
#include "common/about.h"

//-----------------------------------------------------------------------------
const char * const Icd2::VOLTAGE_LABELS[Nb_VoltageTypes] = {
  I18N_NOOP("ICD2 Vpp"), I18N_NOOP("Target Vdd"), I18N_NOOP("Target Vpp")
};

const Icd2::TargetModeData Icd2::TARGET_MODE_DATA[Nb_TargetModes] = {
  { I18N_NOOP("Stopped"),  0x00 },
  { I18N_NOOP("Running"),  0x01 },
  { I18N_NOOP("In Reset"), 0x02 }
};

const Icd2::ResetModeData Icd2::RESET_MODE_DATA[Pic::Nb_ResetModes] = {
  { I18N_NOOP("Reset Held"),     "00" },
  { I18N_NOOP("Reset Released"), "01" }
};

//-----------------------------------------------------------------------------
const Icd2::Hardware::VoltageTypeData Icd2::Hardware::VOLTAGE_TYPE_DATA[Nb_VoltageTypes] = {
 { "37", 0.07988 }, // icd Vpp
 { "34", 0.03908 }, // target Vdd
 { "35", 0.07988 }, // target Vpp
};

const char * const Icd2::Hardware::WRITE_MODE_VALUES[Nb_WriteModes] = {
  "00", // write only
  "01"  // erase then write
};

const Icd2::TestData::Data Icd2::TestData::RESULT_DATA[Nb_ResultTypes] = {
  { I18N_NOOP("Pass"), 0x00 },
  { I18N_NOOP("Low"),  0x01 },
  { I18N_NOOP("High"), 0x80 }
};
const char * const Icd2::TestData::VOLTAGE_LABELS[Nb_VoltageTypes] = {
  I18N_NOOP("Target Vdd"), I18N_NOOP("Module Vpp"), I18N_NOOP("MCLR ground"),
  I18N_NOOP("MCLR Vdd"), I18N_NOOP("MCLR Vpp")
};

Icd2::TestData::TestData()
{
  for (uint k=0; k<Nb_VoltageTypes; k++)
    _voltages[k] = -1;
}

Icd2::TestData::TestData(const QString &rx)
{
  for (uint k=0; k<Nb_VoltageTypes; k++)
    _voltages[k] = fromHex(rx.mid(5 + 2*k, 2));
}

bool Icd2::TestData::pass() const
{
  for (uint k=0; k<Nb_VoltageTypes; k++)
    if ( _voltages[k]!=RESULT_DATA[Pass].value ) return false;
  return true;
}

QString Icd2::TestData::result(VoltageType type) const
{
  if ( _voltages[type]==-1 ) return "--";
  for (uint i=0; i<Nb_ResultTypes; i++)
    if ( _voltages[type]==RESULT_DATA[i].value ) return i18n(RESULT_DATA[i].label);
  return toHex(_voltages[type], 2);
}

QString Icd2::TestData::pretty(VoltageType type) const
{
  return i18n(VOLTAGE_LABELS[type]) + "=" + result(type);
}

//-----------------------------------------------------------------------------
const char *Icd2::Hardware::readCommand(Pic::MemoryRangeType type) const
{
  switch (type) {
    case Pic::MemoryCode:             return "47";
    case Pic::MemoryEeprom:           return "48";
    case Pic::MemoryConfig:           return "49";
    case Pic::MemoryUserId:           return "4A";
    case Pic::MemoryDeviceId:
      if ( device().architecture()==Pic::Arch_30X ) return "49";
      return "4A";
    case Pic::MemoryCal:
    case Pic::MemoryCalBackup:
      if ( device().architecture()==Pic::Arch_16X ) return "49"; // ?
      return "47"; // for baseline only ?
    case Pic::MemoryDebugVector:      return "40"; // for 16F only ?
    case Pic::MemoryProgramExecutive:
    case Pic::MemoryHardwareStack:
    case Pic::Nb_MemoryRangeTypes:    break;
  }
  Q_ASSERT(false);
  return 0;
}

const char *Icd2::Hardware::writeCommand(Pic::MemoryRangeType type) const
{
  switch (type) {
    case Pic::MemoryCode:             return "43";
    case Pic::MemoryEeprom:           return "44";
    case Pic::MemoryConfig:           return "45";
    case Pic::MemoryUserId:           return "46";
    case Pic::MemoryDeviceId:         break;
    case Pic::MemoryCal:
    case Pic::MemoryCalBackup:
      if ( device().architecture()==Pic::Arch_16X ) return "45"; // ?
      return "43"; // for baseline only ?
    case Pic::MemoryDebugVector:      return "41"; // for 16F only ?
    case Pic::MemoryProgramExecutive:
    case Pic::MemoryHardwareStack:
    case Pic::Nb_MemoryRangeTypes:    break;
  }
  Q_ASSERT(false);
  return 0;
}

bool Icd2::Hardware::internalConnectHardware()
{
  return internalConnect("U");
}

bool Icd2::Hardware::uploadFirmware(const Pic::Memory &memory)
{
  if ( !internalConnect("V") ) return false;
  log(Log::Information, "  Uploading firmware to ICD2...");
  uint start = 0x0004, size = 0x1BFB;
  if ( !command("10" + toHex(start, 4) + toHex(size, 4)) ) return false;
  uint nbBytesWord = memory.device().nbBytesWord(Pic::MemoryCode); // should be 2 for 16F876
  Device::Array data = memory.arrayForWriting(Pic::MemoryCode);
  if ( !writeBlock(nbBytesWord, data, start, size) ) return false;
  if ( !receiveResponse() ) return false;
  if ( !internalConnect("U") ) return false;
  return true;
}

bool Icd2::Hardware::setTarget()
{
  // set target family
  const Data &d = data(device().name());
  if ( !command(QCString("2A") + toHex(d.famid, 2)) ) return false;
  // set code range end
  uint end = device().range(Pic::MemoryCode).end;
  if ( device().range(Pic::MemoryCal).start==end ) end++;
  if ( !command("06" + toHex(end, 6)) ) return false;
  return true;
}

bool Icd2::Hardware::setup()
{
  // ??
  if ( device().architecture()==Pic::Arch_30X )
    if ( !command("0900") ) return false;

  // ??
  _port->send("$7F00\x0D", 6);
  QCString s = _port->receive(2);
  if ( s!="02" ) log(Log::Warning, i18n("Unexpected answer ($7F00) from ICD2 (%1).").arg(s.data()));

  // ??
  if ( !command("08") ) return false;
  if ( _rx.mid(5, 2)!="00" ) log(Log::Warning, i18n("Unexpected answer (08) from ICD2 (%1).").arg(_rx.data()));

  return !hasError();
}

bool Icd2::Hardware::sendCommand(const QCString &s)
{
  QCString cs = s.upper();
  //format: <LL.....CC>
  QCString tx = "<";
  tx += toHex(cs.length() + 6, 2);
  tx += cs;
  uchar chk = tx[1] + tx[2];
  for (uint i=0; i<s.length(); i++)
    chk += cs[i];
  tx += toHex(chk, 2);
  tx += '>';
  log(Log::ExtraDebug, QString("send command: '%1'").arg(tx));
  return _port->send(tx.data(), tx.length());
}

bool Icd2::Hardware::receiveResponse()
{
  // <LL...CC>
  _rx = _port->receive(3); // receive the first 2 bytes (i.e. the length)
  if ( _rx.length()!=3 ) {
    log(Log::Error, i18n("Received string too short."));
    return false;
  }
  uint len = fromHex(_rx.mid(1, 2));
  if ( len<3 ) {
    log(Log::Error, i18n("Received length too short."));
    return false;
  }
  _rx += _port->receive(len-3);
  log(Log::ExtraDebug, QString("received answer: '%1'").arg(_rx));
  if ( _rx.length()!=len ) {
    log(Log::Error, i18n("Received string too short."));
    return false;
  }
  if ( _rx[0]!='[' || _rx[len-1]!=']' ) {
    log(Log::Error, i18n("Malformed string received \"%1\"").arg(_rx));
    return false;
  }
  // verify the checksum
  uchar chk = 0;
  for (uint i=1; i<len-3; i++)
    chk += _rx[i];
  if ( chk!=fromHex(_rx.mid(len-3, 2)) ) {
    log(Log::Error, i18n("Bad checksum for received string"));
    return false;
  }
  return true;
}

bool Icd2::Hardware::command(const QCString &command)
{
  if ( hasError() ) return false;
  if ( !sendCommand(command) ) return false;
  if ( !receiveResponse() ) return false;
  if ( command.mid(0, 2)!=_rx.mid(3, 2) ) {
    log(Log::Error, i18n("Wrong return value (\"%1\"; was expecting \"%2\")")
             .arg(_rx.mid(3, 2)).arg(command.mid(0, 2)));
    return false;
  }
  return true;
}

VersionData Icd2::Hardware::getFirmwareVersion()
{
  if ( !command("01") ) return VersionData();
  return VersionData::fromHexString(_rx.mid(5, 6));
}

uint Icd2::Hardware::getFirmwareId()
{
  if ( !command("07") ) return 0;
  return fromHex(_rx.mid(5, 2));
}

VersionData Icd2::Hardware::getDebugExecVersion()
{
  if ( !command("04") ) return VersionData();
  return VersionData::fromHexString(_rx.mid(5, 6));
}

bool Icd2::Hardware::setTargetSelfPowered(bool selfPowered)
{
  return command(QCString("05") + (selfPowered ? "00" : "FF"));
}

double Icd2::Hardware::getVoltage(VoltageType type)
{
  if ( !command(VOLTAGE_TYPE_DATA[type].command) ) return 0.0;
  return VOLTAGE_TYPE_DATA[type].factor * fromHex(_rx.mid(5, 2));
}

Icd2::TargetMode Icd2::Hardware::getTargetMode()
{
  if ( !command("2C") ) return Nb_TargetModes;
  uchar mode = fromHex(_rx.mid(5, 2));
  for (uint i=0; i<Nb_TargetModes; i++)
    if ( mode==TARGET_MODE_DATA[i].value ) return TargetMode(i);
  Q_ASSERT(false);
  return Nb_TargetModes;
}

bool Icd2::Hardware::setTargetReset(Pic::ResetMode mode)
{
  return command(QCString("33") + RESET_MODE_DATA[mode].value);
}

Icd2::TestData Icd2::Hardware::selfTest()
{
  if ( !command("02") ) return TestData();
  return TestData(_rx);
}

Device::Array Icd2::Hardware::readBlock(uint nbBytesWord, uint nbWords)
{
  //qDebug("readBlock %i %s", nbBytesWord, toHex(nbWords, 8).data());
  // receive data
  uint length = 2*nbBytesWord*nbWords+4;
  QCString s;
  uint i = 0;
  while ( i<length ) {
    uint maxSize = (_port->type()==Port::SerialType ? 0x1 : 0x100);
    uint size = QMIN(maxSize, length-i);
    s += _port->receive(size);
    if ( hasError() ) return Device::Array();
    i += size;
  }

  // treat data
  if ( s[0]!='{' && s[s.length()-1]!='}' ) {
    log(Log::Error, i18n("Invalid begin/end character for read block."));
    return Device::Array();
  }
  log(Log::MaxDebug, "received: " + s);
  Device::Array data(nbWords);
  Q_UINT8 chk = 0;
  for (uint i=0; i<nbWords; i++) {
    QCString ts = s.mid(1+2*nbBytesWord*i, 2*nbBytesWord);
    //if ( i<10 ) qDebug("%i: %s", i, ts.data());
    data[i] = 0;
    for (int k=nbBytesWord-1; k>=0; k--) {
      data[i] = data[i] << 8;
      data[i] |= fromHex(ts.mid(2*k, 2));
      chk += ts[2*k] + ts[2*k+1];
    }
  }

  QCString cs = s.mid(s.length()-3, 2);
  if ( chk!=fromHex(cs) ) {
    log(Log::Error, i18n("Bad checksum for read block: %1 (%2 expected).").arg(cs).arg(toHex(chk, 2)));
    return Device::Array();
  }
  return data;
}

Device::Array Icd2::Hardware::readMemory(Pic::MemoryRangeType type, uint wordOffset, uint nbWords)
{
  const char *r = readCommand(type);
  if ( !device().isPresent(type) || r==0 ) return Device::Array();
  uint nbBytesWord = device().nbBytesWord(type);
  uint div = 2;
  if ( type==Pic::MemoryEeprom || nbBytesWord>=2 ) div = 1;
  uint inc = device().addressIncrement(type);
  uint start = device().range(type).start; // address
  uint todo = inc * nbWords; // address
  uint offset = inc * wordOffset; // address
  //qDebug("read size=%s div=%i nbBytes=%i", toHex(size, 8).data(), div, nbBytesWord);
  Device::Array data;
  do {
    uint size = QMIN(todo, 0x1000); // addresses
    uint nb = size / inc; // word
    //qDebug("read %s start=%s size=%s", Pic::MEMORY_RANGE_TYPE_DATA[type].label, toHex(start+offset, 8).data(), toHex(nb, 8).data());
    QCString com = r + toHex(start+offset, 8) + toHex(nb/div, 8);
    if ( !command(com) ) return Device::Array();
    data += readBlock(nbBytesWord, nb);
    if ( data.count()==0 ) return Device::Array();
    if ( !receiveResponse() ) return false;
    offset += size;
    todo -= size;
    if ( type==Pic::MemoryCode || type==Pic::MemoryEeprom ) addProgress(nb);
  } while ( todo!=0 );
  return data;
}

bool Icd2::Hardware::writeBlock(uint nbBytesWord, const Device::Array &data,
                                uint wordOffset, uint nbWords)
{
  //qDebug("writeBlock offset:%s nbWords:%s (size: %s)", toHex(wordOffset, 4).data(), toHex(nbWords, 4).data(), toHex(data.size(), 4).data());
  Q_ASSERT( wordOffset+nbWords<=data.size() );
  // prepare data
  QCString s;
  s += "{";
  uchar chk = 0;
  for (uint i=0; i<nbWords; i++) {
    QCString ts = toHex(data[wordOffset+i], 2*nbBytesWord);
    for (int k=nbBytesWord-1; k>=0; k--) {
      //if ( i<10 || i>=nbWords-10 ) qDebug("send: %i-%i %s", i, k, ts.mid(2*k, 2).data());
      s += ts.mid(2*k, 2);
      chk += ts[2*k] + ts[2*k+1];
    }
  }
  s += toHex(chk, 2);
  s += "}";
  log(Log::MaxDebug, "send: " + s);

  // send data
  uint i = 0;
  while ( i<s.length() ) {
    uint maxSize = (_port->type()==Port::SerialType ? 0x1 : 0x100);
    uint size = QMIN(maxSize, s.length()-i);
    if ( !_port->send(s.data()+i, size) ) return false;
    i += size;
  }

  //qDebug("done sending %i words (chk=%s)", nbWords, toHex(chk, 2).data());
  return true;
}

bool Icd2::Hardware::writeMemory(Pic::MemoryRangeType type, const Device::Array &data,
                       uint wordOffset, uint nbWords)
{
  //qDebug("write memory: offset:%s nbWords:%s (size: %s)", toHex(wordOffset, 4).data(), toHex(nbWords, 4).data(), toHex(data.size(), 4).data());
  const char *w = writeCommand(type);
  if ( !device().isPresent(type) || w==0 ) return true;
  uint nbBytesWord = device().nbBytesWord(type);
  uint div = 2;
  if ( type==Pic::MemoryEeprom || nbBytesWord>=2 ) div = 1;
  uint inc = device().addressIncrement(type);
  uint start = device().range(type).start; // address
  uint todo = inc * nbWords; // address
  uint offset = inc * wordOffset; // address
  //qDebug("write todo=%s div=%i nbBytes=%i dataSize=%i", toHex(todo, 8).data(), div, nbBytesWord, data.size());
  do {
    uint size = QMIN(todo, 0x1000); // address
    uint nb = size / inc; // word
    //qDebug("write %s start=%s nbWords=%s", Pic::MEMORY_RANGE_TYPE_DATA[type].label, toHex(start+offset, 8).data(), toHex(nb, 8).data());
    QCString com = w + toHex(start+offset, 8) + toHex(nb/div, 8);
    if ( !command(com) ) return false;
    if ( !writeBlock(nbBytesWord, data, offset/inc, nb) ) return false;
    if ( !receiveResponse() ) return false;
    offset += size;
    todo -= size;
    if ( type==Pic::MemoryCode || type==Pic::MemoryEeprom ) addProgress(nb);
  } while ( todo!=0 );
  return true;
}

bool Icd2::Hardware::eraseAll()
{
  setTargetReset(Pic::ResetHeld);
  if ( !command("29") ) return false;  // erase (this should take care of the config)
  return true;
}

bool Icd2::Hardware::haltRun()
{
  return command("2E");
}

bool Icd2::Hardware::step()
{
  return command("2F");
}

bool Icd2::Hardware::resumeRun()
{
  return command("30");
}

bool Icd2::Hardware::setWriteMode(WriteMode mode)
{
  return command(QCString("4B") + WRITE_MODE_VALUES[mode]);
}

bool Icd2::Hardware::writeRegister(uint address, uint value, uint nbBytes)
{
  Q_ASSERT( nbBytes==1 || nbBytes==2 );
  if ( !command("1B" + toHex(address, 8) + toHex(nbBytes, 8)) ) return false;
  Device::Array data(nbBytes);
  if ( nbBytes==2 ) {
    data[0] = value >> 8;
    data[1] = value & 0xFF;
  } else data[0] = value;
  if ( !writeBlock(1, data, 0, nbBytes) ) return false;
  return receiveResponse();
}

uint Icd2::Hardware::readRegister(uint address, uint nbBytes)
{
  Q_ASSERT( nbBytes==1 || nbBytes==2 );
  if ( !command("1E" + toHex(address, 8) + toHex(nbBytes, 8)) ) return 0;
  Device::Array data = readBlock(1, nbBytes);
  if ( hasError() ) return 0;
  if ( !receiveResponse() ) return 0;
  if ( nbBytes==2 ) return ( data[0] << 8 | data[1] );
  return data[0];
}

uint Icd2::Hardware::getProgramCounter()
{
  if ( !command("3D") ) return 0;
  return fromHex(_rx.mid(5, 8));
}

//-----------------------------------------------------------------------------
Device::Array Icd2::DeviceSpecific::doRead(Pic::MemoryRangeType type)
{
  return hardware().readMemory(type, 0, device().nbWords(type));
}

bool Icd2::DeviceSpecific::doWrite(Pic::MemoryRangeType type, const Device::Array &data, bool force)
{
  uint wordOffset = (force ? 0 : findNonMaskStart(type, data));
  uint nbWords = 0;
  if ( wordOffset!=data.count() ) {
    uint end = (force ? data.count()-1 : findNonMaskEnd(type, data));
    nbWords = end - wordOffset + 1;
    log(Log::NormalDebug, QString("    start=%1 nbWords=%2 total=%3 force=%4")
      .arg(toHexLabel(wordOffset, device().nbCharsAddress())).arg(toHexLabel(nbWords, device().nbCharsAddress()))
      .arg(toHexLabel(data.count(), device().nbCharsAddress())).arg(force ? "true" : "false"));
    if (!hardware().writeMemory(type, data, wordOffset, nbWords) ) return false;
  }
  if ( type==Pic::MemoryCode || type==Pic::MemoryEeprom ) addProgress(data.count()-nbWords);
  return true;
}

bool Icd2::DeviceSpecific::doErase(bool)
{
  Device::Array data(device().nbWords(Pic::MemoryConfig));
  for (uint i=0; i<data.size(); i++) data[i] = device().config()._words[i].wmask;
  if ( !doWrite(Pic::MemoryConfig, data, true) ) return false;
  return hardware().eraseAll();
}
