/***************************************************************************
 *   Copyright (C) 2005 Nicolas Hadacek <hadacek@kde.org>                  *
 *   Copyright (C) 2002-2003 Stephen Landamore <stephen@landamore.com>     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#ifndef ICD2_H
#define ICD2_H

#include "microchip.h"
#include "devices/pic/pic_prog.h"
#include "devices/pic/pic_memory.h"

namespace Icd2
{
  struct FirmwareData {
    uchar major, mid, minor;
  };
  enum { Nb_Firmwares = 0x0E };
  extern const FirmwareData FIRMWARE_DATA[Nb_Firmwares];
  struct FamilyData {
    const char *name;
    uchar efid;
    const char *debugExec;
  };
  extern const FamilyData FAMILY_DATA[];
  struct Data {
    uint famid;
  };
  extern const Data &data(const QString &device);
  extern uint family(const QString &device);

//-----------------------------------------------------------------------------
  enum VoltageType { IcdVpp = 0, TargetVdd, TargetVpp, Nb_VoltageTypes };
  extern const char * const VOLTAGE_LABELS[Nb_VoltageTypes];

  enum TargetMode { TargetStopped = 0, TargetRunning, TargetInReset, Nb_TargetModes};
  struct TargetModeData {
    const char *label;
    uchar value;
  };
  extern const TargetModeData TARGET_MODE_DATA[Nb_TargetModes];

  struct ResetModeData {
    const char *label, *value;
  };
  extern const ResetModeData RESET_MODE_DATA[Pic::Nb_ResetModes];

  class TestData {
  public:
    TestData();
    TestData(const QString &rx);
    enum VoltageType { TargetVdd = 0, ModuleVpp, MclrGround, MclrVdd, MclrVpp,
                       Nb_VoltageTypes };
    bool pass() const;
    QString result(VoltageType type) const;
    QString pretty(VoltageType type) const;
    static const char * const VOLTAGE_LABELS[Nb_VoltageTypes];

  private:
    int _voltages[Nb_VoltageTypes];
    enum ResultType { Pass = 0, Low, High, Nb_ResultTypes };
    struct Data {
      const char *label;
      uchar value;
    };
    static const Data RESULT_DATA[Nb_ResultTypes];
  };

  enum WriteMode { WriteOnlyMode = 0, EraseWriteMode, Nb_WriteModes };

//-----------------------------------------------------------------------------
class Hardware : public ::Programmer::PicHardware
{
public:
  Hardware(::Programmer::Base &base, Port::Base *port) : ::Programmer::PicHardware(base, port) {}

// initialization
  bool uploadFirmware(const Pic::Memory &memory);
  bool setTarget();
  bool setup();

// status
  VersionData getFirmwareVersion();
  uint getFirmwareId();
  VersionData getDebugExecVersion();
  bool setTargetSelfPowered(bool selfPowered);
  double getVoltage(VoltageType type);
  TargetMode getTargetMode();
  bool setTargetReset(Pic::ResetMode mode);
  TestData selfTest();

// programming
  Device::Array readMemory(Pic::MemoryRangeType type, uint wordOffset, uint nbWords);
  bool writeMemory(Pic::MemoryRangeType type, const Device::Array &data, uint wordOffset, uint nbWords);
  bool eraseAll();
  bool setWriteMode(WriteMode mode);

// debugging
  uint readRegister(uint address, uint nbBytes);
  bool writeRegister(uint address, uint value, uint nbBytes);
  bool resumeRun();
  bool step();
  bool haltRun();
  uint getProgramCounter();

protected:
  bool command(const QCString &command);
  virtual bool internalConnect(const QCString &mode) = 0;

private:
  QCString _rx;

  struct VoltageTypeData {
    const char *command;
    double factor;
  };
  static const VoltageTypeData VOLTAGE_TYPE_DATA[Nb_VoltageTypes];
  static const char * const WRITE_MODE_VALUES[Nb_WriteModes];

  virtual bool internalConnectHardware();
  bool sendCommand(const QCString &s);
  bool receiveResponse();
  QCString receivedData() const { return _rx.mid(5, _rx.length()-8); }
  Device::Array readBlock(uint nbBytesWord, uint nbWords);
  bool writeBlock(uint nbBytesWord, const Device::Array &data, uint wordOffset, uint nbWords);
  const char *readCommand(Pic::MemoryRangeType type) const;
  const char *writeCommand(Pic::MemoryRangeType type) const;
};

//-----------------------------------------------------------------------------
class DeviceSpecific : public ::Programmer::PicDeviceSpecific
{
public:
  DeviceSpecific(::Programmer::Base &base) : ::Programmer::PicDeviceSpecific(base) {}
  Hardware &hardware() { return static_cast<Hardware &>(*_base.hardware()); }
  virtual bool init() { return hardware().setTarget(); }
  virtual void setPowerOff() {}
  virtual void setPowerOn() {}
  virtual void setTargetVddOn(bool on) { hardware().setTargetSelfPowered(!on); }
  virtual bool doErase(bool);
  virtual Device::Array doRead(Pic::MemoryRangeType type);
  virtual bool doWrite(Pic::MemoryRangeType type, const Device::Array &data, bool force);
};

}; // namespace

#endif
