/***************************************************************************
 *   Copyright (C) 2005-2006 Nicolas Hadacek <hadacek@kde.org>             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "icd2_debug.h"

#include <qdir.h>

#include "progs/base/prog_config.h"

//-----------------------------------------------------------------------------
const uint Icd2::Debugger::NOP = 0x0;

Icd2::Debugger::Debugger(const ::Programmer::Group &group, const Pic::Data *data)
  : Icd2::ProgrammerBase(group, data, "icd2_debugger")
{}

bool Icd2::Debugger::internalSetupHardware()
{
  if ( _specific==0 ) return true;

  // find debug executive file
  QDir dir(::Programmer::Config::readFirmwareDirectory(group()));
  if ( !dir.exists() ) {
    log(Log::Error, i18n("Firmware directory not configured."));
    return false;
  }
  const FamilyData &fdata = FAMILY_DATA[family(device()->name())];
  QString filepath = dir.filePath(QString("de") + fdata.debugExec + ".hex");
  log(Log::NormalDebug, QString("  Debug executive file: %1").arg(filepath));
  QFile file(filepath);
  if ( !file.exists() ) {
    log(Log::Error, i18n("Could not find debug executive file \"%1\".").arg(file.name()));
    return false;
  }
  // upload hex file
  if ( !file.open(IO_ReadOnly) ) {
    log(Log::Error, i18n("Could not open firmware file \"%1\".").arg(file.name()));
    return false;
  }
  QString error;
  QStringList warnings;
  Pic::Memory::WarningTypes warningTypes;
  Pic::Memory memory(*device());
  if ( !memory.load(file, error, warningTypes, warnings) ) {
    log(Log::Error, i18n("Could not read debug executive file \"%1\" (%2).").arg(file.name()).arg(error));
    return false;
  }
  if ( warningTypes!=Pic::Memory::NoWarning ) {
    log(Log::Error, i18n("Debug executive file seems incompatible with current device:\n%1.").arg(warnings.join("\n")));
    return false;
  }
  _deArray = memory.arrayForWriting(Pic::MemoryCode);
  _deStart = specific()->findNonMaskStart(Pic::MemoryCode, _deArray);
  _deEnd = specific()->findNonMaskEnd(Pic::MemoryCode, _deArray);

  return Icd2::ProgrammerBase::internalSetupHardware();
}

uint Icd2::Debugger::gotoInstruction(uint address) const
{
  // #### only valid for mid-range devices
  Q_ASSERT( address<(1 << 11) );
  return 0x2800 | address;
}

Pic::Memory Icd2::Debugger::toDebugMemory(const Pic::Memory &mem, bool withDebugExecutive)
{
  Pic::Memory memory = mem;
  memory.setDebugOn(true);
  if ( memory.hasWatchdogTimerOn() ) {
    log(Log::Warning, i18n("Disabling watchdog timer for debugging"));
    memory.setWatchdogTimerOn(false);
  }
  memory.setWord(Pic::MemoryDebugVector, 0, gotoInstruction(_deStart));
  if (withDebugExecutive) {
    for (uint i=_deStart; i<=_deEnd; i++)
      memory.setWord(Pic::MemoryCode, i, _deArray[i]);
  }
  return memory;
}

bool Icd2::Debugger::writeDebugExecutive()
{
  log(Log::Information, i18n("  Write debug executive"));
  uint nbWords = _deEnd - _deStart + 1;
  if ( !hardware().writeMemory(Pic::MemoryCode, _deArray, _deStart, nbWords) ) return false;
  log(Log::Information, i18n("  Verify debug executive"));
  Device::Array data = hardware().readMemory(Pic::MemoryCode, _deStart, nbWords);
  for (uint i=0; i<nbWords; i++) {
    if ( _deArray[_deStart+i]==data[i] ) continue;
    uint inc = device()->addressIncrement(Pic::MemoryCode);
    uint address = device()->range(Pic::MemoryCode).start + inc * (_deStart + i);
    log(Log::Error, i18n("Device memory doesn't match debug executive (at address 0x%2: reading 0x%3 and expecting 0x%4).")
             .arg(toHex(address, device()->nbCharsAddress()))
             .arg(toHex(data[i], device()->nbCharsWord(Pic::MemoryCode))).arg(toHex(_deArray[_deStart+i], device()->nbCharsWord(Pic::MemoryCode))));
    return false;
  }
  return true;
}

bool Icd2::Debugger::programAll(const Pic::Memory &mem)
{
  Pic::Memory memory = toDebugMemory(mem, false);
  if ( !programAndVerifyRange(Pic::MemoryCode, memory) ) return false;
  if ( !writeDebugExecutive() ) return false;
  if ( !programAndVerifyRange(Pic::MemoryDebugVector, memory) ) return false;
  if ( !programAndVerifyRange(Pic::MemoryEeprom, memory) ) return false;
  if ( !programAndVerifyRange(Pic::MemoryUserId, memory) ) return false;
  if ( !programAndVerifyRange(Pic::MemoryConfig, memory) ) return false;
  return true;
}

bool Icd2::Debugger::verifyAll(const Pic::Memory &mem, ::Programmer::VerifyActions actions)
{
  if ( (actions & ::Programmer::BlankCheckVerify) )
    return Icd2::ProgrammerBase::verifyAll(mem, actions);
  Pic::Memory memory = toDebugMemory(mem, true);
  if ( !Icd2::ProgrammerBase::verifyAll(memory, actions) ) return false;
  if ( !verifyRange(Pic::MemoryDebugVector, memory, actions) ) return false;
  return true;
}

bool Icd2::Debugger::waitForTargetMode(TargetMode mode)
{
  TargetMode rmode;
  for (uint i=0; i<20; i++) {
    rmode = hardware().getTargetMode();
    if ( rmode==mode ) return true;
    Port::msleep(200);
  }
  log(Log::Error, QString("Timeout waiting for mode: %1 (target is in mode: %2).")
                  .arg(TARGET_MODE_DATA[mode].label).arg(TARGET_MODE_DATA[rmode].label));
  return false;
}

bool Icd2::Debugger::initDebugging()
{
  _state = ::Programmer::Stopped;
  log(Log::Information, i18n("Setting up debugging session."));

  Device::Array saved = hardware().readMemory(Pic::MemoryCode, 0, 1); // save instruction
  if ( hasError() ) return false;
  if ( saved[0]!=NOP ) log(Log::Warning, i18n("  According to ICD2 manual, instruction at address 0x0 should be \"nop\"."));

  // this seems to be needed...
  hardware().setTargetReset(Pic::ResetReleased);
  (void)hardware().getTargetMode();
  hardware().setTargetReset(Pic::ResetHeld);

  log(Log::NormalDebug, QString("  Target Vdd: %1V").arg(hardware().getVoltage(TargetVdd)));
  hardware().setWriteMode(EraseWriteMode);

  // #### this may only be valid for 16F family !!!
  log(Log::NormalDebug, "  Try to run target with loop at reset vector.");
  Device::Array data(1);
  data[0] = gotoInstruction(0x0); // loop at reset vector
  if ( !hardware().writeMemory(Pic::MemoryCode, data, 0, 1) ) return false;
  hardware().setTargetReset(Pic::ResetReleased); // run device
  if ( !waitForTargetMode(TargetRunning) ) return false;
  hardware().haltRun();
  if ( !waitForTargetMode(TargetStopped) ) return false;
  VersionData vd = hardware().getDebugExecVersion();
  log(Log::Information, i18n("  Debug executive is version %1").arg(vd.pretty()));

  log(Log::NormalDebug, "  Set break point at reset vector and run.");
  setBreakPoint(0x0);
  hardware().setTargetReset(Pic::ResetHeld);
  if ( !waitForTargetMode(TargetInReset) ) return false;
  if ( !hardware().writeMemory(Pic::MemoryCode, saved, 0, 1) ) return false; // restore instruction
  hardware().setTargetReset(Pic::ResetReleased); // run device
  if ( !waitForTargetMode(TargetStopped) ) return false;
  if ( readBreakPoint()!=0x1 ) return false;
  setBreakPoint(0x0);

  hardware().setWriteMode(WriteOnlyMode);
  updateState();
  if ( _pc!=0x1 ) return false;
  log(Log::Information, i18n("Ready to start debugging."));
  _state = ::Programmer::Halted;
  return true;
}

bool Icd2::Debugger::setBreakPoint(uint address)
{
  uint v = address | 0x8000; // read-only INBUG bit...
  return hardware().writeRegister(0x18E, v, 2);
}

uint Icd2::Debugger::readBreakPoint()
{
  return hardware().readRegister(0x18E, 2) & 0x1FFF;
}

bool Icd2::Debugger::updateState()
{
  if ( !ProgrammerBase::updateState() ) return false;
  if ( _state!=::Programmer::Halted ) {
    _pc = Device::INVALID;
    _status = Device::INVALID;
    _wreg = Device::INVALID;
  } else {
    _pc = hardware().getProgramCounter() & 0x1FFF;
    if ( hasError() ) return false;
    _status = hardware().readRegister(0x1F0, 1);
    if ( hasError() ) return false;
    _wreg = hardware().readRegister(0x1BF, 1);
    if ( hasError() ) return false;
  }
  return true;
}

bool Icd2::Debugger::setBreakpoints(const QValueVector<uint> &addresses)
{
  if ( addresses.count()==0 ) return setBreakPoint(0x1FFF); // set breakpoint outside
  log(Log::NormalDebug, QString("Set breakpoint at %1").arg(toHexLabel(addresses[0], device()->nbCharsAddress())));
  return setBreakPoint(addresses[0]);
}

bool Icd2::Debugger::internalRun()
{
  _state = ::Programmer::Running;
  return hardware().resumeRun();
}

bool Icd2::Debugger::internalStop()
{
  if ( !hardware().haltRun() ) return false;
  if ( !waitForTargetMode(TargetStopped) ) {
    log(Log::Warning, i18n("Failed to halt target: try a reset."));
    return reset();
  }
  log(Log::Information, QString("Halted at %1").arg(toHexLabel(_pc, device()->nbCharsAddress())));
  _state = ::Programmer::Halted;
  return updateState();
}

bool Icd2::Debugger::reset()
{
  if ( !hardware().setTargetReset(Pic::ResetHeld) ) return false;
  return initDebugging();
}

uint Icd2::Debugger::pc() const
{
  if ( _state==::Programmer::Halted ) return _pc;
  return Device::INVALID;
}

uint Icd2::Debugger::wreg() const
{
  if ( _state==::Programmer::Halted ) return _wreg;
  return Device::INVALID;
}
void Icd2::Debugger::setWreg(uint value)
{
  hardware().writeRegister(0x1BF, value, 1);
}

bool Icd2::Debugger::internalStep()
{
  if ( !hardware().step() ) return false;
  _state = ::Programmer::Halted;
  return updateState();
}

uint Icd2::Debugger::readRegister(uint i)
{
  QString name = device()->registerName(i);
  if ( name=="PCL" ) return _pc & 0xFF;
  if ( name=="PCLATH" ) return _pc >> 8;
  uint address = i * device()->nbBytesRegister();
  if ( name=="FSR" ) address = 0x1BE;
  return hardware().readRegister(address, 1);
}

bool Icd2::Debugger::writeRegister(uint i, uint value)
{
  QString name = device()->registerName(i);
  uint address = i * device()->nbBytesRegister();
  if ( name=="FSR" ) address = 0x1BE;
  if ( !hardware().writeRegister(address, value, 1) ) return false;
  if ( name=="STATUS" && !updateState() ) return false;
  return true;
}

QString Icd2::Debugger::statusString() const
{
  if ( _state!=::Programmer::Halted ) return QString::null;
  uint bank = (_status & 0x20 ? 1 : 0);
  bank += (_status & 0x40 ? 2 : 0);
  return QString("W:%1   %2 %3 %4   PC:%5   Bank:%6")
         .arg(toHexLabel(_wreg, 2)).arg(_status & 0x4 ? "Z" : "z")
         .arg(_status & 0x2 ? "DC" : "dc").arg(_status & 0x1 ? "C" : "c")
         .arg(toHexLabel(_pc, 4)).arg(bank);
}

//----------------------------------------------------------------------------
void Icd2::DebuggerGroup::addDevice(const Device::Data *data)
{
  if ( FAMILY_DATA[family(data->name())].debugExec==0 ) return;
  Group::addDevice(data);
}
