/***************************************************************************
 *   Copyright (C) 2005-2006 Nicolas Hadacek <hadacek@kde.org>             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "icd2_prog.h"

#include <qdir.h>
#include <klocale.h>

#include "icd2_prog_config.h"
#include "progs/base/prog_config.h"
#include "devices/base/device_list.h"
#include "icd2_serial.h"
#include "icd2_usb.h"
#include "progs/base/gui/prog_config_widget.h"
#include "common/about.h"

//-----------------------------------------------------------------------------
Icd2::ProgrammerBase::ProgrammerBase(const ::Programmer::Group &group, const Pic::Data *data, const char *name)
  : ::Programmer::PicBase(group, data, name)
{
  clear();
}

Icd2::Hardware &Icd2::ProgrammerBase::hardware()
{
  return static_cast<Hardware &>(*_hardware);
}

bool Icd2::ProgrammerBase::internalRun()
{
  _state = ::Programmer::Running;
  return hardware().setTargetReset(Pic::ResetReleased);
}

bool Icd2::ProgrammerBase::internalStop()
{
  _state = ::Programmer::Stopped;
  return hardware().setTargetReset(Pic::ResetHeld);
}

bool Icd2::ProgrammerBase::initProgramming()
{
  _state = ::Programmer::Stopped;
  return hardware().setTargetReset(Pic::ResetHeld);
}

bool Icd2::ProgrammerBase::doUploadFirmware(QFile &file)
{
  const Device::Data &data = *Device::lister().data("16F876");
  Pic::Memory memory(static_cast<const Pic::Data &>(data));
  QString error;
  QStringList warnings;
  Pic::Memory::WarningTypes warningTypes;
  if ( !memory.load(file, error, warningTypes, warnings) ) {
    log(Log::Error, i18n("Could not read firmware file \"%1\" (%2).").arg(file.name()).arg(error));
    return false;
  }
  if ( warningTypes!=Pic::Memory::NoWarning ) {
    log(Log::Error, i18n("Firmware hex file seems incompatible with device 16F876 inside icd2."));
    return false;
  }
  if ( !hardware().uploadFirmware(memory) ) {
    log(Log::Error, i18n("Failed to upload firmware."));
    return false;
  }
  return true;
}

void Icd2::ProgrammerBase::clear()
{
  _firmwareId = 0;
  for (uint i=0; i<Nb_VoltageTypes; i++) _voltages[i] = ::Programmer::UNKNOWN_VOLTAGE;
  _testData = TestData();
}

bool Icd2::ProgrammerBase::readFirmware()
{
  _firmwareVersion = hardware().getFirmwareVersion();
  log(Log::NormalDebug, i18n("  Firmware version is %1").arg(_firmwareVersion.pretty()));
  _firmwareId = hardware().getFirmwareId();
  return !hasError();
}

bool Icd2::ProgrammerBase::readVoltages()
{
  for (uint i=0; i<Nb_VoltageTypes; i++) {
    _voltages[i] = hardware().getVoltage(VoltageType(i));
    log(Log::NormalDebug, QString("  %1 = %2 V").arg(i18n(VOLTAGE_LABELS[i])).arg(_voltages[i]));
  }
  return !hasError();
}

bool Icd2::ProgrammerBase::selfTest(bool ask)
{
  log(Log::NormalDebug, "  Self-test");
  _testData = hardware().selfTest();
  if ( !_testData.pass() ) {
    QString s;
    for (uint i=0; i<TestData::Nb_VoltageTypes; i++) {
      if ( i!=0 ) s += "; ";
      s += _testData.pretty(TestData::VoltageType(i));
    }
    log(Log::Warning, i18n("Self-test failed: %1").arg(s));
    if ( ask && !askContinue(i18n("Self-test failed (%1). Do you want to continue anyway?").arg(s)) ) {
      logUserAbort();
      return false;
    }
  }
  return !hasError();
}

bool Icd2::ProgrammerBase::checkFirmware()
{
  const FamilyData &fdata = FAMILY_DATA[family(device()->name())];
  log(Log::NormalDebug, QString("  Firmware id is %1 and we want %2").arg(_firmwareId).arg(fdata.efid));
  if ( fdata.efid!=_firmwareId ) return false;

  VersionData vd(_firmwareVersion.major(), _firmwareVersion.mid(), 0); // ignore minor
  const FirmwareData &fd = FIRMWARE_DATA[_firmwareId-1];
  VersionData tv(fd.major, fd.mid, 0); // ignore minor
  if ( tv.toUInt()<vd.toUInt() )
    log(Log::Warning, i18n("The firmware version (%1) is higher than the version tested with piklab (%2).\n"
                           "You may experience problems; in such case please use firmwares from MPLAB %3.")
                      .arg(_firmwareVersion.pretty()).arg(tv.pretty()).arg(AboutData::MPLAB_VERSION_TESTED));
  if ( vd.toUInt()<tv.toUInt() )
    log(Log::Warning, i18n("The firmware version (%1) is lower than the version tested with piklab (%2).\n"
                           "You may experience problems; in such case please use firmwares from MPLAB %3.")
                      .arg(_firmwareVersion.pretty()).arg(tv.pretty()).arg(AboutData::MPLAB_VERSION_TESTED));
  return true;
}

bool Icd2::ProgrammerBase::internalSetupHardware()
{
  log(Log::NormalDebug, "  Check firmware");
  if ( !readFirmware() ) return false;
  if ( _specific==0 ) return true;

  if ( !checkFirmware() ) {
    log(Log::Information, i18n("  Incorrect firmware loaded."));

    // find firmware file
    QDir dir(::Programmer::Config::readFirmwareDirectory(group()));
    if ( !dir.exists() ) {
      log(Log::Error, i18n("Firmware directory not configured."));
      return false;
    }
    const FamilyData &fdata = FAMILY_DATA[family(device()->name())];
    QString nameFilter = "ICD" + QString::number(fdata.efid).rightJustify(2, '0') + "??????.hex";
    QStringList files = dir.entryList(nameFilter, QDir::Files);
    if ( files.isEmpty() ) {
      log(Log::Error, i18n("Could not find firmware file \"%1\" in directory \"%2\".").arg(nameFilter).arg(dir.path()));
      return false;
    }

    // upload hex file
    QString filepath = dir.filePath(files[files.count()-1]);
    log(Log::NormalDebug, QString("  Firmware file: %1").arg(filepath));
    QFile file(filepath);
    if ( !file.open(IO_ReadOnly) ) {
      log(Log::Error, i18n("Could not open firmware file \"%1\".").arg(file.name()));
      return false;
    }
    if ( !doUploadFirmware(file) ) return false;

    // check firmware
    if ( !readFirmware() ) return false;
    if ( !checkFirmware() ) {
      log(Log::Error, i18n("Firmware still incorrect after uploading."));
      return false;
    }
    log(Log::Information, i18n("  Firmware succesfully uploaded."));
  }

  log(Log::NormalDebug, "  Set target");
  if ( !_specific->init() ) return false;
  bool selfPowered = ::Programmer::Config::isTargetSelfPowered();
  if ( !selfPowered && device()->architecture()==Pic::Arch_30X
       && !askContinue(i18n("It is not recommended to power dsPICs from ICD2. Continue anyway?")) ) {
    logUserAbort();
    return false;
  }
  log(Log::Information, i18n("  Set target self powered: %1").arg(selfPowered ? "true" : "false"));
  if ( !hardware().setTargetSelfPowered(selfPowered) ) return false;
  if ( !readVoltages() ) return false;
  if ( hardware().getTargetMode()!=TargetInReset ) {
    log(Log::Error, i18n("Device not in reset"));
    return false;
  }
  log(Log::NormalDebug, "  Hold reset");
  if ( !hardware().setTargetReset(Pic::ResetHeld) ) return false;
  if ( !hardware().setup() ) return false;
  if ( !selfTest(true) ) return false;
  return !hasError();
}

bool Icd2::ProgrammerBase::updateState()
{
  TargetMode mode = hardware().getTargetMode();
  if ( hasError() ) return false;
  switch (mode) {
    case TargetStopped: _state = ::Programmer::Halted; break;
    case TargetRunning: _state = ::Programmer::Running; break;
    case TargetInReset: _state = ::Programmer::Stopped; break;
    case Nb_TargetModes: Q_ASSERT(false); break;
  }
  return true;
}

//-----------------------------------------------------------------------------
Icd2::Programmer::Programmer(const ::Programmer::Group &group, const Pic::Data *data)
  : Icd2::ProgrammerBase(group, data, "icd2_programmer")
{}

//----------------------------------------------------------------------------
Programmer::Properties Icd2::Group::properties() const
{
   return ::Programmer::HaveFirmware | ::Programmer::CanUploadFirmware
          | ::Programmer::NeedFirmwareDirectory | ::Programmer::CanReleaseReset;
}

Programmer::Hardware *Icd2::Group::createHardware(::Programmer::Base &base, Port::Type portType, const QString &portDevice) const
{
  if ( portType==Port::SerialType ) return new SerialHardware(base, portDevice);
  return new USBHardware(base);
}

Programmer::DeviceSpecific *Icd2::Group::createDeviceSpecific(::Programmer::Base &base) const
{
  return new DeviceSpecific(base);
}
