/***************************************************************************
 * Copyright (C) 2005 Lorenz Msenlechner & Matthias Kranz                 *
 *                    <icd2linux@hcilab.org>                               *
 * Copyright (C) 2005 Nicolas Hadacek <hadacek@kde.org>                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "icd2_usb.h"

#include "config.h"
#ifdef HAVE_USB
#include <usb.h>
#endif

#include "microchip.h"
#include "common/misc.h"

//------------------------------------------------------------------------------
Icd2::USBPort::USBPort(uint deviceId, Log::Manager &manager)
  : Port::USB(Microchip::VENDOR_ID, deviceId, manager)
{}

uint Icd2::USBPort::interface(const struct usb_device &device) const
{
#ifdef HAVE_USB
  return device.config->interface->altsetting[0].bInterfaceNumber; // #### ????
#else
  return 0;
#endif
}

bool Icd2::USBPort::doSequence(SequenceType type, char *data, uint size)
{
  QByteArray tx(0x12);
  char c = 0;
  tx.fill(c);
  tx[0] = type;
  tx[1] = _seqnum;
  if ( type==Receive || type==Send ) {
    tx[2] = size & 0xFF;
    tx[3] = (size >> 8) & 0xFF;
    tx[0x0A] = 0xC9;
  }
  if ( !write(0x01, tx.data(), tx.size()) ) return false;

  switch (type) {
    case Connect: break;
    case Receive:
      if ( !read(0x82, data, size) ) return false;
      break;
    case Send:
      if ( !write(0x01, data, size) ) return false;
      break;
  }

  QByteArray rx(0x08);
  if ( !read(0x81, rx.data(), rx.size()) ) return false;
  Q_ASSERT( rx[0]==tx[1] );

  _seqnum++;
  if ( _seqnum==0xFF ) _seqnum = 0xC0;

  return true;
}

bool Icd2::USBPort::connectDevice(const QCString &mode)
{
  _seqnum = 0xC1; // preset seqnum

  doSequence(Connect, 0, 0);
  send("Z", 1);
  receive(4);

  doSequence(Connect, 0, 0);
  send(mode, mode.length());
  receive(1);

  return !hasError();
}

QCString Icd2::USBPort::receive(uint size)
{
  QCString s(size+1);
  doSequence(Receive, s.data(), size);
  return s;
}

bool Icd2::USBPort::send(const char *data, uint size)
{
  return doSequence(Send, (char *)data, size);
}

//------------------------------------------------------------------------------
Icd2::USBHardware::USBHardware(::Programmer::Base &base)
  : Hardware(base, new USBPort(ID_CLIENT, base))
{}

bool Icd2::USBHardware::internalConnect(const QCString &mode)
{
  // load control messages for USB device if needed
  //qDebug("need firmware ? %i", USBPort::findDevice(Microchip::VENDOR_ID, ID_FIRMWARE));
  if ( Port::USB::findDevice(Microchip::VENDOR_ID, ID_FIRMWARE) ) {
    USBPort port(ID_FIRMWARE, manager());
    if ( !port.open() ) return false;
    uint i = 0;
    while ( CONTROL_MESSAGE_DATA[i].bytes!=0 ) {
      if ( !port.sendControlMessage(CONTROL_MESSAGE_DATA[i]) ) return false;
      i++;
    }
    port.close();
    for (uint i=0; i<10; i++) {
      //qDebug("client here: firmware ? %i", USBPort::findDevice(Microchip::VENDOR_ID, ID_CLIENT));
      if ( Port::USB::findDevice(Microchip::VENDOR_ID, ID_CLIENT) ) break;
      Port::msleep(1000);
    }
  }

  if ( !_port->open() ) return false;
  return static_cast<USBPort *>(_port)->connectDevice(mode);
}
