/***************************************************************************
 * Copyright (C) 2006 Nicolas Hadacek <hadacek@kde.org>                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "pickit.h"

#include "devices/base/device_group.h"
#include "progs/base/prog_group.h"

//-----------------------------------------------------------------------------
Pickit::Array::Array(uint length)
{
  _data.fill('Z', length);
}

QString Pickit::Array::pretty() const
{
  QString s;
  bool z = true;
  for (int i=length()-1; i>=0; i--) {
    if ( char(_data[i])=='Z' && z ) continue;
    z = false;
    if ( isalnum(char(_data[i])) ) s = char(_data[i]) + s;
    else s = "\\" + QString(toHex(_data[i], 2)) + s;
  }
  return s;
}

//-----------------------------------------------------------------------------
Pickit::USBPort::USBPort(uint deviceId, Log::Manager &manager)
  : Port::USB(Microchip::VENDOR_ID, deviceId, manager)
{}

bool Pickit::USBPort::command(char c)
{
  Array a(arrayLength());
  a._data[0] = c;
  return command(a);
}

bool Pickit::USBPort::command(const char *s)
{
  Array a(arrayLength());
  if (s) {
    Q_ASSERT( strlen(s)<=arrayLength() );
    for (uint i=0; i<strlen(s); i++) a._data[i] = s[i];
  }
  return command(a);
}

bool Pickit::USBPort::command(const Array &cmd)
{
  log(Log::ExtraDebug, QString("send command: \"%1\"").arg(cmd.pretty()));
  return write(writeEndPoint(), (const char *)cmd._data.data(), cmd.length());
}

Pickit::Array Pickit::USBPort::receive()
{
  Array array(arrayLength());
  read(readEndPoint(), (char *)array._data.data(), array.length());
  log(Log::MaxDebug, QString("received: \"%1\"").arg(array.pretty()));
  return array;
}

//-----------------------------------------------------------------------------
Pickit::Hardware::Hardware(::Programmer::Base &base, USBPort *port)
  : ::Programmer::PicHardware(base, port)
{}

bool Pickit::Hardware::internalConnectHardware()
{
  return port().open();
}

Pickit::Mode Pickit::Hardware::getMode(VersionData &version)
{
  if ( !port().command('v') ) return Nb_Modes;
  Array a = port().receive();
  if ( hasError() ) return Nb_Modes;
  if ( (_base.group().properties() & ::Programmer::CanUploadFirmware) && a[5]=='B' ) {
    version = VersionData(0, a[6], a[7]);
    return BootloadMode;
  }
  version = VersionData(a[0], a[1], a[2]);
  return NormalMode;
}

QValueVector<uint> Pickit::Hardware::receiveWords(uint nbBytesWord)
{
  Q_ASSERT( nbBytesWord==1 || nbBytesWord==2 );
  Array a = port().receive();
  QValueVector<uint> words(a.length()/nbBytesWord);
  for (uint i=0; i<words.count(); i++) {
    if ( nbBytesWord==1 ) words[i] = a[i];
    else {
      words[i] = a[2*i];
      words[i] |= a[2*i+1] << 8;
    }
  }
  return words;
}

bool Pickit::Hardware::writeWords(uint max, char c, uint nbBytesWord,
                                  uint &i, const Device::Array &data)
{
  Q_ASSERT( i<data.count() );
  Q_ASSERT( nbBytesWord==1 || nbBytesWord==2 );
  Array cmd = port().array();
  uint n = (nbBytesWord==1 ? 2 : 3);
  Q_ASSERT( n*max<=cmd.length() );
  for (uint k=0; k<max; k++) {
    cmd[n*k] = c;
    cmd[n*k+1] = data[i] & 0xFF;
    if ( nbBytesWord==2 ) cmd[n*k+2] = (data[i] >> 8) & 0xFF;
    i++;
    if ( i>=data.count() ) break;
  }
  return port().command(cmd);
}

//----------------------------------------------------------------------------
void Pickit::DeviceSpecific::setPowerOn()
{
  hardware().port().command(entryMode());
}

void Pickit::DeviceSpecific::setPowerOff()
{
  hardware().port().command('p');
}

void Pickit::DeviceSpecific::setTargetVddOn(bool on)
{
  if (on) hardware().port().command("V1");
  else hardware().port().command("V0");
}

//----------------------------------------------------------------------------
Device::Array Pickit::BMDeviceSpecific::doRead(Pic::MemoryRangeType type)
{
  Device::Array data(device().nbWords(type));
  gotoMemory(type);
  switch (type) {
    case Pic::MemoryConfig:
    case Pic::MemoryCode:
    case Pic::MemoryCal:
    case Pic::MemoryCalBackup:
    case Pic::MemoryUserId:
    case Pic::MemoryDeviceId:
      for (uint i=0; i<data.count();) {
        hardware().port().command('R');
        QValueVector<uint> words = hardware().receiveWords(2);
        for (uint k=0; k<words.count(); k++) {
          data[i] = words[k];
          i++;
          if ( i>=data.count() ) break;
        }
      }
      break;
    case Pic::MemoryEeprom:
      for (uint i=0; i<data.count();) {
        hardware().port().command('r'); // #### not sure this is correct for Pickit1: "rrrrrrrr" used by usb_pickit
        QValueVector<uint> words = hardware().receiveWords(1);
        for (uint k=0; k<words.count(); k++) {
          data[i] = words[k];
          i++;
          if ( i>=data.count() ) break;
        }
      }
      break;
    case Pic::MemoryDebugVector:
    case Pic::MemoryHardwareStack:
    case Pic::MemoryProgramExecutive:
    case Pic::Nb_MemoryRangeTypes: Q_ASSERT(false); break;
  }
  if ( type==Pic::MemoryCode || type==Pic::MemoryEeprom ) addProgress(data.count());
  return data;
}

bool Pickit::BMDeviceSpecific::incrementPC(uint nb)
{
  Pickit::Array cmd = hardware().port().array();
  cmd[0] = 'I';
  cmd[1] = nb & 0xFF;
  cmd[2] = (nb >> 8) & 0xFF;
  return hardware().port().command(cmd);
}

::Programmer::Result Pickit::BMDeviceSpecific::doEraseCode()
{
  if ( !hardware().port().command('E') ) return ::Programmer::Error;
  return ::Programmer::Ok;
}

//----------------------------------------------------------------------------
bool Pickit::Baseline::gotoMemory(Pic::MemoryRangeType type)
{
  switch (type) {
    case Pic::MemoryConfig:
    case Pic::MemoryEeprom: return true;
    case Pic::MemoryCode:
    case Pic::MemoryCal:
    case Pic::MemoryUserId:
    case Pic::MemoryCalBackup: return incrementPC(1+device().range(type).start);
    case Pic::MemoryDeviceId:
    case Pic::MemoryDebugVector:
    case Pic::MemoryHardwareStack:
    case Pic::MemoryProgramExecutive:
    case Pic::Nb_MemoryRangeTypes: break;
  }
  Q_ASSERT(false);
  return false;
}

bool Pickit::Baseline::doWrite(Pic::MemoryRangeType type, const Device::Array &data, bool force)
{
  // #### TODO: speed optimize...
  Q_UNUSED(force);
  gotoMemory(type);
  uint nb = nbWrites(type);
  switch (type) {
    case Pic::MemoryConfig:
    case Pic::MemoryCode:
    case Pic::MemoryCal:
    case Pic::MemoryUserId:
      for (uint i=0; i<data.count(); )
        hardware().writeWords(nb, 'w', 2, i, data);
      break;
    case Pic::MemoryEeprom:
      for (uint i=0; i<data.count(); )
        hardware().writeWords(nb, 'D', 1, i, data);
      break;
    case Pic::MemoryCalBackup:
    case Pic::MemoryDeviceId:
    case Pic::MemoryDebugVector:
    case Pic::MemoryHardwareStack:
    case Pic::MemoryProgramExecutive:
    case Pic::Nb_MemoryRangeTypes: Q_ASSERT(false); return false;
  }
  if ( type==Pic::MemoryCode || type==Pic::MemoryEeprom ) addProgress(data.count());
  return true;
}

//----------------------------------------------------------------------------
bool Pickit::P16F::gotoMemory(Pic::MemoryRangeType type)
{
  Pickit::Array cmd = hardware().port().array();
  cmd[0] = 'C';
  switch (type) {
    case Pic::MemoryCode: return true;
    case Pic::MemoryEeprom: return true;
    case Pic::MemoryUserId: return hardware().port().command(cmd);
    case Pic::MemoryDeviceId:
      cmd[1] = 'I';
      cmd[2] = 0x06;
      cmd[3] = 0x00;
      return hardware().port().command(cmd);
    case Pic::MemoryConfig:
      cmd[1] = 'I';
      cmd[2] = 0x07;
      cmd[3] = 0x00;
      return hardware().port().command(cmd);
    case Pic::MemoryCal:
      if ( device().range(type).start==device().range(Pic::MemoryCode).end+1 )
        return incrementPC(device().range(type).start);
      cmd[1] = 'I';
      cmd[2] = 0x08;
      cmd[3] = 0x00;
      return hardware().port().command(cmd);
    case Pic::MemoryCalBackup:
    case Pic::MemoryDebugVector:
    case Pic::MemoryHardwareStack:
    case Pic::MemoryProgramExecutive:
    case Pic::Nb_MemoryRangeTypes: break;
  }
  Q_ASSERT(false);
  return false;
}


bool Pickit::P16F::doWrite(Pic::MemoryRangeType type, const Device::Array &data, bool force)
{
  // #### TODO: speed optimize...
  Q_UNUSED(force);
  gotoMemory(type);
  uint nb = nbWrites(type);
  switch (type) {
    case Pic::MemoryConfig:
    case Pic::MemoryUserId:
    case Pic::MemoryCal:
    case Pic::MemoryCode:
      for (uint i=0; i<data.count(); )
        hardware().writeWords(nb, 'W', 2, i, data); // #### TODO: speed-up for some chips only... (Pickit1 has '4')
      break;
    case Pic::MemoryEeprom:
      for (uint i=0; i<data.count(); )
        hardware().writeWords(nb, 'D', 1, i, data);
      break;
    case Pic::MemoryCalBackup:
    case Pic::MemoryDeviceId:
    case Pic::MemoryDebugVector:
    case Pic::MemoryHardwareStack:
    case Pic::MemoryProgramExecutive:
    case Pic::Nb_MemoryRangeTypes: Q_ASSERT(false); return false;
  }
  if ( type==Pic::MemoryCode || type==Pic::MemoryEeprom ) addProgress(data.count());
  return true;
}

bool Pickit::P16F::doErase(bool)
{
  Pickit::Array cmd = hardware().port().array();
  cmd[0] = 'C';
  cmd[1] = 'W';
  cmd[2] = 0xFF;
  cmd[3] = 0x3F;
  cmd[4] = 'E';
  cmd[5] = 'e';
  return hardware().port().command(cmd);
}

::Programmer::Result Pickit::P16F::doEraseEeprom()
{
  if ( !hardware().port().command('e') ) return ::Programmer::Error;
  return ::Programmer::Ok;
}
