/***************************************************************************
 * Copyright (C) 2006 Nicolas Hadacek <hadacek@kde.org>                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "pickit2.h"

#include <klocale.h>

//-----------------------------------------------------------------------------
double Pickit2::Hardware::getVdd()
{
  port().command("V\04");
  Pickit::Array a = port().receive();
  double vdd = (a[0] & 0xC0) + a[1] * 256;
  vdd /= 65472.;   // divide by 0xFFC0;
  vdd *= 5.0;      // multiply by voltage reference
  return vdd;
}

bool Pickit2::Hardware::setResetMode(Pic::ResetMode mode)
{
  // #### TODO
  return false;
}

Pickit::Array Pickit2::Hardware::command(uchar command, uint nbBytes, uint address,
                                         uint i, bool longAddress) const
{
  Pickit::Array cmd = port().array();
  cmd[i] = command;
  cmd[i+1] = nbBytes;
  cmd[i+2] = address & 0xFF;
  cmd[i+3] = (address >> 8) & 0xFF;
  if (longAddress) cmd[i+4] = (address >> 16) & 0xFF;
  return cmd;
}

bool Pickit2::Hardware::uploadFirmware(const Pic::Memory &memory)
{
  log(Log::Information, i18n("  Uploading PICkit2 firmware..."));

  log(Log::NormalDebug, "  Enter bootload mode");
  port().command('B');
  for (uint i=0; i<=100; i++) {
    usleep(10000);
    VersionData vd;
    if ( getMode(vd)==Pickit::BootloadMode ) break;
    if ( i==100 ) {
      log(Log::Error, i18n("Timeout getting in bootload mode"));
      return false;
    }
  }

  log(Log::NormalDebug, "  Erase pickit2 device");
  Pickit::Array cmd = command(3, 0xC0, 0x2000); // 0xC0 ??
  port().command(cmd); // erase 0x2000-0x4FFF
  cmd = command(3, 0xC0, 0x5000); // 0xC0 ??
  port().command(cmd); // erase 0x5000-0x7FFF

  log(Log::NormalDebug, "  Write firmware");
  Device::Array data = memory.arrayForWriting(Pic::MemoryCode);
  for (uint i=0; i<data.count(); i+=16) { // 2 bytes/word and 32 bytes/command max
    uint start = 0;
    for (; start<16; start++) if ( data[i + start]!=Device::INVALID ) break;
    if ( start==16 ) continue;
    uint end = 15;
    for (; end>start; end--) if ( data[i + end]!=Device::INVALID ) break;
    uint nb = end - start + 1;
    cmd = command(2, 2*nb, 2*i);
    for (uint k=0; k<nb; k++) {
      cmd[5 + 2*k] = data[i + start + k] & 0xFF;
      cmd[6 + 2*k] = (data[i + start + k] >> 8) & 0xFF;
    }
    port().command(cmd);
  }

  log(Log::NormalDebug, "  Verify firmware");
  for (uint i=0; i<data.count(); i+=16) { // 2 bytes/word and 32 bytes/command max
    uint start = 0;
    for (; start<16; start++) if ( data[i + start]!=Device::INVALID ) break;
    if ( start==16 ) continue;
    uint end = 15;
    for (; end>start; end--) if ( data[i + end]!=Device::INVALID ) break;
    uint nb = end - start + 1;
    cmd = command(1, 2*nb, 2*i);
    port().command(cmd);
    QValueVector<uint> read = receiveWords(2);
    for (uint k=0; k<nb; k++) {
      uint index = i + start + k;
      uint d = read[k];
      uint v = data[index];
      if ( d!=v ) {
        log(Log::Error, i18n("Firmware memory does not match hex file (at address 0x%2: reading 0x%3 and expecting 0x%4).")
                        .arg(toHex(2*index, device().nbCharsAddress()))
                        .arg(toHex(d, device().nbCharsWord(Pic::MemoryCode))).arg(toHex(v, device().nbCharsWord(Pic::MemoryCode))));
        return false;
      }
    }
  }

  log(Log::NormalDebug, "  Firmware verified");
  cmd = command(2, 2, 0x7FFE);
  cmd[5] = 0x55;
  cmd[6] = 0x55;
  port().command(cmd);

  log(Log::NormalDebug, "  reset to normal mode");
  port().command(0xFF);
  for (uint i=0; i<=100; i++) {
    usleep(10000);
    VersionData vd;
    if ( getMode(vd)==Pickit::NormalMode ) break;
    if ( i==100 ) {
      log(Log::Error, i18n("Timeout getting in normal mode"));
      return false;
    }
  }

  return true;
}

//-----------------------------------------------------------------------------
bool Pickit2::Baseline::init()
{
  Pickit::Array cmd = hardware().port().array();
  cmd[0] = 'c';
  cmd[1] = 0x00;   // Load Configuration
  cmd[2] = 0x02;   // Load Program latches
  cmd[3] = 0x03;   // Load Data latches
  cmd[4] = 0x04;   // Read Program memory
  cmd[5] = 0x05;   // Read Data latches
  cmd[6] = 0x06;   // Increment Address
  cmd[7] = 0x08;   // Begin programming internally timed
  cmd[8] = 0x08;   // Begin Programming externally timed
  cmd[9] = 0x0E;   // End externally time programming cycle
  cmd[10] = 0x09;  // Bulk Erase program memory
  cmd[11] = 0x0B;  // Bulk Erase Data memory
  cmd[12] = 0xFF;  // Read Program memory
  cmd[13] = 0xFF;  // Read Data latches
  cmd[14] = 0xFF;  // Increment Address
  cmd[15] = 0xFF;  // Begin programming internally timed
  return hardware().port().command(cmd);
}

//-----------------------------------------------------------------------------
bool Pickit2::P16F::init()
{
  Pickit::Array cmd = hardware().port().array();
  cmd[0] = 'c';
  cmd[1] = 0x00;   // Load Configuration
  cmd[2] = 0x02;   // Load Program latches
  cmd[3] = 0x03;   // Load Data latches
  cmd[4] = 0x04;   // Read Program memory
  cmd[5] = 0x05;   // Read Data latches
  cmd[6] = 0x06;   // Increment Address
  cmd[7] = 0x08;   // Begin programming internally timed
  cmd[8] = 0x18;   // Begin Programming externally timed
  cmd[9] = 0x0A;   // End externally time programming cycle
  cmd[10] = 0x09;  // Bulk Erase program memory
  cmd[11] = 0x0B;  // Bulk Erase Data memory
  cmd[12] = 0xFF;  // Read Program memory
  cmd[13] = 0xFF;  // Read Data latches
  cmd[14] = 0xFF;  // Increment Address
  cmd[15] = 0xFF;  // Begin programming internally timed
  return hardware().port().command(cmd);
}

//-----------------------------------------------------------------------------
bool Pickit2::P16F87XA::init()
{
  Pickit::Array cmd = hardware().port().array();
  cmd[0] = 'c';
  cmd[1] = 0x00;   // Load Configuration
  cmd[2] = 0x02;   // Load Program latches
  cmd[3] = 0x03;   // Load Data latches
  cmd[4] = 0x04;   // Read Program memory
  cmd[5] = 0x05;   // Read Data latches
  cmd[6] = 0x06;   // Increment Address
  cmd[7] = 0x08;   // Begin programming internally timed
  cmd[8] = 0x18;   // Begin Programming externally timed
  cmd[9] = 0x17;   // End externally time programming cycle
  cmd[10] = 0x1F;  // Bulk Erase program memory
  cmd[11] = 0x1F;  // Bulk Erase Data memory
  cmd[12] = 0xFF;  // Read Program memory
  cmd[13] = 0xFF;  // Read Data latches
  cmd[14] = 0xFF;  // Increment Address
  cmd[15] = 0xFF;  // Begin programming internally timed
  return hardware().port().command(cmd);
}

//-----------------------------------------------------------------------------
bool Pickit2::P16F7X::init()
{
  Pickit::Array cmd = hardware().port().array();
  cmd[0] = 'c';
  cmd[1] = 0x00;   // Load Configuration
  cmd[2] = 0x02;   // Load Program latches
  cmd[3] = 0x03;   // Load Data latches
  cmd[4] = 0x04;   // Read Program memory
  cmd[5] = 0x05;   // Read Data latches
  cmd[6] = 0x06;   // Increment Address
  cmd[7] = 0x08;   // Begin programming internally timed
  cmd[8] = 0x08;   // Begin Programming externally timed
  cmd[9] = 0x0E;   // End externally time programming cycle
  cmd[10] = 0x09;  // Bulk Erase program memory
  cmd[11] = 0x0B;  // Bulk Erase Data memory
  cmd[12] = 0xFF;  // Read Program memory
  cmd[13] = 0xFF;  // Read Data latches
  cmd[14] = 0xFF;  // Increment Address
  cmd[15] = 0xFF;  // Begin programming internally timed
  return hardware().port().command(cmd);
}

bool Pickit2::P16F716::init()
{
  Pickit::Array cmd = hardware().port().array();
  cmd[0] = 'c';
  cmd[1] = 0x00;   // Load Configuration
  cmd[2] = 0x02;   // Load Program latches
  cmd[3] = 0x03;   // Load Data latches
  cmd[4] = 0x04;   // Read Program memory
  cmd[5] = 0x05;   // Read Data latches
  cmd[6] = 0x06;   // Increment Address
  cmd[7] = 0x08;   // Begin programming internally timed
  cmd[8] = 0x18;   // Begin Programming externally timed
  cmd[9] = 0x0E;   // End externally time programming cycle
  cmd[10] = 0x09;  // Bulk Erase program memory
  cmd[11] = 0x0B;  // Bulk Erase Data memory
  cmd[12] = 0xFF;  // Read Program memory
  cmd[13] = 0xFF;  // Read Data latches
  cmd[14] = 0xFF;  // Increment Address
  cmd[15] = 0xFF;  // Begin programming internally timed
  return hardware().port().command(cmd);
}

//-----------------------------------------------------------------------------
bool Pickit2::P18F::init()
{
  Pickit::Array cmd = hardware().port().array();
  cmd[0] = 'c';
  cmd[1] = 0x00;   // Load Configuration
  cmd[2] = 0x02;   // Load Program latches
  cmd[3] = 0x03;   // Load Data latches
  cmd[4] = 0x04;   // Read Program memory
  cmd[5] = 0x05;   // Read Data latches
  cmd[6] = 0x06;   // Increment Address
  cmd[7] = 0x08;   // Begin programming internally timed
  cmd[8] = 0x18;   // Begin Programming externally timed
  cmd[9] = 0x0A;   // End externally time programming cycle
  cmd[10] = 0x09;  // Bulk Erase program memory
  cmd[11] = 0x0B;  // Bulk Erase Data memory
  cmd[12] = 0xFF;  // Read Program memory
  cmd[13] = 0xFF;  // Read Data latches
  cmd[14] = 0xFF;  // Increment Address
  cmd[15] = 0xFF;  // Begin programming internally timed
  return hardware().port().command(cmd);
}

bool Pickit2::P18F::doEraseCommand(uint cmd1, uint cmd2)
{
  Pickit::Array cmd = hardware().port().array();
  cmd[0] = 0x85;
  cmd[1] = cmd1;
  cmd[2] = cmd2;
  return hardware().port().command(cmd);
}

::Programmer::Result Pickit2::P18F::doEraseCode()
{
  doEraseCommand(0x81, 0x00); // boot
  for (uint i=0; i<device().config().nbBlocks(); i++)
    doEraseCommand(1 << i, 0x80);
  return ::Programmer::Ok;
}

Device::Array Pickit2::P18F::doRead(Pic::MemoryRangeType type)
{
  Device::Array data(device().nbWords(type));
  Pickit::Array cmd = hardware().port().array();
  switch (type) {
    case Pic::MemoryDeviceId:
    case Pic::MemoryUserId:
    case Pic::MemoryConfig: {
      cmd = hardware().command(0x80, data.count(), device().range(type).start);
      hardware().port().command(cmd);
      QValueVector<uint> words = hardware().receiveWords(1);
      for (uint k=0; k<data.count(); k++) data[k] = words[k];
      break;
    }
    case Pic::MemoryCode:
      for (uint i=0; i<data.count();) {
        for (uint k=0; k<8; k++)
          cmd = hardware().command(0x80, 64, 2*(i + 32 * k), 5*k);
        hardware().port().command(cmd);
        for (uint k=0; k<8; k++) {
          QValueVector<uint> words = hardware().receiveWords(2);
          for (uint k=0; k<32; k++) {
            data[i] = words[k];
            i++;
          }
        }
      }
      break;
    case Pic::MemoryEeprom:
      for (uint i=0; i<data.count();) {
        cmd = hardware().command(0x81, 64, i, 0, false);
        hardware().port().command(cmd);
        QValueVector<uint> words = hardware().receiveWords(1);
        for (uint k=0; k<64; k++) {
          data[i] = words[k];
          i++;
        }
      }
      break;
    case Pic::MemoryCal:
    case Pic::MemoryCalBackup:
    case Pic::MemoryDebugVector:
    case Pic::MemoryHardwareStack:
    case Pic::MemoryProgramExecutive:
    case Pic::Nb_MemoryRangeTypes: Q_ASSERT(false); return false;
  }
  if ( type==Pic::MemoryCode || type==Pic::MemoryEeprom ) addProgress(data.count());
  return data;
}

bool Pickit2::P18F::doWrite(Pic::MemoryRangeType type, const Device::Array &data, bool force)
{
  Q_UNUSED(force);
  Pickit::Array cmd = hardware().port().array();
  switch (type) {
    case Pic::MemoryUserId: {
      cmd = hardware().command(0x82, data.count() / 2, device().range(type).start);
      for (uint i=0; i<data.count(); i++) cmd[i + 5] = data[i] & 0xFF;
      hardware().port().command(cmd);
      break;
    }
    case Pic::MemoryConfig:
      for (uint i=0; i<data.count()/2; i++) {
        cmd[4*i]     = 0x84;
        cmd[1 + 4*i] = 2*i;
        cmd[2 + 4*i] = data[2*i] & 0xFF;
        cmd[3 + 4*i] = data[2*i+1] & 0xFF;
      }
      hardware().port().command(cmd);
      break;
    case Pic::MemoryCode: {
      uint nb = Pickit2::data(device().name()).progWidth;
      for (uint i=0; i<data.count();) {
        cmd = hardware().command(0x82, nb, 2*i);
        for (uint k=0; k<nb; k++) {
          cmd[5 + 2*k] = data[i] & 0xFF;
          cmd[6 + 2*k] = (data[i] >> 8) & 0xFF;
          i++;
        }
        hardware().port().command(cmd);
      }
      break;
    }
    case Pic::MemoryEeprom:
      for (uint i=0; i<data.count();) {
        cmd = hardware().command(0x87, 32, i, 0, false);
        for (uint k=0; k<32; k++) {
          cmd[4+k] = data[i];
          i++;
        }
        hardware().port().command(cmd);
      }
      break;
    case Pic::MemoryDeviceId:
    case Pic::MemoryCal:
    case Pic::MemoryCalBackup:
    case Pic::MemoryDebugVector:
    case Pic::MemoryHardwareStack:
    case Pic::MemoryProgramExecutive:
    case Pic::Nb_MemoryRangeTypes: Q_ASSERT(false); return false;
  }
  return true;
}

::Programmer::Result Pickit2::P18F::doEraseEeprom()
{
  doEraseCommand(0x84, 0x00);
  return ::Programmer::Ok;
}

bool Pickit2::P18F2X20::doEraseCommand(uint cmd1)
{
  Pickit::Array cmd = hardware().port().array();
  cmd[0] = 0x86;
  cmd[1] = cmd1;
  return hardware().port().command(cmd);
}

::Programmer::Result Pickit2::P18F2X20::doEraseCode()
{
  doEraseCommand(0x83); // boot
  for (uint i=0; i<device().config().nbBlocks(); i++)
    doEraseCommand(0x88 + i);
  return ::Programmer::Ok;
}

::Programmer::Result Pickit2::P18F2X20::doEraseEeprom()
{
  doEraseCommand(0x81);
  return ::Programmer::Ok;
}
