/***************************************************************************
 *   Copyright (C) 2005-2006 Nicolas Hadacek <hadacek@kde.org>             *
 *   Copyright (C) 2003 Alain Gibaud <alain.gibaud@free.fr>                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "project.h"

#include <qstringlist.h>
#include <ksimpleconfig.h>
#include <klocale.h>

#include "devices/base/device_list.h"
#include "progs/base/prog_list.h"
#include "toplevel.h"
#include "tools/gputils_config.h"

//----------------------------------------------------------------------------
XmlDataFile::XmlDataFile(const PURL &url, const QString &name)
  : _url(url), _name(name), _document(name)
{
  QDomElement root = _document.createElement(name);
  _document.appendChild(root);
}

bool XmlDataFile::load(QString &error)
{
  PFile file(_url, Log::DontShow);
  if ( !file.openForRead() ) {
    error = file.errorString();
    return false;
  }
  if ( !_document.setContent(file.qfile(), false, &error) ) return false;
  if ( _document.doctype().name()!=_name
       || _document.documentElement().nodeName()!=_name ) {
    error = i18n("File is not of correct type.");
    return false;
  }
  return true;
}

bool XmlDataFile::save(QString &error) const
{
  PFile file(_url, Log::DontShow);
  if ( !file.openForWrite() ) {
    error = file.errorString();
    return false;
  }
  QString s = _document.toString(2);
  file.appendText(s);
  if ( !file.close() ) {
    error = file.errorString();
    return false;
  }
  return true;
}

QDomElement XmlDataFile::findChildElement(QDomElement parent, const QString &name) const
{
  QDomNodeList list = parent.elementsByTagName(name);
  return list.item(0).toElement();
}

QDomElement XmlDataFile::createChildElement(QDomElement parent, const QString &name)
{
  QDomNodeList list = parent.elementsByTagName(name);
  if ( list.count()==0 ) {
    QDomElement element = _document.createElement(name);
    parent.appendChild(element);
    return element;
  }
  return list.item(0).toElement();
}

void XmlDataFile::removeChilds(QDomNode parent) const
{
  QDomNodeList list = parent.childNodes();
  for (uint i=0; i<list.count(); i++)
    parent.removeChild(list.item(i));
}

QString XmlDataFile::value(const QString &group, const QString &key, const QString &defValue) const
{
  QDomElement root = _document.documentElement();
  QDomElement groupElement = findChildElement(root, group);
  if ( groupElement.isNull() ) return defValue;
  QDomElement element = findChildElement(groupElement, key);
  if ( element.isNull() ) return defValue;
  QDomText text = element.firstChild().toText();
  if ( text.isNull() ) return defValue;
  return text.data();
}

void XmlDataFile::setValue(const QString &group, const QString &key, const QString &value)
{
  QDomElement root = _document.documentElement();
  QDomElement groupElement = createChildElement(root, group);
  QDomElement element = createChildElement(groupElement, key);
  removeChilds(element);
  QDomText text = _document.createTextNode(value);
  element.appendChild(text);
}

QStringList XmlDataFile::listValues(const QString &group, const QString &key) const
{
  QStringList list;
  QDomElement root = _document.documentElement();
  QDomElement groupElement = findChildElement(root, group);
  if ( groupElement.isNull() ) return list;
  QDomElement element = findChildElement(groupElement, key);
  if ( element.isNull() ) return list;
  QDomNodeList childs = element.childNodes();
  for (uint i=0; i<childs.count(); i++) {
    QDomText text = childs.item(i).toElement().firstChild().toText();
    if ( text.isNull() ) continue;
    list.append(text.data());
  }
  return list;
}

void XmlDataFile::appendListValue(const QString &group, const QString &key, const QString &value)
{
  QDomElement root = _document.documentElement();
  QDomElement groupElement = createChildElement(root, group);
  QDomElement element = createChildElement(groupElement, key);
  QDomElement item = _document.createElement("item");
  element.appendChild(item);
  QDomText text = _document.createTextNode(value);
  item.appendChild(text);
}

void XmlDataFile::removeListValue(const QString &group, const QString &key, const QString &value)
{
  QDomElement root = _document.documentElement();
  QDomElement groupElement = createChildElement(root, group);
  QDomElement element = createChildElement(groupElement, key);
  QDomNodeList list = element.childNodes();
  for (uint i=0; i<list.count(); i++) {
    QDomElement item = list.item(i).toElement();
    QDomText text = item.firstChild().toText();
    if ( text.isNull() || text.data()!=value ) continue;
    element.removeChild(item);
    break;
  }
}

void XmlDataFile::clearList(const QString &group, const QString &key)
{
  QDomElement root = _document.documentElement();
  QDomElement groupElement = createChildElement(root, group);
  QDomElement element = createChildElement(groupElement, key);
  groupElement.removeChild(element);
}

//----------------------------------------------------------------------------
bool Project::load(QString &error)
{
  if ( _url.fileType()==PURL::Project ) return XmlDataFile::load(error);

  if ( !_url.probablyExists() ) {
    error = i18n("Project file %1 does not exist.").arg(_url.pretty());
    return false;
  }
  PURL tmp = _url;
  _url = _url.toFileType(PURL::Project);
  if ( _url.probablyExists() && XmlDataFile::load(error) ) return true;
  KConfig *config = new KSimpleConfig(tmp.filepath(), false);

  config->setGroup("Files");
  QStringList list = config->readListEntry("inputFiles");
  QStringList::const_iterator it = list.begin();
  for (; it!=list.end(); ++it) addFile(PURL(directory(), *it));

  config->setGroup("General");
  setVersion(config->readEntry("version", "0.1"));
  setDescription(config->readEntry("description", QString::null));

  config->setGroup("Assembler");
  QString device = config->readEntry("target-device");
  if ( device=="*" ) device = MainWindow::programmerGroup().supportedDevices()[0]; // compatibility
  GPUtils::Config *gconfig = new GPUtils::Config(this);
  gconfig->setDevice(Device::lister().checkName(device));
  gconfig->setGPAsmWarningLevel(QMIN(config->readUnsignedNumEntry("warn-level", 0), uint(GPUtils::Config::Nb_WarningLevels)));
  gconfig->setRawIncludeDirs(Tool::Assembler, config->readEntry("include-dir", QString::null));
  gconfig->setRawCustomOptions(Tool::Assembler, config->readEntry("other-options", QString::null));

  config->setGroup("Linker") ;
  QString hexFormatString = config->readEntry("hex-format", HexBuffer::FORMATS[HexBuffer::IHX32]);
  for (uint i=0; i<HexBuffer::Nb_Formats; i++)
    if ( hexFormatString==HexBuffer::FORMATS[i] ) gconfig->setHexFormat(HexBuffer::Format(i));
  gconfig->setRawIncludeDirs(Tool::Linker, config->readEntry("objs-libs-dir", QString::null));
  gconfig->setRawCustomOptions(Tool::Linker, config->readEntry("other-options", QString::null));

  delete gconfig;
  delete config;
  return true;
}

PURL::List Project::openedFiles() const
{
  PURL::List files;
  QStringList list = listValues("general", "opened_files");
  QStringList::const_iterator it = list.begin();
  for (; it!=list.end(); ++it) {
    if ( PURL(*it).isRelative() ) files += PURL(directory(), *it);
    else files += PURL(*it);
  }
  return files;
}

void Project::setOpenedFiles(const PURL::List &list)
{
  clearList("general", "opened_files");
  PURL::List::const_iterator it = list.begin();
  for (; it!=list.end(); ++it) {
    QString s = ((*it).isInto(directory()) ? (*it).relativeTo(directory()) : (*it).filepath());
    appendListValue("general", "opened_files", s);
  }
}

PURL::List Project::absoluteFiles() const
{
  PURL::List abs;
  QStringList files = listValues("general", "files");
  QStringList::const_iterator it = files.begin();
  for (; it!=files.end(); ++it) abs += PURL(directory(), *it);
  return abs;
}

void Project::addFile(const PURL &url)
{
  if ( !url.isInto(directory()) ) {
    Q_ASSERT(false);
    return;
  }
  appendListValue("general", "files", url.relativeTo(directory()));
}

void Project::removeFile(const PURL &url)
{
  if ( !url.isInto(directory()) ) {
    Q_ASSERT(false);
    return;
  }
  removeListValue("general", "files", url.relativeTo(directory()));
}

QStringList Project::objectsForLinker(const QString &extension) const
{
  QStringList objs;
  // objects files corresponding to src files
  PURL::List files = absoluteFiles();
  PURL::List::const_iterator it;
  for (it=files.begin(); it!=files.end(); ++it)
    if ( (*it).data().properties & PURL::Source ) {
      if ( extension.isEmpty() ) objs += (*it).toFileType(PURL::Object).relativeTo(directory());
      else objs += (*it).toExtension(extension).relativeTo(directory());
    }
  // objects
  for (it=files.begin(); it!=files.end(); ++it)
    if ( (*it).fileType()==PURL::Object )
      objs += (*it).filepath();
  return objs;
}

QStringList Project::librariesForLinker(const QString &prefix) const
{
  QStringList libs;
  PURL::List files = absoluteFiles();
  PURL::List::const_iterator it;
  for (it=files.begin(); it!=files.end(); ++it)
    if ( (*it).fileType()==PURL::Library )
      libs += prefix + (*it).filepath();
  return libs;
}

QString Project::version() const
{
  return Compile::Config::globalValue(this, "version", "0.1");
}
void Project::setVersion(const QString &version)
{
  Compile::Config::setGlobalValue(this, "version", version);
}
QString Project::description() const
{
  return Compile::Config::globalValue(this, "description", QString::null);
}
void Project::setDescription(const QString &description)
{
  Compile::Config::setGlobalValue(this, "description", description);
}
