/***************************************************************************
 *   Copyright (C) 2005-2006 Nicolas Hadacek <hadacek@kde.org>             *
 *   Copyright (C) 2003 Alain Gibaud <alain.gibaud@free.fr>                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "project_manager.h"

#include <qdragobject.h>
#include <qpainter.h>
#include <qstyle.h>

#include <klocale.h>
#include <kiconloader.h>
#include <kaction.h>
#include <kpopupmenu.h>
#include <kapplication.h>

#include "project.h"
#include "new_dialogs.h"
#include "project_editor.h"
#include "toplevel.h"
#include "editor_manager.h"
#include "object_view.h"
#include "devices/base/device_list.h"
#include "tools/tool_config.h"
#include "register_view.h"
#include "hex_editor.h"

//----------------------------------------------------------------------------
const char *ProjectManager::HeaderItem::GROUP_LABELS[Nb_Groups] = {
  I18N_NOOP("Device"), I18N_NOOP("Headers"), I18N_NOOP("Linker Script"),
  I18N_NOOP("Sources"), I18N_NOOP("Objects"), I18N_NOOP("Views"),
  I18N_NOOP("Generated")
};

ProjectManager::Group ProjectManager::group(PURL::FileType type)
{
  switch (type) {
    case PURL::CSource:
    case PURL::AsmGPAsm:
    case PURL::AsmPIC30:
    case PURL::AsmPICC: return SourceGroup;
    case PURL::CHeader:
    case PURL::Inc: return HeaderGroup;
    case PURL::Lkr:
    case PURL::Gld: return LinkerScriptGroup;
    case PURL::Library:
    case PURL::Object: return ObjectGroup;
    case PURL::Hex:
    case PURL::Elf:
    case PURL::Map:
    case PURL::Cod:
    case PURL::Coff:
    case PURL::Lst:
    case PURL::Rst:
    case PURL::Sym:
    case PURL::Mem:
    case PURL::Adb:
    case PURL::Lnk:
    case PURL::SDCC_P:
    case PURL::SDCC_D:
    case PURL::Sdb:
    case PURL::Hxl:
    case PURL::Rlf: return ViewGroup;
    case PURL::Project:
    case PURL::PikdevProject:
    case PURL::Unknown:
    case PURL::Nb_FileTypes: break;
  }
  return Nb_Groups;
}

ProjectManager::RootItem::RootItem(KListView *listview)
  : KListViewItem(listview)
{
  setSelectable(false);
  setPixmap(0, PURL::icon(PURL::Project));
  set(PURL(), true);
}

void ProjectManager::RootItem::set(const PURL &url, bool standalone)
{
  _url = url;
  _standalone = standalone;
  if ( _url.isEmpty() ) setText(0, i18n("<no project>"));
  else if (_standalone) setText(0, i18n("Standalone File"));
  else setText(0, _url.basename());
}

ProjectManager::HeaderItem::HeaderItem(RootItem *item, Group group)
  : KListViewItem(item), _group(group)
{
  if ( group!=DeviceGroup) setSelectable(false);
  setText(0, i18n(GROUP_LABELS[group]));
}

ProjectManager::FileItem::FileItem(HeaderItem *item, PURL::FileType ftype, const PURL &url, bool generated)
  : KListViewItem(item), _ftype(ftype), _generated(generated)
{
  setPixmap(0, PURL::icon(ftype));
  set(url);
}

void ProjectManager::FileItem::set(const PURL &url)
{
  _url = url;
  switch (_ftype) {
    case PURL::Hex:  setText(0, i18n("Hex File")); break;
    case PURL::Coff: setText(0, i18n("Disassembly Listing")); break;
    case PURL::Lst:  setText(0, i18n("List")); break;
    case PURL::Map:  setText(0, i18n("Memory Map")); break;
    case PURL::Project:
    case PURL::Nb_FileTypes: Q_ASSERT(false); break;
    default: {
      QString s = url.filename();
      if ( _generated && group(_ftype)==LinkerScriptGroup )
        s += i18n(" (default)");
      setText(0, s); break;
    }
  }
}

void ProjectManager::FileItem::paintCell(QPainter *p, const QColorGroup &cg, int column, int width, int align)
{
  QFont f = p->font();
  f.setItalic(group(_ftype)!=ViewGroup && _generated);
  p->setFont(f);
  KListViewItem::paintCell(p, cg, column, width, align);
}

ProjectManager::RegisterItem::RegisterItem(HeaderItem *item)
  : KListViewItem(item)
{
  setText(0, i18n(Register::LISTVIEW_LABEL));
}

ProjectManager::DeviceItem::DeviceItem(HeaderItem *item)
  : EditListViewItem(item)
{}

QWidget *ProjectManager::DeviceItem::editWidgetFactory(int) const
{
  QComboBox *combo = new QComboBox(false, 0);
  QString device = MainWindow::device();
  combo->clear();
  if ( device.isEmpty() ) {
    const Device::Data *data = MainWindow::deviceData();
    if (data) combo->insertItem(data->name() + " " + i18n(Device::AUTO_DATA.label));
    else combo->insertItem(i18n(Device::AUTO_DATA.label));
  }
  combo->insertStringList(Device::lister().supportedDevices());
  if ( !device.isEmpty() ) combo->setCurrentText(device);
  return combo;
}

void ProjectManager::DeviceItem::updateText()
{
  QString device = MainWindow::device();
  if ( device.isEmpty() ) {
    const Device::Data *data = MainWindow::deviceData();
    if (data) device = data->name() + " " + i18n(Device::AUTO_DATA.label);
    else device = i18n(Device::AUTO_DATA.label);
  }
  setText(0, device);
}

//----------------------------------------------------------------------------
ProjectManager::View::View(QWidget *parent)
  : ListView(parent, "project_manager"), _project(0), _modified(false)
{
  connect(this, SIGNAL(mouseButtonClicked(int, QListViewItem *, const QPoint &, int)),
          SLOT(clicked(int, QListViewItem *, const QPoint &, int)));
  connect(this, SIGNAL(contextMenuRequested(QListViewItem *, const QPoint &, int)),
          SLOT(contextMenu(QListViewItem *, const QPoint &, int)));
  connect(this, SIGNAL(itemRenamed(QListViewItem *, int, const QString &)),
          SLOT(renamed(QListViewItem *, int, const QString &)));

  header()->hide();
  setSorting(-1);
  addColumn(QString::null, 170);
  setFullWidth(true);
  setRootIsDecorated(false);
  setAcceptDrops(true);
  init();
}

ProjectManager::View::~View()
{
  delete _project;
}

void ProjectManager::View::init()
{
  clear();
  _rootItem  = new RootItem(this);
  setCurrentItem(_rootItem);
  KListViewItem *item = new RegisterItem(headerItem(DeviceGroup));
  ensureItemVisible(item);
  _deviceItem = new DeviceItem(headerItem(DeviceGroup));
  ensureItemVisible(_deviceItem);
}

ProjectManager::HeaderItem *ProjectManager::View::findHeaderItem(Group group) const
{
  QListViewItemIterator it(const_cast<View *>(this));
  for(; it.current(); ++it) {
    if ( it.current()->rtti()!=HeaderRtti ) continue;
    HeaderItem *hi = static_cast<HeaderItem *>(it.current());
    if ( hi->group()==group ) return hi;
  }
  return 0;
}

ProjectManager::HeaderItem *ProjectManager::View::headerItem(Group group)
{
  HeaderItem *item = findHeaderItem(group);
  if (item) return item;
  item = new HeaderItem(_rootItem, group);
  // reorder groups...
  HeaderItem *items[Nb_Groups];
  for (uint i=0; i<Nb_Groups; i++) {
    items[i] = findHeaderItem(Group(i));
    if (items[i]) _rootItem->takeItem(items[i]);
  }
  for (int i=Nb_Groups-1; i>=0; i--) {
    if ( items[i]==0 ) continue;
    _rootItem->insertItem(items[i]);
  }
  return item;
}

ProjectManager::FileItem *ProjectManager::View::findFileItem(const PURL &url) const
{
  QListViewItemIterator it(const_cast<View *>(this));
  for(; it.current(); ++it) {
    if ( it.current()->rtti()!=FileRtti ) continue;
    FileItem *fi = static_cast<FileItem *>(it.current());
    if ( fi->url()==url ) return fi;
  }
  return 0;
}

ProjectManager::FileItem *ProjectManager::View::findFileItem(PURL::FileType type) const
{
  QListViewItemIterator it(const_cast<View *>(this));
  for(; it.current(); ++it) {
    if ( it.current()->rtti()!=FileRtti ) continue;
    FileItem *fi = static_cast<FileItem *>(it.current());
    if ( fi->ftype()==type ) return fi;
  }
  return 0;
}

QListViewItem *ProjectManager::View::findItem(const QString &tag) const
{
  QListViewItemIterator it(const_cast<View *>(this));
  for(; it.current(); ++it) {
    switch (Rtti(it.current()->rtti())) {
      case RootRtti:
      case FileRtti:
      case DeviceRtti: continue;
      case HeaderRtti:
      case RegisterRtti: break;
    }
    if ( it.current()->text(0)==tag ) return it.current();
  }
  return 0;
}

void ProjectManager::View::select(const Editor *e)
{
  QListViewItem *item = 0;
  if ( e->url().isEmpty() ) item = findItem(e->tag());
  else item = findFileItem(e->url());
  if (item) setSelected(item, true);
  else clearSelection();
}

void ProjectManager::View::contextMenu(QListViewItem *item, const QPoint &p, int)
{
  if ( item==0 ) return;

  KIconLoader loader;
  QPixmap configIcon = loader.loadIcon("configure", KIcon::Small);
  QPixmap compileIcon = loader.loadIcon("piklab_compile", KIcon::Small);
  QPixmap cleanIcon = loader.loadIcon("trashcan_empty", KIcon::Small);
  QPixmap newIcon = loader.loadIcon("filenew", KIcon::Small);
  QPixmap addIcon = loader.loadIcon("piklab_addfile", KIcon::Small);
  QPixmap addCurrentIcon = loader.loadIcon("piklab_addcurrentfile", KIcon::Small);
  QPixmap newProjectIcon = loader.loadIcon("piklab_createproject", KIcon::Small);
  QPixmap openProjectIcon = loader.loadIcon("piklab_openproject", KIcon::Small);
  QPixmap deleteIcon = loader.loadIcon("editdelete", KIcon::Small);
  KPopupMenu *pop = new KPopupMenu;

  switch (Rtti(item->rtti())) {
    case DeviceRtti:
    case RegisterRtti: break;
    case RootRtti: {
      RootItem *ri = static_cast<RootItem *>(item);
      if ( _project==0 ) {
        if ( ri->url().isEmpty() ) {
          pop->insertItem(newProjectIcon, i18n("New Project..."), 1);
          pop->insertItem(openProjectIcon, i18n("Open Project..."), 2);
          switch ( pop->exec(p)) {
            case 1: MainWindow::self().newProject(); break;
            case 2: MainWindow::self().openProject(); break;
          }
        } else {
          pop->insertTitle(i18n("Standalone File"));
          pop->insertItem(configIcon, i18n("Configure..."), 1);
          switch( pop->exec(p) ) {
            case 1: MainWindow::self().configure(ConfigCenter::Standalone); break;
          }
        }
      } else {
        pop->insertTitle(i18n("Project"));
        pop->insertItem(configIcon, i18n("Options..."),      1);
        pop->insertItem(compileIcon, i18n("Build Project"),  2);
        pop->insertItem(cleanIcon, i18n("Clean Project"),    3);
        pop->insertItem(newIcon, i18n("New Source File..."), 4);
        pop->insertItem(addIcon, i18n("Add Source File..."), 5);
        pop->insertItem(addIcon, i18n("Add Object File..."), 6);
        if ( MainWindow::currentEditor() )
          pop->insertItem(addCurrentIcon,  i18n("Add Current File"), 6);
        switch( pop->exec(p) ) {
          case 1: MainWindow::self().configureProject(); break;
          case 2: MainWindow::self().buildProject(); break;
          case 3: MainWindow::self().cleanBuild(); break;
          case 4: MainWindow::self().newSourceFile(); break;
          case 5: insertSourceFile(); break;
          case 6: insertObjectFile(); break;
          case 7: insertCurrentFile(); break;
        }
      }
      break;
    }
    case HeaderRtti:
      if ( _project==0 ) break;
      switch (static_cast<HeaderItem *>(item)->group()) {
        case SourceGroup:
        case HeaderGroup:
          pop->insertItem(addIcon, i18n("Add Source File..."), 1);
          if ( pop->exec(p)==1 ) insertSourceFile();
          break;
        case ObjectGroup:
          pop->insertItem(addIcon, i18n("Add Object File..."), 1);
          if ( pop->exec(p)==1 ) insertObjectFile();
          break;
        case DeviceGroup:
        case LinkerScriptGroup:
        case ViewGroup:
        case GeneratedGroup:
        case Nb_Groups: break;
      }
      break;
    case FileRtti: {
      if ( _project==0 ) break;
      FileItem *fi = static_cast<FileItem *>(item);
      pop->insertTitle(item->text(0));
      if ( isGenerated(fi->url()) ) break;
      pop->insertItem(deleteIcon, i18n("Remove From Project"), 1);
      if ( pop->exec(p)==1 ) removeFile(fi->url());
      break;
    }
  }
  delete pop;
}

void ProjectManager::View::closeProject()
{
  if ( _project==0 && projectUrl().isEmpty() ) return;
  if (_project) {
    PURL::List files = MainWindow::editorManager().files();
    PURL::List::const_iterator it = files.begin();
    PURL::List opened;
    for (; it!=files.end(); ++it) {
      if ( !isGenerated(*it) ) opened += *it;
      MainWindow::editorManager().closeEditor(*it);
    }
    _project->setOpenedFiles(opened);
    QString error;
    if ( !_project->save(error) )
      MessageBox::detailedSorry(i18n("Could not save project file \"%1\".").arg(_project->url().pretty()), error, Log::Show);
    delete _project;
    _project = 0;
  }
  _modified = false;
  init();
}

void ProjectManager::View::addGeneratedFiles(const PURL &url)
{
  const PURL::List &gen = _projectData[url].generated;
  PURL::List::const_iterator it = gen.begin();
  for (; it!=gen.end(); ++it) {
    if ( !(*it).probablyExists() ) continue;
    addFile(*it, (*it).fileType(), Generated);
  }
}

void ProjectManager::View::setStandalone(const PURL &url, PURL::FileType type)
{
  if ( projectUrl()==url ) return;
  closeProject();
  _rootItem->set(url, true);
  addFile(url, type, Intrinsic);
  _projectData[url].type = type;
  addGeneratedFiles(url);
}

PURL ProjectManager::View::standaloneGenerator(const PURL &url, PURL::FileType &type) const
{
  QMap<PURL, ProjectData>::const_iterator it = _projectData.begin();
  for (; it!=_projectData.end(); ++it) {
    if ( !it.data().generated.contains(url) ) continue;
    if ( !it.key().probablyExists() ) continue;
    type = it.data().type;
    return it.key();
  }
  type = PURL::Nb_FileTypes;
  return url;
}

void ProjectManager::View::insertSourceFile()
{
  Q_ASSERT(_project);
  PURL url = PURL::getOpen(_project->directory().path(), PURL::sourceFilter(), this, i18n("Select Source File"));
  if ( url.isEmpty() ) return;
  insertFile(url);
}

void ProjectManager::View::insertObjectFile()
{
  Q_ASSERT(_project);
  PURL url = PURL::getOpen(_project->directory().path(), PURL::objectFilter(), this, i18n("Select Hex File"));
  if ( url.isEmpty() ) return;
  insertFile(url);
}

void ProjectManager::View::insertFile(const PURL &url)
{
  PURL purl = url;
  if ( !url.isInto(_project->directory()) ) {
    if ( !MessageBox::askContinue(i18n("File \"%1\" is not inside the project directory.").arg(url.pretty()),
                                  i18n("Add File to Project"), i18n("Copy")) ) return;
    PURL dest(_project->directory(), url.filename());
    QString error;
    if ( !url.copyTo(dest, Log::DontShow) ) {
      MessageBox::detailedSorry(i18n("Copying file \"%1\" to project directory failed.").arg(url.pretty()), error, Log::Show);
      return;
    }
    purl = dest;
  }
  if ( _project->absoluteFiles().contains(purl) ) {
    MessageBox::sorry(i18n("File \"%1\" is already in the project.").arg(url.pretty()), Log::Show);
    return;
  }
  _project->addFile(purl);
  addFile(purl, purl.fileType(), Intrinsic);
}

bool ProjectManager::View::openProject()
{
  PURL url = PURL::getOpen(":open_project", PURL::projectFilter(), this, i18n("Select Project file"));
  return openProject(url);
}

void ProjectManager::View::addGeneratedFile(const PURL &url)
{
  addFile(url, url.fileType(), Generated);
}

void ProjectManager::View::addFile(const PURL &url, PURL::FileType type, FileOrigin origin)
{
  if ( contains(url) ) return;
  PURL::FileProperties properties = PURL::DATA[type].properties;
  PURL::SourceType stype = PURL::DATA[type].sourceType;
  Group grp = (origin==Generated && !(properties & PURL::Header) && stype==PURL::Asm ? GeneratedGroup : group(type));
  QListViewItem *item = new FileItem(headerItem(grp), type, url, origin==Generated);
  ensureItemVisible(item);
  PURL::List &gen = _projectData[projectUrl()].generated;
  if ( origin==Generated && !gen.contains(url) ) gen.append(url);
  if ( type==PURL::Hex && _project==0 ) {
    PURL durl("<" + (url.isEmpty() ? i18n("Disassembly") : url.toFileType(PURL::AsmGPAsm, PURL::AddExtension).filename()) + ">");
    if ( findFileItem(durl)==0 ) {
      (void)new FileItem(headerItem(ViewGroup), PURL::Coff, durl, true);
      gen.append(durl);
    }
  }
  if ( _project && origin==Intrinsic ) _modified = true;
}

void ProjectManager::View::removeGenerated()
{
  PURL::List &gen = _projectData[projectUrl()].generated;
  PURL::List::const_iterator it = gen.begin();
  for (; it!=gen.end(); ++it) {
    removeFile(*it);
    MainWindow::editorManager().closeEditor(*it);
  }
  gen.clear();
}

void ProjectManager::View::removeFile(const PURL &url)
{
  if ( _project && !isGenerated(url) ) _project->removeFile(url);
  FileItem *item = findFileItem(url);
  if ( item==0 ) return;
  HeaderItem *group = static_cast<HeaderItem *>(item->parent());
  delete item;
  if ( group->childCount()==0 ) delete group;
  _modified = true;
  emit guiChanged();
}

void ProjectManager::View::clicked(int button, QListViewItem *item, const QPoint &, int)
{
  if ( item==0 ) return;
  if ( button!=LeftButton ) return;
  QString device = MainWindow::device();
  switch (Rtti(item->rtti())) {
    case RootRtti:
    case HeaderRtti: {
      HeaderItem *hi = static_cast<HeaderItem *>(item);
      if ( hi->group()!=DeviceGroup ) break;
      ::BusyCursor bc;
      QString title = i18n("Device");
      Editor * e = MainWindow::editorManager().findEditor(title);
      if ( e==0 ) {
        e = new DeviceChooser::Editor(title, item->text(0), this);
        static_cast<DeviceChooser::Editor *>(e)->setDevice(device);
        MainWindow::editorManager().addEditor(e);
      }
      MainWindow::editorManager().showEditor(e);
      break;
    }
    case RegisterRtti:
      Register::showView();
      break;
    case DeviceRtti: break;
    case FileRtti: {
      FileItem *fi = static_cast<FileItem *>(item);
      if ( !(PURL::DATA[fi->ftype()].properties & PURL::Editable) ) return;
      Editor *e = MainWindow::editorManager().findEditor(fi->url());
      if ( e==0 && fi->ftype()==PURL::Coff && _project==0 ) {
        PURL url = findFileItem(PURL::Hex)->url();
        if ( device==Device::AUTO_DATA.name )
          device = DeviceEditor::guessDeviceFromFile(projectUrl());
        if ( url.isEmpty() ) {
          HexEditor *he = ::qt_cast<HexEditor *>(MainWindow::currentEditor());
          if ( he==0 ) return;
          e = new DisassemblyEditor(fi->url(), *he, device, this);
        } else e = new DisassemblyEditor(fi->url(), url, device, this);
        if ( !e->slotLoad() ) {
          delete e;
          break;
        }
        MainWindow::editorManager().addEditor(e);
        addGeneratedFile(fi->url());
      } else {
        e = MainWindow::editorManager().openEditor(fi->url());
        if ( e && isGenerated(fi->url()) ) e->setReadOnly(true);
      }
      break;
    }
  }
}

void ProjectManager::View::insertCurrentFile()
{
  insertFile(MainWindow::editorManager().currentEditor()->url());
}

bool ProjectManager::View::editProject()
{
  ProjectEditor dialog(*_project, this);
  if ( dialog.exec()!=QDialog::Accepted ) return false;
  _modified = true;
  return true;
}

bool ProjectManager::View::newProject()
{
  NewProjectDialog dialog(this);
  if ( dialog.exec()==QDialog::Rejected ) return false;

  if ( dialog.url().probablyExists() ) {
    MessageBox::sorry(i18n("The project file \"%1\" already exists." ).arg(dialog.url().pretty()), Log::Show);
    return false;
  }
  // create dir if it does not exist
  PDirectory dir = dialog.url().directory();
  if ( !dir.create(Log::Show) ) return false;

  Project *p = new Project(dialog.url());
  Compile::Config::setDevice(p, dialog.device());
  setProject(p);
  static_cast<KRecentFilesAction &>(MainWindow::action("project_open_recent")).addURL(dialog.url().kurl());
  return true;
}

void ProjectManager::View::setProject(Project *project)
{
  closeProject();
  MainWindow::editorManager().closeAllEditors();
  _project = project;
  _rootItem->set(project->url(), false);
}

bool ProjectManager::View::openProject(const PURL &url)
{
  if ( url.isEmpty() ) return false;
  bool reload = ( _project && _project->url()==url );
  if ( reload && !MessageBox::askContinue(i18n("Project already loaded. Reload?"), i18n(" Already Loaded"), i18n("Reload")) ) return false;

  Project *p = new Project(url);
  QString error;
  if ( !p->load(error) ) {
    MessageBox::detailedSorry(i18n("Could not open project file."), error, Log::Show);
    delete p;
    static_cast< KRecentFilesAction &>(MainWindow::action("project_open_recent")).removeURL(url.kurl());
    return false;
  }
  setProject(p);
  PURL::List files = _project->absoluteFiles();
  PURL::List::const_iterator it = files.begin();
  for(; it!=files.end(); ++it) addFile(*it, (*it).fileType(), Intrinsic);
  _projectData[url].type = PURL::Project;
  addGeneratedFiles(url);
  if ( !reload) static_cast<KRecentFilesAction &>(MainWindow::action("project_open_recent")).addURL(url.kurl());
  files = _project->openedFiles();
  for(it = files.begin(); it!=files.end(); ++it) MainWindow::editorManager().openFile(*it);
  return true;
}

void ProjectManager::View::dragEnterEvent(QDragEnterEvent *ev)
{
    // accept if dragged object can be decoded as an URI
    ev->accept( QUriDrag::canDecode(ev) );
}

void ProjectManager::View::dropEvent(QDropEvent *ev)
{
  QStringList files;
  if ( !QUriDrag::decodeLocalFiles(ev, files) ) return;
  PURL url(*files.begin());
  // only one file name and it is a project file
  if ( files.count()==1 && url.fileType()==PURL::Project ) {
    openProject(url);
    return;
  }
  if ( _project==0 ) return;
  // try to include files in the current project
  for (QStringList::Iterator it=files.begin(); it!=files.end(); ++it) {
    PURL url(*it);
    PURL::FileType type = url.fileType();
    PURL::FileProperties properties = PURL::DATA[type].properties;
    if ( (properties & PURL::Source) || (properties & PURL::Header)
         || type==PURL::Library ||  type==PURL::Object ) insertFile(url);
  }
}

bool ProjectManager::View::isGenerated(const PURL &url) const
{
  if ( projectUrl().isEmpty() ) return false;
  const PURL::List gen = _projectData[projectUrl()].generated;
  PURL::List::const_iterator it = gen.begin();
  for (; it!=gen.end(); ++it)
    if ( (*it)==url ) return true;
  return false;
}

void ProjectManager::View::modified(const PURL &url)
{
  FileItem *item = findFileItem(url);
  if ( item && !item->isGenerated() ) _modified = true;
}

void ProjectManager::View::renamed(QListViewItem *item, int, const QString &text)
{
  Q_ASSERT ( item->rtti()==DeviceRtti );
  Compile::Config::setDevice(_project, text);
  emit guiChanged();
}

void ProjectManager::View::updateGui()
{
  stopRenaming();
  _deviceItem->updateText();
}
