/***************************************************************************
 *   Copyright (C) 2005-2006 Nicolas Hadacek <hadacek@kde.org>             *
 *   Copyright (C) 2003 Alain Gibaud <alain.gibaud@free.fr>                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#ifndef PROJECT_MANAGER_H
#define PROJECT_MANAGER_H

#include <qtable.h>

#include "common/gui/purl.h"
#include "common/gui/misc_gui.h"
class Project;
class Editor;
namespace Coff { class SectionParser; }

//-----------------------------------------------------------------------------
namespace ProjectManager
{
  enum Rtti { RootRtti = 1000, HeaderRtti, FileRtti, DeviceRtti, RegisterRtti };
  enum Group { DeviceGroup = 0, HeaderGroup, LinkerScriptGroup, SourceGroup, ObjectGroup, ViewGroup, GeneratedGroup, Nb_Groups };
  extern Group group(PURL::FileType type);

class RootItem : public KListViewItem
{
public:
  RootItem(KListView *listview);
  void set(const PURL &url, bool standAlone);
  const PURL &url() const { return _url; }
  virtual int rtti() const { return RootRtti; }

private:
  PURL _url;
  bool _standalone;
};

class HeaderItem : public KListViewItem
{
public:
  HeaderItem(RootItem *item, Group group);
  virtual int rtti() const { return HeaderRtti; }
  Group group() const { return _group; }

private:
  static const char *GROUP_LABELS[Nb_Groups];
  Group _group;
};

class FileItem : public KListViewItem
{
public:
  FileItem(HeaderItem *item, PURL::FileType type, const PURL &url, bool generated);
  virtual int rtti() const { return FileRtti; }
  const PURL &url() const { return _url; }
  PURL::FileType ftype() const { return _ftype; }
  bool isGenerated() const { return _generated; }

private:
  PURL::FileType _ftype;
  PURL           _url;
  bool           _generated;

  void set(const PURL &url);
  virtual void paintCell(QPainter *p, const QColorGroup &cg, int column, int width, int align);
};

class RegisterItem : public KListViewItem
{
public:
  RegisterItem(HeaderItem *item);
  virtual int rtti() const { return RegisterRtti; }
};

class DeviceItem : public EditListViewItem
{
public:
  DeviceItem(HeaderItem *item);
  void updateText();
  virtual int rtti() const { return DeviceRtti; }

private:
  virtual QWidget *editWidgetFactory(int) const;
  virtual bool alwaysAcceptEdit(int) const { return true; }
};

//-----------------------------------------------------------------------------
class View : public ListView
{
Q_OBJECT
public:
  View(QWidget *parent);
  virtual ~View();

  bool editProject();
  bool newProject();
  void insertFile(const PURL &url);
  bool openProject();
  bool openProject(const PURL &url);
  void closeProject();
  Project *project() const { return _project; }

  void setStandalone(const PURL &url, PURL::FileType type);
  PURL projectUrl() const { return _rootItem->url(); }
  PURL standaloneGenerator(const PURL &url, PURL::FileType &type) const;
  void removeFile(const PURL &url);
  void select(const Editor *e);
  bool contains(const PURL &url) const { return findFileItem(url); }
  void addGeneratedFile(const PURL &url);
  bool isGenerated(const PURL &url) const;
  void removeGenerated();

  bool isModified() const { return _modified; }
  void setModified(bool modified) { _modified = modified; }

public slots:
  void insertSourceFile();
  void insertObjectFile();
  void insertCurrentFile();
  void updateGui();

private slots:
  void contextMenu(QListViewItem *item, const QPoint &pos, int column);
  void clicked(int button, QListViewItem *item, const QPoint &pos, int column);
  void renamed(QListViewItem *item, int column, const QString &text);
  void modified(const PURL &url);

signals:
  void guiChanged();

private:
  Project           *_project;
  RootItem   *_rootItem;
  DeviceItem *_deviceItem;
  class ProjectData {
  public:
    PURL::FileType type;
    PURL::List generated;
  };
  QMap<PURL, ProjectData> _projectData;
  bool _modified;

  void init();
  HeaderItem *findHeaderItem(Group group) const;
  HeaderItem *headerItem(Group group);
  FileItem *findFileItem(const PURL &url) const;
  FileItem *findFileItem(PURL::FileType type) const;
  QListViewItem *findItem(const QString &name) const;
  virtual void dragEnterEvent(QDragEnterEvent *);
  virtual void dropEvent(QDropEvent *ev);
  void addGeneratedFiles(const PURL &url);
  void rightClicked(QListViewItem *item, const QPoint &pos);
  void leftClicked(QListViewItem *item);
  enum FileOrigin { Intrinsic, Generated };
  void addFile(const PURL &url, PURL::FileType type, FileOrigin origin);
  void setProject(Project *project);
  void initListView();
};

}; // namespace

#endif
