/***************************************************************************
 *   Copyright (C) 2005-2006 Nicolas Hadacek <hadacek@kde.org>             *
 *   Copyright (C) 2003-2004 Alain Gibaud <alain.gibaud@free.fr>           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "text_editor.h"

#include <qhbox.h>
#include <qfont.h>
#include <qfontmetrics.h>
#include <qpoint.h>
#include <qevent.h>
#include <qapplication.h>
#include <qfile.h>
#include <qstring.h>
#include <qstringlist.h>
#include <qtextedit.h>
#include <qlayout.h>

#include <klibloader.h>
#include <kpopupmenu.h>
#include <kiconloader.h>
#include <kfiledialog.h>
#include <klocale.h>
#include <kapplication.h>

#include "toplevel.h"

//-----------------------------------------------------------------------------
KBEditor::KBEditor(Capabilities capabilities, const PURL &url, QWidget *parent, const char *name)
  : Editor(LoadSave | Selection | Search | Line | Print | View | UndoRedo | CanConfigure | capabilities,
           url, parent, name), _view(0)
{
  KLibFactory *factory = KLibLoader::self()->factory("libkatepart");
  _document = static_cast<Kate::Document *>(factory->create(this, "kate", "KKBEditor::Document"));

  QVBoxLayout *top = new QVBoxLayout(this, 0, 10);
  _split = new QSplitter(this);
  _split->setFrameStyle(QFrame::TabWidgetPanel |  QFrame::Sunken);
  top->addWidget(_split);

  connect(_document, SIGNAL(undoChanged()), SIGNAL(guiChanged()));
  connect(_document, SIGNAL(selectionChanged()), SIGNAL(guiChanged()));
  connect(_document, SIGNAL(textChanged()), SIGNAL(guiChanged()));
  connect(_document, SIGNAL(hlChanged()), SLOT(highlightChanged()));
  setAcceptDrops(true);
  addView();

  if ( capabilities & Breakpoints ) {
    _document->setPixmap(KTextEditor::MarkInterface::Execution, SmallIcon("piklab_program_counter"));
    _document->setPixmap(KTextEditor::MarkInterface::BreakpointActive, SmallIcon("piklab_breakpoint_active"));
    _document->setPixmap(KTextEditor::MarkInterface::BreakpointDisabled, SmallIcon("piklab_breakpoint_disabled"));
    _document->setPixmap(KTextEditor::MarkInterface::BreakpointReached, SmallIcon("piklab_breakpoint_reached"));
    MainWindow::debugManager().addEditor(this);
  }

  setReadOnly(PURL::DATA[url.fileType()].properties & PURL::ReadOnly);
}

void KBEditor::addView()
{
    KTextEditor::View *v = _document->createView(_split);
    if ( _view==0 ) _view = static_cast<Kate::View *>(v);
    Q_ASSERT(v);
    connect(v, SIGNAL(gotFocus(Kate::View *)), SLOT(gotFocus(Kate::View *)));
    connect(v, SIGNAL(cursorPositionChanged()), SLOT(statusChanged()));
    connect(v, SIGNAL(dropEventPass(QDropEvent *)), SIGNAL(dropEventPass(QDropEvent *)));
    connect(v, SIGNAL(newStatus()), SLOT(statusChanged()));
    v->show();
    v->setFocus();
    v->installEventFilter(this);
    KTextEditor::PopupMenuInterface *pmi = KTextEditor::popupMenuInterface(v);
    pmi->installPopup(&MainWindow::popup("text_editor_context_menu"));

    // dispatch available space between views
    QValueList<int> list = _split->sizes();
    QValueList<int>::Iterator it;
    int sum = 0;
    for (it = list.begin(); it!= list.end(); ++it) sum += *it;
    sum /= list.size();
    for (it = list.begin(); it!=list.end(); ++it) *it = sum;
    _split->setSizes(list);

    emit guiChanged();
}

void KBEditor::gotFocus(Kate::View * v)
{
  _view = v;
  setFocusProxy(v);
  statusChanged();
  emit guiChanged();
}

void KBEditor::removeCurrentView()
{
  delete _view;
  _view = static_cast<Kate::View *>(_document->views().last());
  _document->views().last()->setFocus();
  emit guiChanged();
}

bool KBEditor::state(ValueType type) const
{
  switch (type) {
  case None: return true;
  case Modified: return _document->isModified();
  case MultipleViews: return ( _document->views().count()>2 );
  case UndoEnabled: return ( _document->undoCount()!=0 );
  case RedoEnabled: return ( _document->redoCount()!=0 );
  case SelectionEnabled: return _document->hasSelection();
  case LineNumbersOn: return _view->lineNumbersOn();
  case IconBorderOn: return _view->iconBorder();
  case AutoWrapOn: return _view->dynWordWrap();
  case DosEolOn: return ( _view->getEol()==Dos );
  case ReadWriteEnabled: return _document->isReadWrite();
  case FoldingMarkersOn: return false;
//    Kate::Command *command = Kate::Document::queryCommand("folding-markers");
//    if (command) return ;
  case ValueTypes_End: break;
  }
  Q_ASSERT(false);
  return false;
}

void KBEditor::setState(ValueType type, bool on)
{
  if ( on==state(type) ) return;
  switch (type) {
  case Modified: _document->setModified(on); break;
  case LineNumbersOn: _view->toggleLineNumbersOn(); break;
  case IconBorderOn: _view->toggleIconBorder(); break;
  case AutoWrapOn: _view->setDynWordWrap(on); break;
  case DosEolOn: _view->setEol(on ? Dos : Unix); break;
  case ReadWriteEnabled: _document->setReadWrite(on); break;
  case None:
  case MultipleViews:
  case UndoEnabled:
  case RedoEnabled:
  case SelectionEnabled:
  case FoldingMarkersOn: break; //static_cast<KateView *>(view)->toggleFoldingMarkersOn();
  case ValueTypes_End: Q_ASSERT(false); return;
  }
  statusChanged();
  emit guiChanged();
}

void KBEditor::statusChanged()
{
  uint l, c;
  _view->cursorPosition(&l, &c) ;
  QString text = i18n("Line: %1  Col: %2").arg(l+1).arg(c+1);
  if( isReadOnly() ) text += "  " + i18n("R/O");
  emit statusTextChanged(" " + text + " ");
  if (isModified()) emit modified();
}

uint KBEditor::highlightMode(const QString &name) const
{
  uint mode = 0;
  for (; mode<_document->hlModeCount(); mode++)
    if ( _document->hlModeName(mode)==name ) break;
  return mode;
}

void KBEditor::highlightChanged()
{
  // managed by hand because of collisions with file extensions
  // used by other languages/softwares like other ASM and also PHP...
  const char *name = PURL::DATA[fileType()].highlightModeName;
  if ( name==0 ) return;
  uint mode = highlightMode(name);
  if ( mode==_document->hlModeCount() || _document->hlMode()==mode ) return;
  _document->setHlMode(mode);
}

const KBEditor::MarkTypes KBEditor::AllBreakpoints = KTextEditor::MarkInterface::BreakpointActive
  | KTextEditor::MarkInterface::BreakpointDisabled | KTextEditor::MarkInterface::BreakpointReached;

void KBEditor::setMark(uint line, MarkTypes type)
{
  _document->setMark(line, type);
  setState(IconBorderOn, true);
}

QValueVector<uint> KBEditor::lines(MarkTypes type) const
{
  const QPtrList<KTextEditor::Mark> list = _document->marks();
  QValueVector<uint> l;
  QPtrList<KTextEditor::Mark>::const_iterator it = list.begin();
  for (; it!=list.end(); ++it)
    if ( (*it)->type & type ) l.append((*it)->line);
  return l;
}

void KBEditor::clearMarks(MarkTypes type)
{
  QValueVector<uint> l = lines(type);
  for (uint i=0; i<l.count(); i++)
    _document->removeMark(l[i], type);
}

void KBEditor::clearBreakpoints()
{
  clearMarks(AllBreakpoints);
}

QValueVector<uint> KBEditor::breakpoints() const
{
  return lines(KTextEditor::MarkInterface::BreakpointActive);
}

void KBEditor::toggleBreakpoint()
{
  if ( fileType()==PURL::CSource ) {
    MessageBox::sorry(i18n("Placing breakpoint in C code is not supported yet."), Log::Show);
    return;
  }
  uint line, col;
  _view->cursorPosition(&line, &col);
  toggleBreakpoint(line);
}

void KBEditor::toggleBreakpoint(uint line)
{
  if ( _document->mark(line) & AllBreakpoints ) _document->removeMark(line, AllBreakpoints);
  else {
    // only one breakpoint set
    MainWindow::editorManager().clearMarks(AllBreakpoints);
    setMark(line, KTextEditor::MarkInterface::BreakpointActive);
  }
}

bool KBEditor::load()
{
  if ( !_document->openURL(url().kurl()) ) return false;
  _view->setEol(url().isDosFile() ? Dos : Unix);
  highlightChanged();
  return true;
}

#if 0
void KBEditor::slotChangedText()
{
    //This slot runs the instrucion set help bar,
    // finds the current command word and compares it to the set of instrucions
    //Found in the descript Qstringlist.


    QString currentword;
    QString testval;

    Kate::View *v =  currentView();
    currentword = v->currentTextLine();
    //prepare the string for compareing
    currentword = currentword.simplifyWhiteSpace();
    currentword = currentword.upper();

    for ( QStringList::Iterator it = descript.begin(); it != descript.end(); ++it )
    {
        testval = *it;
        if(testval.startsWith(currentword.left(5)))
        {
            //if (testval.left(5) == currentword.left(5) ) {
            lab_curword->setText(*it);
        }
        //      else {
        //              lab_curword->setText("Pic Instruction Help");
        //      }
    }


}
#endif
#if 0
void KBEditor::populateList()
{
    //Populate the qstringlist with the pic instruction set, and the text to go along with them

    descript    += "CLRF  : Clear F     OPERANDS: f";
    descript    += "ADDWF : Add W to F  OPERANDS: f,d";
    descript    += "ANDWF : AND W with F OPERANDS: f,d";
    descript    += "CLRW  : Clear W     OPERANDS: NONE";
    descript    += "COMF  : Complement F OPERANDS: f,d";
    descript    += "DECF  : Decrement  F OPERANDS: f,d";
    descript    += "DECSSZ: Decrement F, Skip 0 OPERANDS: f,d";
    descript    += "INCF  : Increment F  OPERANDS: f,d";
    descript    += "INCFSZ: Increment F, Skip 0 OPERANDS: f,d";
    descript    += "IORWF : Inclusive OR W with F OPERANDS: f,d";
    descript    += "MOVF  : Move F,  OPERANDS: f,d";
    descript    += "MOVWF : Move W to F OPERANDS: f,d";
    descript    += "NOP   : No Operation OPERANDS: NONE";
    descript    += "RLF   : Rotate Left F through Carry OPERANDS: f,d";
    descript    += "RRF   : Rotate Right F through Carry OPERANDS: f,d";
    descript    += "SUBWF : Subtract W from F OPERANDS: f,d";
    descript    += "SWAPF : Swap Nibbles in F OPERANDS: f,d";
    descript    += "XORWF : Exclusive OR W with F OPERANDS: f,s";
    descript    += "BCF   : Bit Clear F OPERANDS: f,b";
    descript    += "BSF   : Bit Set F   OPERANDS: f,b";
    descript    += "BTFSC : Bit Test F, Skip if Clear OPERANDS: f,b";
    descript    += "BTFSS : Bit Test F, Skip if Set   OPERANDS: f,b";
    descript    += "ADDLW : Add Literal and W OPERANDS: k";
    descript    += "ANDLW : And Literal and W OPERANDS: k";
    descript    += "CLRWDT: Clear Watchdog Timer OPERANDS: NONE";
    descript    += "CALL  : Call Subroutine OPERANDS: k";
    descript    += "GOTO  : Goto address OPERANDS: k";
    descript    += "IORLW : Inclusive OR literal with W OPERANDS: k";
    descript    += "MOVLW : Move Literal with W  OPERANDS: k";
    descript    += "RETFIE: Return From Interrupt OPERANDS: NONE";
    descript    += "RETLW : Return with Literal in W OPERANDS: k";
    descript    += "RETURN: Return from subroutine OPERANDS: NONE";
    descript    += "SLEEP : Go into Standby Mode OPERANDS: NONE";
    descript    += "SUBLW : Subtract W from Literal OPERANDS: k";
    descript    += "XORLW : Exclusive OR Literal with W OPERANDS: k";

}
#endif

//-----------------------------------------------------------------------------
SimpleTextEditor::SimpleTextEditor(Capabilities capabilities, PURL::FileType type, QWidget *parent, const char *name)
  : KBEditor(capabilities, PURL(), parent, name), _type(type), _file(Log::DontShow)
{}

SimpleTextEditor::SimpleTextEditor(Capabilities capabilities, const PURL &url, QWidget *parent, const char *name)
  : KBEditor(capabilities, url, parent, name), _type(url.fileType()), _file(Log::DontShow)
{}

bool SimpleTextEditor::setText(const QString &text)
{
  _file.openForWrite();
  _file.appendText(text);
  _file.close();
  if ( !_document->openURL(_file.url().kurl()) ) return false;
  highlightChanged();
  return true;
}
