/***************************************************************************
 *   Copyright (C) 2005-2006 Nicolas Hadacek <hadacek@kde.org>             *
 *   Copyright (C) 2003-2004 Alain Gibaud <alain.gibaud@free.fr>           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "compile.h"

#include <qapplication.h>
#include <klocale.h>
#include <kprocess.h>
#include <kmessagebox.h>

#include "editor_manager.h"
#include "toplevel.h"
#include "project.h"
#include "config.h"
#include "devices/base/device_list.h"
#include "tools/tool_config.h"

//-----------------------------------------------------------------------------
Compile::LogWidget::LogWidget(QWidget *parent)
  : Log::Widget(parent, "compile_log")
{
  connect(this, SIGNAL(clicked(int, int)), SLOT(lineClicked(int)));
}

void Compile::LogWidget::clear()
{
  Log::Widget::clear();
  _map.clear();
}

void Compile::LogWidget::appendLine(Log::Type type, const QString &message, const PURL &url, uint line)
{
  log(type, message, Log::Delayed);
  if ( !url.isEmpty() ) _map[paragraphs()-1] = Data(url, line);
}

void Compile::LogWidget::appendLine(Log::DebugLevel level, const QString &message, const PURL &url, uint line)
{
  log(level, message, Log::Delayed);
  if ( !url.isEmpty() ) _map[paragraphs()-1] = Data(url, line);
}

void Compile::LogWidget::lineClicked(int line)
{
  if ( !_map.contains(line) ) return;
  Editor *e = MainWindow::editorManager().openEditor(_map[line].url);
  if ( e==0 ) return;
  e->setCursor(_map[line].line, 0);
}

//-----------------------------------------------------------------------------
Compile::Process::Process()
  : _process(0)
{}

void Compile::Process::init(const Tool::Group &group, Tool::Category category, const Project *project)
{
  _group = &group;
  _config = group.createConfig(const_cast<Project *>(project));
  _category = category;
}

Compile::Process::~Process()
{
  delete _config;
}

PDirectory Compile::Process::directory() const
{
  if (_data.project) return _data.project->directory();
  return _data.item.url.directory();
}

bool Compile::Process::start()
{
  _stdout = QString::null;
  _stderr = QString::null;
  QString command = tool()->executable();
  QStringList args = arguments();
  QStringList::const_iterator it = args.begin();
  for (; it!=args.end(); ++it) command += " " + *it;
  emit log(Log::Data, command);
  delete _process;
  _process = new KProcess(this);
  connect(_process, SIGNAL(processExited(KProcess *)), SLOT(done()));
  connect(_process, SIGNAL(receivedStdout(KProcess *, char *, int)), SLOT(receivedStdout(KProcess *, char *, int)));
  connect(_process, SIGNAL(receivedStderr(KProcess *, char *, int)), SLOT(receivedStderr(KProcess *, char *, int)));
  _process->setWorkingDirectory(directory().path());
  *_process << tool()->executable();
  *_process << args;
  return _process->start(KProcess::NotifyOnExit, KProcess::AllOutput);
}

void Compile::Process::receivedStdout(KProcess *, char *buffer, int len)
{
  for (int i=0; i<len; i++) {
    if ( buffer[i]=='\n' ) {
      logStdoutLine(_stdout);
      _stdout = QString::null;
    } else _stdout += QString::fromLatin1(buffer + i, 1);
  }
}

void Compile::Process::receivedStderr(KProcess *, char *buffer, int len)
{
  for (int i=0; i<len; i++) {
    if ( buffer[i]=='\n' ) {
      logStderrLine(_stderr);
      _stderr = QString::null;
    } else _stderr += QString::fromLatin1(buffer + i, 1);
  }
}

void Compile::Process::done()
{
  if ( !_stdout.isEmpty() ) logStdoutLine(_stdout);
  if ( !_stderr.isEmpty() ) logStderrLine(_stderr);
  int code = _process->exitStatus();
  if ( code==0 && successInternal() ) emit success();
  else {
    log(Log::Error, i18n("*** Exited with status: %1 ***").arg(code));
    emit failure();
  }
}

//-----------------------------------------------------------------------------
Compile::Manager::Manager(QObject *parent)
  : QObject(parent, "compile_manager"), _base(0), _project(0)
{}

void Compile::Manager::cleanFile(const PURL &url)
{
  setupFile(Clean, TodoItem(url, false), 0);
}

void Compile::Manager::compileFile(const TodoItem &item, const Project *project)
{
  _label = i18n("Compiling file...");
  setupFile(Clean | (project ? CompileOnly : Build), item, project);
}

void Compile::Manager::setupFile(Operations op, const TodoItem &item, const Project *project)
{
  clear();
  view()->clear();
  _operations = op;
  _type = LinkNormal;
  _project = project;
  _todo.append(item);
  _action = Compiling;
  QTimer::singleShot(0, this, SLOT(start()));
}

void Compile::Manager::cleanProject(const Project &project, LinkType type)
{
  setupProject(Clean, project, type);
}

void Compile::Manager::buildProject(const Project &project, LinkType type)
{
  _label = i18n("Building project...");
  setupProject(Clean | Build, project, type);
}

void Compile::Manager::setupProject(Operations op, const Project &project, LinkType type)
{
  clear();
  view()->clear();
  _operations = op;
  _type = type;
  _project = &project;
  PURL::List urls = sources(project);
  PURL::List::const_iterator it = urls.begin();
  for (; it!=urls.end(); ++it) _todo.append(TodoItem(*it, false));
  _action = Compiling;
  QTimer::singleShot(0, this, SLOT(start()));
}

void Compile::Manager::kill()
{
  if ( clear() ) log(Log::Error, i18n("*** Aborted ***"), Log::Delayed);
  emit failure();
}

bool Compile::Manager::clear()
{
  bool running = ( _base!=0 );
  delete _base;
  _base = 0;
  _project = 0;
  _todo.clear();
  resetError();
  return running;
}

void Compile::Manager::clean(const FileData::List &list)
{
  FileData::List::const_iterator it = list.begin();
  for (; it!=list.end(); ++it)
    if ( (*it).actions & Generated ) (*it).url.del(Log::DontShow);
}

PURL::List Compile::Manager::sources(const Project &project)
{
  PURL::List urls;
  PURL::List files = project.absoluteFiles();
  PURL::List::const_iterator it = files.begin();
  for (; it!=files.end(); ++it)
    if ( ((*it).data().properties & PURL::Source) ) urls += *it;
  return urls;
}

uint Compile::Manager::toolType() const
{
  return Compile::Config::toolType(_project);
}

bool Compile::Manager::needs(Tool::Category category) const
{
  Tool::ActionType atype = (_project ? Tool::Project : Tool::StandaloneFile);
  return Tool::lister().group(toolType()).needs(atype, category);
}

void Compile::Manager::setup(Tool::Category category)
{
  delete _base;
  QString device = Compile::Config::device(_project);
  _base = group().createCompileProcess(category, _project);
  _base->_data = Data(_singleItem, device, _project, _type);
}

bool Compile::Manager::setupCompile()
{
  if ( group().base(Tool::Compiler)==0 ) {
    if ( _operations!=Clean ) {
      KMessageBox::sorry(qApp->mainWidget(), i18n("The selected tool (%1) cannot compile C sources.")
                                             .arg(group().label()));
      log(Log::Error, i18n("*** Aborted ***"), Log::Delayed);
      failureSlot();
    }
    return false;
  }
  if ( !compileOnly() && needs(Tool::Assembler) ) {
    PURL::FileType type = group().sourceFileType(PURL::Asm);
    _todo.append(TodoItem(_singleItem.url.toFileType(type), true));
  }
  setup(Tool::Compiler);
  return true;
}

bool Compile::Manager::setupAssemble()
{
  PURL::FileType type = group().sourceFileType(PURL::Asm);
  if ( _singleItem.url.fileType()!=type ) {
    if ( _operations!=Clean ) {
      QString e = PURL::extensions(type);
      KMessageBox::sorry(qApp->mainWidget(), i18n("The selected assembler (%1) only supports files with extensions %2.")
                                             .arg(group().label()).arg(e));
      log(Log::Error, i18n("*** Aborted ***"), Log::Delayed);
      failureSlot();
    }
    return false;
  }
  setup(Tool::Assembler);
  return true;
}

bool Compile::Manager::setupLink()
{
  _action = Linking;
  if ( !needs(Tool::Linker) ) {
    start();
    return false;
  }
  setup(Tool::Linker);
  return true;
}

bool Compile::Manager::setupBinToHex()
{
  _action = BinToHex;
  if ( !needs(Tool::BinToHex) ) {
    start();
    return false;
  }
  setup(Tool::BinToHex);
  return true;
}

bool Compile::Manager::prepareAction()
{
  switch (_action) {
    case Compiling:
      if ( _todo.count()!=0 ) {
        _singleItem = _todo[0];
        _todo.remove(_todo.begin());
        if ( _singleItem.url.fileType()==PURL::CSource ) return setupCompile();
        return setupAssemble();
      }
      if ( !compileOnly() ) return setupLink();
      break;
    case Linking: return setupBinToHex();
    case BinToHex: break;
  }
  if ( !(_operations & Clean) ) {
    log(Log::Data, i18n("*** Success ***"), Log::Delayed);
    emit success();
  } else if ( _operations!=Clean ) {
    if ( _project==0 || compileOnly() ) setupFile(CompileOnly, _singleItem, _project);
    else setupProject(Build, *_project, _type);
  }
  return false;
}

void Compile::Manager::start()
{
  if ( !prepareAction() ) return;
  if ( !_base->check() ) {
    failureSlot();
    return;
  }
  if ( _operations & Clean ) {
    clean(_base->files());
    QTimer::singleShot(0, this, SLOT(start()));
    return;
  }
  connect(_base, SIGNAL(success()), SLOT(doneSlot()));
  connect(_base, SIGNAL(failure()), SLOT(failureSlot()));
  connect(_base, SIGNAL(log(Log::Type, const QString &, const PURL &, uint)),
          SLOT(logSlot(Log::Type, const QString &, const PURL &, uint)));
  connect(_base, SIGNAL(log(Log::DebugLevel, const QString &, const PURL &, uint)),
          SLOT(logSlot(Log::DebugLevel, const QString &, const PURL &, uint)));
  if ( !_base->start() ) {
    log(Log::Error, i18n("Failed to execute command."), Log::Delayed);
    failureSlot();
  }
}

void Compile::Manager::logSlot(Log::Type type, const QString &message, const PURL &url, uint line)
{
  if ( type==Log::Error ) _error = message;
  static_cast<LogWidget *>(view())->appendLine(type, message, url, line);
}

void Compile::Manager::logSlot(Log::DebugLevel level, const QString &message, const PURL &url, uint line)
{
  static_cast<LogWidget *>(view())->appendLine(level, message, url, line);
}

void Compile::Manager::doneSlot()
{
  if ( hasError() ) {
    failureSlot();
    return;
  }
  FileData::List list = _base->files();
  FileData::List::const_iterator it = list.begin();
  for (; it!=list.end(); ++it) emit updateFile(*it);
  QTimer::singleShot(0, this, SLOT(start()));
}

void Compile::Manager::failureSlot()
{
  clear();
  emit failure();
}
