/***************************************************************************
 *   Copyright (C) 2005-2006 Nicolas Hadacek <hadacek@kde.org>             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#ifndef COMPILE_H
#define COMPILE_H

#include "common/qflags.h"
class KProcess;

#include "common/gui/purl.h"
#include "log_view.h"
#include "devices/base/hex_buffer.h"
#include "tools/tool.h"
class MainWindow;

namespace Compile
{
  class Manager;
  enum FileAction { NoAction = 0, Show = 1, InProject = 2, Generated = 8 };
  Q_DECLARE_FLAGS(FileActions, FileAction)
  Q_DECLARE_OPERATORS_FOR_FLAGS(FileActions)

  class FileData {
  public:
    FileData() {}
    FileData(const PURL &u, FileActions a) : url(u), actions(a) {}
    PURL url;
    FileActions actions;
    class List : public QValueList<FileData> {
    public:
      List() {}
      List(const FileData &data) { append(data); }
    };
  };

//-----------------------------------------------------------------------------
class LogWidget : public Log::Widget
{
  Q_OBJECT
public:
  LogWidget(QWidget *parent);
  void appendLine(Log::Type type, const QString &message, const PURL &url, uint line);
  void appendLine(Log::DebugLevel debug, const QString &message, const PURL &url, uint line);
  virtual void clear();

private slots:
  void lineClicked(int line);

private:
  class Data {
  public:
    Data() {}
    Data(const PURL &u, uint l) : url(u), line(l) {}
    PURL url;
    uint line;
  };
  QMap<uint, Data> _map;
};

//-----------------------------------------------------------------------------
class Process : public QObject
{
Q_OBJECT
public:
  Process();
  void init(const Tool::Group &group, Tool::Category category, const Project *project);
  virtual ~Process();
  virtual bool check() const { return tool()->check(_data.device); }
  virtual bool start();
  virtual FileData::List files() const = 0;

signals:
  void success();
  void failure();
  void log(Log::Type type, const QString &message, const PURL &url = PURL(), uint line = 0);
  void log(Log::DebugLevel level, const QString &message, const PURL &url = PURL(), uint line = 0);

protected:
  const Data &data() const { return _data; }
  PDirectory directory() const;
  virtual QStringList arguments() const { return QStringList(); }
  const Tool::Group &group() const { return *_group; }
  const Tool::Base *tool() const { return group().base(_category); }
  const Config *config() const { return _config; }
  QString _stdout, _stderr;
  virtual void logStdoutLine(const QString &line) { logStderrLine(line); }
  virtual void logStderrLine(const QString &line) = 0;
  virtual bool successInternal() { return true; }

protected slots:
  virtual void receivedStdout(KProcess *p, char *buffer, int len);
  virtual void receivedStderr(KProcess *p, char *buffer, int len);

private slots:
  void done();

private:
  const Tool::Group *_group;
  Tool::Category _category;
  Data           _data;
  KProcess      *_process;
  Config        *_config;

  friend class Compile::Manager;
};

//-----------------------------------------------------------------------------
class Manager : public QObject, public Log::Manager
{
Q_OBJECT
public:
  enum Operation { NoOperation = 0, Clean = 1, CompileOnly = 2, Build = 4 };
  Q_DECLARE_FLAGS(Operations, Operation)

public:
  Manager(QObject *parent);
  QString label() const { return _label; }

  void compileFile(const TodoItem &item, const Project *project);
  void cleanFile(const PURL &url);
  void buildProject(const Project &project, LinkType type);
  void cleanProject(const Project &project, LinkType type);
  void kill();
  bool compileOnly() const { return (_operations & CompileOnly); }

signals:
  void success();
  void failure();
  void updateFile(const Compile::FileData &fdata);

private slots:
  void start();
  void doneSlot();
  void failureSlot();
  void logSlot(Log::Type type, const QString &message, const PURL &url, uint line);
  void logSlot(Log::DebugLevel debug, const QString &message, const PURL &url, uint line);

private:
  Operations _operations;
  enum Action { Compiling, Linking, BinToHex };
  Action     _action;
  QValueList<TodoItem> _todo;
  Process   *_base;
  TodoItem   _singleItem;
  const Project *_project;
  QString    _label;
  LinkType   _type;

  bool startProject();
  bool clear();
  static PURL::List sources(const Project &project);
  static void clean(const FileData::List &list);
  void setupFile(Operations operations, const TodoItem &item, const Project *project);
  void setupProject(Operations operations, const Project &project, LinkType type);
  void setup(Tool::Category category);
  bool setupCompile();
  bool setupAssemble();
  bool setupLink();
  bool setupBinToHex();
  bool prepareAction();
  uint toolType() const;
  const Tool::Group &group() const { return Tool::lister().group(toolType()); }
  bool needs(Tool::Category category) const;
};
Q_DECLARE_OPERATORS_FOR_FLAGS(Manager::Operations)

}; // namespace

#endif
