/***************************************************************************
 *   Copyright (C) 2005-2006 Nicolas Hadacek <hadacek@kde.org>             *
 *   Copyright (C) 2004 Alain Gibaud <alain.gibaud@free.fr>                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "config_widget.h"

#include <qtooltip.h>
#include <qwidgetstack.h>
#include <klocale.h>

#include "devices/base/hex_buffer.h"
#include "tools/tool_config.h"
#include "project.h"
#include "config.h"

//----------------------------------------------------------------------------
GridConfigWidget::GridConfigWidget(QWidget *parent, const char *name)
  : ConfigWidget(parent, name)
{
  QGridLayout *top = new QGridLayout(this, 1, 1, 10, 10);
  top->setRowStretch(1, 1);
  _topLayout = new QGridLayout(1, 1, 10);
  top->addLayout(_topLayout, 0, 0);
}

//----------------------------------------------------------------------------
TabConfigWidget::TabConfigWidget(QWidget *parent, const char *name)
  : ConfigWidget(parent, name)
{
  QGridLayout *top = new QGridLayout(this, 1, 1, 0, 10);
  _tabWidget = new QTabWidget(this);
  top->addWidget(_tabWidget, 0, 0);
}

//----------------------------------------------------------------------------
ToolConfigWidget::ToolConfigWidget(Project *project, const char *name)
  : GridConfigWidget(0, name), _project(project),
    _customOptions(0), _includeDirs(0), _hexFormat(0),
    _config(0)
{
  _container = new Container(this);
  _container->gridLayout()->setColStretch(2, 1);
  _topLayout->addWidget(_container, 0, 0);
  _customOptions = createCustomOptionsEntry();
}

ToolConfigWidget::~ToolConfigWidget()
{
  delete _config;
}

SelectDirectoryWidget *ToolConfigWidget::createDirectoryEntry(const QString &text)
{
  uint row = _container->numRows();
  QLabel *label = new QLabel(text, _container);
  _container->addWidget(label, row, 0);
  SelectDirectoryWidget *sdw = new SelectDirectoryWidget(_project ? _project->directory() : PDirectory(), _container);
  QToolTip::add(sdw, i18n("<qt>By using the pseudovariable $(SRCPATH), directories relative to the source directory can be specified.</qt>"));
  _container->addMultiCellWidget(sdw, row,row, 1,2);
  return sdw;
}

QLineEdit *ToolConfigWidget::createCustomOptionsEntry()
{
  uint row = _container->numRows();
  QLabel *label = new QLabel(i18n("Custom options:"), _container);
  _container->addWidget(label, row, 0);
  QLineEdit *edit = new QLineEdit(_container);
  QToolTip::add(edit, i18n("<qt>By using the pseudovariable $(SRCPATH), directories relative to the source directory can be specified.</qt>")) ;
  _container->addMultiCellWidget(edit, row,row, 1,2);
  return edit;
}

QComboBox *ToolConfigWidget::createHexFormatEntry()
{
  uint row = _container->numRows();
  QLabel *label = new QLabel(i18n("Hex file format:"), _container);
  _container->addWidget(label, row, 0);
  QComboBox *cb = new QComboBox(_container);
  for (uint i=0; i<HexBuffer::Nb_Formats; i++)
    cb->insertItem(HexBuffer::FORMATS[i]);
  cb->insertItem(i18n("as in LIST directive"));
  _container->addWidget(cb, row, 1);
  return cb;
}

void ToolConfigWidget::loadConfig()
{
  ConfigWidget::loadConfig();
  if (_includeDirs) _includeDirs->setDirectories(config().rawIncludeDirs(_category)) ;
  if (_customOptions) _customOptions->setText(config().rawCustomOptions(_category));
  if (_hexFormat) _hexFormat->setCurrentItem(config().hexFormat());
}

void ToolConfigWidget::saveConfig()
{
  ConfigWidget::saveConfig();
  if (_includeDirs) config().setRawIncludeDirs(_category, _includeDirs->directories());
  if (_customOptions) config().setRawCustomOptions(_category, _customOptions->text());
  if (_hexFormat) config().setHexFormat(HexBuffer::Format(_hexFormat->currentItem()));
}

//----------------------------------------------------------------------------
ToolsConfigWidget::ToolsConfigWidget(Project *project, QWidget *parent)
  : GridConfigWidget(parent, "tools_config_widget"), _project(project),
    _stacks(Tool::Nb_Categories)
{
  QLabel *label = new QLabel(i18n("Toolchain:"), this);
  _topLayout->addWidget(label, 0, 0);
  _tool = new QComboBox(this);
  for (uint i=0; i<Tool::lister().nbGroups(); i++)
    _tool->insertItem(Tool::lister().group(i).label());
  connect(_tool, SIGNAL(activated(int)), SLOT(toolChanged()));
  _topLayout->addWidget(_tool, 0, 1);

  _tabWidget = new QTabWidget(this);
  _topLayout->addMultiCellWidget(_tabWidget, 1,1, 0,2);
  for (uint k=0; k<Tool::Nb_Categories; k++) {
    _stacks[k] = new QWidgetStack(_tabWidget);
    for (uint i=0; i<Tool::lister().nbGroups(); i++) {
      ConfigWidget *cw = Tool::lister().group(i).createConfigWidget(Tool::Category(k), project);
      if ( cw==0 ) continue;
      cw->loadConfig();
      _stacks[k]->addWidget(cw, i);
    }
  }

  _topLayout->setColStretch(2, 1);
}

void ToolsConfigWidget::toolChanged()
{
  for (uint k=0; k<Tool::Nb_Categories; k++) {
    _tabWidget->removePage(_stacks[k]);
    _stacks[k]->hide();
    _stacks[k]->raiseWidget(_tool->currentItem());
  }
  for (uint k=0; k<Tool::Nb_Categories; k++) {
    if ( _stacks[k]->widget(_tool->currentItem())==0 ) continue;
    _stacks[k]->show();
    _tabWidget->addTab(_stacks[k], i18n(Tool::CATEGORY_DATA[k].label));
  }
  _tabWidget->setCurrentPage(0);
}

void ToolsConfigWidget::loadConfig()
{
  ConfigWidget::loadConfig();
  _tool->setCurrentItem(Compile::Config::toolType(_project));
  toolChanged();
}

void ToolsConfigWidget::saveConfig()
{
  ConfigWidget::saveConfig();
  Compile::Config::setToolType(_project, _tool->currentItem());
  for (uint k=0; k<Tool::Nb_Categories; k++) {
    if ( _stacks[k]->visibleWidget()==0 ) continue;
    static_cast<ConfigWidget *>(_stacks[k]->visibleWidget())->saveConfig();
  }
}
