/***************************************************************************
 *   Copyright (C) 2005-2006 Nicolas Hadacek <hadacek@kde.org>             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "device_info.h"

#include <qpainter.h>
#include <klocale.h>
#include <kglobal.h>

#include "devices/base/gui/device_group_ui.h"
#include "devices/base/device_list.h"
#include "progs/base/prog_list.h"
#include "tools/tool.h"

QColor Device::statusColor(Device::Status status)
{
  switch (status) {
    case Device::Status_Future: return Qt::blue;
    case Device::Status_IP:     return Qt::green;
    case Device::Status_NR:     return QColor("orange");
    case Device::Status_EOL:    return Qt::red;
    case Device::Status_Unknown:
    case Device::Nb_Status:     break;
  }
  return Qt::black;
}

QString coloredString(const QString &text, QColor color)
{
  return QString("<font color=\"") + color.name() + "\">" + text + "</font>";
}

QString supportedString(bool supported)
{
  return coloredString(supported ? i18n("Supported") : i18n("Unsupported"),
                       supported ? Qt::green :  Qt::red);
}

class Tick {
public:
  Tick() {}
  Tick(double value, double oValue) {
    s = KGlobal::locale()->formatNumber(value, 1);
    min = oValue;
  }
  QString s;
  double min;
};

class TickMap : public QMap<double, Tick>
{
public:
  TickMap() {}
  void add(double value, double oValue) {
    insert(value, Tick(value, oValue), false);
    (*this)[value].min = QMIN((*this)[value].min, oValue);
  }
};

QPixmap drawGraph(const QValueVector<Device::RangeBox> &boxes)
{
  const uint w = 300, h = 200;
  QPixmap pixmap(w, h);
  pixmap.fill(Qt::white);
  QPainter p(&pixmap);
  QFontMetrics f(p.font());
  TickMap xTicks, yTicks;
  xTicks.add(0.0, 0.0);
  yTicks.add(0.0, 0.0);
  for (uint i=0; i<boxes.count(); i++) {
//    qDebug("box #%i: %f=[%f %f] %f=[%f %f]", i, boxes[i].start.x, boxes[i].start.yMin,
//           boxes[i].start.yMax, boxes[i].end.x, boxes[i].end.yMin, boxes[i].end.yMax);
    xTicks.add(boxes[i].start.x, boxes[i].start.yMin);
    xTicks.add(boxes[i].start.x, boxes[i].start.yMax);
    xTicks.add(boxes[i].end.x, boxes[i].end.yMin);
    xTicks.add(boxes[i].end.x, boxes[i].end.yMax);
    yTicks.add(boxes[i].start.yMin, boxes[i].start.x);
    yTicks.add(boxes[i].start.yMax, boxes[i].start.x);
    yTicks.add(boxes[i].end.yMin, boxes[i].end.x);
    yTicks.add(boxes[i].end.yMax, boxes[i].end.x);
  }
  double xMax = 0.0, yMax = 0.0;
  int xStart = 0;
  int yStart = h-1 - f.lineSpacing();
  TickMap::const_iterator it = xTicks.begin();
  for (; it!=xTicks.end(); ++it) {
    xStart = QMAX(xStart, f.width(it.data().s));
    xMax = QMAX(xMax, it.key());
  }
  for (it = yTicks.begin(); it!=yTicks.end(); ++it)
    yMax = QMAX(yMax, it.key());
  int xEnd = w-1 - f.width(xTicks[xMax].s)/2;
  QRect rect = f.boundingRect(yTicks[yMax].s);
  int yEnd = rect.height()/2;

  // draw boxes
  p.setPen(Qt::lightGray);
  p.setBrush(Qt::lightGray);
  for (uint i=0; i<boxes.count(); i++) {
    double ax = double(xEnd - xStart)/xMax;
    double ay = double(yEnd - yStart)/yMax;
    QPointArray pa(4);
    pa.setPoint(0, qRound(ax*boxes[i].start.x), qRound(ay*boxes[i].start.yMin));
    pa.setPoint(1, qRound(ax*boxes[i].end.x), qRound(ay*boxes[i].end.yMin));
    pa.setPoint(2, qRound(ax*boxes[i].end.x), qRound(ay*boxes[i].end.yMax));
    pa.setPoint(3, qRound(ax*boxes[i].start.x), qRound(ay*boxes[i].start.yMax));
    pa.translate(xStart, yStart);
    p.drawPolygon(pa);
  }

  // draw axis
  p.setPen(Qt::black);
  p.drawLine(xStart, yStart, w-1, yStart);
  p.drawLine(xStart, yStart, xStart, 0);

  // draw ticks and lines
  p.setPen(Qt::DotLine);
  for (it = yTicks.begin(); it!=yTicks.end(); ++it) {
    int y1 = yStart + qRound(it.key()*(yEnd-yStart)/yMax);
    QRect rect = f.boundingRect(it.data().s);
    p.drawText(xStart/2-rect.width()/2 , y1+rect.height()/2, it.data().s);
    int xmin = xStart + qRound(it.data().min*(xEnd-xStart)/xMax);
    p.drawLine(xStart, y1, xmin, y1);
  }
  for (it = xTicks.begin(); it!=xTicks.end(); ++it) {
    int x1 = xStart + qRound(it.key()*(xEnd-xStart)/xMax);
    QRect rect = f.boundingRect(it.data().s);
    p.drawText(x1-rect.width()/2, h-1, it.data().s);
    int ymin = yStart + qRound(it.data().min*(yEnd-yStart)/yMax);
    p.drawLine(x1, yStart, x1, ymin);
  }

  return pixmap;
}

QPixmap Device::vddGraph(const QString &xLabel, const QString &yLabel,
                         const QValueVector<Device::RangeBox> &boxes)
{
  uint sp = 10;
  QPixmap graph = drawGraph(boxes);
  QPainter p;
  QFontMetrics f(p.font());
  QPixmap pixmap(graph.width() + sp + f.width(xLabel), graph.height() + sp + f.lineSpacing());
  pixmap.fill(Qt::white);
  copyBlt(&pixmap, 0, f.lineSpacing() + sp, &graph, 0, 0, graph.width(), graph.height());
  p.begin(&pixmap);
  p.setPen(Qt::black);
  p.drawText(0, f.lineSpacing(), yLabel);
  p.drawText(pixmap.width()-1-f.width(xLabel), pixmap.height()-1, xLabel);
  return pixmap;
}

QString Device::htmlInfo(const QString &name, const QString &deviceHref)
{
  const Device::Data *data = Device::lister().data(name);
  Q_ASSERT(data);

  QString doc = "<h1>";
  bool first = true;
  for (uint i=0; i<FrequencyRange::Nb_Specials; i++) {
    FrequencyRange::Special s = FrequencyRange::Special(i);
    for (uint k=0; k<data->frequencyRanges().count(); k++) {
      if ( data->frequencyRanges()[k].special!=s ) continue;
      if (first) first = false;
      else doc += " / ";
      doc += data->fname(s);
      break;
    }
  }
  doc += "</h1>";
  doc += "<table>";
  QString status = coloredString(i18n(Device::STATUS_DATA[data->status()].label), statusColor(data->status()));
  if ( data->alternatives().count() ) {
    QString s;
    for (uint i=0; i<data->alternatives().count(); i++) {
      if ( i!=0 ) s += ", ";
      QString href = deviceHref.arg(data->alternatives()[i].upper());
      s += QString("<a href=\"%1\">%2</a>").arg(href).arg(data->alternatives()[i].upper());
    }
    status += i18n(" (consider using %1)").arg(s);
  }
  doc += htmlTableRow(i18n("Status"), status);
  QString url = "<a href=\"" + data->groupui().informationUrl(*data) + "\">" + name + "</a>";
  doc += htmlTableRow(i18n("Information Page"), url);
  doc += "</table>";

  // memory
  doc += "<hr />";
  doc += data->groupui().informationHtml(*data);

  // supported
  doc += "<hr />";
  doc += "<table>";
  QString tools;
  for (uint i=0; i<Tool::lister().nbGroups(); i++) {
    if ( !Tool::lister().group(i).isSupported(name) ) continue;
    tools += Tool::lister().group(i).label() + "<br />";
  }
  doc += htmlTableRow(i18n("Supported Tools"), tools);
  QString progs;
  for (uint i=0; i<Programmer::lister().nbGroups(); i++) {
    if ( !Programmer::lister().group(i).isSupported(name) ) continue;
    progs += Programmer::lister().group(i).label() + "<br />";
  }
  doc += htmlTableRow(i18n("Supported Programmers"), progs);
  doc += "</table>";

  return doc;
}
