/***************************************************************************
 *   Copyright (C) 2005-2006 Nicolas Hadacek <hadacek@kde.org>             *
 *   Copyright (C) 2003-2004 Alain Gibaud <alain.gibaud@free.fr>           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "gputils.h"

#include <qregexp.h>

#include "tools/gputils_compile.h"
#include "tools/gputils_config_widget.h"
#include "tools/gputils_config.h"
#include "devices/pic/pic_memory.h"
#include "devices/base/device_list.h"

QString GPUtils::toDeviceName(const QString &device)
{
  if ( device.startsWith("PS") ) return device.lower();
  return "p" + device.lower();
}

bool GPUtils::Base::checkExecutable(const QString &exec, bool withWine) const
{
  return !run(exec, "-v", withWine).isNull();
}

QString GPUtils::Group::downloadText() const
{
  return i18n("<qt>GPUtils can be downloaded from <a href=\"http://gputils.sourceforge.net\">this page</a>.<br>"
              "A binary package for Mandriva is available <a href=\"href=http://sourceforge.net/project/showfiles.php?group_id=138852&package_id=181224\">here</a>.</qt>");
}

Tool::Base *GPUtils::Group::baseFactory(Tool::Category c) const
{
  switch (c) {
    case Tool::Compiler:      return 0;
    case Tool::Assembler:     return new Base("gpasm");
    case Tool::Linker:        return new Base("gplink");
    case Tool::BinToHex:      return 0;
    case Tool::ObjectViewer:  return new Base("gpvo");
    case Tool::Disassembler:  return new Base("gpdasm");
    case Tool::Nb_Categories: break;
  }
  Q_ASSERT(false);
  return 0;
}

void GPUtils::Group::initSupported()
{
  const Tool::Base *b = base(Tool::Assembler);
  QString s = b->run(b->executable(), "-l", withWine());
  if ( s.isNull() ) return;
  QStringList devices = QStringList::split(' ', s.simplifyWhiteSpace().upper());
  for (uint i=0; i<devices.count(); i++) {
    QString name = devices[i];
    if ( devices[i].startsWith("P1") ) name = name.mid(1);
    const Device::Data *data = Device::lister().data(name);
    if (data) addDevice(data);
  }
}

PDirectory GPUtils::Group::linkerScriptsDirectory() const
{
  const Tool::Base *b = base(Tool::Linker);
  QString s = b->run(b->executable(), "-h", withWine());
  if ( s.isNull() ) return PDirectory();
  QRegExp re(".*Default linker script path ([^\\n]*)\\n.*");
  if ( !re.exactMatch(s) ) return PDirectory();
  return PDirectory(re.cap(1));
}

PURL GPUtils::Group::linkerScriptUrl(Compile::LinkType type, const QString &device) const
{
  QString basename = device.lower();
  if ( type==Compile::LinkIcd2 ) basename += 'i';
  return PURL(linkerScriptsDirectory(), basename, PURL::Lkr);
}

PDirectory GPUtils::Group::headerFilesDirectory() const
{
  const Tool::Base *b = base(Tool::Assembler);
  QString s = b->run(b->executable(), "-h", withWine());
  if ( s.isNull() ) return PDirectory();
  QRegExp re(".*Default header file path ([^\\n]*)\\n.*");
  if ( !re.exactMatch(s) ) return PDirectory();
  return PDirectory(re.cap(1));
}

bool GPUtils::Group::needs(Tool::ActionType type, Tool::Category category) const
{
  if ( type==Tool::Project && category==Tool::Linker ) return true;
  return false;
}

Compile::Process *GPUtils::Group::processFactory(Tool::Category category, const Project *project) const
{
  switch (category) {
    case Tool::Compiler: break;
    case Tool::Assembler:
      if (project) return new GPUtils::AssembleProjectFile;
      return new GPUtils::AssembleStandaloneFile;
    case Tool::Linker:
      if (project) return new GPUtils::LinkProject;
      break;
    case Tool::BinToHex: break;
    case Tool::ObjectViewer: break; // #### TODO ?
    case Tool::Disassembler: break; // #### TODO ?
    case Tool::Nb_Categories: break;
  }
  Q_ASSERT(false);
  return 0;
}

ToolConfigWidget *GPUtils::Group::configWidgetFactory(Tool::Category category, Project *project) const
{
  switch (category) {
    case Tool::Compiler: break;
    case Tool::Assembler: return new GPUtils::AssemblerConfigWidget(project);
    case Tool::Linker:
      if (project) return new GPUtils::ProjectLinkerConfigWidget(*project);
      break;
    case Tool::BinToHex: break;
    case Tool::ObjectViewer: break;
    case Tool::Disassembler:
      if (project) break;
      return new GPUtils::DisassemblerConfigWidget;
    case Tool::Nb_Categories: Q_ASSERT(false); break;
  }
  return 0;
}

Compile::Config *GPUtils::Group::configFactory(::Project *project) const
{
  return new Config(project);
}

PURL::FileType GPUtils::Group::sourceFileType(PURL::SourceType type) const
{
  switch (type) {
    case PURL::Asm: return PURL::AsmGPAsm;
    case PURL::C:   return PURL::CSource;
    case PURL::Nb_SourceTypes: break;
  }
  Q_ASSERT(false);
  return PURL::Nb_FileTypes;
}

QString GPUtils::Group::generateDeviceConfig(PURL::SourceType, const Device::Memory &memory, bool &ok) const
{
  const Pic::Memory &pmemory = static_cast<const Pic::Memory &>(memory);
  const Pic::Data &data = pmemory.device();
  const Pic::Config &config = data.config();
  ok = true;
  QString s;
  for (uint i=0; i<data.nbWords(Pic::MemoryConfig); i++) {
    uint v = pmemory.normalizedWord(Pic::MemoryConfig, i);
    const Pic::Config::Word &cword = config._words[i];
    if ( cword.masks.count()==0 ) continue;
    s += "  __CONFIG ";
    QString cn = config.configName(i, Pic::Config::ConfigNameAsmMacro);
    if ( !cn.isEmpty() ) s += cn + ", ";
    QStringList cnames;
    for (uint k=0; k<cword.masks.count(); k++) {
      const Pic::Config::Mask &cmask = cword.masks[k];
      if ( cmask.name=="DEBUG" ) continue;
      if ( (cmask.value & cword.pmask)==cmask.value ) continue; // protected bits
      for (int l=cmask.values.count()-1; l>=0; l--) {
        if ( (v & cmask.values[l].value)==cmask.values[l].value ) {
          if ( cmask.values[l].cname.isEmpty() ) ok = false;
          else cnames += cmask.values[l].cname;
          break;
        }
      }
    }
    s += cnames.join(" & ");
    s += "\n";
  }
  return s;
}
