/***************************************************************************
 *   Copyright (C) 2005-2006 Nicolas Hadacek <hadacek@kde.org>             *
 *   Copyright (C) 2003-2004 Alain Gibaud <alain.gibaud@free.fr>           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "gputils_compile.h"

#include <qregexp.h>
#include <klocale.h>

#include "tools/process.h"
#include "config.h"
#include "project.h"
#include "object_view.h"
#include "tools/gputils.h"
#include "gputils_config.h"
#include "devices/base/device_list.h"

using namespace Compile;

//-----------------------------------------------------------------------------
const GPUtils::AssembleFile::MessageData GPUtils::AssembleFile::MESSAGES[] = {
  { "Warning", I18N_NOOP("warning"), Log::Warning     },
  { "Error",   I18N_NOOP("error"),   Log::Error       },
  { "Message", I18N_NOOP("message"), Log::Information },
  { 0,         0,                    Log::Nb_Types    }
};

void GPUtils::AssembleFile::logStderrLine(const QString &line)
{
  QRegExp asmError("([^:]*):([0-9]+):(.+)\\[([0-9]+)\\](.+)");
  if ( asmError.exactMatch(line) ) {
    PURL url(asmError.cap(1));
    if ( !url.isInto(directory()) ) url = PURL(directory(), asmError.cap(1));
    QString type = asmError.cap(3).stripWhiteSpace();
    for (uint i=0; MESSAGES[i].name; i++) {
      if ( type!=MESSAGES[i].name ) continue;
      emit log(MESSAGES[i].type, url.filename() + ":" + asmError.cap(2) + ": "
               + i18n(MESSAGES[i].label) + ": " + asmError.cap(5).stripWhiteSpace(),
               url, asmError.cap(2).toUInt()-1);
      return;
    }
  }
  emit log(Log::Data, line); // unrecognized
}

//-----------------------------------------------------------------------------
QStringList GPUtils::AssembleStandaloneFile::arguments() const
{
  QStringList args = AssembleFile::arguments();
  args += "-L"; // force list
  args += "-o";
  args += data().item.url.toFileType(PURL::Hex).filename();
  uint wl = static_cast<const Config *>(config())->gpasmWarningLevel();
  if ( wl!=Config::Nb_WarningLevels ) {
    args += "-w";
    args += QString::number(wl);
  }
  args += config()->includeDirs(Tool::Assembler, directory(), "-I");
  if( data().device!=Device::AUTO_DATA.name ) args += "-p" + toDeviceName(data().device);
  HexBuffer::Format format = config()->hexFormat();
  if( format!=HexBuffer::Nb_Formats ) {
    args += "-a";
    args += HexBuffer::FORMATS[format];
  }
  args += config()->customOptions(Tool::Assembler, directory());
  args += data().item.url.filename();
  return args;
}

FileData::List GPUtils::AssembleStandaloneFile::files() const
{
  FileData::List list;
  list.append(FileData(data().item.url.toFileType(PURL::Lst), Generated | InProject));
  list.append(FileData(data().item.url.toFileType(PURL::Cod), Generated));
  list.append(FileData(data().item.url.toFileType(PURL::Hex), Generated | Show | InProject));
  return list;
}

//-----------------------------------------------------------------------------
QStringList GPUtils::AssembleProjectFile::arguments() const
{
  QStringList args = AssembleFile::arguments();
  args += "-c"; // relocatable code
  args += config()->includeDirs(Tool::Assembler, data().project->directory(), "-I");
  if ( !data().item.generated ) args += "-p" + toDeviceName(data().device);
  uint wl = static_cast<const Config *>(config())->gpasmWarningLevel() ;
  if( wl!=Config::Nb_WarningLevels ) {
    args += "-w";
    args += QString::number(wl);
  }
  args += config()->customOptions(Tool::Assembler, data().project->directory());
  args += data().item.url.relativeTo(directory());
  return args;
}

FileData::List GPUtils::AssembleProjectFile::files() const
{
  FileData::List list;
  list.append(FileData(data().item.url.toFileType(PURL::Object), Generated));
  list.append(FileData(data().item.url.toFileType(PURL::Lst), Generated));
  return list;
}

//-----------------------------------------------------------------------------
QStringList GPUtils::LinkProject::arguments() const
{
  QStringList args = Process::arguments();
  args += "-o";
  args += data().project->hexUrl().filename();
  args += "-c"; // create coff file
  HexBuffer::Format f = config()->hexFormat();
  if ( f!=HexBuffer::Nb_Formats ) {
    args += "-a";
    args += HexBuffer::FORMATS[f];
  }
  args += "-m";  // with map
  args += config()->includeDirs(Tool::Linker, data().project->directory(), "-I");
  if ( data().linkType==LinkIcd2 ) {
    args += "-s";
    args += group().linkerScriptUrl(LinkIcd2, data().device).filepath();
  }
  args += config()->customOptions(Tool::Linker, data().project->directory());
  args += data().project->objectsForLinker();
  args += data().project->librariesForLinker();
  return args;
}

Log::Type GPUtils::LinkProject::logType(const QString &line, bool &isDebug)
{
  isDebug = false;
  if ( line.startsWith("error:") ) return Log::Error;
  if ( line.startsWith("debug:") ) {
    isDebug = true;
    return Log::Nb_Types;
  }
  if ( line.startsWith("message:") ) return Log::Information;
  if ( line.startsWith("warning:") ) return Log::Warning;
  return Log::Data; // unrecognized
}

void GPUtils::LinkProject::logStderrLine(const QString &line)
{
  bool isDebug;
  Log::Type type = logType(line, isDebug);
  if (isDebug) emit log(Log::NormalDebug, line);
  else emit log(type, line);
}

FileData::List GPUtils::LinkProject::files() const
{
  FileData::List list;
  list.append(FileData(group().linkerScriptUrl(data().linkType, data().device), InProject));
  PURL url = data().project->url();
  list.append(FileData(url.toFileType(PURL::Map), Generated | InProject));
  list.append(FileData(url.toFileType(PURL::Lst), Generated));
  list.append(FileData(url.toFileType(PURL::Cod), Generated));
  list.append(FileData(url.toFileType(PURL::Coff), Generated | InProject));
  list.append(FileData(data().project->hexUrl(), InProject | Generated));
  PURL::List included = Coff::includedFiles(url.toFileType(PURL::Coff));
  PURL::List::const_iterator it = included.begin();
  for (; it!=included.end(); ++it) {
    PURL url = *it;
    if ( url.isRelative() ) url = PURL(directory(), (*it).filepath());
    list.append(FileData(url, InProject));
  }
  return list;
}
