/***************************************************************************
 *   Copyright (C) 2006 Nicolas Hadacek <hadacek@kde.org>                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "pic30.h"

#include "devices/base/device_list.h"
#include "tools/pic30_compile.h"
#include "tools/pic30_config_widget.h"
#include "tools/pic30_config.h"
#include "devices/pic/pic_memory.h"

bool PIC30::Base::checkExecutable(const QString &exec, bool withWine) const
{
  QString option;
  if ( !withWine || _category!=Tool::BinToHex )
    option = "--version";
  return !run(exec, option, withWine).isNull();
}

QString PIC30::Group::downloadText() const
{
  return i18n("<qt>The <a href=\"http://microchip.com/stellent/idcplg?IdcService=SS_GET_PAGE&nodeId=1406&dDocName=en010065&part=SW006012\">PIC30 ToolChain</a> is distributed by Microchip under the GNU Public License.<br>"
              "Sources and packages can be downloaded from <a href=\"http://sourceforge.net/project/showfiles.php?group_id=138852&package_id=181225\">this page</a>.");
}

Tool::Base *PIC30::Group::baseFactory(Tool::Category c) const
{
  switch (c) {
    case Tool::Compiler:      return new Base("pic30-elf-gcc");
    case Tool::Assembler:     return new Base("pic30-elf-as");
    case Tool::Linker:        return new Base("pic30-elf-ld");
    case Tool::ObjectViewer:  return 0;
    case Tool::BinToHex:      return new Base("pic30-elf-bin2hex");
    case Tool::Disassembler:  return 0;
    case Tool::Nb_Categories: break;
  }
  Q_ASSERT(false);
  return 0;
}

void PIC30::Group::initSupported()
{
  QStringList list = Device::lister().supportedDevices();
  for (uint i=0; i<list.count(); i++) {
    const Device::Data *data = Device::lister().data(list[i]);
    if ( data && static_cast<const Pic::Data *>(data)->architecture()==Pic::Arch_30X )
      addDevice(data);
  }
}

bool PIC30::Group::needs(Tool::ActionType type, Tool::Category category) const
{
  if ( type==Tool::Project || category==Tool::Linker || category==Tool::BinToHex ) return true;
  return false;
}

Compile::Process *PIC30::Group::processFactory(Tool::Category category, const Project *project) const
{
  switch (category) {
    case Tool::Compiler:
      if (project) return new PIC30::CompileProjectFile;
      return new PIC30::CompileStandaloneFile;
    case Tool::Assembler:
      if (project) return new PIC30::AssembleProjectFile;
      return new PIC30::AssembleStandaloneFile;
    case Tool::Linker:
      if (project) return new PIC30::LinkProject;
      return new PIC30::LinkStandalone;
    case Tool::BinToHex:
      if (project) return new PIC30::BinToHexProject;
      return new PIC30::BinToHexStandalone;
    case Tool::ObjectViewer: break; // #### TODO ?
    case Tool::Disassembler: break; // #### TODO ?
    case Tool::Nb_Categories: break;
  }
  Q_ASSERT(false);
  return 0;
}

ToolConfigWidget *PIC30::Group::configWidgetFactory(Tool::Category category, Project *project) const
{
  switch (category) {
    case Tool::Compiler: return new PIC30::CompilerConfigWidget(project);
    case Tool::Assembler: return new PIC30::AssemblerConfigWidget(project);
    case Tool::Linker:
      if (project) return new PIC30::LinkerConfigWidget(*project);
      break;
    case Tool::BinToHex: break;
    case Tool::ObjectViewer: break;
    case Tool::Disassembler: break;
    case Tool::Nb_Categories: Q_ASSERT(false); break;
  }
  return 0;
}

Compile::Config *PIC30::Group::configFactory(::Project *project) const
{
  return new Config(project);
}

PURL::FileType PIC30::Group::sourceFileType(PURL::SourceType type) const
{
  switch (type) {
    case PURL::Asm: return PURL::AsmPIC30;
    case PURL::C:   return PURL::CSource;
    case PURL::Nb_SourceTypes: break;
  }
  Q_ASSERT(false);
  return PURL::Nb_FileTypes;
}

QString PIC30::Group::generateDeviceConfig(PURL::SourceType type, const Device::Memory &memory, bool &ok) const
{
  const Pic::Memory &pmemory = static_cast<const Pic::Memory &>(memory);
  const Pic::Data &data = pmemory.device();
  const Pic::Config &config = data.config();
  ok = true;
  QString s;
  for (uint i=0; i<data.nbWords(Pic::MemoryConfig); i++) {
    uint v = pmemory.normalizedWord(Pic::MemoryConfig, i);
    const Pic::Config::Word &cword = config._words[i];
    QStringList cnames;
    for (uint k=0; k<cword.masks.count(); k++) {
      const Pic::Config::Mask &cmask = cword.masks[k];
      if ( (cmask.value & cword.pmask)==cmask.value ) continue; // protected bits
      for (int l=cmask.values.count()-1; l>=0; l--) {
        if ( (v & cmask.values[l].value)==cmask.values[l].value ) {
          if ( !cmask.values[l].cname.isEmpty() ) cnames += cmask.values[l].cname;
          break;
        }
      }
    }
    if ( cnames.isEmpty() ) continue;
    if ( type==PURL::Asm ) s += "config __" + cword.name + ", " + cnames.join(" & ");
    else s += "_" + cword.name + "(" + cnames.join(" & ") + ");";
    s += "\n";
  }
  return s;
}
