/***************************************************************************
 *   Copyright (C) 2006 Nicolas Hadacek <hadacek@kde.org>                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "pic30_compile.h"

#include <qregexp.h>
#include <klocale.h>

#include "tools/process.h"
#include "config.h"
#include "project.h"
#include "pic30_config.h"
#include "devices/base/device_list.h"

using namespace Compile;

//-----------------------------------------------------------------------------
void PIC30::CompileFile::logStderrLine(const QString &line)
{
  // #### TODO
  emit log(Log::Data, line); // unrecognized
}

QStringList PIC30::CompileStandaloneFile::arguments() const
{
  QStringList args = CompileFile::arguments();
  args += config()->includeDirs(Tool::Compiler, directory(), "-I");
  if( data().device!=Device::AUTO_DATA.name ) args += "-mcpu=" + data().device;
  args += config()->customOptions(Tool::Compiler, directory());
  args += data().item.url.filepath();
  return args;
}

FileData::List PIC30::CompileStandaloneFile::files() const
{
  FileData::List list;
  list.append(FileData(data().item.url.toFileType(PURL::AsmPIC30), Generated | InProject));
  //list.append(FileData(data().item.url.toFileType(PURL::Lst), Generated | InProject));
  list.append(FileData(data().item.url.toFileType(PURL::Object), Generated));
  //list.append(FileData(data().item.url.toFileType(PURL::Map), Generated | InProject));
  list.append(FileData(data().item.url.toFileType(PURL::Elf), Generated));
  return list;
}

QStringList PIC30::CompileProjectFile::arguments() const
{
  QStringList args = CompileFile::arguments();
  args += "-S"; // compile only
  args += config()->includeDirs(Tool::Compiler, directory(), "-I");
  args += "-mcpu=" + data().device;
  args += config()->customOptions(Tool::Compiler, directory());
  args += data().item.url.relativeTo(directory());
  return args;
}

FileData::List PIC30::CompileProjectFile::files() const
{
  FileData::List list;
  list.append(FileData(data().item.url.toFileType(PURL::AsmPIC30), Generated | InProject));
  return list;
}

//-----------------------------------------------------------------------------
void PIC30::AssembleFile::logStderrLine(const QString &line)
{
  // #### TODO
  emit log(Log::Data, line); // unrecognized
}

QStringList PIC30::AssembleStandaloneFile::arguments() const
{
  QStringList args = AssembleFile::arguments();
  args += "-a=" + data().item.url.toFileType(PURL::Lst).filename(); // listing
  args += "-o";
  args += data().item.url.toFileType(PURL::Object).filename();
  args += config()->includeDirs(Tool::Assembler, directory(), "-I");
  if( data().device!="*" ) args += "-p" + data().device;
  args += config()->customOptions(Tool::Assembler, directory());
  args += data().item.url.filename();
  return args;
}

FileData::List PIC30::AssembleStandaloneFile::files() const
{
  FileData::List list;
  list.append(FileData(data().item.url.toFileType(PURL::Lst), Generated | InProject));
  list.append(FileData(data().item.url.toFileType(PURL::Object), Generated));
  return list;
}

QStringList PIC30::AssembleProjectFile::arguments() const
{
  QStringList args = AssembleFile::arguments();
  args += "-a=" + data().item.url.toFileType(PURL::Lst).relativeTo(directory()); // listing
  args += "-o";
  args += data().item.url.toFileType(PURL::Object).relativeTo(directory());
  args += config()->includeDirs(Tool::Assembler, directory(), "-I");
  if ( !data().item.generated ) args += "-p" + data().device;
  args += config()->customOptions(Tool::Assembler, directory());
  args += data().item.url.relativeTo(directory());
  return args;
}

FileData::List PIC30::AssembleProjectFile::files() const
{
  FileData::List list;
  list.append(FileData(data().item.url.toFileType(PURL::Lst), Generated));
  list.append(FileData(data().item.url.toFileType(PURL::Object), Generated));
  return list;
}

//-----------------------------------------------------------------------------
void PIC30::Link::logStderrLine(const QString &line)
{
  // #### TODO
  emit log(Log::Data, line); // unrecognized
}

QStringList PIC30::LinkStandalone::arguments() const
{
  QStringList args = Compile::Process::arguments();
  args += "--Map";
  args += data().item.url.toFileType(PURL::Map).filename();
  args += "-o";
  args += data().item.url.toFileType(PURL::Elf).filename();
  args += config()->includeDirs(Tool::Linker, directory(), "-L");
  args += config()->customOptions(Tool::Linker, directory());
  args += data().item.url.toFileType(PURL::Object).filename();
  return args;
}

FileData::List PIC30::LinkStandalone::files() const
{
  FileData::List list;
  //list.append(FileData(linkerScriptUrl(project.device()), InProject));
  list.append(FileData(data().item.url.toFileType(PURL::Map), Generated | InProject));
  list.append(FileData(data().item.url.toFileType(PURL::Elf), Generated));
  return list;
}

QStringList PIC30::LinkProject::arguments() const
{
  QStringList args = Compile::Process::arguments();
  args += "--Map";
  args += data().project->hexUrl().toFileType(PURL::Map).relativeTo(directory());
  args += "-o";
  args += data().project->url().toFileType(PURL::Elf).relativeTo(directory());
  args += config()->includeDirs(Tool::Linker, directory(), "-L");
  args += config()->customOptions(Tool::Linker, directory());
  args += data().project->objectsForLinker();
  args += data().project->librariesForLinker("-l");
  return args;
}

FileData::List PIC30::LinkProject::files() const
{
  FileData::List list;
  //list.append(FileData(linkerScriptUrl(project.device()), InProject));
  list.append(FileData(data().project->hexUrl().toFileType(PURL::Map), Generated | InProject));
  list.append(FileData(data().project->hexUrl().toFileType(PURL::Elf), Generated));
  return list;
}

//-----------------------------------------------------------------------------
void PIC30::BinToHexBase::logStderrLine(const QString &line)
{
  // #### TODO
  emit log(Log::Data, line); // unrecognized
}

QStringList PIC30::BinToHexStandalone::arguments() const
{
  QStringList args = BinToHexBase::arguments();
  PURL url = data().item.url.toFileType(PURL::Elf);
  args += url.filename();
  return args;
}

FileData::List PIC30::BinToHexStandalone::files() const
{
  FileData::List list;
  list.append(FileData(data().item.url.toFileType(PURL::Hex), Generated | InProject));
  return list;
}

QStringList PIC30::BinToHexProject::arguments() const
{
  QStringList args = BinToHexBase::arguments();
  PURL url = data().project->hexUrl().toFileType(PURL::Elf);
  args += url.relativeTo(directory());
  return args;
}

FileData::List PIC30::BinToHexProject::files() const
{
  FileData::List list;
  list.append(FileData(data().project->hexUrl(), Generated | InProject));
  return list;
}
