/***************************************************************************
 *   Copyright (C) 2006 Nicolas Hadacek <hadacek@kde.org>                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "sdcc.h"

#include "devices/base/device_list.h"
#include "tools/sdcc_compile.h"
#include "tools/sdcc_config_widget.h"
#include "tools/sdcc_config.h"
#include "tools/gputils_compile.h"
#include "devices/pic/pic_memory.h"

const SDCC::FamilyData SDCC::FAMILY_DATA[Nb_Families] = {
  { "pic14" },
  { "pic16" }
};

QString SDCC::toDeviceName(const QString &device)
{
  if ( device.startsWith("PS") ) return device.lower();
  return "p" + device.lower();
}

SDCC::Family SDCC::family(const QString &device)
{
  const Device::Data *data = Device::lister().data(device);
  switch (static_cast<const Pic::Data *>(data)->architecture()) {
    case Pic::Arch_10F2XX:
    case Pic::Arch_16C5X:
    case Pic::Arch_16X: return P14;
    case Pic::Arch_17X:
    case Pic::Arch_18X: return P16;
    case Pic::Arch_30X:
    case Pic::Nb_Architectures: break;
  }
  Q_ASSERT(false);
  return Nb_Families;
}

bool SDCC::Compiler::checkExecutable(const QString &exec, bool withWine) const
{
  return !run(exec, "-v", withWine).isNull();
}

QString SDCC::Group::downloadText() const
{
  return i18n("<qt>The Small Devices C Compiler can be downloaded from <a href=\"http://sdcc.sourceforge.net\">this page</a>.<br>"
              "A binary package for Mandriva is available <a href=\"http://sourceforge.net/project/showfiles.php?group_id=138852&package_id=181223\">here</a>.</qt>");
}

Tool::Base *SDCC::Group::baseFactory(Tool::Category c) const
{
  switch(c) {
    case Tool::Compiler:      return new Compiler;
    case Tool::Assembler:     return 0;
    case Tool::Linker:        return 0;
    case Tool::ObjectViewer:  return 0;
    case Tool::Disassembler:  return 0;
    case Tool::BinToHex:      return 0;
    case Tool::Nb_Categories: break;
  }
  Q_ASSERT(false);
  return 0;
}

void SDCC::Group::initSupported()
{
  for (uint i=0; i<SDCC::Nb_Families; i++) {
    QStringList options;
    options += QString("-m") + SDCC::FAMILY_DATA[i].name;
    options += "-p";
    const Tool::Base *b = base(Tool::Compiler);
    QString s = b->run(b->executable(), options, withWine());
    if ( s.isNull() ) continue;
    QStringList devices = QStringList::split(' ', s.simplifyWhiteSpace().upper());
    for (uint i=0; i<devices.count(); i++) {
      QString name = devices[i];
      if ( devices[i].startsWith("P1") ) name = name.mid(1);
      const Device::Data *data = Device::lister().data(name);
      if (data) addDevice(data);
    }
  }
}

bool SDCC::Group::needs(Tool::ActionType type, Tool::Category category) const
{
  if ( type==Tool::Project && (category==Tool::Assembler || category==Tool::Linker) ) return true;
  return false;
}

Compile::Process *SDCC::Group::processFactory(Tool::Category category, const Project *project) const
{
  if ( category==Tool::Compiler ) {
    if (project) return new SDCC::CompileProjectFile;
    return new SDCC::CompileStandaloneFile;
  }
  return Tool::lister().group("gputils")->processFactory(category, project);
}

const Tool::Group *SDCC::Group::group(Tool::Category category) const
{
  if ( category==Tool::Compiler ) return this;
  return Tool::lister().group("gputils");
}

ToolConfigWidget *SDCC::Group::configWidgetFactory(Tool::Category category, Project *project) const
{
  if ( category==Tool::Compiler ) return new SDCC::CompilerConfigWidget(project);
  return Tool::lister().group("gputils")->configWidgetFactory(category, project);
}

Compile::Config *SDCC::Group::configFactory(::Project *project) const
{
  return new Config(project);
}

PURL::FileType SDCC::Group::sourceFileType(PURL::SourceType type) const
{
  switch (type) {
    case PURL::Asm: return PURL::AsmGPAsm;
    case PURL::C:   return PURL::CSource;
    case PURL::Nb_SourceTypes: break;
  }
  Q_ASSERT(false);
  return PURL::Nb_FileTypes;
}

QString SDCC::Group::generateDeviceConfig(PURL::SourceType, const Device::Memory &memory, bool &ok) const
{
  const Pic::Memory &pmemory = static_cast<const Pic::Memory &>(memory);
  const Pic::Data &data = pmemory.device();
  const Pic::Config &config = data.config();
  ok = true;
  QString s = "typedef unsigned int word;\n";
  uint address = data.range(Pic::MemoryConfig).start;
  for (uint i=0; i<data.nbWords(Pic::MemoryConfig); i++) {
    uint v = pmemory.normalizedWord(Pic::MemoryConfig, i);
    const Pic::Config::Word &cword = config._words[i];
    if ( cword.masks.count()==0 ) continue;
    s += "word at ";
    QString cn = config.configName(i, Pic::Config::ConfigNameCMacro);
    if ( cn.isEmpty() ) s += toHexLabel(address, data.nbCharsAddress());
    else s += cn;
    s += " " + config.configName(i, Pic::Config::ConfigNameCVariable) + " = ";
    QStringList cnames;
    for (uint k=0; k<cword.masks.count(); k++) {
      const Pic::Config::Mask &cmask = cword.masks[k];
      if ( cmask.name=="DEBUG" ) continue;
      if ( (cmask.value & cword.pmask)==cmask.value ) continue; // protected bits
      for (int l=cmask.values.count()-1; l>=0; l--) {
        if ( (v & cmask.values[l].value)==cmask.values[l].value ) {
          if ( cmask.values[l].cname.isEmpty() ) ok = false;
          else cnames += cmask.values[l].cname;
          break;
        }
      }
    }
    s += cnames.join(" & ");
    s += ";";
    s += "\n";
    address += data.addressIncrement(Pic::MemoryConfig);
  }
  return s;
}
