/***************************************************************************
 *   Copyright (C) 2005-2006 Nicolas Hadacek <hadacek@kde.org>             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "tool.h"

#include <klocale.h>

#include "devices/base/device_list.h"
#include "process.h"
#include "gputils.h"
#include "sdcc.h"
#include "pic30.h"
#include "picc.h"
#include "tools/misc_gui.h"
#include "tools/tool_config.h"
#include "tools/compile.h"
#include "tools/config_widget.h"

//-----------------------------------------------------------------------------
QString Tool::Base::executable() const
{
  const Compile::Config *config = _group->createConfig(0);
  QString exec = config->executableDirectory().path() + _executable;
  delete config;
  return exec;
}

bool Tool::Base::check(const QString &device) const
{
  if ( !checkExecutable(executable(), _group->withWine()) ) {
    MessageBox::detailedSorry(i18n("<b>%1</b> executable cannot found.").arg(executable()), _group->downloadText(), Log::Show);
    return false;
  }
  if ( !device.isEmpty() && device!=Device::AUTO_DATA.name && !_group->isSupported(device) ) {
    MessageBox::sorry(i18n("The selected tool (%1) does not supported device %2.").arg(_group->label()).arg(device), Log::Show);
    return false;
  }
  return true;
}

QString Tool::Base::run(const QString &exec, const QStringList &options, bool withWine) const
{
  SimpleProcess process(withWine ? "wine" : exec);
  if (withWine) process << exec;
  process << options;
  if ( !process.start() ) return QString::null;
  //qDebug("run: %s (%s) %i", exec.latin1(), options.join(" ").latin1(), process.exitStatus());
  int status = process.exitStatus();
  if ( withWine && status!=0 && status!=255 ) return QString::null; // ??
  return process.received();
}

//-----------------------------------------------------------------------------
void Tool::Group::init()
{
  for (uint i=0; i<Nb_Categories; i++) {
    _bases[i] = baseFactory(Category(i));
    if ( _bases[i]==0 ) continue;
    _bases[i]->_category = Category(i);
    _bases[i]->_group = this;
  }
  ::Group::init();
}

bool Tool::Group::withWine() const
{
  const Compile::Config *config = createConfig(0);
  bool with = config->withWine();
  delete config;
  return with;
}

Compile::Process *Tool::Group::createCompileProcess(Category category, const ::Project *project) const
{
  Compile::Process *p = processFactory(category, project);
  Q_ASSERT(p);
  p->init(*group(category), category, project);
  return p;
}

ToolConfigWidget *Tool::Group::createConfigWidget(Category category, ::Project *project) const
{
  ToolConfigWidget *cw = configWidgetFactory(category, project);
  if ( cw==0 ) return 0;
  cw->_config = group(category)->createConfig(project);
  cw->_category = category;
  return cw;
}

Compile::Config *Tool::Group::createConfig(::Project *project) const
{
  Compile::Config *c = configFactory(project);
  c->_group = this;
  return c;
}

bool Tool::Group::supports(PURL::SourceType type) const
{
  switch (type) {
    case PURL::Asm:    return base(Assembler);
    case PURL::Source: return base(Compiler);
    case PURL::Nb_SourceTypes: break;
  }
  Q_ASSERT(false);
  return false;
}

//----------------------------------------------------------------------------
const Tool::CategoryData Tool::CATEGORY_DATA[Nb_Categories] = {
  { "compiler",      I18N_NOOP("Compiler")      },
  { "assembler",     I18N_NOOP("Assembler")     },
  { "linker",        I18N_NOOP("Linker")        },
  { "bin_to_hex",    I18N_NOOP("Bin to Hex")    },
  { "object_viewer", I18N_NOOP("Object Viewer") },
  { "disassembler",  I18N_NOOP("Disassembler")  }
};

Tool::Lister::Lister()
{
  addGroup(new GPUtils::Group);
  addGroup(new SDCC::Group);
  addGroup(new PIC30::Group);
  addGroup(new PICC::PICCLiteGroup);
  addGroup(new PICC::PICCGroup);
  addGroup(new PICC::PICC18Group);
}

namespace Tool
{
  Lister *_lister = 0;
};

const Tool::Lister &Tool::lister()
{
  if ( _lister==0 ) _lister = new Lister;
  return *_lister;
};
