/***************************************************************************
 *   Copyright (C) 2006 Nicolas Hadacek <hadacek@kde.org>                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "tool_config.h"

#include <kapplication.h>
#include <kconfig.h>

#include "tools/process.h"
#include "tools/gputils_config.h"
#include "toplevel.h"
#include "devices/base/device_list.h"
#include "project.h"

QString Compile::Config::rawIncludeDirs(Tool::Category category) const
{
  return value(category, "includes", "$(SRCPATH)");
}
QStringList Compile::Config::includeDirs(Tool::Category category, const PDirectory &dir, const QString &prefix) const
{
  // ### FIXME: detect multiple directories...
  return prefix + dir.expandMacros(rawIncludeDirs(category));
}
void Compile::Config::setRawIncludeDirs(Tool::Category category, const QString &dirs)
{
  setValue(category, "includes", dirs);
}

QString Compile::Config::rawCustomOptions(Tool::Category category) const
{
  return value(category, "custom_options", QString::null);
}
QStringList Compile::Config::customOptions(Tool::Category category, const PDirectory &dir) const
{
  QString s = dir.expandMacros(rawCustomOptions(category));
  return splitCustomOptions(s);
}
void Compile::Config::setRawCustomOptions(Tool::Category category, const QString &options)
{
  setValue(category, "custom_options", options);
}

HexBuffer::Format Compile::Config::hexFormat() const
{
  QString s = value(Tool::Linker, "format", HexBuffer::FORMATS[HexBuffer::IHX32]);
  for (uint i=0; i<HexBuffer::Nb_Formats; i++)
    if ( s==HexBuffer::FORMATS[i] ) return HexBuffer::Format(i);
  return HexBuffer::Nb_Formats;
}
void Compile::Config::setHexFormat(HexBuffer::Format f)
{
  QString s = (f==HexBuffer::Nb_Formats ? 0 : HexBuffer::FORMATS[f]);
  setValue(Tool::Linker, "format", s);
}

QString Compile::Config::device(const Project *project)
{
  QString device = globalValue(project, "device", QString::null);
  return Device::lister().checkName(device);
}

uint Compile::Config::toolType(const Project *project)
{
  QString s = globalValue(project, "tool", QString::null);
  for (uint i=0; i<Tool::lister().nbGroups(); i++)
    if ( s==Tool::lister().group(i).name() ) return i;
  return 0;
}

void Compile::Config::setValue(Tool::Category category, const QString &key, const QString &value)
{
  Q_ASSERT( category!=Tool::Nb_Categories );
  Q_ASSERT( _project || _group );
  if (_project) setValue(*_project, Tool::CATEGORY_DATA[category].name, key, value);
  else setValue(_group->name(), Tool::CATEGORY_DATA[category].name, key, value);
}
QString Compile::Config::value(Tool::Category category, const QString &key, const QString &defaultValue) const
{
  Q_ASSERT( category!=Tool::Nb_Categories );
  Q_ASSERT( _project || _group );
  if (_project) return value(*_project, Tool::CATEGORY_DATA[category].name, key, defaultValue);
  return value(_group->name(), Tool::CATEGORY_DATA[category].name, key, defaultValue);
}
bool Compile::Config::boolValue(Tool::Category category, const QString &key, bool defaultValue) const
{
  QString s = value(category, key, QString::null);
  if ( s.isNull() ) return defaultValue;
  return !( s=="false" || s=="0" );
}
bool Compile::Config::uintValue(Tool::Category category, const QString &key, uint defaultValue) const
{
  bool ok;
  uint i = value(category, key, QString::null).toUInt(&ok);
  if ( !ok ) return defaultValue;
  return i;
}

QString Compile::Config::globalValue(const Project *project, const QString &key, const QString &defaultValue)
{
  if (project) return value(*project, "general", key, defaultValue);
  return value("general", QString::null, key, defaultValue);
}
void Compile::Config::setGlobalValue(Project *project, const QString &key, const QString &value)
{
  if (project) setValue(*project, "general", key, value);
  else setValue("general", QString::null, key, value);
}

QString Compile::Config::value(const Project &project, const QString &subGroup, const QString &key, const QString &defaultValue)
{
  return project.value(subGroup, key, defaultValue);
}
void Compile::Config::setValue(Project &project, const QString &subGroup, const QString &key, const QString &value)
{
  project.setValue(subGroup, key, value);
}

QString Compile::Config::value(const QString &group, const QString &subGroup, const QString &key, const QString &defaultValue)
{
  KConfig *conf = kapp->config();
  QString grp = (subGroup.isEmpty() ? group : group + '-' + subGroup);
  conf->setGroup(grp);
  return conf->readEntry(key, defaultValue);
}
void Compile::Config::setValue(const QString &group, const QString &subGroup, const QString &key, const QString &value)
{
  KConfig *conf = kapp->config();
  QString grp = (subGroup.isEmpty() ? group : group + '-' + subGroup);
  conf->setGroup(grp);
  conf->writeEntry(key, value);
}

void Compile::Config::setExecutableDirectory(const PDirectory &dir)
{
  setValue(Tool::Compiler, "executable_path", dir.path());
  const_cast<Tool::Group *>(_group)->reinitSupported();
}

bool Compile::Config::withWine() const
{
  Q_ASSERT(_group);
  return value(_group->name(), QString::null, "with_wine", defaultWithWine() ? "true" : "false")=="true";
}

void Compile::Config::setWithWine(bool withWine)
{
  Q_ASSERT(_group);
  setValue(_group->name(), QString::null, "with_wine", withWine ? "true" : "false");
}
