/***************************************************************************
 *   Copyright (C) 2005-2006 Nicolas Hadacek <hadacek@kde.org>             *
 *   Copyright (C) 2003-2004 Alain Gibaud <alain.gibaud@free.fr>           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "toplevel.h"

#include <qpixmap.h>
#include <qiconset.h>
#include <qlayout.h>
#include <qsplitter.h>
#include <qstringlist.h>
#include <qtimer.h>
#include <qprogressbar.h>
#include <qeventloop.h>
#include <qapplication.h>
#include <qpopupmenu.h>

#include <kstatusbar.h>
#include <klocale.h>
#include <kedittoolbar.h>
#include <kapplication.h>
#include <kconfig.h>
#include <kaction.h>
#include <kcmdlineargs.h>
#include <kio/observer.h>
#include <kiconloader.h>

#include "hex_editor.h"
#include "project_manager.h"
#include "config_center.h"
#include "project.h"
#include "global_config.h"
#include "prog_config_widget.h"
#include "editor.h"
#include "device_gui.h"
#include "text_editor.h"
#include "new_dialogs.h"
#include "tools/process.h"
#include "tools/misc_gui.h"
#include "devices/base/device_list.h"
#include "progs/base/prog_config.h"
#include "progs/base/prog_list.h"
#include "progs/base/gui/prog_group_ui.h"
#include "editor_manager.h"
#include "tools/compile.h"
#include "object_view.h"
#include "config_gen.h"
#include "tools/tool_config.h"
#include "watch_view.h"

//----------------------------------------------------------------------------
MainWindow *MainWindow::_self = 0;

const Programmer::Group &MainWindow::programmerGroup()
{
  return GlobalConfig::programmerGroup();
}

KAction &MainWindow::action(const char* name)
{
  KAction *a = _self->KMainWindow::action(name);
  Q_ASSERT(a);
  return *a;
}

QPopupMenu &MainWindow::popup(const char *name)
{
  QPopupMenu *popup = static_cast<QPopupMenu *>(_self->factory()->container(name, _self));
  Q_ASSERT(popup);
  return *popup;
}

QString MainWindow::device()
{
  return Compile::Config::device(_self->project());
}

const Device::Data *MainWindow::deviceData()
{
  QString name = device();
  if ( name==Device::AUTO_DATA.name )
    name = DeviceEditor::guessDeviceFromFile(projectManager().projectUrl());
  return Device::lister().data(name);
}

//----------------------------------------------------------------------------
MainWindow::MainWindow()
  : _configGenerator(0), _watchView(0), _programmer(0), _state(Idle)
{
  Q_ASSERT( _self==0 );
  _self = this;

// status bar
    _actionStatus = new QLabel(statusBar());
    statusBar()->addWidget(_actionStatus);
    _actionProgress = new QProgressBar(statusBar());
    statusBar()->addWidget(_actionProgress);
    _debugStatus = new QLabel(statusBar());
    statusBar()->addWidget(_debugStatus, 0, true);
    _editorStatus = new QLabel(statusBar());
    statusBar()->addWidget(_editorStatus, 0, true);
    _programmerStatus = new QPushButton(statusBar());
    connect(_programmerStatus, SIGNAL(clicked()), SLOT(configureProgrammer()));
    statusBar()->addWidget(_programmerStatus, 0, true);
    _toolStatus = new QPushButton(statusBar());
    connect(_toolStatus, SIGNAL(clicked()), SLOT(configureTool()));
    statusBar()->addWidget(_toolStatus, 0, true);

// interface
    QValueList<int> szv;
    szv += 80;
    szv += 350;
    Splitter *splitv = new Splitter(szv, Horizontal, this, "main_vertical_splitter");
    setCentralWidget(splitv);

    _tabWidget = new QTabWidget(splitv);
    splitv->setCollapsible(_tabWidget, false);
    _projectManager = new ProjectManager::View(_tabWidget);
    _tabWidget->addTab(_projectManager, i18n("Project"));
    connect(_projectManager, SIGNAL(guiChanged()), SLOT(updateGUI()));

    QValueList<int> szh;
    szh += 300;
    szh += 80;
    Splitter *splith = new Splitter(szh, Vertical, splitv, "main_horizontal_splitter");
    splitv->setCollapsible(splith, false);

    _editorManager = new EditorManager(splith);
    splith->setCollapsible(_editorManager, false);
    connect(_editorManager, SIGNAL(guiChanged()), SLOT(updateGUI()));
    connect(_editorManager, SIGNAL(modified(const PURL &)),
            _projectManager, SLOT(modified(const PURL &)));
    connect(_editorManager, SIGNAL(statusChanged(const QString &)),
            _editorStatus, SLOT(setText(const QString &)));

    KIconLoader loader;
    _logTab = new QTabWidget(splith, "log_tab");
    splith->setCollapsible(_logTab, false);
    _compileLog = new Compile::LogWidget(_logTab);
    _compileLog->setFocusPolicy(NoFocus);
    QIconSet iset = loader.loadIconSet("piklab_compile", KIcon::Small, 0);
    _logTab->addTab(_compileLog, iset, i18n("Compile Log"));
    _compileLog->setMinimumHeight(50);
    _programLog = new Log::Widget(_logTab, "program_log");
    _programLog->setFocusPolicy(NoFocus);
    iset = loader.loadIconSet("piklab_burnchip", KIcon::Small, 0);
    _logTab->addTab(_programLog, iset, i18n("Program Log"));

// managers
    _debugManager = new DebugManager(this);
    _debugManager->setView(_programLog);
    connect(_debugManager, SIGNAL(targetStateChanged()), SLOT(updateGUI()));
    connect(_debugManager, SIGNAL(statusChanged(const QString &)),
            _debugStatus, SLOT(setText(const QString &)));
    connect(_debugManager, SIGNAL(actionMessage(const QString &)),
            _actionStatus, SLOT(setText(const QString &)));

    _compileManager = new Compile::Manager(this);
    _compileManager->setView(_compileLog);
    connect(_compileManager, SIGNAL(success()), SLOT(compileSuccess()));
    connect(_compileManager, SIGNAL(failure()), SLOT(compileFailure()));
    connect(_compileManager, SIGNAL(updateFile(const Compile::FileData &)),
            SLOT(updateFile(const Compile::FileData &)));

// actions
    // file actions
    KAction *a = KStdAction::openNew(this, SLOT(newSourceFile()), actionCollection());
    a->setText(i18n("&New Source File..."));
    (void)new KAction(i18n("New Hex File..."), "filenew", 0, this, SLOT(newHexFile()),
                      actionCollection(), "file_new_hex");
    (void)KStdAction::open(this, SLOT(openFile()), actionCollection(), "file_open");
    KRecentFilesAction *recent = KStdAction::openRecent(this, SLOT(openRecentFile(const KURL &)), actionCollection());
    recent->setMaxItems(20);
    recent->loadEntries(kapp->config(), "recent-files");
    (void)new KAction(i18n("Save All"), 0, 0, _editorManager, SLOT(saveAllFiles()),
                      actionCollection(), "file_save_all");
    KStdAction::close(_editorManager, SLOT(closeCurrentEditor()), actionCollection());
    (void)new KAction(i18n("C&lose All"), 0, 0, _editorManager, SLOT(closeAllEditors()),
                      actionCollection(), "file_close_all");
    (void)new KAction(i18n("Close All Others"), 0, 0, _editorManager, SLOT(closeAllOtherEditors()),
                      actionCollection(), "file_close_all_others");

    KStdAction::quit(this, SLOT(close()), actionCollection());

    // editor actions
    for (uint i=0; Editor::actionName(i); i++) {
      const char *name = Editor::actionName(i);
      const Editor::ActionData d = Editor::ACTION_DATA[i];
      if ( d.stdAction==KStdAction::ActionNone ) {
        if (d.toggle) (void)new KToggleAction(i18n(d.label), d.icon, d.key, actionCollection(), name);
        else (void)new KAction(i18n(d.label), d.icon, d.key, actionCollection(), name);
      } else (void)KStdAction::create(d.stdAction, name, 0, 0, actionCollection());
    }

    // project actions
    (void)new KAction(i18n("New Project..."), "piklab_createproject",
                      0, this, SLOT(newProject()), actionCollection(), "project_new");
    (void)new KAction(i18n("Open Project..."), "piklab_openproject",
                      0, this , SLOT(openProject()), actionCollection(), "project_open");
    recent = new KRecentFilesAction(i18n("Open Recent Project"),
                      0, this, SLOT(openRecentProject(const KURL &)), actionCollection(), "project_open_recent");
    recent->setMaxItems(20);
    recent->loadEntries(kapp->config(), "recent-projects");
    (void)new KAction(i18n("Project Options..."), "configure",
                      0, this, SLOT(configureProject()), actionCollection(), "project_options");
    (void)new KAction(i18n("Close Project"), "fileclose",
                      0, this, SLOT(closeProject()), actionCollection(), "project_close");
    (void)new KAction(i18n("Add Source File..."), "piklab_addfile",
                      0, _projectManager, SLOT(insertSourceFile()), actionCollection(), "project_add_source_file");
    (void)new KAction(i18n("Add Object File..."), "piklab_addfile",
                      0, _projectManager, SLOT(insertObjectFile()), actionCollection(), "project_add_object_file");
    (void)new KAction(i18n("Add Current File"), "piklab_addcurrentfile",
                      0, _projectManager, SLOT(insertCurrentFile()), actionCollection(), "project_add_current_file");

    // build actions
    (void)new KAction(i18n("&Build Project"), "piklab_compile",
                 Qt::Key_F8, this, SLOT(buildProject()), actionCollection(), "build_build_project");
    (void)new KAction(i18n("&Compile File"), 0,
                 Qt::Key_F10, this, SLOT(compileFile()), actionCollection(), "build_compile_file");
    (void)new KAction(i18n("Clean"), "trashcan_empty",
                 0, this, SLOT(cleanBuild()), actionCollection(), "build_clean");
    (void)new KAction(i18n("Stop"), "stop",
                 0, this, SLOT(stopBuild()), actionCollection(), "build_stop");

    // programmer actions
    (void)new KAction(i18n("&Connect"), "connect_creating",
                 0, this, SLOT(connectDevice()), actionCollection(), "prog_connect");
    (void)new KToggleAction(i18n("Device Power"), "piklab_power",
                 0, this, SLOT(toggleDevicePower()), actionCollection(), "prog_power");
    (void)new KAction(i18n("&Disconnect"), "connect_no",
                 0, this, SLOT(disconnectDevice()), actionCollection(), "prog_disconnect");
    (void)new KAction(i18n("&Program"), "piklab_burnchip",
                 Qt::Key_F5, this , SLOT(program()), actionCollection(), "prog_program");
    (void)new KAction(i18n("&Verify"), "piklab_verifychip",
                 Qt::Key_F6, this , SLOT(verify()), actionCollection(), "prog_verify");
    (void)new KAction(i18n("&Read"), "piklab_readchip",
                 Qt::Key_F7, this , SLOT(read()), actionCollection(), "prog_read");
    (void)new KAction(i18n("&Erase"), "piklab_erasechip",
                 0, this , SLOT(erase()), actionCollection(), "prog_erase");
    (void)new KAction(i18n("&Blank Check"), "piklab_blankcheck",
                 0, this , SLOT(blankCheck()), actionCollection(), "prog_blank_check");
    (void)new KAction(i18n("&Run"), "launch",
                 Qt::SHIFT + Qt::Key_F9, this, SLOT(progRun()), actionCollection(), "prog_run");
    (void)new KAction(i18n("&Stop"), "piklab_stop",
                 0, this, SLOT(progStop()), actionCollection(), "prog_stop");
    (void)new KAction(i18n("R&estart"), "piklab_restart",
                 0, this, SLOT(progRestart()), actionCollection(), "prog_restart");
    (void)new KAction(i18n("&Advanced..."), 0,
                 0, this , SLOT(progAdvanced()), actionCollection(), "prog_advanced");

    // debugger actions
    (void)new KAction(i18n("&Run"), "piklab_run",
                 Qt::SHIFT + Qt::Key_F9, this, SLOT(debugRun()), actionCollection(), "debug_run");
    (void)new KAction(i18n("&Interrupt"), "piklab_debughalt",
                 0, this, SLOT(debugHalt()), actionCollection(), "debug_halt");
    (void)new KAction(i18n("&Step"), "piklab_debug_step",
                 0, this, SLOT(debugStep()), actionCollection(), "debug_next");
    //(void)new KAction(i18n("Step &In"), "piklab_debug_stepin",
    //             0, this, SLOT(debugStepIn()), actionCollection(), "debug_step_in");
    //(void)new KAction(i18n("Step &Out"), "piklab_debug_stepout",
    //             0, this, SLOT(debugStepOut()), actionCollection(), "debug_step_out");
    (void)new KAction(i18n("R&eset"), "piklab_restart",
                 0, this, SLOT(debugReset()), actionCollection(), "debug_reset");
    (void)new KAction(i18n("Clear All Breakpoints"), 0,
                 0, this, SLOT(clearBreakpoints()), actionCollection(), "clear_breakpoints");

    // tools
    (void)new KAction(i18n("&Pikloops..."), 0,
                 0, this , SLOT(runPikloops()), actionCollection(), "tools_pikloops");
    (void)new KAction(i18n("&Device Information..."), "info",
                 0, this , SLOT(deviceInfo()), actionCollection(), "tools_deviceinfo");
    (void)new KAction(i18n("&Config Generator..."), 0,
                 0, this , SLOT(configGenerator()), actionCollection(), "tools_configgen");

    // settings actions
    (void)new KAction(i18n("Configure Editor..."), Qt::CTRL+Qt::SHIFT+Qt::Key_E,
                      this, SLOT(configureEditor()), actionCollection(), "options_configure_editor");
    (void)KStdAction::preferences(this, SLOT(configure()), actionCollection());

    setupGUI();
    QTimer::singleShot(0, this, SLOT(initialLoading()));
}

void MainWindow::initialLoading()
{
  // once everything is properly set up, open editors or project
  // as specified by the command line
  KCmdLineArgs *args = KCmdLineArgs::parsedArgs();
  for (int i = 0; i<args->count(); i++) {
    PURL url(args->url(i));
    if ( url.fileType()==PURL::Project && _projectManager->project()==0 )
      _projectManager->openProject(url);
    else _editorManager->openEditor(url);
  }

  // now, try to reopen project/files
  // note: project or files specified with command line args
  // have precedence over previous session ones.
  PURL purl = GlobalConfig::openedProject();
  if ( project()==0 ) _projectManager->openProject(purl);
  PURL::List files = GlobalConfig::openedFiles();
  PURL::List::const_iterator it = files.begin();
  for (; it!=files.end(); ++it) {
    qDebug("reopen %s", (*it).pretty().latin1());
    _editorManager->openEditor(*it);
  }
  updateGUI();
}

void MainWindow::saveSettings()
{
  static_cast<KRecentFilesAction &>(action("project_open_recent")).saveEntries(kapp->config(), "recent-projects");
  static_cast<KRecentFilesAction &>(action("file_open_recent")).saveEntries(kapp->config(), "recent-files");
  GlobalConfig::writeOpenedProject(project() ? project()->url() : PURL());
  if ( project() ) _projectManager->closeProject();
  PURL::List toSave;
  const PURL::List files = _editorManager->files();
  PURL::List::const_iterator it = files.begin();
  for (; it!=files.end(); ++it)
    if ( !_projectManager->isGenerated(*it) ) toSave.append(*it);
  GlobalConfig::writeOpenedFiles(toSave);
}

void MainWindow::openRecentFile(const KURL &kurl)
{
  _editorManager->openFile(PURL(kurl));
}

void MainWindow::configureToolbar()
{
  saveMainWindowSettings(KGlobal::config(), "MainWindow");
  KEditToolbar dlg(actionCollection());
  connect(&dlg, SIGNAL(newToolbarConfig()), SLOT(applyToolbarSettings()));
  dlg.exec();
}

void MainWindow::applyToolbarSettings()
{
  createGUI();
  applyMainWindowSettings(KGlobal::config(), "MainWindow");
}

void MainWindow::configureEditor()
{
    Editor *e = _editorManager->currentEditor();
    if (e) e->configureEditor();
}

void MainWindow::configure(ConfigCenter::Type showType)
{
  stopOperations();
  ConfigCenter dialog(showType, this);
  dialog.exec();
  delete _programmer;
  _programmer = 0;
  updateGUI();
  _debugManager->update();
}

void MainWindow::configureTool()
{
  if ( project()==0 ) configure(ConfigCenter::Standalone);
  else configureProject();
}

void MainWindow::runPikloops()
{
  SimpleProcess process("pikloops");
  if ( !process.start() ) {
    MessageBox::detailedSorry(i18n("Cannot run \"pikloops\""),
                              i18n("The Pikloops utility (http://pikloops.tripod.com) is not installed in your system:\n"
                                   " You can download pikloops from http://piklab.sourceforge.net"), Log::Show);
    return;
  }
}

void MainWindow::deviceInfo()
{
  DeviceChooser::Dialog d(device(), DeviceChooser::Info, this);
  d.exec();
}

void MainWindow::configGenerator()
{
  PURL::SourceType stype = (currentEditor() ? PURL::DATA[currentEditor()->fileType()].sourceType : PURL::Asm);
  uint group = Compile::Config::toolType(project());
  ConfigGenerator cg(device(), group, stype, this);
  cg.exec();
}

bool MainWindow::queryClose()
{
  stopOperations();
  setState(Closing);
  saveSettings();
  if ( !_editorManager->closeAllEditors() ) {
    setState(Idle);
    return false;
  }
  return true;
}

void MainWindow::newSourceFile()
{
  NewFileDialog dialog(project(), this);
  if ( dialog.exec()!=QDialog::Accepted ) return;
  if ( dialog.url().probablyExists() ) {
    MessageBox::sorry(i18n("File \"%1\" already exists.").arg(dialog.url().pretty()), Log::Show);
    return;
  }
  if ( !dialog.url().create(Log::Show) ) return;
  _editorManager->openEditor(dialog.url());
  if ( dialog.addToProject() ) _projectManager->insertFile(dialog.url());
}

void MainWindow::newHexFile()
{
  if ( device()==Device::AUTO_DATA.name ) {
    MessageBox::sorry(i18n("You need to specify a device to create a new hex file."), Log::Show);
    return;
  }
  QString s;
  for (uint i=0; true; i++) {
    s = i18n("Hex") + (i==0 ? QString::null : QString::number(i));
    if ( _editorManager->findEditor(s)==0 ) break;
  }
  HexEditor *editor = new HexEditor(s, _editorManager);
  editor->setDevice(device());
  editor->memoryRead();
  _editorManager->addEditor(editor);
}

bool MainWindow::openFile()
{
  QString startDir = (project() ? project()->directory().path() : ":open_file");
  PURL url = PURL::getOpen(startDir, "all/allfiles", this ,i18n("Open File"));
  if ( url.fileType()==PURL::Project || url.fileType()==PURL::PikdevProject ) {
    stopOperations();
    if ( !_projectManager->openProject(url) ) return false;
    updateGUI();
    return true;
  }
  return _editorManager->openFile(url);
}

void MainWindow::setState(State state)
{
  _state = state;
  bool idle = ( _state==Idle );
  switch (_state) {
  case Closing: return;
  case Idle:
    BusyCursor::stop();
    _actionStatus->hide();
    _actionProgress->hide();
    break;
  case Compiling:
    BusyCursor::start();
    _actionStatus->show();
    _actionProgress->show();
    _actionStatus->setText(_compileManager->label());
    _logTab->showPage(_compileLog);
    break;
  case Programming:
    BusyCursor::start();
    _actionStatus->show();
    _actionProgress->show();
    _logTab->showPage(_programLog);
    Q_ASSERT(_programmer);
    _programmer->proxy()->disconnect(this);
    connect(_programmer->proxy(), SIGNAL(setTotalSteps(uint)), SLOT(setTotalSteps(uint)));
    connect(_programmer->proxy(), SIGNAL(addToTotalSteps(uint)), SLOT(addToTotalSteps(uint)));
    connect(_programmer->proxy(), SIGNAL(addProgress(uint)), SLOT(addProgress(uint)));
    break;
  }

  // update editor actions
  _editorManager->updateDisplay();
  action("file_save_all").setEnabled(currentEditor());
  action("file_close").setEnabled(currentEditor());
  action("file_close_all").setEnabled(currentEditor());
  action("file_close_all_others").setEnabled(_editorManager->nbEditors()>1);
  action("options_configure").setEnabled(idle);
  action("options_configure_editor").setEnabled(currentEditor() && (currentEditor()->capabilities() & Editor::CanConfigure) );

  // update project
  PURL::FileType currentType = (currentEditor() ? currentEditor()->fileType() : PURL::Nb_FileTypes);
  bool isSource = ( PURL::DATA[currentType].properties & PURL::Source );
  bool inProject = ( currentEditor() && (currentType==PURL::Nb_FileTypes || currentEditor()->url().isEmpty() || _projectManager->contains(currentEditor()->url())) );
  if ( project()==0 && !inProject ) {
    if ( currentEditor()==0 ) _projectManager->closeProject();
    else if ( isSource ) _projectManager->setStandalone(currentEditor()->url(), currentType);
    else {
      PURL::FileType type;
      PURL purl = _projectManager->standaloneGenerator(currentEditor()->url(), type);
      if ( type!=PURL::Nb_FileTypes ) _projectManager->setStandalone(purl, type);
      else if ( currentType==PURL::Hex ) _projectManager->setStandalone(purl, PURL::Hex);
    }
  }
  if ( currentEditor() ) _projectManager->select(currentEditor());

  // update project actions
  action("project_new").setEnabled(idle);
  action("project_open").setEnabled(idle);
  action("project_close").setEnabled(project() && idle);
  action("project_options").setEnabled(project() && idle);
  action("project_add_source_file").setEnabled(project() && idle);
  action("project_add_object_file").setEnabled(project() && idle);
  action("project_add_current_file").setEnabled(project() && !inProject && idle && isSource);

  // update build actions
  uint tool = Compile::Config::toolType(project());
  _toolStatus->setText(" " + Tool::lister().group(tool).label() + " ");
  bool hexProject = ( _projectManager->projectUrl().fileType()==PURL::Hex );
  action("build_build_project").setEnabled((project() || (inProject && !hexProject) ) && idle);
  action("build_compile_file").setEnabled(!hexProject && isSource && inProject && idle);
  action("build_clean").setEnabled((project() || inProject) && idle);
  action("build_stop").setEnabled(_state==Compiling);

  // update programmer status
  Port::Type ptype = Programmer::Config::readPortType(programmerGroup());
  _programmerStatus->setText(" " + programmerGroup().label() + " (" + i18n(Port::DATA[ptype].label) + ") ");
  QFont f = font();
  bool supported = (deviceData() ? programmerGroup().isSupported(deviceData()->name()) : false);
  f.setItalic(!supported);
  _programmerStatus->setFont(f);
  bool hasHex = ( currentType==PURL::Hex || (project() && _projectManager->contains(project()->hexUrl())) );
  action("prog_read").setEnabled(idle);
  action("prog_program").setEnabled(hasHex && idle);
  action("prog_verify").setEnabled(hasHex && idle);
  action("prog_erase").setEnabled(idle);
  action("prog_blank_check").setEnabled(idle);
  action("prog_connect").setEnabled(idle);
  Programmer::State pstate = (_programmer ? _programmer->state() : Programmer::NotConnected);
  static_cast<KToggleAction &>(action("prog_power")).setEnabled(idle && pstate!=Programmer::NotConnected);
  action("prog_disconnect").setEnabled(idle && pstate!=Programmer::NotConnected);
  bool debugger = programmerGroup().isDebugger();
  bool resetAvailable = ( programmerGroup().properties() & Programmer::CanReleaseReset );
  action("prog_run").setEnabled(idle && resetAvailable && !debugger && pstate!=Programmer::Running);
  action("prog_stop").setEnabled(idle && !debugger && pstate==Programmer::Running);
  action("prog_restart").setEnabled(idle && !debugger && pstate==Programmer::Running);
  const Programmer::GroupUI *pgui = static_cast<const Programmer::GroupUI *>(programmerGroup().gui());
  action("prog_advanced").setEnabled(idle && pgui->hasAdvancedDialog());

  // update debugger status
  if (debugger) {
    if ( _watchView==0 ) {
      _watchView = new Watch::View(_tabWidget);
      _tabWidget->addTab(_watchView, i18n("Watch"));
    }
  } else {
    delete _watchView;
    _watchView = 0;
  }
  action("debug_run").setEnabled(idle && debugger && pstate==Programmer::Halted);
  action("debug_halt").setEnabled(idle && debugger && pstate==Programmer::Running);
  action("debug_next").setEnabled(idle && debugger && pstate==Programmer::Halted);
  action("debug_reset").setEnabled(idle && debugger && ( pstate==Programmer::Running || pstate==Programmer::Halted ));

  _projectManager->updateGui();

  // caption
  QString caption;
  if ( project() ) caption += project()->name();
  if ( currentEditor() ) {
    if ( project() ) caption += " - ";
    caption += currentEditor()->url().filepath();
  }
  setCaption(KApplication::kApplication()->makeStdCaption(caption));

  emit stateChanged();
}

//-----------------------------------------------------------------------------
void MainWindow::compileFile()
{
  if ( project()==0 && programmerGroup().isDebugger() ) {
    MessageBox::sorry(i18n("<qt>It is not possible to compile a stand-alone file for debugging. You need either to create a project or to select another programmer.</qt>"), Log::Show);
    return;
  }
  Editor *e = _editorManager->currentEditor();
  Q_ASSERT( e && (PURL::DATA[e->fileType()].properties & PURL::Source) );
  if ( e->isModified() ) e->save(); // buffer is systematically saved
  stopOperations();
  _compileLog->clear();
  bool generated = _projectManager->isGenerated(e->url());
  // #### FIXME: we should only remove the generated files that are "below" this one...
  if ( !generated ) _projectManager->removeGenerated();
  _compileManager->compileFile(Compile::TodoItem(e->url(), generated), project());
  setState(Compiling);
}

void MainWindow::buildProject()
{
  if ( project()==0 ) {
    compileFile();
    return;
  }
  // save modified buffers
  PURL::List files = project()->absoluteFiles();
  PURL::List::const_iterator it = files.begin();
  for (; it!=files.end(); ++it) {
    // save modified editors
    Editor *e = _editorManager->findEditor(*it);
    if ( e && e->isModified() ) e->save();
  }
  stopOperations();
  _compileLog->clear();
  _projectManager->removeGenerated();
  Compile::LinkType ltype = (programmerGroup().isDebugger() ? Compile::LinkIcd2 : Compile::LinkNormal);
  _compileManager->buildProject(*project(), ltype);
  setState(Compiling);
}

void MainWindow::compileFailure()
{
  setState(Idle);
}

void MainWindow::compileSuccess()
{
  if ( !_compileManager->compileOnly() && project() ) {
    _projectManager->setModified(false);
    _debugManager->init();
    if ( GlobalConfig::programAfterBuild() ) program();
  }
  setState(Idle);
}

void MainWindow::updateFile(const Compile::FileData &fdata)
{
  if ( fdata.actions & Compile::InProject )
    _projectManager->addGeneratedFile(fdata.url);
  if ( fdata.actions & Compile::Show ) {
    Editor *e = _editorManager->openEditor(fdata.url);
    if (e) e->setReadOnly(true);
  }
}

void MainWindow::cleanBuild()
{
  stopOperations();
  _compileLog->clear();
  if ( project() ) {
    Compile::LinkType ltype = (programmerGroup().isDebugger() ? Compile::LinkIcd2 : Compile::LinkNormal);
    _compileManager->cleanProject(*project(), ltype);
  } else {
    PURL::FileType type;
    PURL url = _projectManager->standaloneGenerator(_editorManager->currentEditor()->url(), type);
    _compileManager->cleanFile(url);
  }
  _projectManager->removeGenerated();
}

void MainWindow::stopBuild()
{
  _compileManager->kill();
}

//-----------------------------------------------------------------------------
void MainWindow::keyPressEvent(QKeyEvent *e)
{
  if ( e->key()==Key_Escape ) stopOperations();
}

void MainWindow::stopOperations()
{
  if ( _state==Programming ) {
    MessageBox::sorry(i18n("Programming in progress. Cannot be aborted."), Log::Show);
    return;
  }
  _compileManager->kill();
  if (_programmer) _programmer->disconnectHardware();
  _debugManager->clear();
}

void MainWindow::newProject()
{
  stopOperations();
  _projectManager->newProject();
  updateGUI();
  _compileLog->clear();
}

void MainWindow::openProject()
{
  stopOperations();
  _projectManager->openProject();
  updateGUI();
  _compileLog->clear();
}

void MainWindow::openRecentProject(const KURL &url)
{
  stopOperations();
  _projectManager->openProject(PURL(url));
  updateGUI();
  _compileLog->clear();
}

void MainWindow::configureProject()
{
  stopOperations();
  _projectManager->editProject();
  updateGUI();
}

void MainWindow::closeProject()
{
  stopOperations();
  _projectManager->closeProject();
  updateGUI();
  _compileLog->clear();
}

//----------------------------------------------------------------------------
void MainWindow::createProgrammer(const Device::Data *data)
{
  delete _programmer;
  Port::Type ptype = Programmer::Config::readPortType(programmerGroup());
  QString portDevice = Programmer::Config::readPortDevice(programmerGroup(), ptype);
  _programmer = programmerGroup().createProgrammer(data, ptype, portDevice);
  _programmer->setView(_programLog);
  connect(_programmer, SIGNAL(actionMessage(const QString &)),
          _actionStatus, SLOT(setText(const QString &)));
}

bool MainWindow::initProgramming()
{
  const Device::Data *data = deviceData();
  if ( data==0 ) {
    MessageBox::sorry(i18n("You need to specify the device for programming."), Log::Show);
    return false;
  }
  if ( !programmerGroup().isSupported(data->name()) ) {
    MessageBox::sorry(i18n("The current programmer \"%1\" does not support device \"%2\".")
                      .arg(programmerGroup().label()).arg(data->name()), Log::Show);
    return false;
  }
  if ( _programmer==0 || _programmer->device()!=data ) createProgrammer(data);
  setState(Programming);
  return true;
}

void MainWindow::endProgramming()
{
  _debugManager->update();
  setState(Idle);
}

HexEditor *MainWindow::getHexEditor()
{
  if ( _editorManager->currentEditor() && _editorManager->currentEditor()->fileType()==PURL::Hex )
    return static_cast<HexEditor *>(_editorManager->currentEditor());
  HexEditor *editor = static_cast<HexEditor *>(_editorManager->openEditor(project()->hexUrl()));
  if ( editor==0 ) return 0;
  if ( _projectManager->isModified() ) {
    if ( !MessageBox::askContinue(i18n("<qt>The project hex file may not be up-to-date since some project files have been modified. Continue anyway ?</qt>")) )
    return 0;
  }
  editor->setReadOnly(true);
  return editor;
}

void MainWindow::program()
{
  HexEditor *editor = getHexEditor();
  if ( editor==0 ) return;
  if ( programmerGroup().isDebugger() && !_projectManager->contains(editor->url()) ) {
    MessageBox::sorry(i18n("<qt>It is not possible to start a debugging session with an hex file not generated by the current project. Please close the current project first.</qt>"), Log::Show);
    return;
  }
  if ( !initProgramming() ) return;
  _programmer->program(*editor->memory());
  endProgramming();
}

void MainWindow::verify()
{
  HexEditor *editor = getHexEditor();
  if ( editor==0 ) return;
  if ( !initProgramming() ) return;
  _programmer->verify(*editor->memory());
  endProgramming();
}

void MainWindow::read()
{
  if ( !initProgramming() ) return;
  QString s = i18n("Read");
  Editor *e = _editorManager->findEditor(s);
  if (e) _editorManager->closeEditor(e, true);
  HexEditor *editor = new HexEditor(s, _editorManager);
  editor->setDevice(device());
  if ( _programmer->read(*editor->memory()) ) {
    editor->memoryRead();
    _editorManager->addEditor(editor);
  } else delete editor;
  endProgramming();
}

void MainWindow::erase()
{
  if ( !initProgramming() ) return;
  _programmer->erase();
  endProgramming();
}

void MainWindow::blankCheck()
{
  if ( !initProgramming() ) return;
  _programmer->blankCheck();
  endProgramming();
}

void MainWindow::connectDevice()
{
  if ( !initProgramming() ) return;
  _programmer->disconnectHardware();
  _programmer->connectDevice();
  endProgramming();
}

void MainWindow::toggleDevicePower()
{
  if ( !initProgramming() ) {
    static_cast<KToggleAction &>(action("prog_power")).setChecked(false);
    return;
  }
  bool on = static_cast<KToggleAction &>(action("prog_power")).isChecked();
  _actionStatus->setText(i18n("Toggle Device Power..."));
  _programmer->setTargetVddOn(on);
  endProgramming();
}

void MainWindow::disconnectDevice()
{
  _actionStatus->setText(i18n("Disconnecting..."));
  _programmer->setTargetVddOn(false);
  _programmer->disconnectHardware();
  endProgramming();
}

void MainWindow::progRun()
{
  if ( !initProgramming() ) return;
  _programmer->run();
  setState(Idle);
}

void MainWindow::progStop()
{
  _programLog->log(Log::Information, i18n("Stop."));
  _programmer->stop();
  endProgramming();
}

void MainWindow::progRestart()
{
  _programLog->log(Log::Information, i18n("Restart..."));
  _programmer->stop();
  Port::msleep(200);
  _programmer->run();
  endProgramming();
}

void MainWindow::progAdvanced()
{
  const Programmer::GroupUI *groupui = static_cast<const Programmer::GroupUI *>(programmerGroup().gui());
  const Device::Data *data = deviceData();
  if ( data && !programmerGroup().isSupported(data->name()) ) data = 0;
  bool create = ( _programmer==0 );
  if (create) createProgrammer(data);
  _logTab->showPage(_programLog);
  Programmer::AdvancedDialog *dialog = groupui->createAdvancedDialog(*_programmer, this);
  Q_ASSERT(dialog);
  dialog->updateDisplay();
  dialog->exec();
  delete dialog;
  if ( create && data==0 ) { // don't want to keep a simpleConnected programmer around
    delete _programmer;
    _programmer = 0;
  }
  updateGUI();
}

void MainWindow::debugRun()
{
  _debugManager->run();
  setState(Idle);
}

void MainWindow::debugHalt()
{
  _debugManager->halt();
  setState(Idle);
}

void MainWindow::debugReset()
{
  _programLog->log(Log::Information, i18n("Reset."));
  _debugManager->reset();
  setState(Idle);
}

void MainWindow::debugStep()
{
  _debugManager->step();
  setState(Idle);
}

void MainWindow::setTotalSteps(uint total)
{
  //KIO::Job *job = new KIO::SimpleJob(KURL(), 0, 0, false);
  //int id = Observer::self()->newJob(job, true);
  //Observer::self()->slotTotalSize(job, total);
  //Observer::self()->slotInfoMessage(job, "test");
  //qDebug("set total steps: %i", total);
  _actionProgress->setTotalSteps(total);
  _actionProgress->setProgress(0);
  QApplication::eventLoop()->processEvents(QEventLoop::ExcludeUserInput);
}

void MainWindow::addToTotalSteps(uint add)
{
  _actionProgress->setTotalSteps(_actionProgress->totalSteps() + add);
  //qDebug("add to total steps: %i %i", add, _actionProgress->totalSteps());
}

void MainWindow::addProgress(uint add)
{
  //qDebug("add progress %i %i/%i", add, _actionProgress->progress() + add, _actionProgress->totalSteps());
  _actionProgress->setProgress(_actionProgress->progress() + add);
  QApplication::eventLoop()->processEvents(QEventLoop::ExcludeUserInput); // #### DANGER !!!!
}

void MainWindow::clearBreakpoints()
{
  _editorManager->clearMarks(KBEditor::AllBreakpoints);
}
